Package: openCyto
Type: Package
Title: Hierarchical Gating Pipeline for flow cytometry data
Version: 2.24.0
Date: 2026-03-12
Author: Mike Jiang, John Ramey, Greg Finak, Raphael Gottardo
Maintainer: Mike Jiang <mike@ozette.com>
Description: This package is designed to facilitate the automated gating
    methods in sequential way to mimic the manual gating strategy.
License: AGPL-3.0-only
LazyLoad: yes
Imports: methods, Biobase, BiocGenerics, flowCore(>= 1.99.17), flowViz,
        ncdfFlow(>= 2.11.34), flowWorkspace(>= 3.99.1), flowClust(>=
        3.11.4), RBGL, graph, data.table, RColorBrewer, grDevices
Suggests: flowWorkspaceData, knitr, rmarkdown, markdown, testthat,
        utils, tools, parallel, ggcyto, CytoML, flowStats(>= 4.5.2),
        MASS
biocViews: ImmunoOncology, FlowCytometry, DataImport, Preprocessing,
        DataRepresentation
Encoding: UTF-8
VignetteBuilder: knitr
RoxygenNote: 7.3.2
LinkingTo: cpp11, BH(>= 1.62.0-1)
git_url: https://git.bioconductor.org/packages/openCyto
git_branch: RELEASE_3_23
git_last_commit: 8a4e8f8
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 23:41:52 UTC; biocbuild
Depends: R (>= 3.5.0)

# Generated by roxygen2: do not edit by hand

S3method(as.data.table,gatingTemplate)
S3method(gt_gating,dummyMethod)
S3method(gt_gating,gatingTemplate)
S3method(pop_add,ocRectRefGate)
export(CytoExploreR_.argDeparser)
export(CytoExploreR_.preprocess_csv)
export(CytoExploreR_.preprocess_row)
export(add_pop)
export(add_pop_init)
export(delete.helperGates)
export(fast_rlm)
export(flowClust.1d)
export(flowClust.2d)
export(gate_flowClust_1d)
export(gate_flowClust_2d)
export(gate_flowclust_1d)
export(gate_flowclust_2d)
export(gate_mindensity)
export(gate_mindensity2)
export(gate_quad_sequential)
export(gate_quad_tmix)
export(gate_quantile)
export(gating)
export(gatingTemplate)
export(generate_gate_template)
export(get.helperGates)
export(gh_generate_template)
export(gs_add_gating_method)
export(gs_add_gating_method_init)
export(gs_remove_gating_method)
export(gt_delete_helpergates)
export(gt_gating)
export(gt_get_children)
export(gt_get_gate)
export(gt_get_helpergates)
export(gt_get_nodes)
export(gt_get_parent)
export(gt_list_methods)
export(gt_toggle_helpergates)
export(listgtMethods)
export(mindensity)
export(mindensity2)
export(prior_flowClust)
export(prior_flowclust)
export(quadGate.seq)
export(quadGate.tmix)
export(quantileGate)
export(registerPlugins)
export(register_plugins)
export(remove_pop)
export(templateGen)
export(toggle.helperGates)
exportClasses(gatingTemplate)
exportMethods(getChildren)
exportMethods(getGate)
exportMethods(getNodes)
exportMethods(getParent)
exportMethods(plot)
exportMethods(show)
import(data.table)
import(flowWorkspace)
import(utils)
importClassesFrom(Biobase,AssayData)
importClassesFrom(flowCore,ellipsoidGate)
importClassesFrom(flowCore,filterList)
importClassesFrom(flowCore,polygonGate)
importClassesFrom(flowCore,rectangleGate)
importClassesFrom(graph,graph)
importClassesFrom(graph,graphBase)
importClassesFrom(graph,graphNEL)
importClassesFrom(methods,ANY)
importClassesFrom(methods,character)
importClassesFrom(methods,data.frame)
importClassesFrom(methods,environment)
importClassesFrom(methods,list)
importClassesFrom(methods,logical)
importClassesFrom(methods,matrix)
importClassesFrom(methods,missing)
importClassesFrom(methods,numeric)
importClassesFrom(methods,oldClass)
importClassesFrom(ncdfFlow,ncdfFlowList)
importClassesFrom(ncdfFlow,ncdfFlowSet)
importFrom(BiocGenerics,dims)
importFrom(RBGL,tsort)
importFrom(RColorBrewer,brewer.pal)
importFrom(data.table,fwrite)
importFrom(flowClust,dmvt)
importFrom(flowClust,dmvtmix)
importFrom(flowClust,flowClust)
importFrom(flowClust,getEstimates)
importFrom(flowClust,rbox)
importFrom(flowClust,tmixFilter)
importFrom(flowCore,"%in%")
importFrom(flowCore,"exprs<-")
importFrom(flowCore,"filterDetails<-")
importFrom(flowCore,"parameters<-")
importFrom(flowCore,colnames)
importFrom(flowCore,exprs)
importFrom(flowCore,filters)
importFrom(flowCore,getChannelMarker)
importFrom(flowCore,identifier)
importFrom(flowCore,parameters)
importFrom(flowCore,polygonGate)
importFrom(flowCore,quadGate)
importFrom(flowCore,rectangleGate)
importFrom(flowWorkspace,cf_lock)
importFrom(flowWorkspace,cf_unlock)
importFrom(flowWorkspace,gh_get_pop_paths)
importFrom(flowWorkspace,gh_get_transformations)
importFrom(flowWorkspace,gh_pop_get_data)
importFrom(flowWorkspace,gh_pop_get_gate)
importFrom(flowWorkspace,gh_pop_is_negated)
importFrom(flowWorkspace,gh_pop_set_visibility)
importFrom(flowWorkspace,gs_get_pop_paths)
importFrom(flowWorkspace,gs_pop_get_children)
importFrom(flowWorkspace,identifier)
importFrom(flowWorkspace,pData)
importFrom(flowWorkspace,realize_view)
importFrom(flowWorkspace,sampleNames)
importFrom(flowWorkspace,transform)
importFrom(flowWorkspace,transformerList)
importFrom(grDevices,chull)
importFrom(graph,"edgeData<-")
importFrom(graph,"edgeDataDefaults<-")
importFrom(graph,"nodeData<-")
importFrom(graph,"nodeDataDefaults<-")
importFrom(graph,addEdge)
importFrom(graph,edgeData)
importFrom(graph,edgeDataDefaults)
importFrom(graph,edges)
importFrom(graph,graphNEL)
importFrom(graph,inEdges)
importFrom(graph,nodeData)
importFrom(graph,nodeDataDefaults)
importFrom(graph,nodes)
importFrom(graph,subGraph)
importFrom(methods,as)
importFrom(methods,coerce)
importFrom(methods,show)
useDynLib(openCyto,.registration = TRUE)

openCyto [<img src="logo_mid.png"/>](http://github.com/RGLab/openCyto)
========
[![Build Status](https://github.com/RGLab/openCyto/workflows/build/badge.svg?branch=master)](https://github.com/RGLab/openCyto/actions)

An R package that providing an automated data analysis pipeline for flow cytometry.


### Reporting Bugs or Issues
- Use the issue template in github when creating a new issue. 
- Follow the instructions in the template (do your background reading).
- Search and verify that the issue hasn't already been addressed.
- Check the Bioconductor support site. 
- Make sure your flow packages are up to date.
- THEN if your issue persists, file a bug report.

Otherwise, we may close your issue without responding.

## Documentation Resources
The [github-pages](http://opencyto.org) site documentation is a bit outdated.

The [package vignettes](http://www.bioconductor.org/packages/devel/bioc/html/openCyto.html) are the current best resource to 
get started with the openCyto.

- [An introduction to the openCyto package](https://bioconductor.org/packages/devel/bioc/vignettes/openCyto/inst/doc/openCytoVignette.html)
- [How to use different auto gating functions](https://bioconductor.org/packages/devel/bioc/vignettes/openCyto/inst/doc/HowToAutoGating.html)
- [How to write a csv gating template](https://bioconductor.org/packages/devel/bioc/vignettes/openCyto/inst/doc/HowToWriteCSVTemplate.html)

The `add_pop()` API is a good interactive approach to building up a template, population by population. It takes arguments found in the `csv` template, performs the gating of a `GatingSet` or `GatingHierarchy` and returns a line of text that can be added to a `csv` template.

#' reading flowFrame from csv spreadsheet.It is for internal usage.
#' @param file \code{character} csv file name that contains the fluorescence intensities
#' @param stains \code{character} a vector specifying the stains for channels. Default is \code{NA}
#' @return a \code{flowFrame} object
#' @importFrom flowCore parameters<-
#' @noRd
.read.FCS.csv <- function(file, stains = NA) {
    mat <- as.matrix(read.csv(file, check.names = FALSE))

    fr <- new("flowFrame", exprs = mat)

    pd <- pData(parameters(fr))
    pd$desc <- as.character(pd$desc)
    pd$name <- as.character(pd$name)

    ## update the desc with marker name
    if (!is.na(stains)) {
        ind <- match(names(stains), pd$name)
        pd[ind, ]$desc <- as.character(stains)

        ## update SSC and FSC description with NA
        ind <- grepl("[F|S]SC", pd$desc)
        pd[ind, ]$desc <- NA
    } else {
        pd$desc <- NA
    }

    ## update minRange with -111 for proper display of the data
    pd$minRange[pd$minRange < (-111)] <- -111
    pData(parameters(fr)) <- pd
    fr
}
#' reading flowSet from multiple csv spreadsheets.It is for internal usage.
#' @param files \code{character} csv file names
#' @param ... arguments passed to \link{read.FCS.csv}
#' @return a \code{flowSet} object
#' @noRd
.read.flowSet.csv <- function(files, ...) {
    fs <- flowSet(lapply(files, .read.FCS.csv, ...))
    sampleNames(fs) <- basename(files)
    fs
}

#' Creates a matrix of points on an ellipse from a fitted flowClust model.
#'
#' The ellipse is constructed from a contour from the fitted flowClust model.
#'
#' By default, the contour level is extracted from the \code{ruleOutliers} slot
#' in the \code{filter}. The user can override this level by specifying value
#' between 0 and 1 in \code{quantile}.
#'
#' @param filter object containing the fitted \code{flowClust} model.
#' @param include the mixture component in the fitted \code{flowClust} model for
#' which the contour (ellipse) is returned
#' @param ecol \code{numeric} to be documented
#' @param elty \code{numeric} to be documented
#' @param quantile the contour level of the ellipse. See details.
#' @param npoints the number of points on the ellipse
#' @param subset the dimensions of the mixture component to return
#' @param \code{...} additional parameters
#' @return matrix containing the points of the ellipse from the flowClust contour
#' @importFrom flowClust rbox
#' @noRd
.getEllipse <- function(filter = NULL, include = seq_len(filter@K), ecol = 1, elty = 1,
                        quantile = NULL, npoints = 50, subset = c(1, 2), ...) {
    .Defunct(".getEllipseGate")
    # Sets the quantile of the ellipse.
    if (is.null(quantile)) {
        quantile <- filter@ruleOutliers[2]
    } else {
        if (!is.numeric(quantile) || quantile < 0 || quantile > 1) {
            stop("The 'quantile' must be a numeric value between 0 and 1.")
        }
    }

    # py is the degrees of freedom?
    py <- 2
    ecol <- matrix(ecol, length(include))
    elty <- matrix(elty, length(include))

    if (all(filter@nu != Inf)) {
        if (filter@ruleOutliers[1] == 0) {
            # 0 means quantile
            cc <- py * qf(p = quantile, py, filter@nu)
        } else {
            # 1 means u.cutoff
            cc <- ((filter@nu + py) / quantile - filter@nu)
        }
    } else {
        cc <- qchisq(p = quantile, py)
    }

    j <- 0

    # Does trans exist in the extra parameter list?
    # If not, set it to true by default
    ellipsis <- as.environment(list(...))
    if (exists("trans", envir = ellipsis)) {
        trans <- get("trans", ellipsis)
    } else {
        trans <- 1
    }

    # Test for trans==0 when lambda is defined to get around the off
    # by one bug due to the reverse box-cox transformation
    if ((length(filter@lambda) > 0) && trans == 0) {
        lambda <- rep(filter@lambda, length.out = filter@K)
    } else {
        lambda <- numeric(0)
    }
    cc <- rep(cc, length.out = filter@K)
    for (i in include) {
        eigenPair <- eigen(filter@sigma[i, subset, subset])
        l1 <- sqrt(eigenPair$values[1]) * sqrt(cc)
        l2 <- sqrt(eigenPair$values[2]) * sqrt(cc)
        angle <- atan(eigenPair$vectors[2, 1] / eigenPair$vectors[1, 1]) * 180 / pi

        if ((length(filter@lambda) > 0) & trans == 1) {
            res <- rbox(flowClust:::.ellipsePoints(
                a = l1[i], b = l2[i], alpha = angle,
                loc = filter@mu[i, subset], n = npoints
            ), lambda[i])
        } else {
            res <- flowClust:::.ellipsePoints(
                a = l1[i], b = l2[i], alpha = angle,
                loc = filter@mu[i, subset], n = npoints
            )
        }
    }
    res
}
#' revised based on .getEllipse.
#' try to construct ellipsoidGate directly from tmixFilter result instead of polygon fitting
#' when trans = 0
#' @noRd
.getEllipseGate <- function(filter = NULL, include = seq_len(filter@K), ecol = 1, elty = 1,
                            quantile = NULL, npoints = 50, subset = c(1, 2), ...) {

    # Sets the quantile of the ellipse.
    if (is.null(quantile)) {
        quantile <- filter@ruleOutliers[2]
    } else {
        if (!is.numeric(quantile) || quantile < 0 || quantile > 1) {
            stop("The 'quantile' must be a numeric value between 0 and 1.")
        }
    }

    # py is the degrees of freedom?
    py <- 2
    ecol <- matrix(ecol, length(include))
    elty <- matrix(elty, length(include))

    if (all(filter@nu != Inf)) {
        if (filter@ruleOutliers[1] == 0) {
            # 0 means quantile
            cc <- py * qf(p = quantile, py, filter@nu)
        } else {
            # 1 means u.cutoff
            cc <- ((filter@nu + py) / quantile - filter@nu)
        }
    } else {
        cc <- qchisq(p = quantile, py)
    }

    j <- 0

    # Does trans exist in the extra parameter list?
    # If not, set it to true by default
    ellipsis <- as.environment(list(...))
    if (exists("trans", envir = ellipsis)) {
        trans <- get("trans", ellipsis)
    } else {
        trans <- 1
    }

    # Test for trans==0 when lambda is defined to get around the off
    # by one bug due to the reverse box-cox transformation
    if ((length(filter@lambda) > 0) && trans == 0) {
        lambda <- rep(filter@lambda, length.out = filter@K)
    } else {
        lambda <- numeric(0)
    }
    cc <- rep(cc, length.out = filter@K)
    cc <- sqrt(cc)
    coln <- as.vector(filter@varNames)[subset]
    for (i in include) {
        cov.mat <- filter@sigma[i, subset, subset]

        # fit polygon points
        if ((length(filter@lambda) > 0) & trans == 1) {
            if (length(filter@lambda) > 1) {
                lambda <- filter@lambda
            } else {
                lambda <- rep(filter@lambda, filter@K)
            }
            eigenPair <- eigen(cov.mat)
            l1 <- sqrt(eigenPair$values[1]) * cc
            l2 <- sqrt(eigenPair$values[2]) * cc
            angle <- atan(eigenPair$vectors[2, 1] / eigenPair$vectors[1, 1]) * 180 / pi

            contour_ellipse <- rbox(flowClust:::.ellipsePoints(
                a = l1[i], b = l2[i], alpha = angle,
                loc = filter@mu[i, subset], n = npoints
            ), lambda[i])
            res <- flowCore::polygonGate(
                .gate = matrix(contour_ellipse,
                    ncol = 2,
                    dimnames = list(NULL, coln)
                )
            )
        } else {
            # construct ellipsoidGate directly from covaraince matrix and mu
            dimnames(cov.mat) <- list(coln, coln)
            res <- flowCore:::ellipsoidGate(cov.mat, mean = filter@mu[i, subset], distance = cc[1])
        }
    }
    res
}

#' Removes any observation from the given flowFrame object that has values
#' outside the given range for the specified channels
#'
#' The minimum/maximum values are ignored if \code{NULL}.
#'
#' @param flow_frame a \code{flowFrame} object
#' @param min a numeric vector that sets the lower bounds for data filtering
#' @param max a numeric vector that sets the upper bounds for data filtering
#' @param channels \code{character} specifying which channel to operate on
#' @return a \code{flowFrame} object
#' @examples
#' \dontrun{
#' library(flowClust)
#' data(rituximab)
#' # Consider the range of values for FSC.H and SSC.H
#' summary(rituximab)
#'
#' # Truncates any observations with FSC.H outside [100, 950]
#' rituximab2 <- .truncate_flowframe(rituximab, channels = "FSC.H", min = 100, max = 950)
#' summary(rituximab2)
#' # Next, truncates any observations with SSC.H outside [50, 1000]
#' rituximab3 <- .truncate_flowframe(rituximab2, channels = "SSC.H", min = 50, max = 1000)
#' summary(rituximab3)
#'
#' # Instead, truncates both channels at the same time
#' rituximab4 <- .truncate_flowframe(rituximab,
#'     channels = c("FSC.H", "SSC.H"),
#'     min = c(100, 50), max = c(950, 1000)
#' )
#' summary(rituximab4)
#' }
#'
#' @noRd
.truncate_flowframe <- function(flow_frame, channels, min = NULL, max = NULL) {
    channels <- as.character(channels)
    num_channels <- length(channels)

    # For comparison purposes, we update the min and max values to -Inf and Inf,
    # respectively, if NULL.
    if (is.null(min)) {
        min <- rep(-Inf, num_channels)
    }
    if (is.null(max)) {
        max <- rep(Inf, num_channels)
    }

    if (!(num_channels == length(min) && num_channels == length(max))) {
        stop("The lengths of 'min' and 'max' must match the number of 'channels' given.")
    }

    gate_coordinates <- lapply(seq_len(num_channels), function(i) {
        c(min[i], max[i])
    })
    names(gate_coordinates) <- channels

    truncate_filter <- rectangleGate(gate_coordinates)
    Subset(flow_frame, truncate_filter)
}

#' Removes any observation from each flowFrame object within the flowSet that has
#' values outside the given range for the specified channels
#'
#' The minimum/maximum values are ignored if \code{NULL}.
#'
#' @param flow_set a \code{flowSet} object
#' @inheritParams .truncate_flowframe
#' @return a \code{flowSet} object
#' @noRd
.truncate_flowset <- function(flow_set, channels, min = NULL, max = NULL) {
    channels <- as.character(channels)
    num_channels <- length(channels)

    # For comparison purposes, we update the min and max values to -Inf and Inf,
    # respectively, if NULL.
    if (is.null(min)) {
        min <- rep(-Inf, num_channels)
    }
    if (is.null(max)) {
        max <- rep(Inf, num_channels)
    }

    if (!(num_channels == length(min) && num_channels == length(max))) {
        stop("The lengths of 'min' and 'max' must match the number of 'channels' given.")
    }

    gate_coordinates <- lapply(seq_len(num_channels), function(i) {
        c(min[i], max[i])
    })
    names(gate_coordinates) <- channels

    truncate_filter <- rectangleGate(gate_coordinates)
    Subset(flow_set, truncate_filter)
}


#' Computes the quantile from flowClust for a given vector of probabilties
#'
#' We estimate the quantile from a \code{flowClust} fit with a combination of
#' numerical integration and a root-finding method. We are effectively
#' estimating the cumulative distribution function (CDF) of the mixture density
#' estimated by \code{flowClust}.
#'
#' Because we are using numerical methods, we also need an \code{interval} of
#' values in which we will attempt to find the specified quantile.
#'
#' @param p vector of probabilities
#' @param object an object containing the \code{flowClust} fit
#' @param interval a vector of length 2 containing the end-points of the interval
#' of values to find the quantile
#' @param ... Additional arguments that are passed to \code{uniroot} to find the
#' quantile.
#' @return the quantile corresponding to the specified probabilities
#' @noRd
.quantile_flowClust <- function(p, object, interval, ...) {
    cdf_target <- function(x, p, object) {
        cdf_values <- sapply(seq_len(object@K), function(k) {
            nu <- ifelse(length(object@nu) == 1, object@nu, object@nu[k])
            lambda <- ifelse(length(object@lambda) == 1, object@lambda, object@lambda[k])

            # TODO: Incorporate the Box-Cox transformation (i.e., box(qt(...), lambda =
            # lambda)) into quantile The case of 'lambda = 1' may be not be trivial -- this
            # case is largely ignored in flowClust.
            pt((x - object@mu[k]) / sqrt(object@sigma[k]), df = nu)
        })
        weighted.mean(cdf_values, w = object@w) - p
    }

    uniroot(cdf_target, interval = interval, p = p, object = object, ...)$root
}

#' Extracts the quadrants of a quadGate as a list of rectangleGates
#'
#' The quadrants are numbered in a clockwise manner with the top-left quadrant
#' numbered 1, the top-right quadrant numbered 2, and so on.
#'
#' @param quad_gate a \code{quadGate} object
#' @param markers character vector of the marker names for the x- and y-axes
#' @param channels character vector of the channel names for the x- and y-axes
#' @param quadrants a vector indicating the quadrants to extract
#' @return a \code{filters} object containing a list of the rectangle gates
#' @noRd
.quadGate2rectangleGates <- function(quad_gate, markers, channels, quadrants = 1:4) {
    x_gate <- quad_gate@boundary[1]
    y_gate <- quad_gate@boundary[2]

    gates_list <- list()

    # Top-left quadrant
    gates <- list(c(-Inf, x_gate), c(y_gate, Inf))
    names(gates) <- channels
    gates_list[[paste0(markers[1], "-", markers[2], "+")]] <- rectangleGate(gates)

    # Top-right quadrant
    gates <- list(c(x_gate, Inf), c(y_gate, Inf))
    names(gates) <- channels
    gates_list[[paste0(markers[1], "+", markers[2], "+")]] <- rectangleGate(gates)

    # Lower-right quadrant
    gates <- list(c(x_gate, Inf), c(-Inf, y_gate))
    names(gates) <- channels
    gates_list[[paste0(markers[1], "+", markers[2], "-")]] <- rectangleGate(gates)

    # Lower-left quadrant
    gates <- list(c(-Inf, x_gate), c(-Inf, y_gate))
    names(gates) <- channels
    gates_list[[paste0(markers[1], "-", markers[2], "-")]] <- rectangleGate(gates)

    filters(gates_list[quadrants])
}

#' Constructs a vector of all the combinations of A & B & C
#'
#' The \code{permutations} function is from the \code{gregmisc} package on CRAN.
#' @param markers character vector of marker names
#' @return vector containing all combinations of the markers
#' @examples
#' .polyfunction_nodes(c("IFNg", "IL2", "TNFa", "GzB", "CD57"))
#' @noRd
.polyfunction_nodes <- function(markers) {
    markers <- paste0(markers, "+")
    num_markers <- length(markers)
    and_list <- as.vector(permutations(n = 1, r = num_markers - 1, c("&"), repeats.allowed = TRUE))
    isnot_list <- permutations(n = 2, r = num_markers, c("!", ""), repeats.allowed = TRUE)
    apply(isnot_list, 1, function(isnot_row) {
        isnot_row[-1] <- paste0(and_list, isnot_row[-1])
        paste(paste0(isnot_row, markers), collapse = "")
    })
}

#' Finds values of a vector in an interval
#'
#' @param x numeric vector
#' @param interval numeric vector of length 2
#' @return numeric vector containing the values of \code{x} between
#' \code{interval}. If no values are found, \code{NA} is returned.
#' @examples
#' z <- seq.int(1, 9, by = 2)
#' .between_interval(z, interval = c(2, 8))
#' @noRd
.between_interval <- function(x, interval) {
    x <- x[findInterval(x, interval) == 1]
    if (length(x) == 0) {
        x <- NA
    }
    x
}

#' Constructs a 2x2 rotation matrix for a given angle
#' @param theta \code{numeric} the degree of rotation that ensures the angle between the x-axis and the eigenvector is between 0 and pi
#' @noRd
.rotation_matrix <- function(theta) {
    matrix(c(cos(theta), sin(theta), -sin(theta), cos(theta)), nrow = 2)
}

#' Validate the arguments to flowClust and tell the user early if there are problems
#' For now we make sure that the K argument to flowClust and to prior_flowClust are present and agree.
#' @param df a \code{data.frame} of the template
#' @return a \code{data.frame}
#' @noRd
.validateFlowClustArgs <- function(df) {
    inds <- which(df$gating_method %like% "flowClust")
    for (i in inds) {
        fc_args <- .argParser(df[i, "gating_args"], TRUE)
        prepro_args <- .argParser(df[i, "preprocessing_args"], TRUE)
        if (length(fc_args) == 0) {
            # default K
            fc_args$K <- 2
        }

        if (exists("K", fc_args)) {
            K <- fc_args$K
        }

        if (length(prepro_args) == 0) {
            prepro_args$K <- K
        } else if (exists("K", prepro_args)) {
            if (!(prepro_args$K == K)) {
                warning("K in preprocessing_args doesn't match K in gating_args for flowClust.\nOverridiing with K from flowClust.")
            }
            prepro_args$K <- K
        }
        df[i, "preprocessing_args"] <- gsub("\\)$", "", gsub("list\\(", "", deparse(prepro_args, control = c())))
        df[i, "gating_args"] <- gsub("\\)$", "", gsub("list\\(", "", deparse(fc_args, control = c())))
    }
    df
}

#' Centers a vector of data using the mode of the kernel density estimate
#'
#' @param x numeric vector
#' @param ... additional arguments passed to \code{\link{density}}
#' @return numeric vector containing the centered data
#' @noRd
.center_mode <- function(x, ...) {
    x <- as.vector(x)
    density_x <- density(x, ...)
    mode <- density_x$x[which.max(density_x$y)]

    x <- as.vector(scale(x, center = mode, scale = FALSE))
    attributes(x) <- list(`mode` = mode)
    x
}

#' Scales a vector of data using the Huber robust estimator for mean and
#' standard deviation
#'
#' This function is an analog to \code{\link{scale}} but using Huber robust
#' estimators instead of the usual sample mean and standard deviation.
#'
#' @param x numeric vector
#' @param center logical value. Should \code{x} be centered?
#' @param scale logical value. Should \code{x} be scaled?
#' @return numeric vector containing the scaled data
#' @noRd
.scale_huber <- function(x, center = TRUE, scale = TRUE) {
    x <- as.vector(x)
    sd <- mad(x)
    huber_x <- robust_m_estimator(x, sd)

    # If 'center' is set to TRUE, we center 'x' by the Huber robust location
    # estimator.
    center_x <- FALSE
    if (center) {
        center_x <- huber_x
    }

    # If 'scale' is set to TRUE, we scale 'x' by the Huber robust standard
    # deviation estimator.
    scale_x <- FALSE
    if (scale) {
        scale_x <- sd
    }

    x <- as.vector(base::scale(x, center = center_x, scale = scale_x))

    if (!center) {
        center_x <- NULL
    }
    if (!scale) {
        scale_x <- NULL
    }
    attributes(x) <- list(center = center_x, scale = scale_x)
    x
}

#' Standardizes a channel within a \code{flowSet} object using the mode of the
#' kernel density estimate and the Huber estimator of the standard deviation
#'
#' @param fs a \code{flowSet} object
#' @param channel the channel to standardize
#' @param data \code{logical} indicating whether to return the transformed flow data.
#'
#' @return the \code{transformation} list, center, scale and optionally the transformed \code{flowFrame}
#' @noRd
.standardize_flowFrame <- function(fr, channel, data = TRUE) {
    x <- exprs(fr)[, channel]

    if (length(x) >= 2) {
        # First, centers the values by the mode of the kernel density estimate.
        x <- .center_mode(x)
        mode <- attr(x, "mode")

        # Scales the marker cells by the Huber estimator of the standard deviation.
        x <- .scale_huber(x, center = FALSE)
        sd_huber <- attr(x, "scale")

        exprs(fr)[, channel] <- x
    } else {
        mode <- NA
        sd_huber <- NA
    }
    res <- list(center = mode, scale = sd_huber)

    if (data) {
          res$flow_frame <- fr
      }

    attr(res, "openCyto_preprocessing") <- "standardize"


    res
}
#' convert gate to a filterResult
#'
#' used for computing the gate indices on the fly
#'
#' @param fr flowFrame
#' @param channel \code{character} channel to used for gate
#' @param gate \code{filter} object
#' @param positive \code{logical}
#' @noRd
.gateToFilterResult <- function(fr, channel, gate, positive) {
    x <- exprs(fr)[, channel]
    gate_coordinates <- c(gate@min, gate@max)
    cutpoint <- gate_coordinates[!is.infinite(gate_coordinates)]
    nCount <- length(x)
    # deal with dummy gate where both boundaries are Inf
    if (length(cutpoint) == 0) {
        if (gate@min < gate@max) {
            ind <- rep(positive, nCount)
        } else {
            ind <- rep(!positive, nCount)
        }
    }
    if (positive) {
          ind <- x >= cutpoint
      } else {
          ind <- x < cutpoint
      }
    #  fres <- as(ind, "filterResult")
    #  filterDetails(fres, identifier(gate)) <- gate
    # pack into bit vector to ease the traffic

    fres <- as(gate, "ocRectangleGate")
    fres@ind <- ncdfFlow:::toBitVec(ind)
    fres
}

#' CytoExploreR exports
#' 
#' Exported wrappers of internal functions for use by CytoExploreR
#' 
#' @name CytoExploreR_exports
NULL

#' @name CytoExploreR_.argDeparser
#' @keywords internal
#' @rdname CytoExploreR_exports
#' @export
CytoExploreR_.argDeparser <- function(args, split = TRUE){
  .argDeparser(args, split = split)
}

#' @name CytoExploreR_.preprocess_csv
#' @keywords internal
#' @rdname CytoExploreR_exports
#' @export
CytoExploreR_.preprocess_csv <- function(dt, strict = TRUE){
  .preprocess_csv(dt, strict = strict)
}

#' @name CytoExploreR_.preprocess_row
#' @keywords internal
#' @rdname CytoExploreR_exports
#' @export
CytoExploreR_.preprocess_row <- function(this_row, strict = TRUE){
  .preprocess_row(this_row, strict = strict)
}

## ----setup, include=FALSE-----------------------------------------------------
library(knitr)
opts_chunk$set(out.extra='style="display:block; margin: auto"', fig.align="center", message = FALSE, warning = FALSE, knitr.table.format = "markdown")
assign("depthtrigger", 60, data.table:::.global)

## ----init, echo=F-------------------------------------------------------------
library(openCyto)
library(data.table)

template_row <- data.table(alias = "TH", pop = "+-"
                          , parent = "cd3", dims = "cd4,cd8"
                          , gating_method = "mindensity", gating_args = ""
                          , collapseDataForGating = "TRUE", groupBy = "4"
                          , preprocessing_method = "", preprocessing_args = ""
                      )



## ----A+/-(before), echo = F, results = "asis"---------------------------------
	this_row <- copy(template_row)	
	this_row[, pop := "+/-"]
	this_row[, alias := "*"]
	this_row[, dims := "cd4"]
	kable(this_row[, 1:6, with = F])

## ----A+/-(after), echo = F, results = "asis"----------------------------------
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])
      

## ----A+B+(before), echo = F, results = "asis"---------------------------------
	this_row <- copy(template_row)	
	this_row[, pop := "+-"]
	this_row[, alias := "T helper"]
	kable(this_row[, 1:6, with = F])

## ----A+B+(after), echo = F, results = "asis"----------------------------------
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])

## ----A+/-B+/-(before), echo = F, results = "asis"-----------------------------
	this_row <- copy(template_row)	
	this_row[, pop := "+/-+/-"]
	this_row[, alias := "*"]
	kable(this_row[, 1:6, with = F])

## ----A+/-B+/-(after), echo = F, results = "asis"------------------------------
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])

## ----multiPop(before), echo = F, results = "asis"-----------------------------
	this_row <- copy(template_row)	
	this_row[, pop := "*"]
	this_row[, alias := "CD4,CD8"]
	this_row[, gating_method := "curv2gate"]
	kable(this_row[, 1:6, with = F])

## ----multiPop(after), echo = F, results = "asis"------------------------------
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])

## ----multiPop_noExpand(before), echo = F, results = "asis"--------------------
	this_row[, alias := "*"]
	kable(this_row[, 1:6, with = F])

## ----multiple_parents, echo=F, results='asis'---------------------------------
	this_row[, alias := "IL2+"]
  this_row[, pop := "+"]
  this_row[, parent := "cd4,cd8"]
  this_row[, dims := "IL2"]
  this_row[, gating_method := "refGate"]
  this_row[, gating_args := "cd3/IL2+"]
	kable(this_row[, 1:6, with = F])

## ----multiple_parents_expand, echo=F, results='asis'--------------------------
kable(openCyto:::.split_multi_parents(this_row)[, 1:6, with = F])

## ----finalizer, include=FALSE-------------------------------------------------
assign("depthtrigger", 3, data.table:::.global)

#' @name openCyto.options
#' @description  Get/set some global options for openCyto
#' @rdname openCyto.options
#' @aliases openCyto.options
#' @title Some global options for openCyto
#' See examples for the meaning of these options and how to get/set them.
#' @examples 
#' opt <- getOption("openCyto")
#' #the threshold of minimum cell events required for the gating algorithm to proceed
#' opt[["gating"]][["minEvents"]]
#' #to change the threshold
#' opt[["gating"]][["minEvents"]] <- 100
#' options(openCyto = opt)
#' 
#' #switch off the validity check flags(Not recommended)
#' opt[["check.pop"]] <- FALSE
#' options(openCyto = opt)
#' 
NULL

.onLoad <- function(libname, pkgname) 
{
  options("openCyto" = list(gating = list(minEvents = 0 # minimum events needed in order for gating to proceed
                                          )
                            , check.pop = TRUE
                            )
          )
}

# For keeping track of add_pop history
add_pop_history <- new.env(parent = emptyenv())
add_pop_history$records <- list()

#' @title Deprecated functions in package \pkg{openCyto}.
#' @templateVar old add_pop
#' @templateVar new gs_add_gating_method
#' @template template-depr_pkg
NULL
#' apply a gating method to the \code{GatingSet}
#' 
#' When interacting with the existing gated data, this function provides an alternative way to interact with the GatingSet 
#' by supplying the gating description directly through arguments without the need to write the complete
#' csv gating template. 
#' 
#' Calls to \code{gs_add_gating_method} can also be easily reversed with \code{\link{gs_remove_gating_method}}. Note, however, that it is not possible
#' to differentiate between different \code{GatingSet} objects loaded from the same directory with 
#' \code{\link[flowWorkspace]{load_gs}} within a session. Thus, to guarantee a clean history for \code{gs_remove_gating_method}, 
#' it is necessary to call \code{\link{gs_add_gating_method_init}} on the loaded \code{GatingSet} immediately after re-loading it. 
#' See the documentation for \code{\link{gs_add_gating_method_init}} for more details. 
#' This will not be an issue for \code{GatingSet} objects created directly using the constructor. 
#'      
#' @name gs_add_gating_method
#' @aliases add_pop
#' @param gs GatingSet or GatingSetList
#' @param alias,pop,parent,dims,gating_method,gating_args,collapseDataForGating,groupBy,preprocessing_method,preprocessing_args see details in \link[openCyto:gatingTemplate-class]{gatingTemplate}
#' @param strip_extra_quotes \code{logical} Extra quotes are added to strings by fread. This causes problems with parsing R strings to expressions in some cases. Default FALSE for usual behaviour. TRUE should be passed if parsing gating_args fails.
#' @param ... other arguments
#' \itemize{
#'      \item{mc.cores}{ passed to \code{multicore} package for parallel computing}
#'      \item{parallel_type}{  \code{character} specifying the parallel type. The valid options are "none", "multicore", "cluster".}
#'      \item{cl}{ \code{cluster} object passed to \code{parallel} package (when \code{parallel_type} is "cluster")}
#'      }
#' @seealso \code{\link{gs_remove_gating_method}} \code{\link{gs_add_gating_method_init}}
#' @examples 
#' \dontrun{
#' # add quad gates 
#' gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
#' 
#' # polyfunctional gates (boolean combinations of exsiting marginal gates)
#' gs_add_gating_method(gs, gating_method = "polyFunctions", parent = "cd8", gating_args = "cd8/IFNg:cd8/IL2:cd8/TNFa")
#' 
#' #boolGate method
#' gs_add_gating_method(gs, alias = "IL2orIFNg", gating_method = "boolGate", parent = "cd4", gating_args = "cd4/IL2|cd4/IFNg") 
#' }
#' @export
gs_add_gating_method <- function(gs, alias = "*"
                       , pop = "+"
                       , parent
                       , dims = NA
                       , gating_method
                       , gating_args = NA
                       , collapseDataForGating = NA
                       , groupBy = NA
                       , preprocessing_method = NA
                       , preprocessing_args = NA
                       , strip_extra_quotes = FALSE
                       , ...) {
  #still check this new pop                     
  .validity_check_alias(alias)
  
  #generate the dummy template based on the existing gating hierarchy
  dt <- as.data.table(gh_generate_template(gs[[1]]))
  pre_add_state <- gs_get_pop_paths(gs[[1]])
  
  if(nrow(dt)>0){
    #Can't use the existing dummy_gate since it is dedicated as dummy_ref gate generated by multiPos entry (alias = '*')
    #which requires the ref node to be explicitly supplied
    dt[, gating_method := "dummy"]
  }
  if(is.list(gating_args))
  {
    gating_args <- .argDeparser(gating_args)      
  }
  if(is.list(preprocessing_args))
  {
    preprocessing_args <- .argDeparser(preprocessing_args)      
  }
  
  thisRow <- data.table(alias = alias
                        , pop = pop
                        , parent = parent
                        , dims = dims
                        , gating_method = gating_method
                        , gating_args =  gating_args
                        , collapseDataForGating = collapseDataForGating
                        , groupBy = groupBy
                        , preprocessing_method = preprocessing_method
                        , preprocessing_args = preprocessing_args
  )
  if(nrow(thisRow)>1)
    stop("Can't add multiple rows!Please make sure each argument is of length 1.")
  #there's a weird bug where rbinding a 0-row dt and a non-zero row dt returns > 4M rows.
  if(nrow(dt)>0){       
    dt <- rbind(dt, thisRow)   
  }else{
    dt = thisRow
  }
  
  tmp <- tempfile(fileext = ".csv")
  write.csv(dt, tmp, row.names = F)
  
  #skip the validity check on the other entries
  # Pass ... to gatingTemplate to allow strip_extra_quotes to be passed
  suppressMessages(gt <- gatingTemplate(tmp, strict = FALSE,strip_extra_quotes = strip_extra_quotes))
  message("...")
  suppressMessages(gt_gating(gt, gs, ...))
  message("done")
  post_add_state <- gs_get_pop_paths(gs[[1]])
  
  ## Add records if everything succeeded
  # Find record for this gating set or create it if necessary
  if(!(identifier(gs) %in% names(add_pop_history$records))){
    add_pop_history$records[[identifier(gs)]] <- list()
    # Fresh record, so make the pre_add snapshot the first
    # Otherwise, it's already there from the last call to gs_add_gating_method)
    add_pop_history$records[[identifier(gs)]][[1]] <- pre_add_state
    
    ## If it's a GatingSetList, make this the first snapshot for each of its GatingSets if needed
    if(is(gs, "GatingSetList")){
      lapply(gs, function(x){
        if(!(identifier(x) %in% names(add_pop_history$records))){
          add_pop_history$records[[identifier(x)]] <- list()
          add_pop_history$records[[identifier(x)]][[1]] <- pre_add_state
        }
      })
    }
  }
  
  # Push on the new record
  add_pop_history$records[[identifier(gs)]][[length(add_pop_history$records[[identifier(gs)]])+1]] <- post_add_state
  
  # If it's a GatingSetList, push on the new record for each of its GatingSets
  if(is(gs, "GatingSetList")){
    lapply(gs, function(x){
      add_pop_history$records[[identifier(x)]][[length(add_pop_history$records[[identifier(x)]])+1]] <- post_add_state
    })
  }
  
  invisible(thisRow)
}

#' @export
add_pop <- function(gs, alias = "*"
                    , pop = "+"
                    , parent
                    , dims = NA
                    , gating_method
                    , gating_args = NA
                    , collapseDataForGating = NA
                    , groupBy = NA
                    , preprocessing_method = NA
                    , preprocessing_args = NA
                    , strip_extra_quotes = FALSE
                    , ...){
  .Deprecated("gs_add_gating_method")
  gs_add_gating_method(gs, alias, pop, parent, dims, gating_method
             , gating_args, collapseDataForGating, groupBy
             , preprocessing_method, preprocessing_args
             , strip_extra_quotes, ...)
}


#' @templateVar old add_pop_init
#' @templateVar new gs_add_gating_method_init
#' @template template-depr_pkg
NULL
#' Clear history of \code{gs_add_gating_method} calls for a given \code{GatingSet} or \code{GatingSetList}
#' 
#' Repeated calls to the \code{\link{load_gs}} method in the same session
#' will yield indistinguishable objects that can result in overlapping history
#' of \code{\link{gs_add_gating_method}} calls. This method allows for the history to be cleared
#' if the user would like to reload the \code{GatingSet} and start fresh. Calling
#' \code{gs_add_gating_method_init} without an argument will clear the entire \code{gs_add_gating_method} history.
#' 
#' @name gs_add_gating_method_init
#' @aliases add_pop_init
#' @usage 
#' gs_add_gating_method_init(gs)
#' @param gs a \code{GatingSet} or \code{GatingSetList}. Can be omitted to clean entire \code{gs_add_gating_method} history.
#' 
#' @examples
#' \dontrun{
#' # load in a GatingSet
#' gs <- load_gs(path)
#' # Add some nodes using gs_add_gating_method
#' gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
#' gs_add_gating_method(gs, gating_method = "polyFunctions", parent = "cd8", gating_args = "cd8/IFNg:cd8/IL2:cd8/TNFa")
#' # Remove the effect of the last gs_add_gating_method call using gs_remove_gating_method (note that the first call's effects remain)
#' gs_remove_gating_method(gs)
#' # Re-load the GatingSet to start over
#' gs <- load_gs(path)
#' 
#' # At this point, gs will still see the history of the first gs_add_gating_method call above
#' # which will cause problems for later calls to gs_remove_gating_method.
#' # To fix that, just call gs_add_gating_method_init() to start a clean history
#' gs_add_gating_method_init(gs)
#' # Now you can continue using gs_add_gating_method and gs_remove_gating_method from scratch
#' gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
#' }
#' 
#' @export
gs_add_gating_method_init <- function(gs = NULL){
  if(!is.null(gs)){
    if(is(gs, "GatingSetList")){
      lapply(gs, function(x) add_pop_history$records[[identifier(x)]] <- NULL)
    }
    add_pop_history$records[[identifier(gs)]] <- NULL
  }else{
    add_pop_history$records <- list() 
  }
}

#' @export
add_pop_init <- function(gs = NULL){
  .Deprecated("gs_add_gating_method_init")
  gs_add_gating_method_init(gs)
}

context("negated gate")

test_that("negated ellipse gate", {
  gs <- gs_copy_tree_only(gs)
  #grab an existing ellipse gate for the demo purpose
  ellipse_gate <- gh_pop_get_gate(gs[[1]], "CD4")
  #create a dummy gating function that simply return this static gate
  dummy_func <- function(fr, pp_res, channels, ...)
  {
    return (ellipse_gate)
  }
  #register it as the plugin gating method
  register_plugins(dummy_func, "cd4gate")
  #apply it to the gs
  gs_add_gating_method(gs, alias = "cd4_negated", parent = "CD3+", pop = "-", dims = "CD4,CD8", gating_method = "cd4gate")
  
  expect_true(flowWorkspace::gh_pop_is_negated(gs[[1]],"cd4_negated"))
  expect_equal(gh_pop_get_proportion(gs[[1]], "cd4_negated"), 0.3753648)
  
  #add 2d ref gate  
  gs_add_gating_method(gs, alias = "cd4_ref", parent = "CD3+", pop = "+", dims = "CD4,CD8", gating_method = "refGate",gating_args = "CD4")
  expect_false(flowWorkspace::gh_pop_is_negated(gs[[1]],"cd4_ref"))
  expect_equal(gh_pop_get_proportion(gs[[1]], "CD4"), gh_pop_get_proportion(gs[[1]], "cd4_ref"))
    
  gs_add_gating_method(gs, alias = "cd4_ref_negated", parent = "CD3+", pop = "-", dims = "CD4,CD8", gating_method = "refGate",gating_args = "CD4")
  expect_true(flowWorkspace::gh_pop_is_negated(gs[[1]],"cd4_ref_negated"))
  expect_equal(gh_pop_get_proportion(gs[[1]], "cd4_negated"), gh_pop_get_proportion(gs[[1]], "cd4_ref_negated"))
})

context("refgate")

test_that("refGate", {
  gs <- gs_copy_tree_only(gs)
  #clear existing gates
  gs_pop_remove("not debris", gs = gs)
  
  #add 1d gates
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "root", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "CD8", pop = "+", parent = "root", dims = "CD8", gating_method = "mindensity")
  #add quadgates through refGates
  gs_add_gating_method(gs, pop = "+/-+/-", parent = "root", dims = "CD4,CD8", gating_method = "refGate", gating_args = "CD4+:CD8")
  nodes <- gs_get_pop_paths(gs)[5:8]
  stats1 <- gh_pop_compare_stats(gs[[1]], subpopulations = nodes)
  for(node in nodes)
    gs_pop_remove(node, gs = gs)
  #refer to cd4-
  gs_add_gating_method(gs, pop = "+/-+/-", parent = "root", dims = "CD4,CD8", gating_method = "refGate", gating_args = "CD4-:CD8")
  stats2 <- gh_pop_compare_stats(gs[[1]], subpopulations = nodes)
  
  expect_equal(stats1, stats2)
  
})

context("gating method that takes no dimensions")

test_that(" gating method that simply return Logical vector", {
  gs <- GatingSet(GvHD[1])
  # Add a gate
  gs_pop_add(
      gs,
        parent = "root",
        gate = rectangleGate(
              "FSC-H"=c(225,500), 
                  "SSC-H"=c(15, 150),
                  filterId="test"
        )
  )
        
  #create a dummy gating function that mimic sampling method
  dummy_func <- function(fr, pp_res, channels, ...)
  {
    n = nrow(fr)
    res = rep(TRUE,n)
    return (res)
  }
  #register it as the plugin gating method
  register_plugins(dummy_func, "sample_method")
  #apply it to the gs
  gs_add_gating_method(gs, alias = "A", parent = "root", pop = "*", dims = "", gating_method = "sample_method")
  
  expect_equal(gs_get_pop_paths(gs),c("root","/test","/A"))
  expect_is(gh_pop_get_gate(gs[[1]], "A"), "booleanFilter")
})
  
test_that("clustering gating method that returns factor representing multiple clusters generated by the clustering", {
    gs <- GatingSet(GvHD[1])
  
      #create a dummy gating function that mimic clustering method
  dummy_func <- function(fr, pp_res, channels, ...)
  {
    n = nrow(fr)
    res = rep(1,n)
    res[1:10] = 2
    res = factor(res)
    return (res)
  }
  #register it as the plugin gating method
  register_plugins(dummy_func, "clustering_B")
  # Note that alias must be *, the generated clustering populations are named after gating method and factor level
  gs_add_gating_method(gs, alias = "*", parent = "root", pop = "*", dims = "", gating_method = "clustering_B")
  
  expect_equal(gs_get_pop_paths(gs),c("root","/clustering_B_1","/clustering_B_2"))
  expect_is(gh_pop_get_gate(gs[[1]], "clustering_B_1"), "booleanFilter")  
})


test_that("clustering gating method that returns multipleFilterResult", {
  gs <- GatingSet(GvHD[1])
  
  #create a dummy gating function that mimic clustering method
  dummy_func <- function(fr, pp_res, channels, ...)
  {
    n = nrow(fr)
    res = rep(1,n)
    res[1:10] = 2
    res = factor(res)
    # Now we  explicitly name these populations
    levels(res) = c("cluster_A", "cluster_B")
    res = as(res, "filterResult")
    
    return (res)
  }
  #register it as the plugin gating method
  register_plugins(dummy_func, "clustering_method1")
  gs_add_gating_method(gs, alias = "*", parent = "root", pop = "*", dims = "", gating_method = "clustering_method1")
  
  expect_equal(gs_get_pop_paths(gs),c("root","/cluster_A","/cluster_B"))
  expect_is(gh_pop_get_gate(gs[[1]], "cluster_B"), "booleanFilter")  
})

context("gs_remove_gating_method")

gs1 <- gs_copy_tree_only(gs)
gs2 <- gs_copy_tree_only(gs)

test_that("gs_remove_gating_method testing", {
  snapshot_pre <- gs_get_pop_paths(gs)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  snapshot_1 <- gs_get_pop_paths(gs)
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  snapshot_2 <- gs_get_pop_paths(gs)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.95")
  snapshot_3 <- gs_get_pop_paths(gs)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_2)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_1)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_pre)
  expect_error(gs_remove_gating_method(gs))
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.98")
  expect_equal(gs_get_pop_paths(gs), snapshot_3)
  gs_remove_gating_method(gs)
  gs_remove_gating_method(gs)
  gs_remove_gating_method(gs)

  # Just to prevent GatingSetList constructor from protesting overlapping samples
  sampleNames(gs2) <- "CytoTrol_Cytotrol_2.fcs"
  gslist <- GatingSetList(list(gs1, gs2))

  expect_equal(gs_get_pop_paths(gslist[[1]]), snapshot_pre)
  gs_add_gating_method(gslist, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gslist[[1]]), snapshot_1)
  gs_add_gating_method(gslist, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gslist[[1]]), snapshot_2)
  gs_add_gating_method(gslist, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.95")
  expect_equal(gs_get_pop_paths(gslist[[1]]), snapshot_3)
  gs_remove_gating_method(gslist)
  expect_equal(gs_get_pop_paths(gslist[[2]]), snapshot_2)
  gs_remove_gating_method(gslist)
  expect_equal(gs_get_pop_paths(gslist[[2]]), snapshot_1)
  gs_remove_gating_method(gslist)
  expect_equal(gs_get_pop_paths(gslist[[2]]), snapshot_pre)
  expect_error(gs_remove_gating_method(gslist))
  gs_add_gating_method(gslist, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gslist, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  gs_add_gating_method(gslist, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.98")
  expect_equal(gs_get_pop_paths(gslist[[2]]), snapshot_3)
  gs_remove_gating_method(gslist)
  gs_remove_gating_method(gslist)
  gs_remove_gating_method(gslist)
  
  # Repeat test for one of the component GatingSets to make sure it carried the adds and removals correctly
  expect_equal(gs_get_pop_paths(gs2), snapshot_pre)
  gs_add_gating_method(gs2, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs2), snapshot_1)
  gs_add_gating_method(gs2, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs2), snapshot_2)
  gs_add_gating_method(gs2, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.95")
  expect_equal(gs_get_pop_paths(gs2), snapshot_3)
  gs_remove_gating_method(gs2)
  expect_equal(gs_get_pop_paths(gs2), snapshot_2)
  gs_remove_gating_method(gs2)
  expect_equal(gs_get_pop_paths(gs2), snapshot_1)
  gs_remove_gating_method(gs2)
  expect_equal(gs_get_pop_paths(gs2), snapshot_pre)
  expect_error(gs_remove_gating_method(gs2))
  gs_add_gating_method(gs2, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gs2, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  gs_add_gating_method(gs2, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.98")
  expect_equal(gs_get_pop_paths(gs2), snapshot_3)
  gs_remove_gating_method(gs2)
  gs_remove_gating_method(gs2)
  gs_remove_gating_method(gs2)
  
  # Repeat test for loaded GatingSet with same guid to make sure starting from blank slate
  # after removals
  gs <- load_gs(list.files(dataDir, pattern = "gs_manual",full = TRUE))
  expect_equal(gs_get_pop_paths(gs), snapshot_pre)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs), snapshot_1)
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs), snapshot_2)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.95")
  expect_equal(gs_get_pop_paths(gs), snapshot_3)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_2)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_1)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_pre)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.98")
  expect_equal(gs_get_pop_paths(gs), snapshot_3)

# Repeat test for loaded GatingSet with same guid to make sure starting from blank slate
# after gs_add_gating_method_init()
  gs_add_gating_method_init()
  gs <- load_gs(list.files(dataDir, pattern = "gs_manual",full = TRUE))
  expect_equal(gs_get_pop_paths(gs), snapshot_pre)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs), snapshot_1)
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  expect_equal(gs_get_pop_paths(gs), snapshot_2)
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.95")
  expect_equal(gs_get_pop_paths(gs), snapshot_3)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_2)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_1)
  gs_remove_gating_method(gs)
  expect_equal(gs_get_pop_paths(gs), snapshot_pre)
  expect_error(gs_remove_gating_method(gs))
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris", dims = "CD4", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "test_gate", parent = "/not debris/singlets", dims = "CD8,CD4", pop = "+-", gating_method = "mindensity")
  gs_add_gating_method(gs, alias = "*", pop = "+/-", parent = "/not debris/singlets/CD3+/DNT", dims = "CCR7", 
          gating_method = "gate_quantile", gating_args = "probs = 0.98")
  expect_equal(gs_get_pop_paths(gs), snapshot_3)
})

context("polyfunction")

test_that("polyfunction testing", {
  skip("no longer supported")
  gs <- GatingSet(GvHD[1])
  gs <- transform(gs, estimateLogicle(gs[[1]], colnames(gs)[3:7]))
  gs_add_gating_method(gs, alias = "FL1+", parent = "root", dims = "FL1", gating_method = "tailgate")
  gs_add_gating_method(gs, alias = "FL3+", parent = "root", dims = "FL3", gating_method = "tailgate")
  gs_add_gating_method(gs, alias = "FL4+", parent = "root", dims = "FL4", gating_method = "tailgate")
  gs_add_gating_method(gs, alias = "*", parent = "root", dims = "*", gating_method = "polyfunctions", gating_args = "FL1+:FL3+:FL4+")
  expect_equal(gs_get_pop_paths(gs)[-(1:4)], c("/FL1+&FL3+&FL4+", "/FL1+&FL3+&!FL4+", "/FL1+&!FL3+&FL4+", "/FL1+&!FL3+&!FL4+", "/!FL1+&FL3+&FL4+", "/!FL1+&FL3+&!FL4+", "/!FL1+&!FL3+&FL4+", "/!FL1+&!FL3+&!FL4+"))
  node <- "/FL1+&!FL3+&!FL4+"
  expect_equal(gh_pop_get_count(gs[[1]], node), sum(gh_pop_get_indices(gs[[1]], "FL1+")&!gh_pop_get_indices(gs[[1]], "FL3+")&!gh_pop_get_indices(gs[[1]], "FL4+")))
  })

#' @templateVar old toggle.helperGates
#' @templateVar new gt_toggle_helpergates
#' @template template-depr_pkg
NULL
#' toggle/delete the hidden flag of the helper gates
#' 
#' The helper gates are defined as the referred gates in csv template. And all the chidlren of referred gates are also 
#' referred gates thus they are considered the helper gates and can usually be hidden to simply the final gating tree.
#' 
#' Note that delete action is NOT reversible.
#' 
#' @name gt_toggle_helpergates
#' @aliases toggle.helperGates
#' @param gt gatingTemplate object
#' @param gs GatingSet
#' @export
#' @importFrom flowWorkspace gh_pop_set_visibility gs_get_pop_paths gs_pop_get_children
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(gtFile)
#' #run the gating
#' gt_gating(gt, gs)
#' #hide the gates that are not of interest
#' gt_toggle_helpergates(gt, gs) 
#' #or simply remove them if you are sure they will not be useful in future
#' gt_delete_helpergates(gt, gs) 
#' }
gt_toggle_helpergates <- function(gt, gs){
  helperGates <- gt_get_helpergates(gt, gs)
  nonHiddenNodes <- gs_get_pop_paths(gs, showHidden = FALSE, path = "full")
  for(i in helperGates){
      if(i%in%nonHiddenNodes)
		  gs_pop_set_visibility(gs, i, FALSE)
      else
		  gs_pop_set_visibility(gs, i, TRUE)
  }
}

#' @export
toggle.helperGates <- function(gt, gs){
  .Deprecated("gt_toggle_helpergates")
  gt_toggle_helpergates(gt, gs)
}

#' @templateVar old get.helperGates
#' @templateVar new gt_get_helpergates
#' @template template-depr_pkg
NULL
#' @rdname gt_toggle_helpergates
#' @aliases get.helperGates
#' @export
gt_get_helpergates <- function(gt, gs){
  gated.nodes <- gs_get_pop_paths(gs, showHidden = TRUE, path = "full")
  referror <- names(gt_get_nodes(gt, only.names = T))[-1]
  referror <- intersect(gated.nodes, referror)#restrict to the gated nodes
  
  referree <- sapply(referror, function(node){
    gt_get_parent(gt, node, isRef = TRUE)
  }, USE.NAMES = FALSE)
  referree <- unique(unlist(referree))
  referree <- intersect(gated.nodes, referree)#restrict to the gated nodes
  gh <- gs[[1]]
  isHelper <- sapply(referree, function(node){
    children <- gs_pop_get_children(gh, node)
    if(length(children) == 0)
      return(TRUE)
    else{
      all(children %in% referree)
    }
    
  }, USE.NAMES = FALSE)
  if(length(isHelper) > 0)
  {
    referree[isHelper]  
  }else
  character(0)
  
}

#' @export
get.helperGates <- function(gt, gs){
  .Deprecated("gt_get_helpergates")
  gt_get_helpergates(gt, gs)
}

#' @templateVar old delete.helperGates
#' @templateVar new gt_delete_helpergates
#' @template template-depr_pkg
NULL
#' @rdname gt_toggle_helpergates
#' @aliases delete.helperGates
#' @export
gt_delete_helpergates <- function(gt, gs){
  helperGates <- gt_get_helpergates(gt, gs)
  
  for(i in helperGates){
      if(i%in%gs_get_pop_paths(gs))
        gs_pop_remove(i, gs = gs)
  }
}

#' @export
delete.helperGates <- function(gt, gs){
  .Deprecated("gt_delete_helpergates")
  gt_delete_helpergates(gt, gs)
}

context("gating...")

gatingResults <- readRDS("gatingResults.rds")
# gatingResults <- readRDS("tests/testthat/gatingResults.rds")

localPath <- "~/rglab/workspace/openCyto"

skip_if_not(dir.exists(file.path(localPath, "misc")))
library(parallel)
register_plugins(flowStats:::.tailgate, "tailgate")
test_that("tcell", {
      
      gt_tcell <- gatingTemplate(gtFile)
      
      gs <- load_gs(file.path(localPath,"misc/testSuite/gs-tcell"))
      
      expect_warning(gating(gt_tcell, gs, mc.core = 1, parallel_type = "multicore"), regexp = "HLA is partially matched")
      
      #test toggle helperGates
      
      expect_equal(length(gs_get_pop_paths(gs)), 29)
      helperGates <- gt_get_helpergates(gt_tcell, gs)
      expect_true(setequal(helperGates, c('/nonDebris/singlets/lymph/cd3/cd4+',
                                          '/nonDebris/singlets/lymph/cd3/cd8+',
                                          '/nonDebris/singlets/lymph/cd3/cd4+cd8-/CD45_neg',
                                          '/nonDebris/singlets/lymph/cd3/cd4+cd8-/CD45_neg/CCR7_gate',
                                          '/nonDebris/singlets/lymph/cd3/cd4+cd8-/CD38+',
                                          '/nonDebris/singlets/lymph/cd3/cd4+cd8-/HLA+',
                                          '/nonDebris/singlets/lymph/cd3/cd4-cd8+/CCR7+',
                                          '/nonDebris/singlets/lymph/cd3/cd4-cd8+/CD45RA+',
                                          '/nonDebris/singlets/lymph/cd3/cd4-cd8+/CD38+',
                                          '/nonDebris/singlets/lymph/cd3/cd4-cd8+/HLA+')))
      #hide
      gt_toggle_helpergates(gt_tcell, gs)
      expect_equal(length(gs_get_pop_paths(gs)), 19)
      expect_equal(length(gs_get_pop_paths(gs, showHidden = TRUE)), 29)
      #recover
      gt_toggle_helpergates(gt_tcell, gs)
      expect_equal(length(gs_get_pop_paths(gs)), 29)
    
      #rm helper gates
      gs1 <- gs_copy_tree_only(gs)
      gt_delete_helpergates(gt_tcell, gs1)
      expect_equal(length(gs_get_pop_paths(gs1, showHidden = TRUE)), 19)
      
      
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expectRes <- gatingResults[["gating_tcell"]]
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      expect_equivalent(gs_pop_get_stats(gs[2], path = "full", type = "percent")[,percent], expectRes[thisRes[,pop],2], tol = 0.04)
      #test the interactive gating API
      nodes <- gs_pop_get_children(gs[[1]], "cd4-cd8+")
      for(node in nodes)
        gs_pop_remove(gs, node)
      opt <- getOption("openCyto")
      opt[["check.pop"]] <- FALSE
      options(openCyto = opt)
      expect_warning(add_pop(gs, gating_method = "tailgate", dims = "CD38,HLA"
              , parent = "cd4-cd8+", pop = "CD38+HLA+"
              , alias = "activated cd8", preprocessing_method = "standardize_flowset")
        , regexp = "HLA is partially matched")
      opt[["check.pop"]] <- TRUE
      options(openCyto = opt)
      #test new .mindensity2 wrapper
      gs_add_gating_method(gs, gating_method = "gate_mindensity2", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "+/-+/-")
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      
      expect_true(all(c("cd4-cd8+/CD38+", "cd4-cd8+/HLA+") %in% gs_get_pop_paths(gs, path = "auto")))
      
     
      #use channel in pop and stains in dims
      for(node in nodes[7:9])
        gs_pop_remove(gs, node)
      expect_warning(add_pop(gs, gating_method = "tailgate"
                      , dims = "CD38,HLA"
                      , parent = "cd4-cd8+"
                      , pop = "-+"
                      , alias = "activated cd8"
                      , preprocessing_method = "standardize_flowset"
                        )      
      , regexp = "HLA is partially matched")
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      
      
      #pure +-
      for(node in nodes[7:9])
        gs_pop_remove(gs, node)
      expect_warning(
        add_pop(gs, gating_method = "tailgate"
              , dims = "CD38,HLA"
              , parent = "cd4-cd8+"
              , pop = "-+"
              , alias = "activated cd8"
              , preprocessing_method = "standardize_flowset"
            )      
      , regexp = "HLA is partially matched")
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      
      
      #test keep.helperGates = FALSE
      nodes <- gs_pop_get_children(gs[[1]], "cd4-cd8+")
      for(node in nodes)
        gs_pop_remove(gs, node)
      expect_warning(
        add_pop(gs, gating_method = "tailgate", dims = "CD38,HLA"
                , parent = "cd4-cd8+", pop = "++"
                , alias = "activated cd8", keep.helperGates = FALSE)
        , regexp = "HLA is partially matched")
      expect_false(any(c("cd4-cd8+/CD38+", "cd4-cd8+/HLA+") %in% gs_get_pop_paths(gs, path = "auto")))
      expect_equal(gs_pop_get_children(gs[[1]], "cd4-cd8+", path = "auto"), "activated cd8")
      
    })

test_that("tcell--asinhtGml2", {
  
  fcs <- list.files(pattern = "CytoTrol*", system.file("extdata", package = "flowWorkspaceData"), full.names = TRUE)
  fs <- load_cytoset_from_fcs(fcs)  
  gs <- GatingSet(fs)
  #compensate
  comp <- compensation(spillover(fs[[1]])[["SPILL"]])
  gs <- compensate(gs, comp)
  #transform
  trans <- asinhtGml2_trans()
  chnls <- as.vector(parameters(comp))
  trans.list <- transformerList(chnls, trans)
  gs <- transform(gs, trans.list)
  
  # autoplot(gs_pop_get_data(gs), chnls[1])
  
  #load gating template
  localPath <- "~/rglab/workspace/openCyto"
  gtFile <- system.file("extdata/gating_template/tcell.csv", package = "openCyto")
  #modify scale-specific parameters
  dt <- fread(gtFile)
  dt[c(5, 8), gating_args := NA]
  tmp <- tempfile()
  write.csv(dt, file = tmp, row.names = FALSE)
  gt_tcell <- gatingTemplate(tmp)
  library(parallel)
  expect_warning(gating(gt_tcell, gs, mc.core = 1, parallel_type = "multicore"), regexp = "HLA is partially matched")
  # autoplot(gs[[1]])  
  
  thisRes <- gs_pop_get_count_fast(gs, path = "full")
  expectRes <- gatingResults[["gating_tcell_asinhtGml2"]]
  expect_equal(thisRes, expectRes, tol = 0.018)

})

test_that("ICS", {
      
      
      gtfile <- system.file("extdata/gating_template/ICS.csv", package = "openCyto")
      gt <- gatingTemplate(gtfile)
      
      
      gs <- load_gs(file.path(localPath,"misc/testSuite/gs-ICS"))[1]
	  gs_pop_remove(gs, "s")
	  #TODO:investigate segfault associated with multicore and L#125 asinhtGml2_trans
      
      expect_warning(gating(gt, gs
                            # , mc.core = 2
                            # , parallel_type = "multicore"
                            ), regexp = "Pacific Blue-A is partially matched")
      
      expectRes <- gatingResults[["gating_ICS"]]
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      
      #test add_pop with polyFunctions
      
      #test add_pop with boolean method
      pop <- "IL2orIFNg"
      gs_pop_remove(gs, pop)
      gs_add_gating_method(gs, alias = pop, gating_method = "boolGate", parent = "cd4", gating_args = "cd4/IL2|cd4/IFNg")
      
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.04)
      
    })

test_that("treg", {
      
      skip("TO investigate the failure")
      gtfile <- system.file("extdata/gating_template/treg.csv", package = "openCyto")
      gt <- gatingTemplate(gtfile)
      
      gs <- load_gs(file.path(localPath,"misc/testSuite/gs-treg"))
      gs_pop_remove(gs, "boundary")
      #TODO: fix mc.core=3 error
      expect_warning(gating(gt, gs, mc.core = 1, parallel_type = "multicore"), "did not converge")
      
      expectRes <- gatingResults[["gating_treg"]]
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.025)
      
    })

test_that("bcell", {
      
      
      gtfile <- system.file("extdata/gating_template/bcell.csv", package = "openCyto")
      gt <- gatingTemplate(gtfile)
      
      gs <- load_gs(path = file.path(localPath,"misc/testSuite/gs-bcell"))
      gs_pop_remove(gs, "boundary")
      expect_warning(gating(gt, gs, mc.core = 1, parallel_type = "multicore"), "did not converge")
      
      expectRes <- gatingResults[["gating_bcell"]]
      thisRes <- gs_pop_get_stats(gs[1], path = "full", type = "percent")
      expect_equivalent(thisRes[,percent], expectRes[thisRes[,pop],1], tol = 0.08)
      
    })

#' plot a \code{fcFilterList}
#' 
#' It is usually called by \code{plot} method for \code{fcTree} instead of directly by users.
#' 
#' @param x \code{fcFilterList}
#' @param y \code{character} channel name
#' @param samples \code{character} a vector of sample names to be plotted
#' @param posteriors \code{logical} indicating whether posteriors should be plotted
#' @param xlim,ylim scale settings for x,y axises
#' @param lwd line width
#' @param breaks passed to \link{hist}
#' @param data \code{GatingSet} object
#' @param node \code{character} population name associated with the \code{fcFilterList}
#' @param ... other arguments passed to base \code{plot}
#' @examples
#' \dontrun{ 
#'  env1<-new.env(parent=emptyenv())
#'  #gt is a gatingTemplate, gs is a GatingSet
#'  gt_gating(gt,gs,env1) #the flowClust gating results are stored in env1 
#'  plot(env1$fct,"nonDebris",post=T) #plot the priors as well as posteriors for the "nonDebris" gate
#' }
setMethod("plot", sig = c("fcFilterList", "ANY"),
          definition = function(x, y, samples = NULL, posteriors = FALSE
        , xlim = NULL, ylim = NULL, node = NULL, data = NULL, breaks = 20, lwd = 1, ...) {
#            browser()
  prior1 <- priors(x[[1]], y)
  
  if (is.null(y)) {
    priorNames <- names(prior1)
    if (length(prior1) == 1) {
      y <- priorNames
    } else {
      stop("Need to specify which channel:", paste(priorNames, collapse = "or"))
    }
  }

  # refetch prior according to modified y
  prior1 <- priors(x[[1]], y)  

  # try to set x,y lim from
  minX <- min(unlist(lapply(x, function(curFilter) posteriors(x = curFilter, y = y)$min)))
  maxX <- max(unlist(lapply(x, function(curFilter) posteriors(x = curFilter, y = y)$max)))
  x_dens <- seq(minX, maxX, length = 1000)

  # TODO: right now the priors from the first sample is used to determine y scale
  K <- nrow(prior1$Mu0)
  prior_density <- lapply(seq_len(K), function(k) dnorm(x_dens, mean = prior1$Mu0[k], 
    sd = sqrt(prior1$Omega0[k])))
  minY <- min(unlist(lapply(prior_density, min)))
  maxY <- max(unlist(lapply(prior_density, max)))
  if(is.null(xlim)){
    xlim <- c(minX, maxX)
  }
  if(is.null(ylim)){
    ylim <- c(minY, maxY)
  }
  plot(x = NULL, type = "n", xlim = xlim, ylim = ylim, xlab = y, 
    ylab = "", ...)
  
  # plot post
  if (posteriors) {
    if (is.null(samples)) 
      samples <- names(x)
    for (samp in samples) {
      curFilter <- x[[samp]]
#      browser()
      if(!is.null(data)){
        fr <- gh_pop_get_data(data[[samp]],node)
        thisChnl <- priorNames
        hist(exprs(fr)[,thisChnl],add=T,prob=T,breaks=breaks)  
      }
      curPost <- posteriors(x = curFilter, y = y)
      
      curPrior <- priors(x = curFilter, y = y)
      K <- nrow(curPrior$Mu0)
      prior_density <- lapply(seq_len(K), function(k) dnorm(x_dens, mean = curPrior$Mu0[k], 
        sd = sqrt(curPrior$Omega0[k])))
#      browser()
      for (k in seq_len(K)) {
        lines(x_dens, prior_density[[k]], col = rainbow(K)[k], lty = 2, lwd = lwd)
        
        posterior_density <- curPost$w[k]*flowClust::dmvt(x_dens, mu = curPost$mu[k, ], sigma = curPost$sigma[k, 
          , ], nu = curPost$nu, lambda = curPost$lambda)$value
        lines(x_dens, posterior_density, col = rainbow(K)[k], lwd = lwd)

        # plot gate
        g <- c(curFilter@min[y], curFilter@max[y])
        ind <- which(!is.infinite(g))
        abline(v = g[ind], col = "grey")
      }
    }
  }
  
  
  
})

#' @templateVar old getNodes
#' @templateVar new gt_get_nodes
#' @template template-depr_pkg
NULL
#' get nodes from \link[openCyto:gatingTemplate-class]{gatingTemplate} object
#' 
#' @name gt_get_nodes
#' @aliases getNodes getNodes,gatingTemplate-method
#' @param x \code{gatingTemplate}
#' @param y \code{character} node index. When \code{missing}, return all the nodes
#' @param order \code{character} specifying the order of nodes. options are "default", "bfs", "dfs", "tsort"
#' @param only.names \code{logical} specifiying whether user wants to get the entire \code{gtPopulation} object or just the name of the population node
#'   
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' gt_get_nodes(gt)[1:2]
#' gt_get_nodes(gt, only.names = TRUE)
#' gt_get_nodes(gt, "/nonDebris")
#' }
#' @export
gt_get_nodes <- function(x, y
                             , order = c("default", "bfs", "dfs", "tsort")
                             , only.names = FALSE) {
  
  if (missing(y)){
    res <- nodeData(x, attr = "pop")
    order <- match.arg(order)
    if(order != "default"){
      nodeIds <- eval(substitute(f1(x),list(f1=as.symbol(order))))
      if(order == "dfs")
        nodeIds <- nodeIds$discovered
      res <- res[nodeIds]
    }
  }else
  {
    res <- nodeData(x, y, "pop")
  }
  if(only.names){
    res <- sapply(res,alias)
  }
  if(length(res) == 1 && class(res) == "list")
    res <- res[[1]]
  res
}

#' @export
setMethod("getNodes", signature = c("gatingTemplate"),
          definition = function(x, y
                                  , order = c("default", "bfs", "dfs", "tsort")
                                  , only.names = FALSE) {
            .Deprecated("gt_get_nodes")
            gt_get_nodes(x, y, order, only.names)
})

#' @templateVar old getChildren
#' @templateVar new gt_get_children
#' @template template-depr_pkg
NULL
#' get children nodes
#'
#' @name gt_get_children
#' @aliases getChildren getChildren,gatingTemplate,character-method
#' @param obj \code{gatingTemplate}
#' @param y \code{character} parent node path  
#' @export 
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' 
#' gt_get_nodes(gt, "/nonDebris")
#' gt_get_children(gt, "/nonDebris") 
#' }
#' @importClassesFrom methods character ANY data.frame environment list logical matrix missing numeric oldClass
gt_get_children <- function(obj, y){
  edges(obj, y)[[1]]
}

#' @export
setMethod("getChildren", signature = c("gatingTemplate", "character"),
          definition = function(obj, y) {
            .Deprecated("gt_get_children")
            gt_get_children(obj, y)
})

#' @templateVar old getParent
#' @templateVar new gt_get_parent
#' @template template-depr_pkg
NULL
#' get parent nodes
#' 
#' @param obj \code{gatingTemplate}
#' @param y \code{character} child node path
#' @param isRef \code{logical} whether show the reference node besides the parent node
#' 
#' @name gt_get_parent
#' @aliases getParent getParent,gatingTemplate,character-method
#' @importFrom graph inEdges
#' @aliases getParent,gatingTemplate,character-method
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' 
#' gt_get_nodes(gt, "/nonDebris")
#' gt_get_parent(gt, "/nonDebris/singlets") 
#' }
#' @export 
gt_get_parent <- function(obj, y, isRef = FALSE) {
  #            browser()
  src <- inEdges(y, obj)[[1]]
  isRefs <- unlist(lapply(src,function(thisSrc){
    thisEdge <- edgeData(obj,thisSrc,y)
    thisEdge[[1]]$isReference 
  }), recursive = F)
  #  browser()
  if(isRef)
    src[isRefs]
  else
    src[!isRefs]
}

#' @export
setMethod("getParent", signature = c("gatingTemplate", "character"),
          definition = function(obj, y, isRef = FALSE) {
            .Deprecated("gt_get_parent")
            gt_get_parent(obj, y, isRef)
})

#' @templateVar old getGate
#' @templateVar new gt_get_gate
#' @template template-depr_pkg
NULL
#' get gating method from the node
#' 
#' @name gt_get_gate
#' @aliases getGate,gatingTemplate,character-method
#' @param obj \code{gatingTemplate}
#' @param y \code{character} parent node path
#' @param z \code{character} child node path
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' gt_get_nodes(gt, only.names = TRUE)
#' gt_get_nodes(gt, "/nonDebris")
#' gt_get_children(gt, "/nonDebris")
#' gt_get_gate(gt, "/nonDebris", "/nonDebris/singlets")
#' }
#' @export
gt_get_gate <- function(obj, y, z) {
  edgeData(obj, from = y, to = z, attr = "gtMethod")[[1]]
}

#' @export
setMethod("getGate", signature = c("gatingTemplate", "character"),
          definition = function(obj, y, z) {
            .Deprecated("gt_get_gate")
            gt_get_gate(obj, y, z)
})

setGeneric("ppMethod", function(obj, y, ...) standardGeneric("ppMethod"))
#' get preprocessing method from the node
#' 
#' get preprocessing method from the node
#' @param obj \code{gatingTemplate}
#' @param y \code{character} parent node path
#' @param z \code{character} child node path
#' @aliases ppMethod,gatingTemplate,character-method
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' ppMethod(gt,  "/nonDebris/singlets",  "/nonDebris/singlets/lymph")
#' }
setMethod("ppMethod", signature = c("gatingTemplate", "character"),
    definition = function(obj, y, z) {
      edgeData(obj, from = y, to = z, attr = "ppMethod")[[1]]
    })

#' show method for gatingTemplate
#' 
#' show method for gatingTemplate
#' 
#' @param object \code{gatingTemplate}
#' @aliases 
#' show,gatingTemplate-method
#' @export 
#' @importFrom methods show
setMethod("show", signature = c("gatingTemplate"),
          definition = function(object) {
  cat("--- Gating Template: ")
  cat(object@name)
  cat("\n")
  cat("\twith ", length(object@nodes), " populations defined\n")
})

#' plot the gating scheme
#' 
#' plot the gating scheme using Rgraphviz
#' 
#' @param x \code{gatingTemplate} object
#' @param y either \code{character} specifying the root node which can be used to visualize only the subgraph 
#'              or \code{missing} which display the entire gating scheme
#' @param ... other arguments 
#'  
#'      graphAttr, nodeAttr:  graph rendering attributes passed to \link[Rgraphviz:renderGraph]{renderGraph}
#'      showRef \code{logical}: whether to display the reference gates. Sometime it maybe helpful to 
#'                              hide all those reference gates which are not the cell population of interest 
#'                              and used primarily for generating other population nodes.
#' @export 
#' @importFrom RColorBrewer brewer.pal
#' @aliases 
#' plot,gatingTemplate,missing-method
#' plot,gatingTemplate,character-method
#' plot,gatingTemplate-method
#' plot,gatingTemplate,ANY-method
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' plot(gt) #plot entire tree
#' plot(gt, "lymph") #only plot the subtree rooted from "lymph"
#' }
setMethod("plot",c("gatingTemplate","missing"),function(x,y,...){
      .plotTree(x,...)
      
    })

#FIXME:somehow the edge attributes are not correctly assigned to each edge respectively
.plotTree<-function(x
                      , graphAttr = list(rankdir = "LR", page = c(8.5, 11)) 
                      , nodeAttr = list(fixedsize = FALSE, shape = "ellipse")
                      , showRef = TRUE
                      , ...){
#          browser()          
  # get gating method name attr from edges
  hasEdge <- length(edgeData(x)) > 0
  if(hasEdge){
    gm.names <- unlist(lapply(edgeData(x, attr = "gtMethod"), names))
    gm.types <- unique(gm.names)
    
    ref.edges <- unlist(edgeData(x, attr = "isReference"))
    ref.edges <- ref.edges[ref.edges]
    
    
    # fix the name attr
    e.colnames <- gsub("\\|", "~", names(gm.names))
    ref.colnames <- gsub("\\|", "~", names(ref.edges))
    
    #only care about the ref edges that are not conflicting with gm edges
    ref.colnames <- ref.colnames[!ref.colnames%in%e.colnames]
    
    
    
    # encode the method name with color
    nMethods <- length(gm.types)
    gm.col <- brewer.pal(nMethods, name = "Dark2")
    names(gm.col) <- gm.types
    eCol.gm <- gm.col[gm.names]
    
    # restore names
    names(eCol.gm) <- e.colnames  
    
    eCols <- eCol.gm
    
    #add ref edges
    eCol.ref <- sapply(ref.colnames,function(i)return("grey"))
    if(length(eCol.ref)>0)
      eCols <- c(eCols, eCol.ref)
    
    # encode edge style
    gm.isPoly <- unlist(lapply(gm.names, function(y) {
              ifelse(y == "polyFunctions", "poly", "regular")
            }))
    edge.styles <- c("solid", "dashed")
    names(edge.styles) <- unique(gm.isPoly)
    eStyles.gm <- edge.styles[gm.isPoly]
    names(eStyles.gm) <- e.colnames
    
    eStyles <- eStyles.gm
    
    eStyles.ref <- sapply(ref.colnames,function(i)ifelse(showRef,"dotted","blank"))
    if(length(eStyles.ref)>0)
      eStyles <- c(eStyles,eStyles.ref)
    
    eAttrs <- list(color = eCols, lty = eStyles)
    
  }else{
    eAttrs <- list()
  }
    
  # encode the node shape and color with isSubsets
  nodeTypes <- unlist(lapply(nodeData(x, attr = "pop"), function(y) {
            ifelse(class(y) == "gtSubsets", "subset", "pop")
          }))
  n.colnames <- names(nodeTypes)
  nodeType <- unique(nodeTypes)
  
  # color
  subset.col <- brewer.pal(9, name = "Set3")[c(9, 7)]
  names(subset.col) <- nodeType
  nFillCol <- subset.col[nodeTypes]
  names(nFillCol) <- n.colnames
  
  # line type (somehow it doesn't work here)
  LineType <- c("solid", "dotted")
  names(LineType) <- nodeType
  nLtys <- LineType[nodeTypes]
  names(nLtys) <- n.colnames
  
  nLabels <- unlist(lapply(nodeData(x, attr = "pop"), function(thisPop)paste(alias(thisPop), collapse = ",")))
  nAttrs <- list(label = nLabels, fillcolor = nFillCol, lty = nLtys)
  
  plot(as(x,"graphNEL"), nodeAttrs = nAttrs, edgeAttrs = eAttrs
                  ,attrs = list(graph = graphAttr,node = nodeAttr)
    )
  
  if(hasEdge){
    plot.space = par()[["usr"]]
    x1 = plot.space[1]
    x2 = plot.space[2]
    y1 = plot.space[3]
    y2 = plot.space[4]
    
    legend.lty <- ifelse(gm.types == "polyFunctions", "poly", "regular")
    legend(x1 + 100, y2 - 100, legend = gm.types, title = "Gating Methods",
        col = gm.col, lty = edge.styles[legend.lty], cex = 0.8)
  }
}

setMethod("plot", signature = c("gatingTemplate"),
          definition = function(x, y = missing,...) {
  
})

setMethod("plot",c("gatingTemplate","character"),function(x,y,...){
      
      #convert alias to nodeID
      allNodes <- gt_get_nodes(x)
      allAlias <- unlist(lapply(allNodes, alias), recursive = F)
      nodeInd <- match(y, allAlias)
      thisNode <- allNodes[nodeInd]
      thisId <- names(thisNode)
#      browser()
      if(length(y)==1){#use it as the root
        
        nodelist <- new.env(parent=emptyenv())
        nodelist$v <- character()
        flowWorkspace:::.getAllDescendants (x, thisId, nodelist)  
        
        
#        browser()
        nodelist$v <- c(nodelist$v,thisId)
        nodelist$v <- unique(nodelist$v)
        #assume the number y is consistent with  R graph node name: N_x
        subNode_Ind <- nodelist$v
        
      }else{
        #when y is a vector, use it to subset the graph
        subNode_Ind <- thisId
      }
      
#      subNodes <- paste("N",subNode_Ind-1,sep="_")
      if(length(subNode_Ind)<=1){
        stop("Rgraphviz doesn't know how to plot leaf node!")
      }
      x <- subGraph(subNode_Ind, x)
      .plotTree(x,...)
      
    })

#' get nodes from \code{fcTree}
#'
#' @param x \code{fcTree}
#' @param y \code{character} node name
#' @aliases getNodes,fcTree-method
setMethod("getNodes", signature = c("fcTree"), definition = function(x, y) {
  if (missing(y)) {
    nodeData(x)
  } else {
    nodeData(x, y)
  }
})
#' get gates saved in \code{fcTree}
#'
#' @param obj \code{fcTree}
#' @param y \code{character} node name
#' @param ... other arguments (not used)
#' @aliases getGate,fcTree,character-method
setMethod("getGate", sig = c("fcTree", "character"),
    definition = function(obj, y,  ...) {
      # get filterList
      matchedNode <-.getNode(obj,y,...)
      flist <- matchedNode[["fList"]]
      flist
    })

.getNode<-function(x,y,..){
  nodes <- getNodes(x)
  
  allAlias <- lapply(nodes, function(curNode) alias(curNode$pop))
  ind <- which(allAlias %in% y)
  if (length(ind) > 1) {
    stop("Population '", y, "' is ambiguous!")
  } else if (length(ind) == 0) {
    stop("Population '", y, "' is not found!")
  } else {
    matchedNode <- nodes[[ind]]
  }
  matchedNode
}

#' plot the flowClust gating results
#' 
#' This provides the priors and posteriors as well as the gates for the purpose of debugging flowClust gating algorithm
#' 
#' @param x \code{fcTree}
#' @param y \code{character} node name in the \code{fcTree}
#' @param channel \code{character} specifying the channel.
#' @param data \code{GatingSet} that the \code{fcTree} is associated with
#' @param ... other arguments
setMethod("plot", sig = c("fcTree", "character"),
          definition = function(x, y, channel = NULL, data = NULL,...) {
  # get filterList
#      browser()
    flist <- getGate(x,y)
    matchedNode <-.getNode(x,y)
    if(!is.null(data))
      parentNode <- gs_pop_get_parent(data[[1]],y)
  plot(x = flist, y = channel, main = matchedNode$pop@name, node =parentNode, data = data, ...)
})

#setMethod("plot", sig = c("fcTree", "numeric"),
#          definition = function(x, y, channel = NULL, ...) {
#  y <- as.character(y)
#  nodes <- getNodes(x)
#  matchedNode <- nodes[[y]]
#  fcObj <- matchedNode$fcObj
#  
#  plot(x = fcObj, y = channel, main = matchedNode$pop@name, ...)
#}) 

#' OpenCyto plugin for custom gating functions
#'
#' \code{.gate_custom()} is an openCyto plugin that accepts any custom gating
#' function through the \code{FUN} argument to allow the use of any custom
#' gating function without the need to write additional openCyto plugins.
#'
#' @param fr a \code{flowFrame} or \code{cytoframe} object containing the data
#'   to be gated by \code{FUN}.
#' @param pp_res the output of the \code{.pp_gate_custom()} pre-processing
#'   method.
#' @param channels vector of channels names to indicate the parameters to use
#'   for gating. \code{channels} are parsed by OpenCyto using the \code{dims}
#'   specified in the \code{gatingTemplate} and therefore \code{channels} can
#'   only include one or two parameters. The \code{params} argument is reserved
#'   for cases where more than two channels are to be passed to the gating
#'   function.
#' @param FUN a gating function or name of a gating function supplied as a
#'   character string. \code{FUN} is applied to the data following
#'   \code{inverse} transformations and/or scaling along with any additional
#'   arguments specified through \code{...}.
#' @param input indicates how the data extracted from the \code{cytoframe}
#'   should be formatted prior to passing it to the gating function supplied to
#'   \code{FUN}, set to \code{"cytoframe"} by default. Options include
#'   \code{"matrix"}, \code{"data.table"} or \code{"cytoframe"}.
#' @param inverse logical indicating whether inverse data transformations should
#'   be applied to the extracted data prior to passing it to the gating
#'   function, set to FALSE by default.
#' @param scale can be either a logical or the name of a scaling function to
#'   apply to each channel of the \code{cytoframe} prior to applying the gating
#'   function supplied through \code{FUN}, set to FALSE by default. Min/max
#'   scaling can be applied by setting \code{scale = "range'}, which is the
#'   default behavior when \code{scale = TRUE}.
#' @param slot provides a mechanism by which gate objects can be extracted from
#'   the outputs of gating functions that return complex data structures (e.g.
#'   list objects with multiple elements). \code{slot} can either be the name of
#'   a slot or a the name of a function extracts or formats the output of the
#'   gating function to return a valid gate object.
#' @param ... additional arguments passed to \code{FUN} for gating.
#'
#' @return a gate object as returned by \code{FUN}.
#' 
#' @importFrom flowCore exprs exprs<-
#' @importFrom flowWorkspace realize_view cf_unlock cf_lock transform
#'
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#'
#' @noRd
.gate_custom <- function(fr,
                         pp_res,
                         channels,
                         FUN,
                         input = "cytoframe",
                         inverse = FALSE,
                         scale = FALSE,
                         slot  = NULL,
                         ...) {
  
  # NOTE: openCyto currently supports 1D or 2D gates only!
  
  # extract transformers from pp_res
  trans <- pp_res$trans
  
  # data requires scaling or inverse transformations
  if(!scale %in% FALSE | (inverse & any(channels %in% names(trans)))) {
    # cytoframe must be copied and unlocked for editing
    if(inherits(fr, "cytoframe")) {
      fr <- realize_view(fr)
      cf_unlock(fr)
    }
    # inverse transformations
    if(inverse & any(channels %in% names(trans))) {
      fr <- transform(
        fr,
        lapply(trans, `[[`, "inverse")
      )
    }
    # scaling
    if(!scale %in% FALSE) {
      # default range scale
      if(scale %in% TRUE) {
        scale <- "range"
      }
      # range (min/max) scaling
      if(scale %in% "range") {
        scale <- function(x) {
          (x - min(x))/diff(range(x))
        }
      # scaling function 
      } else {
        scale <- .fun_match(
          scale
        )
      }
      # apply scaling function
      exprs(fr)[, channels] <- apply(
        exprs(fr)[, channels],
        2,
        scale
      )
    }
    # lock copied cytoframe for editing
    if(inherits(fr, "cytoframe")) {
      cf_lock(fr)
    }
  }
  
  # input - data extraction required
  if(!input %in% "cytoframe") {
    # extract matrix
    fr <- exprs(fr[, channels])
    # coerce to vector | data.table | data.frame | matrix
    fr <- do.call(
      "as",
      list(
        fr,
        input
      )
    )
  }
  
  # extract gating arguments
  gating_args <- list(
    fr,
    ...
  )
  
  # get gating function
  FUN <- .fun_match(
    FUN
  )
  
  # apply gating function
  gate <- do.call(
    FUN,
    gating_args
  )
  
  # extract gate object
  if(!is.null(slot)) {
    # slot name
    if(slot %in% names(gate)) {
      gate <- gate[[slot]]
    # slot function
    } else {
      gate <- do.call(
        .fun_match(slot),
        list(gate)
      )
    }
  }
  
  # return gate object
  return(gate)
  
}

#' OpenCyto pre-processing plugin to transfer transformers to gating function
#'
#' @param fs a cytoset containing samples for the current sample group.
#' @param gs a GatingSet.
#' @param gm a gating method.
#' @param channels the names of the channels to use or gating.
#' @param groupBy grouing variables separated by colon to be used to separate
#'   samples into groups.
#' @param isCollapse logical indicating whether the data is collapsed prior to
#'   gating.
#' @param trans an optional \code{transformerList} containing the transformers
#'   applied to the GatingSet, automatically extracted from the\code{GatingSet}
#'   directly.
#' @param ... not in use.
#'
#' @return returns a list of pre-processing arguments including \code{trans}
#'   which contains the transformers applied to the \code{GatiingSet}.
#'
#' @importFrom flowWorkspace gh_get_transformations transformerList
#'
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#'
#' @noRd
.pp_gate_custom <- function(fs,
                            gs,
                            gm,
                            channels,
                            groupBy = NA,
                            isCollapse = NA,
                            trans = list(),
                            ...) {
  
  # extract transformers
  if(!inherits(trans, "transformerList")) {
    trans <- gh_get_transformations(
      gs[[1]],
      only.function = FALSE
    )
    if(length(trans) > 0) {
      trans <- transformerList(
        names(trans),
        trans
      )
    }
  }
  
  # return required arguments
  return(
    list(
      "trans" = trans
    )
  )
  
}

#' Internal function to source a namespaced function provided by name
#'
#' @param FUN character vector specifying the name of the function to source,
#'   may be prefixed with a namespace using either \code{pkg::FUN} for exported
#'   or \code{pkg:::FUN} for internal functions respectively.
#' @param ... additional arguments passed to \code{match.fun()}.
#'
#' @return the sourced function.
#' 
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#' 
#' @examples
#' FUN <- .fun_match(
#'   "stats::median"
#' )
#' 
#' @noRd
.fun_match <- function(FUN,
                       ...) {
  
  # name of function -> function
  if(is.character(FUN)) {
    # namespaced function from another package
    if(grepl(":{2,3}", FUN)) {
      # split into namespace and function
      FUN <- unlist(
        strsplit(
          FUN, 
          ":{2,3}"
        )
      )
      # source function from namespace
      FUN <- get(
        FUN[2],
        envir = asNamespace(FUN[1]),
        mode = "function"
      )
    }
    # get function
    FUN <- match.fun(
      FUN, 
      descend = TRUE
    )
  }
  
  # return a function to call
  return(FUN)
  
}

#' Creates a singlet polygon gate using the prediction bands from a robust linear model
#'
#' We construct a singlet gate by applying a robust linear model. By default, we model the forward-scatter height
#' (FSC-H)as a function of forward-scatter area (FSC-A). If \code{sidescatter}
#' is given, forward-scatter height is as a function of \code{area} +
#' \code{sidescatter} + \code{sidescatter / area}.
#'
#' Because \code{rlm} relies on iteratively reweighted least
#' squares (IRLS), the runtime to construct a singlet gate is dependent in part
#' on the number of observations in \code{x}. To improve the runtime, we provide
#' an option to subsample randomly a subset of \code{x}. A percentage of
#' observations to subsample can be given in \code{subsample_pct}. By default, no
#' subsampling is applied.
#' 
#' @param x a \code{\link[flowCore:flowFrame-class]{flowFrame}} object
#' @param area character giving the channel name that records the signal
#' intensity as peak area
#' @param height character giving the channel name that records the signal
#' intensity as peak heightchannel name of height
#' @param sidescatter character giving an optional channel name for the
#' sidescatter signal. By default, ignored.
#' @param prediction_level a numeric value between 0 and 1 specifying the level
#' to use for the prediction bands
#' @param subsample_pct a numeric value between 0 and 1 indicating the percentage
#' of observations that should be randomly selected from \code{x} to construct
#' the gate. By default, no subsampling is performed.
#' @param wider_gate logical value. If \code{TRUE}, the prediction bands used to
#' construct the singlet gate use the robust fitted weights, which increase
#' prediction uncertainty, especially for large FSC-A. This leads to wider gates,
#' which are sometimes desired.
#' @param filterId the name for the filter that is returned
#' @param maxit the limit on the number of IWLS iterations 
#' @param ... additional arguments (not used)
#' @return a \code{\link[flowCore]{polygonGate}} object with the singlet gate
#' @rdname gate_singlet
gate_singlet <- function(x, area = "FSC-A", height = "FSC-H", sidescatter = NULL,
                        prediction_level = 0.99, subsample_pct = NULL,
                        wider_gate = FALSE, filterId = "singlet", maxit = 5, ...) {
  flowCore:::checkClass(x, "flowFrame")
  flowCore:::checkClass(area, "character")
  flowCore:::checkClass(height, "character")
  if (!is.null(sidescatter)) {
    flowCore:::checkClass(sidescatter, "character")
  }
  if (length(area) + length(height) != 2) {
    stop("Each of 'area' and 'height' must be 'character' vectors of length 1.")
  }

  x <- exprs(x[, c(area, height, sidescatter)])
  channel_names <- c(area, height)
  area <- make.names(area)
  height <- make.names(height)
  colnames(x)[1:2]  <- c(area, height)
  if (!is.null(sidescatter)) 
  {
    sidescatter <- make.names(sidescatter)
    colnames(x)[3] <- sidescatter
  }
    
  # If specified, subsample from 'x'
  if (!is.null(subsample_pct)) {
    subsample_pct <- as.numeric(subsample_pct)
    if (subsample_pct <= 0 || subsample_pct > 1) {
      warning("The subsampling percentage must be between 0 and 1. ",
              "Setting percentage to default value...", call. = FALSE)
      subsample_pct <- 1
    }
    n <- nrow(x)
    x <- x[sample(x = seq_len(nrow(x)), size = subsample_pct * n), ]
  }
  # Creates polygon gate based on the prediction bands at the minimum and maximum
  # 'area' observation using the trained robust linear model.
  which_min <- which.min(x[,area])
  which_max <- which.max(x[,area])
  x_extrema <- x[c(which_min, which_max), ]

  y <- x[, height]
  x <- x[, c(area, sidescatter), drop = FALSE]
  x <- cbind(1, x)#add default weight column
  rlm_formula <- paste(make.names(height), make.names(area), sep = " ~ ")
  
  if (!is.null(sidescatter)) {
    sidescatter <- make.names(sidescatter)
    x <- cbind(x, x[, sidescatter] / x[, area])
    ssc_ratio <- paste0("I(", sidescatter, " / ", area, ")")
    colnames(x)[4] <- ssc_ratio
    
    rlm_formula <- paste(rlm_formula, sidescatter, ssc_ratio, sep = " + ")
  }
  rlm_formula <- as.formula(rlm_formula)
  
  rlm_fit <- withCallingHandlers(fast_rlm(x, y, maxit = maxit, ...),
                                 warning = function(w){
                                   if(grepl("failed to converge", conditionMessage(w))){
                                     invokeRestart("muffleWarning")
                                   }
                                 })

  # if (!rlm_fit$converged) {
  #   warning("The IRLS algorithm employed in 'rlm' did not converge.")
  # }

  
  # Prediction weights. By default, these are weighted equally. If 'wider_gate'
  # is set, the weights from 'rlm' are used. The weight for the maximum FSC-A
  # is usually much smaller and yields a more uncertain prediction interval.
  # This leads to a wider gate.
  prediction_weights <- c(1, 1)
  if (wider_gate) {
    prediction_weights <- rlm_fit$w[c(which_min, which_max)]
  }

  #not really needed since default weights are 1s
  #but somehow this recompute qr yield slightly different predicted
  #vals (5e-7 < err < 2e-6), which is actually small enough for gates to be considered as ignorable
  #but for the sake of comparison to MASS version, we keep this step for now
  rlm_fit$weights <- rep(1, nrow(x))
  rlm_fit$qr <- qr(sqrt(rlm_fit$weights) * rlm_fit$x)
  
  #strip rlm class to dispatch stats::predict.lm
  class(rlm_fit) <- "lm"
  rlm_fit$terms <- rlm_formula
   predictions <- predict(rlm_fit, scale = rlm_fit$s, data.frame(x_extrema), interval = "prediction",
                         level = prediction_level, weights = prediction_weights)

  # Create a matrix of the vertices using the prediction bands at the minimum
  # and maximum values of x. The ordering matters. Must go clockwise.
  # Otherwise, the polygon is not convex and makes an X-shape.
  gate_vertices <- rbind(cbind(x_extrema[, area][1], predictions[1, "lwr"]),
                         cbind(x_extrema[, area][1], predictions[1, "upr"]),
                         cbind(x_extrema[, area][2], predictions[2, "upr"]),
                         cbind(x_extrema[, area][2], predictions[2, "lwr"]))
  colnames(gate_vertices) <- channel_names

  polygonGate(gate_vertices, filterId = filterId)
}

singletGate <- gate_singlet

#' robust linear model using an M estimator 
#' 
#' rewritten in c++, till eval stats::lm.wfit r function in underlying cpp11 code 
#' It is internally used for singletGate, thus its output format may not be generic enough for common model fitting .
#' e.g. it doesn't take formula as input
#' 
#' @param x matrix with first column as weight (default can be 1s), the rest columns are predict variable
#' @param y numeric vector as response 
#' @param maxit maximum iterations
#' @export
#' @examples
#' n <- 1e3
#' x <- seq_len(n)
#' y <- x * 2.5 - 1.3 + rnorm(n, sd = 30)
#' names(y) <- x
#' x <- cbind(1, x)
#' r2 <- fast_rlm(x, y)
fast_rlm <- function(x, y, maxit = 20)
{
    
    fit_result <- rlm_cpp(x, y, maxit)
    fitted <- drop(x %*% fit_result$coefficients)
    fit <- list(coefficients = fit_result$coefficients
                , residuals = y - fitted
                , wresid = fit_result$residuals,
                # effects = fit_result$effects,
                rank = fit_result$rank
                , fitted.values = fitted
                # assign = fit_result$assign
                # ,  qr = fit_result$qr
                # , df.residual = NA
                , w = fit_result$w
                ,s = fit_result$scale
                ,weights = NULL
                , converged = fit_result$done
                , x = x
                )
    class(fit) <- c("rlm", "lm")
    fit
}

test_that("fast_rlm", {

  
  set.seed(1)
  n <- 1e3
  x <- seq_len(n)
  y <- x * 2.5 - 1.3 + rnorm(n, sd = 30)
  df <- data.frame(x, y)
  #convert to rlm data format
  names(y) <- x
  x <- cbind(1, x)
  # r1 <- MASS::rlm(y~x, df)
  system.time(r1 <- MASS::rlm(x, y))
  system.time(r2 <- fast_rlm(x, y))
  # r2$coefficients is unnamed
  expect_equivalent(r1$coefficients, r2$coefficients)
  sel = c("coefficients"
          , "residuals"
          , "fitted.values"
          , "weights"
          , "rank"
          , "converged"
          , "wresid"
          , "w" ,"x", "s"
  )
  expect_equivalent(r1[sel], r2[sel])
  
  
  })

      

#' the class that carries event indices as well

setClass("ocRectangleGate", contains = "rectangleGate", representation(ind = "raw"))

#' bypass the default flowWorkspace:::.addGate 
#' 
#' to support adding gate along with indices without loading flow data and computing
#' 
#' however it is proven that logical indices are too big to be efficiently passed around
#' 
#' @param gh \code{GatingHierarchy} see \link[flowWorkspace]{add} in \code{flowWorkspace} package
#' @param gate \code{ocRectangleGate} or \code{logicalFilterResult}
#' @param recompute \code{logical} see \link[flowWorkspace]{add} in \code{flowWorkspace} package
#' @param ... see \link[flowWorkspace]{add} in \code{flowWorkspace} package
#' @rdname add
pop_add.ocRectangleGate <- function(gate, gh, recompute, ... )
    {
      #unpack the bit vector
      indices <- ncdfFlow:::toLogical(gate@ind)
      #ignore the recompute flag and force it to be skipped
      nodeID <- flowWorkspace:::.addGate(gh, filter_to_list(gate), ...)
      sn <- sampleNames(gh)
      ptr <- gh@pointer
      flowWorkspace:::cpp_setIndices(ptr, sn, nodeID-1, indices)
      
      
    }


#' special gate type that mix the rectangleGate with boolean gate
setClass("ocRectRefGate", contains = c("rectangleGate", "booleanFilter"))
#' constructor for ocRectRefGate
#' @param rectGate \code{rectangleGate}
#' @param boolExprs \code{character} boolean expression of reference nodes
ocRectRefGate <- function(rectGate, boolExprs){
  
  bf <- eval(substitute(booleanFilter(x), list(x = as.symbol(boolExprs))))
  g <- as(rectGate, "ocRectRefGate")
  g@expr <- bf@expr
  g@deparse <- bf@deparse
  g
}
	
#' byPass the default .addGate 
#' 
#' to support adding rectangleGate yet gating through boolean operations 
#' without loading flow data
#' 
#' @export 
#' @rdname add
pop_add.ocRectRefGate <- function(gate, gh, recompute, ... )
    {
      rectFilterObj <- flowWorkspace:::filter_to_list.rectangleGate(gate)
      boolFilterObj <- flowWorkspace:::filter_to_list.booleanFilter(gate)
      #ignore the recompute flag and force it to be skipped
      nodeID <- flowWorkspace:::.addGate(gh, rectFilterObj, ...)
      sn <- sampleNames(gh)
      ptr <- gh@pointer
      flowWorkspace:::cpp_boolGating(ptr, sn, boolFilterObj, nodeID - 1)
      nodeID
    }




#' fast version of add gates to gatingset (bypassing some R checks)
#' 
#' used by gating_polyFunctions
#' @noRd 
.addGate_fast <- function(gs, filter, name = NULL, parent = "root", negated = FALSE){
  
  #preprocess filter
  filterObj <- filter_to_list(filter)
#  browser()
  if(is.null(name))
    name<-filterObj$filterId
  #replace the slash with colon 
  #since forward slash is reserved for gating path
  if(grepl("/",name)){
    old_name <- name
    name <- gsub("/",":",name)
    warning(old_name, " is replaced with ", name)
  }
  
  
  gh<-gs[[1]]
  ##get node ID
  
  filterObj$negated<-negated
  
  
  if(class(gs) == "GatingSetList"){
    nodeIDs <- lapply(gs, function(thisGS){
          samples <- sampleNames(thisGS)
          lapply(samples,function(sample){
                
                nodeID <- flowWorkspace:::cpp_addGate(thisGS@pointer,sample,filterObj, parent,name)
                nodeID+1
              })
        }, level = 1)
    nodeIDs <- unlist(nodeIDs)
    
  }else{
    samples <- sampleNames(gs)
    nodeIDs<-lapply(samples,function(sample){
          
          nodeID <- flowWorkspace:::cpp_addGate(gs@pointer,sample,filterObj, parent,name)
          nodeID+1
        })  
  }
  
  
  nodeID<-nodeIDs[[1]]
  
  if(!all(sapply(nodeIDs[-1],function(x)identical(x,nodeID))))
    stop("nodeID are not identical across samples!")
  nodeID
}

#' a function to find interesting features in density (most notably the minimum)
#' as precursor to creating a 1D gate
#' 
#' @author Greg Finak, Phu T. Van
#' @param D a \code{density} containing the data to operate on
#' @param gate_range a \code{character} specifying the data range to operate on
#' @param adjust a \code{numeric} specifying the amount of smoothing to be used
#' @param plot a \code{boolean} specifying whether to output a plot
#' @noRd 
.improvedMindensity <- function(D,adjust=2,gate_range=NA, plot = FALSE, ...){
  # construct the density from data and adjust params we were given
  dens <- density(D,adjust=adjust)
  
  # restrict data to range
  if (!is.null(gate_range)) {
    if (length(gate_range) == 2 & gate_range[1] < gate_range[2]) {
      filter <- dens$x > gate_range[1] & dens$x < gate_range[2]
      dens$x <- dens$x[filter]
      dens$y <- dens$y[filter]
    }else{
      #no range provided, do nothing
    }
  }
  
  sp <- smooth.spline(dens$x,dens$y)
  pred <- predict(sp)
  
  d1 <- predict(sp,deriv = 1)
  d2 <- predict(sp,deriv = 2)
  d3 <- predict(sp,deriv = 3)
  
  # find features
  inf1 <- sign(d2$y[-length(d2$y)])>0&sign(d2$y[-1L])<0
  inf2 <- sign(d2$y[-length(d2$y)])<0&sign(d2$y[-1L])>0
  minima <- sign(d1$y[-1])>0&sign(d1$y[-length(d1$y)])<0
  maxima <- sign(d1$y[-1])<0&sign(d1$y[-length(d1$y)])>0
  shoulders <- sign(d3$y[-1])<0&sign(d3$y[-length(d3$y)])>0
  
  minima_xcoords <- sp$x[which(minima)] # x-coords of minima
  maxima_xcoords <- sp$x[which(maxima)] # x-coords of maxima
  
  minima_ycoords <- sp$y[which(sp$x %in% minima_xcoords)] # y-coords of minima
  maxima_ycoords <- sp$y[which(sp$x %in% maxima_xcoords)] # y-coords of maxima
  
  #### TO-DO: determine a range in X-coord where the peaks are tiny (< some data-derived threshold)
  # exclude features from this range when considering where to cut. 
  # this is motivated by cyTOF data, where there is a long tail of extreme values that 
  # cannot be eliminated by higher `adjust` values, and shouldn't be eliminated by gate_range()
  # 
  
  if (length(which(minima == TRUE)) == 0){ # no minima found, look through shoulders
    # if there is a peak, pick first shoulder to the right of peak  
    if (length(which(maxima == TRUE)) == 1  ){ 
      pkidx <- which(maxima == TRUE)
      pt <- sp$x[median(which(shoulders)[which(shoulders) > pkidx])]
      if (is.na(pt)){
        # there is only one peak, and there is no shoulder to the right of it
        # pick an inflection point as a last resort
        pt <- sp$x[median(which(inf2)[which(inf2) > pkidx])]
      }
      if( is.na(pt)){
        # There also is no inflection point to the right of it. Resort
        # to the overall median shoulder as the cutpoint like below
        pt <- sp$x[median(which(shoulders))]
      }
      
    }  else {
      # no peak (or multiple peaks), select overall median shoulder as cutpoint (for now...)
      pt <- sp$x[median(which(shoulders))]    
    }
    
    # This is an absolute last resort to make sure this block never returns an NA
    # due to a lack of shoulders and inflection points. In this case just return the
    # simple min
    if( is.na(pt)){
      pt <- sp$x[which.min(sp$y)]
    }
    
  } else if (length(which(minima == TRUE)) > 1) { # multiple minima
    
    m <- min(sp$y[which(sp$x %in% minima_xcoords)])  # pick the minima with lowest y
    pt <- sp$x[which(sp$y == m)]
    
    
  } else if (length(which(minima == TRUE)) == 1){ # only 1 minima, use it as cut point
    pt <- sp$x[which(minima)]
    
  }
  
  .plots = function(){
    abline(v = d2$x[which(inf1)],col="red")       # red    == inflection point
    abline(v = d2$x[which(inf2)],col="green")     # green  == inflection point
    abline(v = d2$x[which(maxima)],col="blue")    # blue   == maxima
    abline(v = d2$x[which(minima)],col="orange")  # orange == minima
    abline(v = d2$x[which(shoulders)],col="pink") # pink   == shoulders
  }
  if(plot){
    par(mfrow=c(2,1))
    plot(sp,type="l",main="features")
    .plots()
    plot(sp,type="l",main="final_cut")
    abline(v = pt, col="black", lwd=2)
    par(mfrow=c(1,1)) # be nice, restore plot settings
  }    
  
  return(list(density = sp,
              inf_rising = sp$x[which(inf1)],
              inf_falling = sp$x[which(inf2)],
              maxima = maxima_xcoords,
              minima = minima_xcoords,
              maxima_heights = maxima_ycoords,
              minima_heights = minima_ycoords,
              shoulders = sp$x[which(shoulders)], 
              final_cut = pt))
  
}


#' An improved version of mindensity used to determines a cutpoint as the minimum point 
#' of a kernel density estimate between two peaks.
#' 
#' Analogous to the original openCyto::mindensity(), mindensity2 operates on a standard flowFrame. Its behavior is closely modeled on 
#' the original mindensity() whenever possible. However, the underlying peak-finding algorithm 
#' (improvedMindensity) behaves significantly differently. 
#' 
#' @author Greg Finak, Phu T. Van
#' 
#' @name gate_mindensity2
#' @aliases mindensity2
#' @param fr a \code{flowFrame} object
#' @param channel the channel to operate on
#' @param filterId a name to refer to this filter
#' @param gate_range numeric vector of length 2. If given, this sets the bounds
#' on the gate applied. 
#' @param min a numeric value that sets the lower boundary for data filtering
#' @param max a numeric value that sets the upper boundary for data filtering
#' @param peaks \code{numeric} vector. If not given , then perform peak detection first by .find_peaks
#' @param ... Additional arguments for peak detection.
#' @return a \code{rectangleGate} object based on the minimum density cutpoint
#' @examples
#' \dontrun{
#'  gate <- gate_mindensity2(fr, channel = "APC-A") # fr is a flowFrame
#' }
#' @export
gate_mindensity2 <- function(fr, channel, filterId = "", 
                         gate_range = NULL, min = NULL, max = NULL, peaks = NULL, 
                         ...) {
  if (missing(channel) || length(channel) != 1) {
    stop("A single channel must be specified.")
  }
  # if there is 
  if (!(is.null(min) && is.null(max))) {
    fr <- .truncate_flowframe(fr, channels = channel, min = min, 
                              max = max)
  }
  x <- exprs(fr[,channel])
  # this line does the actual searching of the cut point
  g <- .improvedMindensity(D=x,gate_range=gate_range,...)
  
  coords <- list(c(g$final_cut, Inf))
  # name the gate and return it
  names(coords) <- channel
  return(rectangleGate(coords, filterId = filterId)) 
}

#' @export
mindensity2 <- function(fr, channel, filterId = "", 
                        gate_range = NULL, min = NULL, max = NULL, peaks = NULL, 
                        ...){
  .Deprecated("gate_mindensity2")
  gate_mindensity2(fr, channel, filterId, gate_range, min, max, peaks, ...)
}

#' wrapper for mindensity2
#' 
#' It does some parameter preprocessing before calling the mindensity
#' 

#' @param ... arguments to be passed to \link{mindensity}
#' @inheritParams .gate_flowclust_1d 
#' 
#' @return a \code{filter} object
#' @noRd 
.gate_mindensity2 <- function(fr, pp_res, channels, ...) {
  
  if(length(channels) != 1)
    stop("invalid number of channels for mindensity!")
  gate <- gate_mindensity2(fr, channel = channels, ...)
  gate
}

.mindensity2 <- .gate_mindensity2

context("gating functions")
data(GvHD)
fr <- GvHD[[1]]
fr <- transform(fr, transformList(c("FL1-H"),  log))

test_that("gate_singlet", {
  fr <- read.FCS(system.file("extdata/CytoTrol_CytoTrol_1.fcs", package = "flowWorkspaceData"))
  g <- gate_singlet(fr, "FSC-A", "FSC-H"
                    # , robust = F
                    # , prediction_level =0.9
                    )
  expect_equal(g, polygonGate(structure(c(24601, 24601
                                          , 262143, 262143
                                          , 11476.9417996388
                                          , 48163.8684594367
                                          , 236614.412689126, 199926.505500422)
                                          , .Dim = c(4L, 2L)
                                          , .Dimnames = list(NULL, c("FSC-A", "FSC-H")))
                                , filterId = "singlet"), tol = 5e-07)
  g <- gate_singlet(fr, "FSC-A", "FSC-H"
                    , sidescatter = "SSC-A"
                    # , prediction_level =0.9
  )
  expect_equal(g, polygonGate(structure(c(24601, 24601
                                          , 262143, 262143
                                          , 16988.0208566085, 41393.3256789415
                                          , 227799.565338899, 203392.940551338)
                                        , .Dim = c(4L, 2L)
                                        , .Dimnames = list(NULL, c("FSC-A", "FSC-H")))
                              , filterId = "singlet"), tol = 5e-06)
  # library(ggcyto)
  # autoplot(fr, "FSC-A", "FSC-H") + geom_gate(g) + geom_stats()
})

test_that("gate_mindensity2", {
      
      expect_equal(gate_mindensity2(fr, "SSC-H"), rectangleGate(`SSC-H` = c(901.120, Inf), filterId = ""), tol = 5e-07)
      expect_equal(gate_mindensity2(fr, "FL1-H"), rectangleGate(`FL1-H` = c(5.216064, Inf), filterId = ""), tol = 5e-07)
      #the latest pull from phu seems to change the cutpoint for this call
      #      expect_equal(mindensity2(fr, "FL2-H"), rectangleGate(`FL2-H` = c(511.0989417, Inf), filterId = ""), tol = 5e-07)
      expect_equal(gate_mindensity2(fr, "FL2-H"), rectangleGate(`FL2-H` = c(4777.69089, Inf), filterId = ""), tol = 5e-07)
      
      
    })

#' @templateVar old listgtMethods
#' @templateVar new gt_list_methods
#' @template template-depr_pkg
NULL
#'The environment holding the names of registered methods
#' @noRd 
.openCyto_plugin_method_lookup <- new.env()
.openCyto_plugin_method_lookup[["registered_methods"]] <- list(gating = character(0), preprocessing = character(0))

#' Environment holding dynamically registered plugin functions
#' @noRd
.openCyto_plugin_functions <- new.env(parent = emptyenv())

#' Resolve a plugin function by name
#'
#' First checks the plugin functions environment for dynamically registered
#' functions, then falls back to the package namespace for built-in functions.
#' @param name character function name (with leading dot)
#' @return the function object
#' @noRd
.resolve_plugin_function <- function(name) {
  if (exists(name, envir = .openCyto_plugin_functions, inherits = FALSE)) {
    return(get(name, envir = .openCyto_plugin_functions, inherits = FALSE))
  }
  ns <- getNamespace("openCyto")
  if (exists(name, envir = ns, inherits = FALSE)) {
    return(get(name, envir = ns, inherits = FALSE))
  }
  stop(sprintf("Plugin function '%s' not found", name))
}
.DEFAULT_GT <- c("quantileGate", "gate_quantile", "rangeGate","flowClust.2d", "gate_flowclust_2d", "mindensity", "gate_mindensity"
                 , "mindensity2", "gate_mindensity2", "flowClust.1d", "gate_flowclust_1d", "boundary","singletGate"
                 , "quadGate.tmix", "gate_quad_tmix", "quadGate.seq", "gate_quad_sequential", "gate_template", "gate_custom"
                 )
.DEFAULT_PP <- c("prior_flowClust", "warpSet", "standardize_flowset", "pp_gate_template", "pp_gate_custom")
#'Print a list of the registered gating methods
#'@name gt_list_methods
#'@aliases listgtMethods
#'@return Does not return anything. Prints a list of the available gating methods.
#'@export
gt_list_methods <- function(){
  res <- .getPluginMethods()
  cat("Gating Functions:\n")
  for(i in res[["gating"]])
      cat("=== ", i,  "\n")
  cat("Preprocessing Functions:\n")
  for(i in res[["preprocessing"]])
    cat("=== ", i,  "\n")
}

#'@export
listgtMethods <- function(){
  .Deprecated("gt_list_methods")
}

#'Is the method registered
#'
#'This will strip the preceding dot.
#' @noRd 
.isRegistered <- function(name){
  return(gsub("^\\.","", name)%in%unlist(.getPluginMethods()))
}

#'return a list of registered and default gating methods
#' @noRd 
.getPluginMethods <- function(x){
#  ns <- getNamespace("openCyto")
#  plugins <- getFromNamespace(".openCyto_plugin_method_lookup", ns = ns)
  
  plugin_gt <- .openCyto_plugin_method_lookup[["registered_methods"]][["gating"]]
  plugin_pp <- .openCyto_plugin_method_lookup[["registered_methods"]][["preprocessing"]]
  list(gating = c(plugin_gt, .DEFAULT_GT) 
      , preprocessing = c(plugin_pp, .DEFAULT_PP)
      )
  
}
#'Check the formal arguments of a gating function.
#'
#'The formal arguments need to match a certain template
#'We check that they do or do not.
#' @noRd 
.checkFormals <- function(frmls = NA, type = c("gating", "preprocessing")){
  
  type <- match.arg(type, c("gating", "preprocessing"))
  # we don't need to check ... since it is up to either
  # wrapper function or the actual gating function to handle its 
  # own formals
  if(type == "gating"){
    expected <- c("fr","pp_res","channels")
  }else{
    
    expected <- c("fs","gs", "gm", "channels", "groupBy", "isCollapse")
  }
  isMatched <- all(expected %in% names(frmls))
  if(!isMatched)
  {
    message("Formals of function don't match expected template: ")
    message("function(", paste(expected, collapse = ", "), ", ...)")
    return(FALSE)
  }
  return(TRUE)
}

registerGatingFunction <- function(fun=NA,methodName, dep=NA){
  .Defunct("registerPlugins")
}
#' @templateVar old registerPlugins
#' @templateVar new register_plugins
#' @template template-depr_pkg
NULL
#' Register a gating or preprocessing function with OpenCyto
#'
#' Function registers a new gating or preprocessing method with openCyto so that it may be used in the 
#' csv template.
#'
#' @name register_plugins
#' @aliases registerGatingFunction registerPlugins
#' @param fun \code{function} to be registered
#' @param methodName \code{character} name of the gating or preprocessing method
#' @param dep \code{character} name of the library dependency required for the plugin method to work.
#' @param ... other arguments
#'         type \code{character} specifying the type of registering method. Should be either "gating" or "preprocessing".
#' 
#' @return \code{logical} TRUE if successful and prints a message. FALSE otherwise.
#' @useDynLib openCyto,.registration = TRUE
#' @details The \code{fun} argument should be a wrapper function definition for the gating or preprocessing method. 
#'                          Gating method must have formal arguments:
#' 
#'                           fr a \code{flowFrame}
#' 
#'                           pp_res a pre-processing result
#' 
#'                           xChannel \code{character} (optional)
#' 
#'                           yChannel \code{character} (required)
#' 
#'                           filterId \code{character}
#' 
#'                           ... ellipses for the additional parameters.
#' 
#'                          Preprocessing method must have formal arguments:
#' 
#'                          fs a \code{flowSet} that stores the flow data (could be subgrouped data if \code{groupBy} column is defined in the csv template
#' 
#'                          gs a \code{GatingSet}
#'  
#'                          gm a \code{gtMethod} object that stores the information from gating method
#' 
#'                          xChannel \code{character} (required)
#' 
#'                          yChannel \code{character} (required)
#' 
#'                           ... ellipses for the additional parameters.
#' 
#' The gating function must return a filter (i.e. polygonGate or other instance) from flowCore.
#' The preprocessing can return anything and it will be passed on to the gating function. So it is up to gating function to use and interpret the results of preprocessing.
#' Not all formal parameters need to be used. Additional arguments are passed via the ... and can be processed in the wrapper
#' 
#' @import utils
#' @export
register_plugins <- function(fun = NA, methodName, dep = NA, ...){
  
  if(!is.na(dep)){
    if(is.character(dep)){
      if(system.file(package = dep) == ""){
        message(sprintf("Can't register %s with dependency on %s, because dependency is not installed.",methodName,dep))
        return(FALSE)
      }
    }else{
      warning("If provided, dep must be a character naming the package dependency.")
      return(FALSE)
    }
  }
  if(!is.function(fun)){
    warning("You need to put the fun in function! (argument fun is not a function)")
    return(FALSE)
  }else{
    ###Check the formal arguments
    frmls <- formals(fun)
    
    if(.checkFormals(frmls = frmls, ...)){
      if(.register(fun = fun, methodName = methodName, ...)){
        message(sprintf("Registered %s",methodName))
      }
      
    }else{
      warning("Can't register function")
      return(FALSE)
    }
  }
  return(TRUE)
}

#' @export
registerPlugins <- function(fun = NA, methodName, dep = NA, ...){
  .Deprecated("register_plugins")
  register_plugins(fun, methodName, dep, ...)
}


#'Register a gating function for OpenCyto
#' @noRd 
.register <- function(fun = NA,methodName, type = c("gating", "preprocessing")){
  
  type <- match.arg(type, c("gating", "preprocessing"))
  
  methodName <- paste0(".",methodName)
  
  # Store function in the plugin functions environment
  assign(methodName, fun, envir = .openCyto_plugin_functions)
  
  #add to the plugin method list
  
  toAdd <- gsub("^\\.","",methodName)
  current <- .openCyto_plugin_method_lookup[["registered_methods"]][[type]]
  
  if(length(current) == 0)
    found <- FALSE
  else if(any(grepl(toAdd, current)))
    found <- TRUE
  else
    found <- FALSE
  
  if(!found)
    .openCyto_plugin_method_lookup[["registered_methods"]][[type]] <- c(current, toAdd)  
  
  return(TRUE)
}

#' only for internal usage (debug)
#' @noRd 
.unregister <- function(methodName, type = c("gating", "preprocessing")){

  methodName <- paste0(".",methodName)
  type <- match.arg(type, c("gating", "preprocessing"))
  
  
  toRm <- gsub("^\\.","",methodName)
  
  current <- .openCyto_plugin_method_lookup[["registered_methods"]][[type]]
  
  if(length(current) == 0)
    found <- FALSE
  else{
    ind <- grepl(toRm, current)
    if(any(ind))
      found <- TRUE
    else
      found <- FALSE
  } 
  
  if(found)
    .openCyto_plugin_method_lookup[["registered_methods"]][[type]] <- current[!ind]
  
  # Remove from plugin functions environment
  if (exists(methodName, envir = .openCyto_plugin_functions, inherits = FALSE)) {
    rm(list = methodName, envir = .openCyto_plugin_functions)
  }
  
  return(TRUE)
}

context("plugin Framework")

test_that(".isRegistered", {
      
      expect_true(.isRegistered("mindensity"))
      expect_true(.isRegistered("warpSet"))
      
      expect_true(.isRegistered(".mindensity"))
      expect_true(.isRegistered(".warpSet"))
      
      expect_false(.isRegistered("myFunc"))
    })
test_that("register_plugins", {
      
      dummyMethod <- function(){}
      #depdency not installed
      expect_false(openCyto:::.isRegistered("dummyMethod"))
      expect_message(thisRes <- register_plugins(fun=dummyMethod,methodName="dummyMethod",dep="dummyPkg")
                    , "Can't register dummyMethod with dependency on dummyPkg, because dependency is not installed.")
      expect_false(thisRes)
      expect_false(openCyto:::.isRegistered("dummyMethod"))
      
      #invalid formal
      myGatingFunc <- function(fr1, pp_res, yChannel, ...){
        cat("dummy gating function")
      }
      expect_warning(expect_message(thisRes <- register_plugins(fun = myGatingFunc,methodName="myGatingFunc"), "Formals of function don't match expected template"))
      expect_false(thisRes)
      expect_false(.isRegistered("myGatingFunc"))

      #successfully registered without rep
      myGatingFunc <- function(fr, pp_res, channels, ...){
        cat("dummy gating function")
      }
      
      expect_message(thisRes <- register_plugins(fun = myGatingFunc, methodName = "myGatingFunc"), "Registered myGatingFunc")
      expect_true(thisRes)
      expect_true(.isRegistered("myGatingFunc"))
      expect_true(.unregister("myGatingFunc"))
      expect_false(.isRegistered("myGatingFunc"))

      
      #pp method
      myPreprocessingFunc <- function(fr, pp_res, yChannel, ...){
        cat("dummy preprocessing function")
      }
      
      expect_error(thisRes <- register_plugins(type = "pp", fun = myPreprocessingFunc, methodName = "myPreprocessingFunc"), "'arg' should be one of ")
      expect_warning(expect_message(thisRes <- register_plugins(type = "preprocess", fun = myPreprocessingFunc, methodName = "myPreprocessingFunc"), "Formals of function don't match expected template"))
      expect_false(.isRegistered("myPreprocessingFunc"))
      
      
      myPreprocessingFunc <- function(fs, gs, gm, xChannel, yChannel, ...){
        cat("dummy preprocessing function")
      }
      expect_warning(expect_message(thisRes <- register_plugins(type = "preprocess", fun = myPreprocessingFunc, methodName = "myPreprocessingFunc")
                    , "Formals of function don't match")      )
      expect_false(openCyto:::.isRegistered("myPreprocessingFunc"))
                
                
      myPreprocessingFunc <- function(fs, gs, gm, channels, groupBy, isCollapse, ...){
        cat("dummy preprocessing function")
      }                
      expect_message(thisRes <- register_plugins(type = "preprocess", fun = myPreprocessingFunc, methodName = "myPreprocessingFunc")
                    , "Registered myPreprocessingFunc")
      expect_true(.isRegistered("myPreprocessingFunc"))
      expect_true(.unregister("myPreprocessingFunc", type = "pre"))
      expect_false(.isRegistered("myPreprocessingFunc"))
      
    })

#' convert a gatingTemplate object to a data.table
#' 
#' It is the inverse function of gatingTemplate constructor.
#' 
#' @name as.data.table
#' @param x gatingTemplate object
#' @param keep.rownames not used
#' @return a data.table
#' @export
as.data.table.gatingTemplate <- function(x, keep.rownames = FALSE){
  
  
  # gate each node 
  gt_nodes <- gt_get_nodes(x, order = "tsort")[-1]
  
  
  res <- lapply(gt_nodes, function(node){
    
    # get parent node to gate
    nodePath <- node@id
    parent <- gt_get_parent(x, nodePath)
    # extract gate method from one edge(since multiple edge to the same node is
    # redudant)
    this_gate <- gt_get_gate(x, parent, nodePath)
    gating_args <- parameters(this_gate)
    #collapse into string when neccessary
    split <- !extends(class(this_gate), "refGate")
    gating_args <- .argDeparser(gating_args, split)
    
    #get preprocessing method
    this_ppm <- openCyto:::ppMethod(x, parent, nodePath)
    
    #preprocessing
    if(class(this_ppm) == "ppMethod")
    {
      preprocessing_method <- names(this_ppm)
      preprocessing_args <- parameters(this_ppm)
      preprocessing_args <- .argDeparser(preprocessing_args)
    }else{
      preprocessing_args <- preprocessing_method <- ""
    }
    
    data.table(alias = alias(node)
                          , pop = names(node)
                          , parent = parent
                          , dims = this_gate@dims
                          , gating_method = names(this_gate)
                          , gating_args =  gating_args
                          , collapseDataForGating = openCyto:::isCollapse(this_gate)
                          , groupBy = openCyto:::groupBy(this_gate)
                          , preprocessing_method = preprocessing_method
                          , preprocessing_args = preprocessing_args
                      )    
    
  })
  
  rbindlist(res)
  
}

#' deparse a list(named) of expression into a string 
#' inverse function of argParser
#' @noRd 
.argDeparser <- function(args, split = TRUE){
  if(split){
    args <- sapply(names(args), function(argn){
          
          argv <- deparse(args[[argn]])
          argv <- gsub("\"", "'", argv) #restore dquote to squote
          argv <- paste(argv, collapse = "")
          paste(argn, argv, sep = " = ")
        })
    
    
    paste(args, collapse = ", ")  
  }else
    as.character(args[[1]])
    
}

#' get population name
#' @param x \code{gtPopulation} object
#' @aliases names,gtPopulation-method
setMethod("names", signature = c("gtPopulation"), definition = function(x) {
  x@name
})

# the original S4 method has been suffering from S4 masking issue,thus we keep it as private functioninstead
# get population alias
# @param object \code{gtPopulation} 
#setMethod("alias", signature = c("gtPopulation"), definition = function(object) {
alias <- function(object){
  object@alias
}

 

# rewrite huber estimator
robust_m_estimator <- function(x, sd)
{
  stopifnot(sd != 0) 
  Winsorizes_k = 1.5
  convergence_tol = 1e-6
  x <- x[!is.na(x)]
  n <- length(x)
  mu <- median(x)
  Winsorizes_k <- sd * Winsorizes_k
  convergence_tol <- sd * convergence_tol
  while(TRUE)
  {
    p <- mu - Winsorizes_k
    y <- sapply(x, function(i)max(p, i))
    q <- mu + Winsorizes_k
    y <- sapply(y, function(i)min(q, i))
    res <- mean(y)
    if(abs(mu - res) < convergence_tol)
      return(mu)
    mu = res
  }
  
}

context("misc functions")

test_that("fast collapsing for ncdfFlowSet", {
      library(openCyto)
      data(GvHD)
      fs <- GvHD[pData(GvHD)$Patient %in% 6:7][1:4]
      suppressMessages(ncfs <- ncdfFlowSet(fs))
      
      fr <- as(fs, "flowFrame")
      fr1 <- as(ncfs, "flowFrame")
      expect_is(fr1, "flowFrame")
      expect_equal(fr@exprs[, -9], fr1@exprs, tol = 2e-7)
      
    })

test_that("start/stop.at argument in gt_gating function", {
  library(openCyto)
  
  dataDir <- system.file("extdata",package="flowWorkspaceData")
  files <- list.files(dataDir, pattern = "a2004_O1T2pb05i_A1_A01.fcs$",full = TRUE)
  fs <- read.flowSet(files[1])
  
  compMat <- spillover(fs[[1]]) %>% compensation
  fs_comp <- compensate(fs, compMat)
  chnls <- parameters(compMat)
  transFuncts <- estimateLogicle(fs[[1]], channels = chnls)
  fs_trans <- transform(fs_comp, transFuncts)
  
  gs <- GatingSet(fs_trans)
  gt <- gatingTemplate(system.file("extdata/gating_template/testGatingTemplate.csv", package = "openCyto"))
  gt_gating(gt, gs, stop.at = "/boundary/singlet/Alexa700_gate")
  expect_setequal(gs_get_pop_paths(gs),c("root", "/boundary", "/boundary/singlet", "/boundary/singlet/PerCPCY5_5_gate", "/boundary/singlet/APC_gate", "/boundary/singlet/AmCyan_gate"))  
  gt_gating(gt, gs, start = "/boundary/singlet/Alexa700_gate")
  expect_setequal(gs_get_pop_paths(gs),c("root", "/boundary", "/boundary/singlet"
                                        , "/boundary/singlet/PerCPCY5_5_gate"
                                        , "/boundary/singlet/APC_gate"
                                        , "/boundary/singlet/AmCyan_gate"
                                        , "/boundary/singlet/Lymph"
                                        , "/boundary/singlet/Alexa700_gate"
                                        , "/boundary/singlet/Lymph/Alexa700pPerCPCY5_5p", "/boundary/singletRefGate", "/boundary/singletRefGate/singletRefGate_LymphRefGate"))  
})

test_that("robust mean", {
  data(GvHD)
  fr <- GvHD[[1]]
  x <- exprs(fr)[,1]
  sd <- mad(x)
  res <- robust_m_estimator(x, sd)
  
  expect_equal(sd, 148.26, tol = 2e-7)
  expect_equal(res, 229.625, tol = 2e-7)
  
  set.seed(1)
  x = rnorm(100)
  sd <- mad(x)
  res <- robust_m_estimator(x, sd)
  expect_equal(sd, 0.8700003, tol = 2e-7)
  expect_equal(res, 0.1153234, tol = 2e-7)
  
})

test_that("solve_LSAP", {
  
  x <- matrix(c(5, 1, 4, 3, 5, 2, 2, 4, 4), nrow = 3)
  x
  y <- as.vector(solve_LSAP(x))
  expect_equal(y, c(3,1,2))  
  
  x = structure(c(120.19673111448, 109.212815628447
                  , 693.431895464961, 611.636808127566)
                , .Dim = c(2L, 2L)
                , .Dimnames = list(c("1", "2"), c("1", "2")))
  y <- as.vector(solve_LSAP(x))
  expect_equal(y, c(1,2))  
  
  x = structure(c(49.9764377352581, 150.456415722488, 998.26447559313, 882.252993034892)
                , .Dim = c(2L, 2L)
                , .Dimnames = list(c("1", "2"), c("1", "2")))
  y <- as.vector(solve_LSAP(x))
  expect_equal(y, c(1,2))  
  
})

#' Hierarchical Gating Pipeline for flow cytometry data.
#' 
#' openCyto is a package designed to facilitate the automated gating methods in sequential way 
#' to mimic the manual gating strategy.
#'
#'
#' \tabular{ll}{
#' Package: \tab openCyto\cr
#' Type: \tab Package\cr
#' Version: \tab 1.2.8\cr
#' Date: \tab 2014-04-10\cr
#' License: \tab GPL (>= 2)\cr
#' LazyLoad: \tab yes\cr
#' }
#'
#' @author
#' Mike Jiang \email{wjiang2@@fhcrc.org},
#' John Ramey \email{jramey@@fhcrc.org},
#' Greg Finak \email{gfinak@@fhcrc.org}
#'
#' Maintainer: Mike Jiang \email{wjiang2@@fhcrc.org}
#' @name openCyto
#' @title Hierarchical Gating Pipeline for flow cytometry data
#' @keywords package
#' @examples
#' \dontrun{gatingTemplate('test.csv')}
#' @seealso See \code{\link[openCyto]{gt_gating}}, 
#' \code{\link{gate_flowclust_1d}}, 
#' for an overview of gating functions.
"_PACKAGE"

#' a class storing the gating method and population information in a graphNEL object
#' 
#' Each cell population is stored in graph node and is connected with its parent population 
#' or its reference node for boolGate or refGate.
#' 
#' @rdname gatingTemplate-class 
#' @export 
#' @importClassesFrom graph graphNEL graphBase graph
#' @importClassesFrom Biobase AssayData
#' @name gatingTemplate-class
setClass("gatingTemplate", contains = "graphNEL", representation(name = "character"))

	
#' a virtual class that represents the gating result generated by flowClust gating function
#' 
#' Bascially it extends flowCore 'filter classes to have extra slot to store priors and posteriors
#' @name fcFilter-class
setClass("fcFilter", representation("VIRTUAL", priors = "list", posteriors = "list"))

#' a concrete class that reprents the polygonGate generated by flowClust
#' 
#' It stores priors and posteriors as well as the actual polygonGate.
#' @importClassesFrom flowCore polygonGate
#' @name fcPolygonGate-class
setClass("fcPolygonGate", contains = c("fcFilter", "polygonGate"))

#' a concrete class that reprents the ellipsoidGate generated by flowClust
#' 
#' It stores priors and posteriors as well as the actual ellipsoidGate.
#' @importClassesFrom flowCore ellipsoidGate
#' @name fcEllipsoidGate-class
setClass("fcEllipsoidGate", contains = c("fcFilter", "ellipsoidGate"))


#' a concrete class that reprents the rectangleGate generated by flowClust
#' 
#' It stores priors and posteriors as well as the actual rectangleGate.
#' @importClassesFrom flowCore rectangleGate
#' @name fcRectangleGate-class
setClass("fcRectangleGate", contains = c("fcFilter", "rectangleGate"))

#' constuctor for \code{fcRectangleGate}
#' 
#' @param x a \code{rectangleGate} object
#' @param priors a \code{list} storing priors
#' @param posts a \code{list} storing posteriors
fcRectangleGate <- function(x, priors, posts) {
  res <- as(x, "fcRectangleGate")
  res@priors <- priors
  res@posteriors <- posts
  res
}

#' constuctor for \code{fcPolygonGate}
#' @param x a \code{polygonGate} object
#' @inheritParams fcRectangleGate
fcPolygonGate <- function(x, priors, posts) {
  res <- as(x, "fcPolygonGate")
  res@priors <- priors
  res@posteriors <- posts
  res
}

#' constuctor for \code{fcEllipsoidGate}
#' @param x a \code{ellipsoidGate} object
#' @inheritParams fcRectangleGate
fcEllipsoidGate <- function(x, priors, posts) {
  res <- as(x, "fcEllipsoidGate")
  res@priors <- priors
  res@posteriors <- posts
  res
}


#' a class that extends \code{filterList} class.
#' 
#' Each filter in the filterList must extends the \code{fcFilter} class
#' @importClassesFrom flowCore filterList
#' @name fcFilterList-class
setClass("fcFilterList", contains = "filterList")

#' constuctor for \code{fcFilterList}
#' 
#' @param x \code{list} of \code{fcFilter} (i.e. \code{fcPolygonGate} or \code{fcRectangleGate})
#' 
fcFilterList <- function(x) {
  
  if (!all(unlist(lapply(x, function(i) extends(class(i), "fcFilter"))))) {
    stop("not all filters are fcFilter!")
  }
  if (class(x) == "list") {
    x <- filterList(x)
  }
  
  sname <- names(x)
  x <- as(x, "fcFilterList")
  
  attr(x, "names") <- sname
  x
}

 
#' A class to represent a flowClust tree.
#' 
#' It is a graphNEL used as a container to store priors and posteriors for each flowClust gate that can be
#' visualized for the purpose of fine-tunning parameters for flowClust algorithm
#' 
#' @name fcTree-class
setClass("fcTree", contains = "gatingTemplate")

#' constructor of \code{fcTree}
#' 
#' It adds an extra node data slot "fList"(which is a \code{filterList} object) to the \code{gatingTemplate}
#'  
#' @param gt a \code{gatingTemplate} object
fcTree <- function(gt) {
  res <- as(gt, "fcTree")
  nodeDataDefaults(res, "fList") <- new("filterList")
  res
}

#' A class to represent a gating method.
#' 
#' A gating method object contains the specifics for generating the gates.
#' 
#' @section Slots:
#'  \describe{
#' 
#'      \item{name}{ a \code{character} specifying the name of the gating method}
#
#'      \item{dims}{ a \code{character} vector specifying the dimensions (channels or markers) of the gate}
#' 
#'      \item{args}{ a \code{list} specifying the arguments passed to gating function}
#' 
#'      \item{groupBy}{ a \code{character} or \code{integer} specifying how to group the data.
#'                   If \code{character}, group the data by the study variables (columns in \code{pData}).
#'                    If \code{integer},  group the data by every \code{N} samples.
#'                   }
#' 
#'      \item{collapse}{ a \code{logical} specifying wether to collapse the data within group before gating.
#'                  it is only valid when \code{groupBy} is specified}
#'  }
#' 
#' 
#' 
#' @rdname gtMethod-class 
#' @name gtMethod-class
#' @aliases gtMethod
#' @examples 
#'  \dontrun{
#'      gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#'      gh_pop_get_gate(gt, '2', '3')
#' }
setClass("gtMethod", representation(name = "character"
                                    , dims = "character"
                                    , args = "list"
                                    , groupBy = "ANY"
                                    , collapse = "logical"
                                    )
          )

          
#' A class to represent a preprocessing method.
#' 
#' It extends \code{gtMethod} class.
#' 
#' @name ppMethod-class
#' @aliases ppMethod
#' @examples 
#'  \dontrun{
#'      gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#'      ppMethod(gt, '3', '4')
#' }     
setClass("ppMethod", contains = "gtMethod")


#' A class to represent a reference gating method.
#' 
#' It extends \code{gtMethod} class.
#' 
#' @section Slots:
#' \describe{ 
#'  \item{refNodes}{ \code{character} specifying the reference nodes}
#' }
#' @name refGate-class      
setClass("refGate", contains = "gtMethod", representation(refNodes = "character"))

#' A class to represent a dummy gating method that does nothing but serves as reference to be refered by other population
#' 
#' It is generated automatically by the csv template preprocessing to handle the gating function that returns multiple gates. 
setClass("dummyMethod", contains = "refGate")


#' A class to represent a boolean gating method.
#' 
#' It extends \code{refGate} class.
#' @name boolMethod-class
setClass("boolMethod", contains = "refGate")

#' A class to represent a polyFunctions gating method.
#' 
#' It extends \code{boolMethod} class and will be expanded to multiple \code{boolMethod} object.
#' 
#' @name polyFunctions-class
setClass("polyFunctions", contains = "boolMethod")

#' A class to represent a cell population that will be generated by a gating method.
#' 
#' @section Slots: 
#' \describe{
#' 
#'  \item{id}{ \code{numeric} unique ID that is consistent with node label of graphNEL in gating template} 
#'  \item{name}{ \code{character} the name of population}
#'  \item{alias}{ \code{character} the more user friendly name of population}
#' }
#' @name gtPopulation-class
#' @examples 
#'  \dontrun{
#'      gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#'       
#'      gt_get_nodes(gt, '2')
#' }
setClass("gtPopulation", representation(id = "character", name = "character",
                                        alias = "character"
                                  )
                              )
                              
                              
#' A class representing a group of cell populations.
#' 
#' It extends \code{gtPopulation} class.
#' @name gtSubsets-class
setClass("gtSubsets", contains = "gtPopulation")

#' A function to tell wether a gating method is \code{polyFunctions}
#' 
#' @rdname isPolyfunctional
#' @param gm an object that extends \code{gtMethod}  
#' @noRd 
.isPolyfunctional <- function(gm) {
  # grepl('^\\[\\:.+\\:\\]$',x)
  class(gm) == "polyFunctions"
}

#' gating arguments parser
#' 
#' parsing the arguments read from `args` columns of csv template into 
#' list of paired arguments
#' @noRd 
.argParser <- function(txt, split = TRUE) {
  # trim whitespaces at beginning and the end of string
  txt <- gsub("^\\s*|\\s*$", "", txt)

  if (split) {
    paired_args <- paste("c(", txt, ")")
    paired_args <- try(parse(text = paired_args), silent = TRUE)
    if (class(paired_args) == "try-error") {
      errmsg <- attr(paired_args, "condition")
      msg <- conditionMessage(errmsg)
      stop("invalid gating argument:\n", msg)
    }
    
    paired_args <- as.list(as.list(paired_args)[[1]])[-1]
    names(paired_args) <- names(paired_args)
  } else {
    if(nchar(txt) >0){
      paired_args <- as.symbol(txt)
      paired_args <- list(paired_args)  
    }else
      stop("argument is empty!")
    
  }
  
  paired_args
}


 

#' @title
#' gatingTemplate constructor 
#' 
#' @description 
#' It parses the csv file that specifies the gating scheme for a particular staining pannel. 
#' 
#' @details 
#' This csv must have the following columns:
#' 
#' 'alias': a name used label the cell population, the path composed by the alias and its precedent nodes (e.g. /root/A/B/alias) has to be uniquely identifiable.
#'          So alias can not contain '/' character, which is reserved as path delimiter.
#'  
#' 'pop': population patterns of '+/-` or '+/-+/-', which tells the algorithm which side (postive or negative) of 1d gate or which quadrant of 2d gate to be kept.
#'                         
#' 'parent': the parent population alias, its path has to be uniquely identifiable.
#'  
#' 'dims': characters seperated by comma specifying the dimensions(1d or 2d) used for gating. 
#' It can be either channel name or stained marker name (or the substrings of channel/marker names as long as they are uniquely identifiable.).
#'  
#' 'gating_method': the name of the gating function (e.g. 'flowClust'). It is invoked by a wrapper function that has the identical function name prefixed with a dot.(e.g. '.flowClust')
#'     
#' 'gating_args': the named arguments passed to gating function (Note that double quotes are often used as text delimiter by some csv editors. So try to use single quote instead if needed.)
#'  
#' 'collapseDataForGating': When TRUE, data is collapsed (within groups if 'groupBy' specified) before gating and the gate is replicated across collapsed samples.
#'  When set FALSE (or blank),then 'groupBy' argument is only used by 'preprocessing' and ignored by gating.
#'    
#' 'groupBy': If given, samples are split into groups by the unique combinations of study variable (i.e. column names of pData,e.g."PTID:VISITNO").
#'  when split is numeric, then samples are grouped by every N samples 
#' 
#' 'preprocessing_method': the name of the preprocessing function(e.g. 'prior_flowclust'). It is invoked by a wrapper function that has the identical function name prefixed with a dot.(e.g. '.prior_flowclust')
#'  the preprocessing results are then passed to gating wrapper function through 'pps_res' argument.
#'       
#' 'preprocessing_args': the named arguments passed to preprocessing function.
#' 
#'     
#' @rdname gatingTemplate-class 
#' @examples
#' \dontrun{ 
#'   gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#'   plot(gt)
#' }
#' 
#' @export
setGeneric("gatingTemplate", function(x, ...) standardGeneric("gatingTemplate"))

#' @aliases gatingTemplate,character-method
#' @param x \code{character} csv file name or a \code{data.table}
#' @param name \code{character} the label of the gating template
#' @param strict \code{logical} whether to perform validity check(special characters) on the alias column. By default it is(and should be) turned on for the regular template parsing. 
#'                              But sometime it is useful to turned it off to bypass the check for the dummy nodes(e.g. the csv template generated by 'gh_generate_template'  with some existing boolean gates that has '!' or ':' symbol).
#' @param strip_extra_quotes \code{logical} Extra quotes are added to strings by fread. This causes problems with parsing R strings to expressions in some cases. Default FALSE for usual behaviour. TRUE should be passed if parsing gating_args fails.
#' @param ... other arguments passed to \code{data.table::fread}
#' @importFrom graph graphNEL addEdge nodes edges nodeDataDefaults nodeData edgeDataDefaults addEdge edgeData subGraph
#' @rdname gatingTemplate-class
setMethod("gatingTemplate", signature(x = "character"), function(x, name = "default", strict = TRUE, strip_extra_quotes=FALSE,...) {
      dt <- fread(x, ...)
      # empty gatingTemplate error
      if(nrow(dt) == 0) {
        stop(
          paste0(
            "Cannot create gatingTemplate from ",
            x, 
            " as it contains no gating entries."
          )
        )
      }
      dt <- .preprocess_csv(dt, strict = strict)
      
      #append the isMultiPops column based on pop name
      dt[, isMultiPops := FALSE]
      dt[pop == "*", isMultiPops := TRUE]
	if(strip_extra_quotes)
      		{
			dt[,gating_args:=gsub("\"\"","\"",gating_args)]
		}
      .gatingTemplate(dt, name = name)
    })

#' @aliases gatingTemplate,data.table-method
#' @rdname gatingTemplate-class
setMethod(
  "gatingTemplate", 
  signature(x = "data.table"), 
  function(x, 
           name = "default", 
           strict = TRUE, 
           strip_extra_quotes = FALSE,
           ...) {
    if(nrow(x) == 0) {
      stop(
        paste0(
          "Cannot create gatingTemplate from this data.table as it ",
          "contains no gating entries."
        )
      )
    }
    dt <- .preprocess_csv(x, strict = strict)
    #append the isMultiPops column based on pop name
    dt[, isMultiPops := FALSE]
    dt[pop == "*", isMultiPops := TRUE]
    if(strip_extra_quotes)
    {
      dt[,gating_args:=gsub("\"\"","\"",gating_args)]
    }
    .gatingTemplate(dt, name = name)
  }
)

#' @importFrom graph nodeDataDefaults<- edgeDataDefaults<- nodeData<- edgeData<-
#' @noRd 
.gatingTemplate <- function(dt, name = "default"){  
#  browser()
  # create graph with root node
#  browser()
  g <- graphNEL(nodes = "root", edgemode = "directed")
  g <- as(g, "gatingTemplate")
  nodeDataDefaults(g, "pop") <- ""
  edgeDataDefaults(g, "gtMethod") <- ""
  edgeDataDefaults(g, "ppMethod") <- ""
  edgeDataDefaults(g, "isReference") <- FALSE
  
  # add default root
  nodeData(g, "root", "pop") <- new("gtPopulation", id = "root", name = "root", alias = "root")

  # parse each row
  nEdges <- nrow(dt)
  edgs <- vector("list", nEdges)
  for (i in 1:nEdges) {
    
    thisRow <- dt[i,]
    # extract info from dataframe
    parent <- thisRow[,parent][[1]]
    
    # get parent ID
    
    curPop <- thisRow[,alias][[1]]
    
    if(grepl("/", curPop))
      stop("Population name(or alias) '", curPop , "' contains '/', which is reserved as gating path delimiter!")
    
    curNodePath <- paste(parent, curPop, sep = "/")
    curNodePath <- sub("root", "", curNodePath)
    curPopName <- thisRow[, pop][[1]]
    isMultiPops <- thisRow[, isMultiPops]
    #try to split alias for the gating function that returns multi-pops
    if(isMultiPops)
        curPop <- trimws(unlist(strsplit(split = ",", curPop)))
    # create pop object
    curNode <- new("gtPopulation", id = curNodePath, name = curPopName, 
                        alias = curPop
                )
                      
    # create gating method object
    cur_method <- thisRow[, gating_method][[1]]
    cur_args <- thisRow[, gating_args][[1]]
    cur_dims <- thisRow[, dims][[1]]
    
    cur_collapse <- thisRow[, collapseDataForGating][[1]]
    if(cur_collapse == "")
      cur_collapse <- FALSE
    cur_collapse <- as.logical(cur_collapse)
    
    if(is.na(cur_collapse))
      stop("Invalid `collapseDataForGating` flag!")
    
    cur_groupBy <- thisRow[, groupBy][[1]]
    cur_method_name_pattern <- paste0("^", cur_method , "$")
    # do not parse args for refGate-like gate since they might break the current
    # parse due to the +/- | &,! symbols
    if (any(grepl(cur_method_name_pattern,  c("boolGate", "polyfunctions", "refGate", "dummy_gate"), ignore.case = TRUE))) {
      split_args <- FALSE
    } else {
      split_args <- TRUE
    }
    
    cur_args <- .argParser(cur_args, split_args)
    
    gm <- new("gtMethod"
                , name = cur_method
                , dims = cur_dims
                , args = cur_args
                , collapse = cur_collapse
                , groupBy = cur_groupBy
              )
    # specialize gtMethod as needed
    if (grepl(cur_method_name_pattern , "boolGate", ignore.case = TRUE)) {
      gm <- as(gm, "boolMethod")
    } else if (grepl(cur_method_name_pattern , "polyFunctions", ignore.case = TRUE)) {
      gm <- as(gm, "polyFunctions")
    } else if (grepl(cur_method_name_pattern , "refGate", ignore.case = TRUE)) {
      gm <- as(gm, "refGate")
      if(nchar(cur_dims) == 0){
        stop("No dimensions defined for refGate!")
      }
    }
    if (grepl(cur_method_name_pattern , "dummy_gate", ignore.case = TRUE)) 
      gm <- as(gm, "dummyMethod")
    
    #preprocessing object
    cur_pp_Method <- thisRow[, preprocessing_method][[1]]
    cur_pp_args <- thisRow[, preprocessing_args][[1]]
    cur_pp_args <- .argParser(cur_pp_args, TRUE)
    
    if(nchar(cur_pp_Method) > 0)
      ppm <- new("ppMethod"
                  , name = cur_pp_Method
                  , dims = cur_dims
                  , args = cur_pp_args
                  , collapse = cur_collapse
                  , groupBy = cur_groupBy
                )
    
    message("Adding population:", basename(curNodePath))
#    browser()
    # add current node to graph
    g_updated <- graph::addNode(curNodePath, g)
    
#    if (!extends(class(gm), "refGate")) {
      # add edge from parent
    g_updated <- addEdge(parent, curNodePath, g_updated)
    
    #add preprcessing method to the edge
    if(nchar(cur_pp_Method) > 0)
      edgeData(g_updated, parent, curNodePath, "ppMethod") <- ppm
    
    ##########################################
    # refGate-like methods need extra parsing
    ##########################################
    if (extends(class(gm), "refGate")) {
      
      # get argument
      args <- gm@args[[1]]
      args <- deparse(args)
      
      # parsing reference nodes
      if (class(gm) == "boolMethod") {
        # strip ! symbols
        args <- gsub("!", "", args)
        # split by logical operator when regular boolean gates
        refNodes <- strsplit(args, "&|\\|")[[1]]
      } else {
        # split by colon for refGate or polyfunctional boolean gates
        refNodes <- strsplit(args, "\\:")[[1]]
      }
      
      # update refNodes slot of gm object
      gm@refNodes <- refNodes
      
      # specialize the node type for polyfunctions
      if (class(gm) == "polyFunctions") {
#        browser()
        curNode <- as(curNode, "gtSubsets")
      }
      
      # add edges from reference nodes (only used for tsort)
      for (ref_node in refNodes) {
        
        # add the edge from it
        g_updated <- addEdge(.getFullPath(ref_node, dt), curNodePath, g_updated)

        # flag the edge 
        edgeData(g_updated, .getFullPath(ref_node, dt), curNodePath, "isReference") <- TRUE
      }
    }
    
    # attach the gm object to the parent edge
    edgeData(g_updated, parent, curNodePath, "gtMethod") <- gm
    # flag the edge 
    edgeData(g_updated, parent, curNodePath, "isReference") <- FALSE
    # add the current population object to the current node
    nodeData(g_updated, curNodePath, "pop") <- curNode
    # update graph
    g <- g_updated
  }
  
  g@name <- name
  g
}
 

#adapted from plyr::list_to_array
id_var<-
function (x, drop = FALSE) 
{
  if (length(x) == 0) 
    return(structure(integer(), n = 0L))
  if (!is.null(attr(x, "n")) && !drop) 
    return(x)
  if (is.factor(x) && !drop) {
    x <- addNA(x, ifany = TRUE)
    id <- as.integer(x)
    n <- length(levels(x))
  }
  else {
    levels <- sort(unique(x), na.last = TRUE)
    id <- match(x, levels)
    n <- max(id)
  }
  structure(id, n = n)
}
id <-
function (.variables, drop = FALSE) 
{
  lengths <- vapply(.variables, length, integer(1))
  .variables <- .variables[lengths != 0]
  if (length(.variables) == 0) {
    n <- nrow(.variables) %||% 0L
    return(structure(seq_len(n), n = n))
  }
  if (length(.variables) == 1) {
    return(id_var(.variables[[1]], drop = drop))
  }
  ids <- rev(lapply(.variables, id_var, drop = drop))
  p <- length(ids)
  ndistinct <- vapply(ids, attr, "n", FUN.VALUE = numeric(1), 
                      USE.NAMES = FALSE)
  n <- prod(ndistinct)
  if (n > 2^31) {
    char_id <- do.call("paste", c(ids, sep = "\r"))
    res <- match(char_id, unique(char_id))
  }
  else {
    combs <- c(1, cumprod(ndistinct[-p]))
    mat <- do.call("cbind", ids)
    res <- c((mat - 1L) %*% combs + 1L)
  }
  attr(res, "n") <- n
  if (drop) {
    id_var(res, drop = TRUE)
  }
  else {
    structure(as.integer(res), n = attr(res, "n"))
  }
}
.dims <- function (x) {length(amv_dim(x))}

amv_dim <- function (x) {
  if (is.vector(x)) length(x) else dim(x)}

amv_dimnames <-function (x) 
{
  d <- if (is.vector(x)) 
    list(names(x))
  else dimnames(x)
  if (is.null(d)) 
    d <- rep(list(NULL), .dims(x))
  null_names <- which(unlist(lapply(d, is.null)))
  d[null_names] <- lapply(null_names, function(i) seq_len(amv_dim(x)[i]))
  d
}
list_to_array <- function (res, labels = NULL) 
  {
    if (length(res) == 0) 
      return(vector())
    n <- length(res)
    # atomic <- sapply(res, is.atomic)
    # if (all(atomic) || all(!atomic)) {
      # dlength <- unique.default(llply(res, .dims))
      # if (length(dlength) != 1)
      #   stop("Results must have the same number of dimensions.")
      # dims <- unique(do.call("rbind", llply(res, amv_dim)))
      # if (is.null(dims))
      #   stop("Results must have one or more dimensions.",
      #        call. = FALSE)
      # if (nrow(dims) != 1)
      #   stop("Results must have the same dimensions.", call. = FALSE)
      res_dim <- amv_dim(res[[1]])
      res_labels <- amv_dimnames(res[[1]])
      if (any(vapply(res_labels, anyDuplicated, integer(1)) != 
              0)) {
        warning("Duplicate names not supported.", call. = FALSE)
      }
      res_index <- expand.grid(res_labels)
      res <- unlist(res, use.names = FALSE, recursive = FALSE)
    # }
    # else {
    #   stop("Results must have compatible types.")
    # }
    if (is.null(labels)) {
      labels <- data.frame(X = seq_len(n))
      in_labels <- list(NULL)
      in_dim <- n
    }
    else {
      in_labels <- lapply(labels, function(x) if (is.factor(x)) 
        levels(x)
        else sort(unique(x), na.last = TRUE))
      in_dim <- sapply(in_labels, length)
    }
    index_old <- rep(id(rev(labels)), each = nrow(res_index))
    index_new <- rep(id(rev(res_index)), nrow(labels))
    index <- (index_new - 1) * prod(in_dim) + index_old
    out_dim <- unname(c(in_dim, res_dim))
    out_labels <- c(in_labels, res_labels)
    n <- prod(out_dim)
    if (length(index) < n) {
      overall <- match(1:n, index, nomatch = NA)
    }
    else {
      overall <- order(index)
    }
    out_array <- res[overall]
    dim(out_array) <- out_dim
    dimnames(out_array) <- out_labels
    out_array
  }


.aaply <- function(x, margin, FUNC){
  res <- unlist(apply(x, margin, function(i) {
    
    list(FUNC(i))
  }), recursive = F) 
  list_to_array(res)
}

#' show method for gtMethod
#' 
#' @param object \code{gtMethod}
#' show method for gtMethod
setMethod("show", signature = c("gtMethod"), definition = function(object) {
  # cat('Gating Method: ')
  cat(names(object))
  cat("(")
  chnls <- dims(object)
  cat(paste(paste(names(chnls), chnls, sep = "="), collapse = ","))
  cat(") \n")
#  cat(parameters(object))
})

#' show method for boolMethod
#' 
#' show method for boolMethod
#' @param object \code{boolMethod}
setMethod("show", signature = c("boolMethod"), definition = function(object) {
  cat(paste(class(object), "(", parameters(object), ")", sep = ""))
  cat("\n")
})

#' get gating method name
#' 
#' @param x \code{gtMethod} 
#' @aliases names,gtMethod-method
#' @examples 
#' \dontrun{
#' gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
#' 
#' gtMthd <- gt_get_gate(gt, "/nonDebris/singlets",  "/nonDebris/singlets/lymph")
#' names(gtMthd) 
#' dims(gtMthd)
#' parameters(gtMthd)
#' isCollapse(gtMthd)
#' groupBy(gtMthd)
#' 
#' gtPop <- gt_get_nodes(gt, "/nonDebris/singlets/lymph/cd3/cd4+cd8-/CD38+")
#' names(gtPop)
#' alias(gtPop)
#' }
setMethod("names", signature = c("gtMethod"), definition = function(x) {
  x@name
})

#' get gating method dimensions
#' 
#' @param x \code{gtMethod}
#' @importFrom BiocGenerics dims
#' @aliases dims,gtMethod-method
setMethod("dims", signature = c("gtMethod"), definition = function(x) {
  dims <- trimws(strsplit(x@dims, ",")[[1]])
#  if (length(dims) == 1) 
#    dims <- c(NA, dims) else if (length(dims) != 2) 
#    stop("invalid dimensions!")
#  names(dims) <- c("xChannel", "yChannel")
  dims
})

#' get parameters of the gating method/function
#' 
#' @param object \code{gtMethod}
#' @importFrom flowCore parameters
#' @aliases parameters,gtMethod-method
setMethod("parameters", signature = c("gtMethod"), definition = function(object) {
  object@args
})

setGeneric("isCollapse",function(object,...)standardGeneric("isCollapse"))

#' get the flag that determines whether gating method is applied on collapsed data 
#' 
#' When TRUE, the flow data(multiple flowFrames) is collapsed into one and the gating method is applied on the collapsed data.
#' Once the gate is generated, it is thenreplicated and applied to the each single flowFrame.
#'   
#' @return \code{logical} 
#' @param object \code{gtMethod} 
#' @aliases isCollapse,gtMethod-method
setMethod("isCollapse", signature = c("gtMethod"), definition = function(object) {
      object@collapse
    })

setGeneric("groupBy",function(object,...)standardGeneric("groupBy"))
#' get the grouping variable for the gating method
#' 
#' When specified, the flow data is grouped by the grouping variable (column names in pData).
#' Within each group, when \code{isCollapse} is set to TRUE, the gating method is applied to the collapsed data.
#' Otherwise, it is done indepentently for each indiviudal sample(flowFrame).
#' Grouping variable is also used by \code{preprocessing method}.
#' 
#' @param object \code{gtMethod}
#' 
#' @aliases groupBy,gtMethod-method
setMethod("groupBy", signature = c("gtMethod"), definition = function(object) {
      object@groupBy
    })

#' @templateVar old templateGen
#' @templateVar new gh_generate_template
#' @template template-depr_pkg
NULL
#' generate a partially complete csv template from the existing gating hierarchy 
#' 
#' To ease the process of replicating the existing (usually a manual one) gating schemes, 
#' this function populate an empty gating template with the 'alias', 'pop', 'parent' and 'dims' 
#' columns that exacted from an \code{GatingHierarchy}, and leave the other columns (e.g. `gating_method`) blank.
#' So users can make changes to that template instead of writing from scratch.
#' 
#' @name gh_generate_template
#' @aliases templateGen 
#' @param gh a \code{GatingHierarchy} likely parsed from a xml workspace
#' @return a gating template in \code{data.frame} format that requires further edition after output to csv
#' @examples 
#' library(flowWorkspace)
#' dataDir <- system.file("extdata",package="flowWorkspaceData")
#' gs <- load_gs(list.files(dataDir, pattern = "gs_manual",full = TRUE))
#' gh_generate_template(gs[[1]])
#' @export 
gh_generate_template <- function(gh){
  nodes <- gs_get_pop_paths(gh, order = "tsort")
  dt = do.call(rbind, lapply(nodes[-1], function(thisNode){
        thisGate <- gh_pop_get_gate(gh, thisNode)
        dims <- paste(as.vector(parameters(thisGate)), collapse = ",")
        parent <- gs_pop_get_parent(gh, thisNode)
        alias <- basename(thisNode)
        pop <- alias
        data.frame(alias = alias
            , pop = "+"
            , parent = parent
            , dims = dims
            , gating_method = NA_character_
            , gating_args = NA_character_
            , collapseDataForGating = NA
            , groupBy = NA_character_
            , preprocessing_method = NA_character_
            , preprocessing_args = NA_character_
        )
      })
  )
  if(is.null(dt)||ncol(dt)==0){
    cn = c("alias","pop","parent","dims","gating_method","gating_args","collapseDataForGating","groupBy","preprocessing_method","preprocessing_args")
    dt = vector('list',10)
    names(dt) = cn
    dt = as.data.table(dt)
  }
  return(dt)
}

#' @export
templateGen <- function(gh){
  .Deprecated("gh_generate_template")
}

#' prepend all ancester nodes to construct the full path for the given node 
#' 
#' @param ref_node \code{character} the node to work with
#' @param dt \code{data.table} that stores the current preprocessed template
#' @noRd 
.getFullPath <- function(ref_node, dt){
  
  ref_node <- trimws(ref_node)
  #prepend root if start with /
  if(substr(ref_node, 1, 1) == "/")
    ref_node <- paste0("root", ref_node)
  
  tokens <- strsplit(ref_node, "/")[[1]]
  res_path <- NULL
  
#browser()
  dt_toSearch <- dt
  # start to match the tokens in the path
  while (length(tokens) > 0) {
    #pop the current one
    curToken <- tokens[1]
    tokens <- tokens[-1]
    if(curToken == "root")
      res_path <- c(res_path, "root")
    else{
      # toMatch <- gsub("\\+", "\\\\\\+", curToken)
      toMatch <- paste0("\\Q",curToken,"\\E")
      toMatch <- paste0("^",toMatch,"$")
      ind <- grep(toMatch, dt_toSearch[, alias])
      if(length(ind) == 0)
        stop("Not able to to find reference to: ", curToken)
      else if(length(ind) > 1)
        stop("Non-unique reference to: ", curToken)
      else{
        # prepend its parent to make it full path
        thisParent <- dt_toSearch[ind, parent]
        if(thisParent == "root")
          curToken <- paste0("/", curToken)
        else
          curToken <- paste(thisParent, curToken, sep = "/")
        
        #only save the full path of the first token 
        if(is.null(res_path))
          toSave <- curToken
        else
          toSave <- basename(curToken)
        
        res_path <- c(res_path, toSave)
      }
    }
#    browser()
    #subset the data frame by parent
    dt_toSearch <- dt[parent == curToken, ]
    
  }
  
  
  res_path <- paste(res_path, collapse = "/")
#  browser()    
  if(res_path == "root")
    res_path <- "root"
  else
    res_path <- sub("root", "", res_path)
  res_path
  
}

#' validity check of alias column
#' @noRd 
.validity_check_alias <- function(alias){
  if(grepl("[\\|\\&|\\:|\\/]", alias))
    stop(alias , "contains illegal character: |,&,:,/")
  
}
#' check if alias is unique within template
#' 
#' @param new_dt \code{data.tab;e} contains the preprocessed rows
#' @param alias \code{character} the alias to be checked
#' @param this_parent \code{character} the full gating path of the parent node  
#' @return NULL when pass the check, otherwise, throw the error
#' @noRd 
.unique_check_alias <- function(new_dt, alias, this_parent) {
  #skip the unexpanded multi-pop pattern 
  #since it is not supposed to be refered 
  if(alias == "*")
    return(NULL)
  siblings <- new_dt[parent == this_parent, alias]
  toMatch <- gsub("\\+", "\\\\\\+", alias)
  toMatch <- paste0("^",toMatch,"$")
  
  matched_sibs <- grep(toMatch, siblings)
  
  if (length(matched_sibs) > 0) {
    stop(alias, " is not unique within ", this_parent)
  }
}

#' split the rows that has multiple parents
#' @noRd 
.split_multi_parents <- function(dt)
{
  #can't use apply since it will coerce dt to df and prepend extra space to int as it convert the column to character
  res<- lapply(1:nrow(dt), function(i){
    
    row <- dt[i,]
    parent <- row[, parent]
    if(!grepl(",", parent))
      row
    else
    {
      message("splitting the row that has multiple parents: '", parent, "'")
      rbindlist(
        lapply(strsplit(parent, split = ",")[[1]], function(p){
          r <- copy(row)
          r[, parent := p]
          r
        })
      ) 
    }
     
  })
 
  rbindlist(res)
}

#' preprocess the csv template
#' 
#' It parses the data table sequentially and does the valdidity checking and expansion row by row.
#'  
#' It expands the definition of gates or construct reference gates when necessary
#' @param dt \code{data.table} loaded directly from csv gating template
#' @param strict whether validity check for pop alias. It is turned on in the normal template parsing. when used with add_pop it is turned off to bypass the check on some existing boolean gates (that has ! : symbols).  
#' @return a preprocessed(expanded when applicable) \code{data.frame}
#' @import data.table
#' @noRd 
.preprocess_csv <- function(dt, strict = TRUE) {
  dt <- .split_multi_parents(dt)
  #only parse these columns(other columns may be used by user for other purpose e.g. comments)
  dt <- dt[, list(alias
          , pop
          , parent
          , dims
          , gating_method
          , gating_args
          , collapseDataForGating
          , groupBy
          , preprocessing_method
          , preprocessing_args
      )
  ]  
  
  #coerce most columns to character 
  #because data.table tends to read the empty columns as logical NAs which causes trouble in parsing later on
  
  for(col in colnames(dt))
  {
    if(col != "collapseDataForGating")
    {
      newCol <- dt[[col]]
      dt[, (col):= as.character(newCol)]
    }
  }
  
  new_dt <- dt[0, ]
  
  for (i in 1:nrow(dt)) {
    this_row <- dt[i, , drop = FALSE]
    if(strict)
      .validity_check_alias(this_row[, alias])
    #update parent with full path
    this_row[, parent := .getFullPath(this_row[, parent], new_dt)]
    #preprocess the current row
    res <- .preprocess_row(this_row, strict = strict)
    
    #check the uniqueness of alias within res
    aliasVec <- res[,alias]
    dupVec <- duplicated(aliasVec)
    if(any(dupVec))
      stop(aliasVec[dupVec], "is duplicated!")
    
    #validity check of alias within the context of new_df
    apply(res, 1, function(row){
          thisAlias <- row[["alias"]]
          if(strict)
            .validity_check_alias(thisAlias)
          .unique_check_alias(new_dt, thisAlias, row[["parent"]])
        })
    
    new_dt <- rbindlist(list(new_dt, res)) 
  }
#  browser()
  
  new_dt[is.na(new_dt)] <- ""
  new_dt
  
}    

#' The actual preprocessing logic for each entry of template
#' 
#' Here are the major preprocessing tasks:
#' 1. validity check for 'alias' (special character and uniqueness check)
#' 2. validity check for the numer of parameters('dim' column)
#' 3. dispatch 'flowClust' method to either 'gate_flowclust_1d' or 'gate_flowclust_2d' based on the 'pop' and 'dims' columns
#' 4. expand the single row to multiple rows when applicable. There are basically two types of expansion:
#'      4.1. expand to two 1d gates and one rectangelGate when 'pop' name is defined as quadrant pattern (e.g. "A+B+")  and 'gating_method' is not "refGate"
#'      4.2. expand to multiple gates when 'pop' is defined with '+/-' (e.g. "A+/-" or "A+/-B+/-") 
#' 
#' @param this_row a single-row \code{data.table}
#' @return  \code{data.table}
#' @noRd 
.preprocess_row <- function(this_row, strict = TRUE){
  #make sure it doesn't tamper the input
  this_row <- copy(this_row)
  
  alias <- this_row[1, alias]
  popName <- this_row[1, pop]
  dims <- this_row[1, dims]
  gm <- this_row[1, gating_method]
  
  if(gm == "dummy") #skip dummy since it doesn't need any processing
    return(this_row)
  
  if(popName == "*" ){
    if(alias == "*"|| dims == "")
      return(this_row)
    else
      return(.gen_dummy_ref_gate(this_row))
    }
    
  if(isTRUE(getOption("openCyto")[["check.pop"]])&&!grepl("^(([\\+-])|(\\+/-)|(-/\\+)){1,2}$",popName))
    stop("'pop': ", popName, " should only contain the + and - symbols and must conform to the valid +/- or +/-+/- patterns!
			\n Please remove any other letters and correct the pop pattern!
			\n Or type ?openCyto.options to see how you can turn off the 'check.pop' flag in options('openCyto') to bypass this validiy check (Not recommended)")
  
  dim_count <- length(strsplit(split = ",", dims)[[1]])
  if (!dim_count %in% c(1, 2)) {
    if (!(dim_count == 0 && gm %in% c("polyFunctions", "boolGate", "refGate"))) {
      stop(popName, " has invalid number of dimensions: ", dim_count)
    }
  }
  
  one_pop_token <- "[\\+-]"
  pop_name_pat <- "[^\\+-]*" #now we allow empty pop name with pure +/- signs  "[^\\+-]+"
  one_pop_pat <- paste(pop_name_pat, one_pop_token, sep = "")
  
  two_pop_token <- "(\\+/-)|(-/\\+)"
  two_pop_pat <- paste(pop_name_pat, "(", two_pop_token, ")", sep = "")
#    browser()
  if (grepl(paste0("^", one_pop_pat, "$"), popName)) {
    # A+ no expansion(simply update flowClust gm)
    if (gm %in% c("flowClust", "gate_flowclust")) {
      if (dim_count == 1) {
        this_row[1, gating_method := "gate_flowclust_1d"] 
      } else {
        this_row[1, gating_method := "gate_flowclust_2d"] 
      }
    }
    res <- this_row
    #strip the other letters from legacy popName
    res[, pop := ifelse(grepl("-$", popName), "-", "+")]
    
  } else if (grepl(paste("^", two_pop_pat, "$", sep = ""), popName)) {
    # A+/-
    
    if (gm %in% c("flowClust", "gate_flowclust")) {
      if (dim_count == 1) {
        this_row[1, gating_method := "gate_flowclust_1d"]
      } else {
        this_row[1, gating_method := "gate_flowclust_2d"]
      }
    }
    # expand to two rows
    new_pops <- c("+", "-")
    
    if(dim_count == 1){
      new_alias <- paste(dims, new_pops, sep = "")  
      if(!trimws(alias) %in% c("", "*"))
        message("alias '", alias, "' is ignored since pop names are auto-generated from 'dims' when pop = '+/-' ")
    }else if(dim_count == 2)
    {
      
      if(alias == "*")
        stop("Please provide a proper alias name because dims(", dims, ") can't be used to derive the expanded pops!")
      new_alias <- paste(alias, new_pops, sep = "")  
      
    }else
      stop("Don't know how to handle ", popName, " for multi-dimensional gating! ")
    
    message("expanding pop: ", popName, " to ", paste(new_alias, collapse = "/"), "\n")
    
    
    
    res <- rbindlist(list(this_row, this_row))
    
    # update 1d gate
    res[1, alias := new_alias[1]]
    res[1, pop := new_pops[1]]
    
    
    # update ref gate
    refNode <- file.path(this_row[, parent], new_alias[1])
    res[2, alias := new_alias[2]]
    res[2, pop := new_pops[2]]
    res[2, gating_method := "refGate"]
    res[2, gating_args := refNode]
    res[2, preprocessing_method := ""]
    res[2, preprocessing_args := ""]
    
  } else if (grepl(paste("^(", one_pop_pat, "){2}$", sep = ""), popName)) {
    # A+B+
    split_terms <- .splitTerms(pop_pat = one_pop_pat, two_pop_token, popName, dims)
    new_pops <- as.character(outer(split_terms[[1]][["pop"]], split_terms[[2]][["pop"]], paste,sep = ""))
    new_alias <- as.character(outer(split_terms[[1]][["alias"]], split_terms[[2]][["alias"]], paste,sep = ""))
    
    if (gm == "refGate") {
      # no expansion
      res <- this_row
      res[, pop := new_pops]
    } else {
      if(gm %in% c("quadGate.seq", "gate_quad_sequential", "quadGate.tmix", "gate_quad_tmix")){
        if(!trimws(alias) %in% c("", "*"))
          message("alias '", alias, "' is ignored since pop names are auto-generated from 'dims' for gate_quad_ methods")
        this_row[1, alias := "*"]
        res <- this_row
      }else{
        if (gm %in% c("flowClust", "gate_flowclust")) {
          if (dim_count == 2) {
            message("expanding pop: ", popName, "\n")
            
            this_row[1, gating_method := "gate_flowclust_1d"]
          } else {
            stop("dimensions '", dims, "' is not consistent with pop name '", 
                 popName, "'")
          }
        }
        # needs to be split into two 1d gates and one refgate
        
        # create 1d gate for each dim
        res_1d <- .gen_1dgate(this_row, strict = strict)
        res_ref <- .gen_refGate(new_pops, this_row, ref_nodes = res_1d[, alias], alias = this_row[, alias], strict = strict)
        res <- rbindlist(list(res_1d, res_ref)) 
      }
    }
  } else if (grepl(paste0("^(", two_pop_pat, "){2}$"), popName) ||
      grepl(paste0("^", two_pop_pat, one_pop_pat, "$"), popName) ||
      grepl(paste0("^", one_pop_pat, two_pop_pat, "$"), popName)) {
    # A+/-B+/-
    message("expanding pop: ", popName, "\n")
    two_or_one_pop_pat <- paste0("(", two_pop_pat, ")|(", one_pop_pat, ")")
    split_terms <- .splitTerms(pop_pat = two_or_one_pop_pat, two_pop_token, popName, dims)
    new_pops <- as.character(outer(split_terms[[1]][["pop"]], split_terms[[2]][["pop"]], paste,sep = ""))
    new_alias <- as.character(outer(split_terms[[1]][["alias"]], split_terms[[2]][["alias"]], paste,sep = ""))
    if (gm == "refGate") {
      res <- .gen_refGate(new_pops, this_row = this_row, alias = new_alias, strict = strict)
      
    } else {
      if(gm %in% c("quadGate.seq", "gate_quad_sequential", "quadGate.tmix", "gate_quad_tmix")){
        if(!trimws(alias) %in% c("", "*"))
          message("alias '", alias, "' is ignored since pop names are auto-generated from 'dims' gate_quad_ methods")
        this_row[1, alias := "*"]
        res <- this_row
      }else{
        if (gm %in% c("flowClust", "gate_flowclust")) {
          if (dim_count == 2) {
            this_row[1, gating_method := "gate_flowclust_1d"]
            
          } else {
            stop("dimensions '", dims, "' is not consistent with pop name '", 
                 popName, "'")
          }
        }
        
        # create 1d gate for each dim
        res_1d <- .gen_1dgate(this_row, strict = strict)
        res_ref <- .gen_refGate(new_pops, this_row, ref_nodes = res_1d[, alias], alias = new_alias, strict = strict)
        res <- rbindlist(list(res_1d, res_ref))
      }
    }
    
  } else 
    stop("invalid population pattern '", popName, "'")
  res
}


#' split the population pattern into multiple population names
#' 
#' currently only works for A+/-B+/- .
#' TODO:  support A+/-  
#' @noRd 
.splitTerms <- function(pop_pat, two_pop_token, popName, dims) {
  dims_vec <- trimws(strsplit(split = ",", dims)[[1]])
  if(length(dims_vec)!=2)
    stop("Can't split the population pattern '", popName, "' into multiple population names due to invalid 'dims' column: '", dims, "'" )
  term_pos <- gregexpr(pop_pat, popName)[[1]]
  x_term <- substr(popName, 1, term_pos[2] - 1)
  y_term <- substr(popName, term_pos[2], nchar(popName))
  terms <- c(x_term, y_term)
  
  splitted_terms <- mapply(terms, dims_vec, FUN = function(cur_term, dim_name) {
        token_pos <- gregexpr(two_pop_token, cur_term)[[1]]
        if (token_pos > 0) {
          # dim_name <- substr(cur_term, 1, token_pos - 1)
          splitted_term <- list(alias = paste(dim_name, c("+", "-"), sep = "")
                                , pop = c("+", "-")
                            )
          
        } else {
          nlen <- nchar(cur_term)
          cur_token <- substr(cur_term, nlen, nlen)
          splitted_term <- list(alias = paste0(dim_name, cur_token)
                                , pop = cur_token
                            )
        }
        splitted_term
      }, SIMPLIFY = FALSE)
  splitted_terms
}

#' generate some dummy rows that just serves as reference without any gating method associated
#' @noRd 
.gen_dummy_ref_gate <- function(this_row){
  
  alias <- this_row[, alias]
  
  refNode <- file.path(this_row[, parent], alias)
  
  pops <- trimws(unlist(strsplit(split = ",", alias)))
  new_rows <- lapply(pops, function(thisPop){
                          dummy_row <- copy(this_row)
                          dummy_row[, alias := thisPop]
                          dummy_row[, pop := ""]
                          dummy_row[, gating_method := "dummy_gate"]
                          dummy_row[, gating_args := refNode]
                          dummy_row[, collapseDataForGating := ""]
                          dummy_row[, groupBy := ""]
                          dummy_row[, preprocessing_method := ""]
                          dummy_row[, preprocessing_args := ""]
                          dummy_row
                        })
  
  rbindlist(list(this_row,rbindlist(new_rows)))    
}
#' convert to 1d gating based on the population pattern (A+/-B+/-)
#' @noRd 
.gen_1dgate <- function(this_row, strict = TRUE) {
  
  dims.dims <- trimws(strsplit(split = ",", this_row[, dims])[[1]])
  
  res <- do.call(rbind, lapply(dims.dims, function(cur_dim) {
        new_pop_name <- paste(cur_dim, "+", sep = "")
        this_parent <- this_row[, parent]
        if(strict)
          .validity_check_alias(new_pop_name)
        
        data.table(alias = new_pop_name
            , pop = "+"
            , parent = this_parent
            , dims = cur_dim 
            , gating_method = this_row[, gating_method]
            , gating_args = this_row[, gating_args]
            , collapseDataForGating = this_row[, collapseDataForGating]
            , groupBy = this_row[, groupBy]
            , preprocessing_method = this_row[, preprocessing_method]
            , preprocessing_args = this_row[, preprocessing_args]
        )
      }))
#  rownames(res) <- NULL
  as.data.table(res)
}
#' generate reference gate based on the splitted population patterns(A+/-B+/-) 
#' @noRd 
.gen_refGate <- function(new_pops, this_row, ref_nodes = NULL, alias, strict = TRUE) {
  this_parent <- this_row[, parent]
  
  if (is.null(ref_nodes)) {
    # simply copy ref args from the row
    ref_args <- this_row[, gating_args]
  } else {
    # use the new generated 1d pops to construct ref args
    # prepend the path to ref_nodes 
    ref_nodes <- file.path(this_parent, ref_nodes)
    ref_args <- paste(ref_nodes, collapse = ":")
  }
  
  
  
  # create ref gate for each new_pop )
  do.call(rbind, mapply(new_pops, alias, FUN = function(new_pop, cur_alias) {
            if(strict)
              .validity_check_alias(cur_alias)
            data.table(alias = cur_alias
                , pop = new_pop
                , parent = this_parent
                , dims = this_row[, dims] 
                , gating_method = "refGate"
                , gating_args = ref_args
                , collapseDataForGating = ""
                , groupBy = ""
                , preprocessing_method = ""
                , preprocessing_args = ""
            )
          }, SIMPLIFY = FALSE))
}

context("Gating Template")

one_pop_token <- "[\\+-]"
pop_name_pat <- "[^\\+-]*"
one_pop_pat <- paste(pop_name_pat, one_pop_token, sep = "")

two_pop_token <- "(\\+/-)|(-/\\+)"
two_pop_pat <- paste(pop_name_pat, "(", two_pop_token, ")", sep = "")
two_or_one_pop_pat <- paste0("(", two_pop_pat, ")|(", one_pop_pat, ")")
#test_that("templateGen", {
#      
#      templateGen(gh)
#    })

dt <- fread(gtFile)


test_that(".preprocess_csv", {
      
      suppressMessages(
          preprocessed_dt <<- .preprocess_csv(dt)
      )
  # Now we deprecate tail gate, so update legacy result
      expectResults[["preprocess_csv"]][c(11:12,14:15),gating_method:="gate_mindensity"]
      expectResults[["preprocess_csv"]][c(11:12,14:15),gating_method:="gate_mindensity"]
      expectResults[["preprocess_csv"]][c(11:12,14:15),gating_args:=""]
      expect_equivalent(preprocessed_dt[,-2, with = F], expectResults[["preprocess_csv"]][,-2, with = F])
      
    })

#check.pop feature breaks this test
# test_that(".gatingTemplate", {
#       preprocessed_dt[, isMultiPops := FALSE]
#       
#       suppressMessages(thisRes <- .gatingTemplate(preprocessed_dt, name="default"))
#       expect_equal(thisRes[,-2, with = F], expectResults[["tcell"]][,-2, with = F])
#     })

test_that(".getFullPath", {
      ref_node <- "root"
      expect_equal(.getFullPath(ref_node, preprocessed_dt), "root")
      
      ref_node <- "cd3"
      expect_equal(.getFullPath(ref_node, preprocessed_dt), "/nonDebris/singlets/lymph/cd3")
      
      ref_node <- "lymph/cd3"
      expect_equal(.getFullPath(ref_node, preprocessed_dt), "/nonDebris/singlets/lymph/cd3")
      
      ref_node <- "/lymph/cd3"
      expect_error(.getFullPath(ref_node, preprocessed_dt), "Not able to to find reference")
      
      ref_node <- "cd31"
      expect_error(.getFullPath(ref_node, preprocessed_dt), "Not able to to find reference")
      
    })

#used by multiple test cases
template_row <- data.table(alias = "TH", pop = "c+-"
                          , parent = "cd3", dims = "cd4,cd8"
                          , gating_method = "mindensity", gating_args = ""
                          , collapseDataForGating = "TRUE", groupBy = "4"
                          , preprocessing_method = "", preprocessing_args = ""
                      )

test_that(".validity_check_alias", {
      
      #illegal character |,&,:,/
      errMsg <- "contains illegal character:"
      expect_error(.validity_check_alias("cd4+:"), errMsg)
      expect_error(.validity_check_alias("cd3/cd4+"), errMsg)
      expect_error(.validity_check_alias("cd8&cd4+"), errMsg)
      expect_error(.validity_check_alias("cd4+|cd8"), errMsg)
      
    })

test_that(".unique_check_alias", {
      
      
      parentPath <- "/nonDebris/singlets/lymph/cd3"
      
      expect_null(.unique_check_alias(preprocessed_dt, alias = "cd4", this_parent = parentPath))
      expect_error(.unique_check_alias(preprocessed_dt, alias = "cd4+", this_parent = parentPath),  "not unique")

      #multi-pops
      #with pop names specified
      this_row <- copy(template_row)
      this_row[, alias := "cd4,cd8"]
      this_row[, pop := "*"]
      
      expect_error(.unique_check_alias(this_row, alias = "cd4,cd8", this_parent = "cd3"),  "not unique")
      expect_null(.unique_check_alias(this_row, alias = "cd4", this_parent = "cd3"))
      
      #no pop names specified
      this_row[, alias := "*"]
      expect_null(.unique_check_alias(this_row, alias = "*", this_parent = "cd3"))
    })

test_that(".splitTerms", {
      
      dims <- "A,B"
      # A+B+
      expectRes <- list(list(alias ="A+", pop = "+")
                        , list(alias = "B+", pop = "+")
      )
      expect_equivalent(.splitTerms(one_pop_pat, two_pop_token, "A+B+", dims), expectRes)
      expect_equivalent(.splitTerms(one_pop_pat, two_pop_token, "++", dims), expectRes)
    
      #A+/-B+
      
      expect_equivalent(.splitTerms(two_or_one_pop_pat, two_pop_token, "A+B+", dims), expectRes)
    
      expectRes <- list(list(alias =c("A+", "A-"), pop = c("+", "-"))
                        , list(alias = "B+", pop = "+")
      )
      
      expect_equivalent(.splitTerms(two_or_one_pop_pat, two_pop_token, "A+/-B+", dims), expectRes)                      
    
      #+/-+  
      expect_equivalent(.splitTerms(two_or_one_pop_pat, two_pop_token, "+/-+", dims), expectRes)                      
    })

test_that(".gen_dummy_ref_gate", {
      
      this_row <- copy(template_row)
      this_row[, alias := "cd4,cd8"]
      this_row[, pop := "*"]
      
      dummy_row1 <- copy(this_row)
      dummy_row1[, alias := "cd4"]
      dummy_row1[, pop := ""]
      dummy_row1[, gating_method := "dummy_gate"]
      dummy_row1[, gating_args := file.path("cd3/cd4,cd8")]
      dummy_row1[, collapseDataForGating := ""]
      dummy_row1[, groupBy := ""]
      dummy_row1[, preprocessing_method := ""]
      dummy_row1[, preprocessing_args := ""]
      
      dummy_row2 <- copy(dummy_row1)
      dummy_row2[, alias := "cd8"]
      
      expectRes <- rbindlist(list(this_row, dummy_row1, dummy_row2)) 
      expect_equal(.gen_dummy_ref_gate(this_row), expectRes)
    })

test_that(".gen_1dgate", {
      
      #A+B+                            
      this_row <- copy(template_row)
      expectRes <- rbindlist(list(this_row, this_row))
      expectRes[1, alias := "cd4+"]
      expectRes[1, pop := "+"]
      expectRes[1, dims := "cd4"]
      expectRes[2, alias := "cd8+"]
      expectRes[2, pop := "+"]
      expectRes[2, dims := "cd8"]
      
      thisRes <- .gen_1dgate(this_row)
      expect_equal(thisRes, expectRes)
      
      #A-/B+/-
      this_row <- copy(template_row)
      this_row[, pop := "-+/-"]
      thisRes <- .gen_1dgate(this_row)
      expect_equal(thisRes, expectRes)
      
      #inconsistency between dims and pops
      this_row <- copy(template_row)
      this_row[, pop := "-+"]
      expect_equal(.gen_1dgate(this_row), expectRes)
    })

test_that(".gen_refGate", {
      
      
      #A+B+                            
      this_row <- copy(template_row)
      this_row[,  pop := "+-"]
      expectRes <- copy(this_row)
      
      expectRes[, gating_method := "refGate"]
      expectRes[, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[, collapseDataForGating := ""]
      expectRes[, groupBy := ""]
      
      thisRes <- .gen_refGate("+-", this_row, c("cd4+", "cd8+"), this_row[, alias])
      expect_equal(thisRes, expectRes)
      
      #A-/B+/-
      this_row <- copy(template_row)
      this_row[, pop := "-+/-"]
      expectRes <- rbindlist(list(this_row, this_row))
      expectRes[1, alias := "cd4-cd8+"]
      expectRes[1, pop := "-+"]
      expectRes[1, gating_method := "refGate"]
      expectRes[1, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[1, collapseDataForGating := ""]
      expectRes[1, groupBy := ""]
      expectRes[2, alias := "cd4-cd8-"]
      expectRes[2, pop := "--"]
      expectRes[2, gating_method := "refGate"]
      expectRes[2, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[2, collapseDataForGating := ""]
      expectRes[2, groupBy := ""]
      thisRes <- .gen_refGate(c("-+" ,"--"),alias=c("cd4-cd8+" ,"cd4-cd8-"), this_row, c("cd4+", "cd8+"))
      expect_equivalent(thisRes, expectRes)
      
      
    })

test_that(".preprocess_row", {
      
      #return as it is
      this_row <- copy(template_row)
      this_row[, pop := "+"]
      expectRes <- this_row
      
      
      #invalid pop name
      this_row <- copy(template_row)
      this_row[, pop := "++-"]
      expect_error(.preprocess_row(this_row), "pop")
      
      
      #A+/- 1d
      ########################
      this_row <- copy(template_row)
      this_row[, pop := "+/-"]
      this_row[, dims := "cd4"]
      
      expectRes <- rbindlist(list(this_row, this_row))
      expectRes[1, alias := "cd4+"]
      expectRes[1, pop := "+"]
      expectRes[2, alias := "cd4-"]
      expectRes[2, pop := "-"] 
      expectRes[2, gating_method := "refGate"]
      expectRes[2, gating_args := "cd3/cd4+"]
      
      expect_equal(suppressMessages(.preprocess_row(this_row)), expectRes)
      
      
      #A+/- 2d
      ########################
      this_row <- copy(template_row)
      this_row[, pop := "+/-"]
      
      expectRes <- rbindlist(list(this_row, this_row))
      expectRes[1, alias := "TH+"]
      expectRes[1, pop := "+"]
      expectRes[2, alias := "TH-"]
      expectRes[2, pop := "-"] 
      expectRes[2, gating_method := "refGate"]
      expectRes[2, gating_args := "cd3/TH+"]
      
      expect_equal(suppressMessages(.preprocess_row(this_row)), expectRes)
      
      
      #3d
      this_row[, dims := "cd3,cd4,cd8"]
      expect_error(suppressMessages(.preprocess_row(this_row)), regexp = "invalid number of dim")
      
      #A+B+
      ########################
      this_row <- copy(template_row)
      this_row[, pop := "+-"]
      this_row[, dims := "cd4,cd8"]
      
      expectRes <- rbindlist(list(this_row, this_row, this_row))
      expectRes[1, alias := "cd4+"]
      expectRes[1, pop := "+"]
      expectRes[1, dims := "cd4"]
      expectRes[2, alias := "cd8+"]
      expectRes[2, pop := "+"]
      expectRes[2, dims := "cd8"]
      expectRes[3, gating_method := "refGate"]
      expectRes[3, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[3, collapseDataForGating := ""]
      expectRes[3, groupBy := ""]
      expect_equal(suppressMessages(.preprocess_row(this_row)), expectRes)
      
      
      #A+B+ (refGate)
      ########################
      this_row <- copy(template_row)
      this_row[, pop := "+-"]
      this_row[, dims := "cd4,cd8"]
      this_row[, gating_method := "refGate"]
      expect_equal(.preprocess_row(this_row), this_row)
      
      #A+/-B+/-
      ########################
      this_row <- copy(template_row)
      this_row[, pop := "+/--"]
      this_row[, dims := "cd4,cd8"]
      
      expectRes <- rbindlist(list(this_row, this_row, this_row, this_row))
      expectRes[1, alias := "cd4+"]
      expectRes[1, pop := "+"]
      expectRes[1, dims := "cd4"]
      expectRes[2, alias := "cd8+"]
      expectRes[2, pop := "+"]
      expectRes[2, dims := "cd8"]
      expectRes[3, alias := "cd4+cd8-"]
      expectRes[3, pop := "+-"]
      expectRes[3, gating_method := "refGate"]
      expectRes[3, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[3, collapseDataForGating := ""]
      expectRes[3, groupBy := ""]
      expectRes[4, alias := "cd4-cd8-"]
      expectRes[4, pop := "--"]
      expectRes[4, gating_method := "refGate"]
      expectRes[4, gating_args := "cd3/cd4+:cd3/cd8+"]
      expectRes[4, collapseDataForGating := ""]
      expectRes[4, groupBy := ""]
      expect_equal(suppressMessages(.preprocess_row(this_row)), expectRes)
      
      
      #gate_flowclust_1d
      this_row <- copy(template_row)
      this_row[, pop := "+"]
      this_row[, dims := "cd4"]
      this_row[, gating_method := "gate_flowclust"]
      
      expectRes <- copy(this_row)
      expectRes[, gating_method := "gate_flowclust_1d"]
      expectRes[, pop := "+"]
      expect_equal(.preprocess_row(this_row), expectRes)
      
      #gate_flowclust_2d
      this_row <- copy(template_row)
      this_row[, pop := "+"]
      this_row[, gating_method := "gate_flowclust"]
      expectRes <- copy(this_row)
      expectRes[, gating_method := "gate_flowclust_2d"]
      expect_equal(.preprocess_row(this_row), expectRes)
      
      #multi-pops
      #no expansion    
      this_row <- copy(template_row)
      this_row[, alias := "*"]
      this_row[, pop := "*"]
      expect_equal(.preprocess_row(this_row), this_row)
      
      #with expansion
      this_row[, alias := "cd4,cd8"]
      expect_equal(.preprocess_row(this_row), .gen_dummy_ref_gate(this_row))
    })


test_that("gatingTemplate constructor", {
  thisPath <- system.file(package = "openCyto")
#  thisPath <- file.path(thisPath, "inst")
  gtfiles <- list.files(file.path(thisPath, "extdata/gating_template"), full = TRUE)[-c(5:6)]
  options(scipen=999)
  for(thisFile in gtfiles){
    
    templateName <- file_path_sans_ext(basename(thisFile))
    if(templateName == "ICS")
      next #skip polyfunctional test and this csv template already gets tested in test-gating-suite.R
    message(templateName)
    suppressWarnings(suppressMessages(
        thisRes <- gatingTemplate(thisFile, skip = 0)
    ))

    expectGt <- expectResults[[templateName]]
    #update expectResult due to the new change in pop columns and cytoverse API
    # browser()
    for(node in nodes(expectGt))
    {
      popname <- names(nodeData(expectGt)[[node]][["pop"]])
      if(popname!="root")
        nodeData(expectGt,node, "pop")[[1]]@name <- gsub("[^-\\+/]", "", popname)
      for(child in gt_get_children(expectGt, node)){
        if(isS4(edgeData(expectGt, node, child, "gtMethod")[[1]])){
          methodname <- edgeData(expectGt, node, child, "gtMethod")[[1]]@name
          edgeData(expectGt, node, child, "gtMethod")[[1]]@name <- gsub("flowClust\\.", "gate_flowclust_", methodname)
        }
      }
    }
    expect_equal(thisRes, expectGt, info = templateName)
    # browser()
    #test as.data.table method
    dt.orig <- fread(thisFile, skip = 0)
    dt.orig <- openCyto:::.preprocess_csv(dt.orig)
    
    dt.new <- as.data.table(thisRes)
    
    #sort by alias
    keys <- c("alias", "parent")
    setkeyv(dt.orig, keys)
    setkeyv(dt.new, keys)
    #standardize collapseDataForGating col
    dt.orig[, collapseDataForGating:=as.logical(ifelse(collapseDataForGating == "", FALSE, collapseDataForGating))]
    
    
    #skip gating_args (for tcell panel) since there is format changes due to the scientifc notation (1e3 -->1000)
    #and we don't want to change this csv since it is used elsewehere
    if(templateName == "tcell")        
      expect_equal(dt.orig[, -6, with = FALSE], dt.new[, -6, with = FALSE])
    else
      expect_equal(dt.orig, dt.new)

  }
  
}) 


        

#' @templateVar old gating
#' @templateVar new gt_gating
#' @template template-depr_pkg
NULL

#' @importFrom flowWorkspace gh_pop_is_negated
#' @export
gt_gating <- function(x, y, ...){
  UseMethod("gt_gating")
}
#' @export
gating <- function(x, y, ...){
  .Deprecated("gt_gating")
  gt_gating(x,y,...)
}

#' Applies a gatingTemplate to a GatingSet.
#'
#' It loads the gating methods by topological order and applies them to \code{GatingSet}.
#' 
#' @name gt_gating
#' @usage gt_gating(x, y, ...)
#' @aliases 
#' gating
#' gating,gatingTemplate,GatingSet-method
#' gt_gating,gatingTemplate,GatingSet-method
#' @param x a \code{gatingTemplate} object
#' @param y a \code{GatingSet} object
#' @param ... 
#'  \itemize{
#'      \item{start}{ a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating process will start from. It is useful to quickly skip some gates and go directly to the target population in the testing run. Default is "root".}
#'      \item{stop.at}{ a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating prcoess will stop at. Default is NULL, indicating the end of gating tree.}
#'      \item{keep.helperGates}{a \code{logical} flag indicating whether to keep the intermediate helper gates that are automatically generated by openCyto. Default is TRUE.}
#'      \item{mc.cores}{ passed to \code{multicore} package for parallel computing}
#'      \item{parallel_type}{  \code{character} specifying the parallel type. The valid options are "none", "multicore", "cluster".}
#'      \item{cl}{ \code{cluster} object passed to \code{parallel} package (when \code{parallel_type} is "cluster")}
#'  }  
#' 
#' 
#' @return 
#' Nothing. As the side effect, gates generated by gating methods are saved in \code{GatingSet}. 
#' @examples
#' \dontrun{
#'  gt <- gatingTemplate(file.path(path, "data/ICStemplate.csv"), "ICS")
#'  gs <- GatingSet(fs) #fs is a flowSet/ncdfFlowSet
#'  gt_gating(gt, gs)
#'  gt_gating(gt, gs, stop.at = "v") #proceed the gating until population 'v'
#'  gt_gating(gt, gs, start = "v") # start from 'v'
#'  gt_gating(gt, gs, parallel_type = "multicore", mc.cores = 8) #parallel gating using multicore
#'  #parallel gating by using cluster
#'  cl1 <- makeCluster (8, type = "MPI")
#'  gt_gating(gt, gs, parallel_type = "cluster", cl = cl1)
#'  stopCluster ( cl1 )
#' }
#' @export
gt_gating.gatingTemplate <- function(x, y, ...) {
  .gating_gatingTemplate(x, y, ...)
}
# gt_gating.gatingTemplate <- function(x, y, env_fct = NULL, ...) {
#       .gating_gatingTemplate(x,y,env_fct,...)
#     }

#' internal function (gating_gatingTemplate)
#'  
#' @param stop.at a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating prcoess will stop at.
#' @param start a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating prcoess will start from. It is useful to quickly skip some gates and go directly to the target population in the testing run. 
#' @param env_fct a \code{environment} that contains \code{fcTree} object named as 'fct'. If NULL (by default), no \code{fcTree} will be constructed. It is currently reserved for the internal debugging.
#' @param ... other arguments passed to the gatingMethod-specific \code{gating} methods.
#' @importFrom RBGL tsort
#' @keywords internal
#' @noRd 
.gating_gatingTemplate <- function(x, y, env_fct = NULL, start = "root", stop.at = NULL, keep.helperGates = TRUE, ...) {
  gt <- x
  if (!is.null(env_fct)) {
    # use the fcTree if already exists
    if (exists("fct", env_fct)) {
      fct <- get("fct", env_fct)
    } else {
      # create one from gt if not
      fct <- fcTree(gt)
      assign("fct", fct, env_fct)
    }
    
  }
  
  nodePaths <- names(sapply(gt_get_nodes(gt),alias))
    #validity check for stop.at argument
  if(!is.null(stop.at)){
    #escape meta character
    stop.pat <- gsub("\\+", "\\\\+", stop.at)
    stop.pat <- paste0(stop.pat, "$")
    if(substr(stop.at, 1,1) != "/")
      stop.pat <- paste0("/", stop.pat) #prepend path delimiter to avoid partial node name matching
    
    idx <- grep(stop.pat, nodePaths)
    if(length(idx) == 0)
      stop("Can't find stop point: ", stop.at)
    else if(length(idx) > 1)
      stop("ambiguous stop point: ", stop.at)
  }
  # gate each node 
  gt_nodes <- tsort(gt)[-1]#by the topological order

  #try to skip some nodes to save time
  if(start != "root"){
    #escape meta character
    start.pat <- gsub("\\+", "\\\\+", start)
    start.pat <- paste0(start.pat, "$") #treat it as terminal node
    if(substr(start, 1,1) != "/")
      start.pat <- paste0("/", start.pat) #prepend path delimiter to avoid partial node name matching
    matchInd <- grep(start.pat, gt_nodes)
    
    if(length(matchInd) == 0)
      stop("Can't find start point: ", start)
    else if(length(matchInd) > 1)
      stop("ambiguous start point: ", start)
    
    gt_nodes <- gt_nodes[matchInd:length(gt_nodes)]
  }  
    
                   
  for (node in gt_nodes) {
    
    # get parent node to gate
    gt_node_pop <- gt_get_nodes(gt, node)
    parent <- gt_get_parent(gt, node)
    
    if(!is.null(stop.at)){
      nodeName <- alias(gt_node_pop)
      nodePath <- file.path(parent, nodeName)
      matchInd <- grep(stop.pat, nodePath)
      if(length(matchInd) == 1)
      {
        message("stop at: ",stop.at)
        break
      }
      
    }    
    # extract gate method from one edge(since multiple edge to the same node is
    # redudant)
    this_gate <- gt_get_gate(gt, parent, node)
    
    #get preprocessing method
    this_ppm <- ppMethod(gt, parent, node)
    
    parentInd <- match(parent, gs_get_pop_paths(y[[1]], showHidden = TRUE))
    if (is.na(parentInd)) 
      stop("parent node '", parent, "' not gated yet!")
    
    #preprocessing
    pp_res <- NULL
#    browser()
    if(class(this_ppm) == "ppMethod")
      pp_res <- preprocessing(x = this_ppm, y, parent = parent, gtPop = gt_node_pop, gm = this_gate, ...)
#    browser()
    # pass the pops and gate to gating routine
    filterObj <- gt_gating(x = this_gate, y, parent = parent, gtPop = gt_node_pop, pp_res = pp_res, ...)
    if(!keep.helperGates)
    {
      gt_delete_helpergates(x, y)
    }
    # update fct
    if (!is.null(env_fct) && !is.null(filterObj)) {
      nodeData(env_fct$fct, node, "fList")[[1]] <- filterObj
    }  
    
    
  }
  message("finished.")
}


#' apply a \link[openCyto:gtMethod-class]{gtMethod} to the \code{GatingSet}
#' 
#' The actual workhorse of most gating methods
#' 
#' @param x \code{gtMethod}
#' @param y \code{GatingSet}
#' @param ... other arguments 
#' @aliases
#' gating,gtMethod,GatingSet-method 
#' gating,gtMethod,GatingSetList-method
#' @keywords internal
#' @noRd
gt_gating.gtMethod <- function(x, y, ...) {
  .gating_gtMethod(x,y,...)
}

#' evalRd tag is currently not working because the output of eval string isn't positioned properly in Rd file
#' @noRd
roxygen_parameter <- function() {
  paste(" 
          gtPop a \\code{gtPopulation} object that contains the information of the cell population that is going to be generated by this gating method.
          parent\\code{character} specifying the parent node within\\code{GatingSet} to which the new popoulation is to be attached.
          pp_res\\code{list} the preprocessing results to be used by the gating method. The names of the list has to be matched to the sample names of the data.
          mc.cores passed to\\code{multicore} package for parallel computing
          parallel_type\\code{character} specifying the parallel type. The valid options are 'none', 'multicore', 'cluster'.
          cl \\code{cluster} object passed to \\code{parallel} package")
}
#' internal function (gating_gtMethod)
#' 
#' It is a generic gating function that does:
#' 1. parse the gating parameters
#' 2. group the data when applicable 
#' 3. apply parallelism when applicable  
#' 4. pass the flow data(maybe grouped) and preprocessing results to the adaptor function ".gating_adaptor"
#' 5. collect the gates and add to GatingSet object
#' 
#' 
#' @param x \code{gtMethod}
#' @param y \code{GatingSet}
#' @evalRd roxygen_parameter()
#' @import flowWorkspace
#' @importFrom flowCore getChannelMarker
#' @noRd
.gating_gtMethod <- function(x, y, gtPop, parent, pp_res 
            , mc.cores = getOption("mc.cores", 2L), parallel_type = c("none", "multicore", "cluster"), cl = NULL
            ,  ...) {
  
  requireNamespace("parallel")
  gFunc_args <- parameters(x)

  # HOTFIX: This resolve an error when args is a named list with name NA and object NA.
  # The resulting error occurs down below and is:
  # Error in thisCall[[arg]] <- args[[arg]] : subscript out of bounds
  if (!is.null(names(gFunc_args))) {
    gFunc_args <- gFunc_args[!is.na(names(gFunc_args))]
  }
  
  ## Optionally omit arguments that should not be used in this context
  if(x@name == "gate_tail"){
    to_omit <- na.omit(match("positive", names(gFunc_args)))
    if(length(to_omit) > 0)
      gFunc_args <- gFunc_args[-to_omit]
  }
  ##  
  gating_method_name =names(x)
  gm <- paste0(".", gating_method_name)
  
  dims <- dims(x)
  
  is_1d_gate <- length(dims) == 1
  
  popAlias <- alias(gtPop)
  popName <- names(gtPop)
  popId <- gtPop@id
  gs_nodes <- basename(gs_pop_get_children(y[[1]], parent))
  
  if (length(gs_nodes) == 0)
    isGated <- FALSE
  else
    isGated <- any(popAlias %in% gs_nodes)
    
  if(!isGated)
  {
    message("Gating for '", popAlias, "'")
    
    parent_data <- gs_pop_get_data(y, parent)
    parallel_type <- match.arg(parallel_type)
    ## get the accurate channel name by matching to the fr
    frm <- parent_data[[1, use.exprs = FALSE]]
    channels <-  unname(sapply(dims, function(channel)as.character(getChannelMarker(frm, channel)$name)))
    if(length(channels) > 0) 
    parent_data <- parent_data[, channels] #it is more efficient to only pass the channels of interest
    # Splits the flow set into a list.
    # By default, each element in the list is a flowSet containing one flow frame,
    # corresponding to the individual sample names.
    # If 'split' is given, we split using the unique combinations within pData.
    # In this case 'split' is specified with column names of the pData.
    # For example, "PTID:VISITNO"
    # when split is numeric, do the grouping by every N samples
    groupBy <- groupBy(x)
    isCollapse <- isCollapse(x)
    if (groupBy != "" && isCollapse) {
      #when x@collapse == FALSE,then ignore groupBy argument since grouping is only used for collapsed gating
      split_by <- as.character(groupBy)
      split_by_num <- as.numeric(split_by)
      #split by every N samples
      if(!is.na(split_by_num)){
            nSamples <- length(parent_data)
            if(nSamples==1){
              split_by <- 1
            }else{
              split_by <-  sample(rep(1:nSamples, each = split_by_num, length.out= nSamples))  
            }
            
      }else{
        #split by study variables
        split_by <- strsplit(split_by, ":")[[1]]
        split_by <- apply(pData(parent_data)[, split_by, drop = FALSE], 1, paste, collapse = ":")
        split_by <- as.character(split_by)
      }
    } else {
      split_by <- sampleNames(parent_data)
    }

    fslist <- split(parent_data, split_by)
    
    if(is.null(pp_res))
      pp_res <- sapply(names(fslist),function(i)pp_res)
    else
      pp_res <- pp_res[names(fslist)] #reorder pp_res to make sure it is consistent with fslist
    # construct method call
    thisCall <- substitute(f1(fslist,pp_res))
    thisCall[["FUN"]] <- as.symbol(".gating_adaptor")
    # args to be passed to gating_adaptor
    args <- list()
    args[["gFunc"]] <- gm  #set gating method
    args[["popAlias"]] <- popAlias  
    args[["channels"]] <- channels 
    
    negated <- popName=="-"
    args[["gFunc_args"]] <- gFunc_args
    
    thisCall[["MoreArgs"]] <- args
    
    ## choose serial or parallel mode
    
      
    if (parallel_type == "multicore") {
      message("Running in parallel mode with ", mc.cores, " cores.")
      thisCall[[1]] <- quote(mcmapply)
      thisCall[["mc.cores"]] <- mc.cores
    }else if(parallel_type == "cluster"){
      if(is.null(cl))
          stop("cluster object 'cl' is empty!")
        thisCall[[1]] <- quote(clusterMap)
        thisCall[["cl"]] <- cl
        thisCall[["fun"]] <- thisCall[["FUN"]] 
        thisCall[["FUN"]] <- NULL
        thisCall[["SIMPLIFY"]] <- TRUE
     }else {
      thisCall[[1]] <- quote(mapply)  #select loop mode
  
     }
    
    flist <- eval(thisCall)

    # Handles the case that 'flist' is a list of lists.
    #   The outer lists correspond to the split by pData factors.
    #   The inner lists contain the actual gates.
    #     The order do not necessarily match up with sampleNames()
    # Unforunately, we cannot simply use 'unlist' because the list element names
    # are mangled.
    
    if (all(sapply(flist, is.list))) {
      flist <- do.call(c, unname(flist))
    }
    #check failed sample 
    #eventually we want to handle this properly (like inserting dummy gates)
    #in order for the other samples proceed the gating
    failed <- sapply(flist, function(i)extends(class(i), "character"))
    if(any(failed)){
      print(flist[failed])
      stop("some samples failed!")
      
    }
    
    #this is flowClust-specific operation, which
    # be abstracted out of this framework
    if (extends(class(flist[[1]]), "fcFilter")) {
      fcflist <- try(fcFilterList(flist), silent = TRUE)
      if(!is(fcflist, "try-error"))
        flist <- fcflist
      }
    
    is_clustering_results = is(flist[[1]], "factor")
    
    if (is_clustering_results) { 
        if(popAlias != "*")
            warning("popAlias is set to gating method name ",gating_method_name," because gating results are factors")
        popAlias = gating_method_name
    }
    
    
    if(length(popAlias) == 1){
      #when Alias is meta character, then pass NULL
      # to add method which uses filterId slot to name the populations
      if(popAlias == "*") {
        popAlias <- NULL  
      }
    }

    # For gate_quad methods, need to filter down to just the gates that were asked for
    if(names(x) %in% c("quadGate.seq", "gate_quad_sequential", "quadGate.tmix", "gate_quad_tmix")){
      pops <- gtPop@name
      pops <- gsub("([\\+-])([^/$])", "\\1&\\2", pops)
      pops <- strsplit(pops, "&")[[1]]
      pops <- strsplit(pops, "/")
      pops <- paste0(rep(pops[[1]], each=length(pops[[2]])), pops[[2]])
      pops <- match(pops, c("-+", "++", "+-", "--"))
      flist <- lapply(flist, function(sublist) filters(sublist[pops]))
    }
    
    gs_node_id <- gs_pop_add(y, flist, parent = parent, name = popAlias, validityCheck = FALSE, negated = negated)
    if(!is.null(popAlias) &&!is_clustering_results)
      recompute(y, file.path(parent, popAlias))
    else
      recompute(y, parent)
	message("done.")
    
  }else{
    
    message("Skip gating! Population '", paste(popAlias, collapse = ","), "' already exists.")
    flist <- NULL
  }
  
  flist
}

# apply a \code{boolMethod} to the \code{GatingSet}
#' @rdname gt_gating
#' @aliases
#' gating,boolMethod,GatingSet-method
#' gating,boolMethod,GatingSetList-method
#' @noRd
gt_gating.boolMethod <- function(x, y, ...) {
  .gating_boolMethod(x,y,...)
}

#' internal function (gating_boolMethod)
#' 
#'  @param y \code{GatingSet}
#'  @param gtPop a \code{gtPopulation} object that contains the information of the cell population that is going to be generated by this gating method.
#'  @param  parent \code{character} specifying the parent node within \code{GatingSet} to which the new popoulation is to be attached.
#' @noRd 
.gating_boolMethod <- function(x, y, gtPop, parent, ...) {
  
  args <- parameters(x)[[1]]
  gm <- paste0(".", names(x))
  popAlias <- alias(gtPop)
  popName <- names(gtPop)
  popId <- gtPop@id
  
  gs_nodes <- basename(gs_pop_get_children(y[[1]], parent))
  tNodes <- deparse(args)
  if (!(popAlias %in% gs_nodes)) {
    message(popAlias, " gating...")
    bf <- eval(substitute(booleanFilter(x), list(x = args)))
    bf@filterId <- tNodes
    #set recompute to FALSE because we want recompute method take over the
    #computing job since it is smart on determining whehter flow data needs to be loaded
    #for boolean gates
    invisible(gs_node_id <- gs_pop_add(y, bf, parent = parent, name = popAlias))
    newNode <- file.path(parent, popAlias)
    invisible(recompute(y, newNode))
    message("done.")
  } else {
    message("Skip gating! Population '", popAlias, "' already exists.")
  }
  
  # gs_node_id
  NULL
}

# apply a \code{polyFunctions} gating method to the \code{GatingSet}
# 
# It generates a batch of \code{boolMethod}s based on the expression defined in \code{polyFunctions} objects.
# It is a convenience way to generate different boolean combinations of cytokine gates. 
#' @rdname gt_gating
#' @aliases
#' gating,polyFunctions,GatingSet-method
#' gating,polyFunctions,GatingSetList-method
#' @noRd
gt_gating.polyFunctions <- function(x, y, ...) {
  .gating_polyFunctions(x,y,...)
}

#' internal function (gating_polyFunctions)
#' 
#'  @param y \code{GatingSet}
#'  @param gtPop a \code{gtPopulation} object that contains the information of the cell population that is going to be generated by this gating method.
#'  @param  parent \code{character} specifying the parent node within \code{GatingSet} to which the new popoulation is to be attached.
#' @noRd 
.gating_polyFunctions <- function(x, y, gtPop, parent, ...) {
  .Defunct()
  refNodes <- x@refNodes
  popAlias <- alias(gtPop)
  message("Population '", paste(popAlias, collapse = ","), "'")
  
  nMarkers <- length(refNodes)

  ## all the comibnations of A & B & C
  opList <- permutations(n = 1, r = nMarkers - 1, c("&"), repeats.allowed = TRUE)
  isNotList <- permutations(n = 2, r = nMarkers, c("!", ""), repeats.allowed = TRUE)
  polyExprsList <- apply(opList, 1, function(curOps) {
    apply(isNotList, 1, function(curIsNot) {
      polyExprs <- curIsNot
      polyExprs[-1] <- paste0(curOps, curIsNot[-1])
      
      paste(paste0(polyExprs, refNodes), collapse = "")
    })
  })
  polyExprsList <- as.vector(polyExprsList)
  
  gs_nodes <- basename(gs_pop_get_children(y[[1]], parent))
  # actual gating
  lapply(polyExprsList, function(polyExpr) {
    
    #replace the slash with colon 
    #since forward slash is reserved for gating path
    if(grepl("/",polyExpr)){
      old_name <- polyExpr
      new_name <- gsub("/",":",polyExpr)
      warning(old_name, " is replaced with ", new_name)
    }else
      new_name <- polyExpr
#    browser()
    
    isExist <- new_name %in% gs_nodes
    if (!isExist) {
      message("adding ", new_name, " ...")
      
      bf <- eval(substitute(booleanFilter(v), list(v = as.symbol(polyExpr))))

      invisible(gs_node_id <- .addGate_fast(y, bf, parent = parent, name = polyExpr))
      
  
    } else {
      message("Skip!Population '", new_name, "' already exists.")
    }        
  })
  #to reduce overhead,compute from parent node once instead of do it multiple times for each individual new bool gate
    
  invisible(recompute(y, parent))
  message("done.")
  
  list()
}

# apply a \code{refGate} to the \code{GatingSet}
#' @rdname gt_gating
#' 
#' @aliases
#' gating,refGate,GatingSet-method
#' gating,refGate,GatingSetList-method
#' gating,dummyMethod,GatingSet-method
#' gating,dummyMethod,GatingSetList-method
#' @noRd
gt_gating.refGate <- function(x, y, ...) {
  .gating_refGate(x, y, ...)
}

#' @export
gt_gating.dummyMethod <- function(x, y, ...) {
  #do nothing
}

#' internal function (gating_refGate)
#' 
#'  @param y \code{GatingSet}
#'  @param gtPop a \code{gtPopulation} object that contains the information of the cell population that is going to be generated by this gating method.
#'  @param  parent \code{character} specifying the parent node within \code{GatingSet} to which the new popoulation is to be attached.
#' @noRd 
.gating_refGate <- function(x, y, gtPop, parent, ...) {
#  negated <- FALSE
  refNodes <- x@refNodes
  popAlias <- alias(gtPop)
  popName <- names(gtPop)
  dims <- dims(x)
  
  my_gh <- y[[1]] 
  gs_nodes <- basename(gs_pop_get_children(my_gh, parent))
  if (length(gs_nodes) == 0 || !popAlias %in% gs_nodes) {
    
    message("Population '", paste(popAlias, collapse = ","), "'")
    if (length(refNodes) > 2) {
      stop("Not sure how to construct gate from more than 2 reference nodes!")
    }
    
    fr <- gh_pop_get_data(my_gh, use.exprs = FALSE)
    
    #check if parent node is shared 
    #to determine whether simply grab the indices from ref nodes without recompute
    ref_parents <- sapply(refNodes, function(refNode)gs_pop_get_parent(my_gh, refNode))
    isSameParent <- all(ref_parents == parent)
    flist <- flowWorkspace::lapply(y, function(gh) {
          
       glist <- lapply(refNodes, function(refNode) {
          
          gh_pop_get_gate(gh, refNode)
        })
       
      # standardize the names for the gate parameters and dims
      gate_params <- unlist(lapply(glist, function(g) {
        cur_param <- parameters(g)
        getChannelMarker(fr, cur_param)["name"]
      }))
      negated_2d_gate <- FALSE
      if(length(glist)==1){
        #1d ref gate
        dims <- dims[!is.na(dims)]
        nDims <- length(dims)
        
        
        if(nDims==2){
          if (popName == "-") {
            negated_2d_gate <- TRUE                             
            }
          
          fres <- glist[[1]]
        }else{
          dim_params <-  getChannelMarker(fr, dims)["name"]
          y_g <- glist[[1]]               
          y_coord <- c(y_g@min, y_g@max)
          cut.y <- y_coord[!is.infinite(y_coord)]
          
          
          if (popName == "+") {
            #handle all infinite coordinates
            if(length(cut.y) == 0)
              cut.y <- -Inf
            gate_coordinates <- list(c(cut.y, Inf))
          } else if(popName=="-"){
            if(length(cut.y) == 0)
              cut.y <- Inf
            gate_coordinates <- list(c(-Inf, cut.y))
          }else{
            stop("unknown population pattern, ",popName)
          }
          names(gate_coordinates) <- as.character(dim_params)
          
          fres <- rectangleGate(gate_coordinates)  
        }
        
      }else{
          #2d quad gate
          dim_params <- unlist(lapply(dims, function(dim) {
                    getChannelMarker(fr, dim)["name"]
                  }))
          
          # match the gate param to dims to find gates for x, y dimensions
          x_ref_id <- match(dim_params[1], gate_params)
          y_ref_id <- match(dim_params[2], gate_params)
          x_g <- glist[[x_ref_id]]
          y_g <- glist[[y_ref_id]]
                        
          
          # pick the non-infinite coordinate as the cut points
          x_coord <- c(x_g@min, x_g@max)
          y_coord <- c(y_g@min, y_g@max)
          x_inf_vec <- is.infinite(x_coord)
          y_inf_vec <- is.infinite(y_coord)
          cut.x <- x_coord[!x_inf_vec]
          cut.y <- y_coord[!y_inf_vec]
          
          # In order, the following vector has the patterns:
          # 1. top left (-+)
          # 2. top right (++)
          # 3. bottom right (+-)
          # 4. bottom left (--)
          quadPatterns <- c("-+", "++", "+-", "--")
          
          quadInd <- match(popName, quadPatterns)
          
          is_negated <- lapply(refNodes, function(refNode) {
            
            gh_pop_is_negated(gh, refNode)
          })          
          x_ref <- refNodes[x_ref_id] 
          y_ref <- refNodes[y_ref_id]
          x_negated <- is_negated[[x_ref_id]]
          y_negated <- is_negated[[y_ref_id]]
          #take into account of (-Inf, x) for bool operations in ocRectRefGate
          if(x_coord[x_inf_vec] < 0)
            x_negated <- !x_negated
          if(y_coord[y_inf_vec] < 0)
            y_negated <- !y_negated
          #standardize event ind to x+ and y+
          #by checking the positive sign of both reference gates
          #to prepare for the bool operation later
          if(length(cut.x) == 0){
          #handle dummy ref gate that has both boudnary as Inf
#            x_event_ind <- TRUE           
            x_ref <- NULL
          }
          if(length(cut.y) == 0){
#            y_event_ind <- TRUE
              y_ref <- NULL
          }
          
#          browser()
          # construct rectangleGate from reference cuts
          if (quadInd == 1) {#-+
            #handle all infinite coordinates
            if(length(cut.x) == 0)
              cut.x <- Inf
            else
            {
              if(!x_negated)
                x_ref <- paste0("!", x_ref) 
            }
              
            if(length(cut.y) == 0)
              cut.y <- -Inf
            else
            {
              if(y_negated)
                y_ref <- paste0("!", y_ref)
            }
            coord <- list(c(-Inf, cut.x), c(cut.y, Inf))
            
          } else if (quadInd == 2) {#++
            #handle all infinite coordinates
            if(length(cut.x) == 0)
              cut.x <- -Inf
            else
            {
              if(x_negated)
                x_ref <- paste0("!", x_ref) 
            }
            if(length(cut.y) == 0)
              cut.y <- -Inf
            else
            {
              if(y_negated)
                y_ref <- paste0("!", y_ref) 
            }
            coord <- list(c(cut.x, Inf), c(cut.y, Inf))
            
          } else if (quadInd == 3) {#+-
            #handle all infinite coordinates
            if(length(cut.x) == 0)
              cut.x <- -Inf
            else
            {
              if(x_negated)
                x_ref <- paste0("!", x_ref) 
            }
            if(length(cut.y) == 0)
              cut.y <- Inf
            else
            {
              if(!y_negated)
                y_ref <- paste0("!", y_ref) 
            }
            coord <- list(c(cut.x, Inf), c(-Inf, cut.y))
             
          } else if (quadInd == 4) {#--
            #handle all infinite coordinates
            if(length(cut.x) == 0)
              cut.x <- Inf
            else
            {
              if(!x_negated)
                x_ref <- paste0("!", x_ref) 
            }
            if(length(cut.y) == 0)
              cut.y <- Inf
            else
            {
              if(!y_negated)
                y_ref <- paste0("!", y_ref) 
            }
            coord <- list(c(-Inf, cut.x), c(-Inf, cut.y))
            
          } else stop("Pop names does not match to any quadrant pattern!")
          
          names(coord) <- as.character(dim_params)
          fres <- rectangleGate(coord)
          if(isSameParent&&!is.null(x_ref)&&!is.null(y_ref)){
            fres <- ocRectRefGate(fres, paste0(x_ref, "&", y_ref))
          }
          
      }
      attr(fres, "negated_2d_gate") <- negated_2d_gate
      fres  
      
    })
    negated_2d_gate <- attr(flist[[1]], "negated_2d_gate")   
    flist <- filterList(flist)
    gs_node_id <- gs_pop_add(y, flist, parent = parent, name = popAlias, validityCheck = FALSE, negated = negated_2d_gate)
    if(!is(flist[[1]], "ocRectRefGate"))
		recompute(y, file.path(parent, popAlias))
	
   
  } else {
    message("Skip gating! Population '", popAlias, "' already exists.")
    flist <- NULL
  }

  message("done.")
  flist
}

#' Pre-processing plugin for manual/templated gating
#'
#' @param fs a \code{flowSet} object containing the samples within a group to be
#'   gated.
#' @param gs the original \code{GatingSet} containing all samples with attached
#'   compensation, transformations and gates.
#' @param gm the name of the gating method.
#' @param channels markers or channels to use for gating.
#' @param groupBy a colon separated character string specifying the names of the
#'   variables to use to group the samples prior to gating, set to NA by default
#'   to group all samples. 
#' @param isCollapse logical indicating whether the data for the current group
#'   is collapsed prior to gating, set to NA by default.
#' @param ... not in use.
#'
#' @return \code{pp_gate_template()} returns the name of the current group so it
#'   can be used to extract the appropriate gate(s) in \code{gate_template()}.
#'
#' @importFrom flowWorkspace sampleNames pData
#'
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#'
#' @noRd
.pp_gate_template <- function(fs,
                              gs,
                              gm,
                              channels,
                              groupBy = NA,
                              isCollapse = NA,
                              ...) {
  
  # gates are created for each unique group in groupBy
  # fs contains samples for current group to gate
  # we obtain group index by searching for group containing all samples in fs
  # group is identified by name to ensure correct matching
  # names used in gates must include group variables separated by " "
  
  # prepare groupBy - split out grouping variables
  if(is.character(groupBy)) {
    if(nchar(groupBy) == 0) {
      groupBy <- NA
    } else {
      groupBy <- unlist(strsplit(groupBy, ":"))
    }
  }
  
  # experiment details
  pd <- pData(gs)
  
  # grouping
  if(!all(is.na(groupBy))) {
    # variable names supplied
    if(all(grepl("^[A-Za-z]+$", groupBy, ignore.case = TRUE))) {
      # check for valid variable names
      if(!all(groupBy %in% colnames(pd))) {
        stop(
          "'groupBy' references variables that don't exist in the GatingSet!"
        )
      }
      # split experiment details by grouping variables
      groups <- split(
        pd, pd[, groupBy],
        sep = " ",
        lex.order = TRUE,
        drop = TRUE
      )
      # locate group containing all samples in fs -> group name
      ind <- names(groups)[
        unlist(
          lapply(
            groups,
            function(group) {
              all(sampleNames(fs) %in% c(rownames(group), group$name))
            }
          )
        )
      ]
    # numeric grouping is not supported - inconsistent gate order
    } else {
      if(groupBy == length(gs)) {
        ind <- 1
      } else {
        message(
          "Numeric groupBy is not supported. Grouping all samples."
        )
        ind <- 1
      }
    }
    # no grouping - use first set of gates
  } else {
    ind <- 1
  }
  
  # return group index or name (preferred)
  return(ind)
  
}

#' Gating plugin for manual/templated gating
#'
#' @param fr a \code{flowFrame} object containing the collapsed data for a
#'   sample group to be gated.
#' @param pp_res indicates the name of the current sample group as provided by
#'   \code{pp_gate_template()}.
#' @param channels names of the channels or markers in which the supplied
#'   \code{gate(s)} were constructed.
#' @param gate list of filters per sample group each named based on the levels
#'   of the grouping variables separated by \code{" "}.
#'   \code{generate_gate_template()} should be used to create the
#'   \code{gatingTemplate} so that the gates can be appropriately formatted and
#'   passed to \code{gate_template()} through the gatingTemplate
#'   \code{gating_args}.
#' @param ... not in use.
#'
#' @return a list of filters for the current sample group to be used for gating.
#'
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#'
#' @noRd
.gate_template <- function(fr,
                          pp_res,
                          channels,
                          gate,
                          ...) {
  
  
  # gate = list of filters per sample group (all groups)
  # pp_res = name of sample group to extract gate(s) from gate
  
  # no sample groups - apply first set of gates to all samples
  if(is.null(pp_res)) {
    ind <- 1
  }
  
  # pp_res is index of gate(s) to extract for sample fr
  if(is.numeric(pp_res)) {
    ind <- pp_res
  # pp_res is the name of sample group
  } else if(is.character(pp_res)) {
    # attempt to locate gate for sample group
    ind <- match(
      pp_res,
      names(gate)
    )
    # sample group doesn't exist in the gatingTemplate
    if(all(is.na(ind))) {
      stop(
        "The gatingTemplate doesn't contain gates for the ",
        pp_res,
        " sample group!"
      )
    }
  }
  
  # extract and apply matching set of gates
  return(gate[[ind]])
  
}

#' Store gates from a GatingSet in a gatingTemplate for templated gating
#'
#' \code{generate_gate_template()} iterates through samples groups in a
#' GatingSet to extract and store existing gates into an openCyto
#' \code{gatingTemplate} that can be easily applied to new GatingSets using
#' \code{gt_gating()}.
#'
#' @param x a \code{GatingHierarchy} or \code{GatingSet} object.
#' @param groupBy a vector of variable names that exist in
#'   \code{colnames(pData(x))} to split samples into groups before extracting
#'   the gates specified by \code{nodes}, set to NA by default to produce
#'   consensus gates across all samples.
#' @param nodes a vector of full node paths that exist in \code{x} for which
#'   templated gating entries are required, set to all nodes in \code{x} by
#'   default.
#' @param gatingTemplate logical to indicate whether a \code{gatingTemplate}
#'   object should be returned, set to FALSE by default. \code{gatingTemplate}
#'   objects can only be created when all nodes in \code{x} are specified.
#' @param save_as name of the CSV file to which the gatingTemplate entries
#'   should be written, set to NULL to prevent the gatingTemplate from being
#'   written to a CSV file.
#' @param ... additional arguments passed to \code{\link{gatingTemplate-class}}.
#'
#' @return either a \code{data.table} or \code{gatingTemplate} object containing
#'   the requested gatingTemplate entries for downstream templated gating using
#'   \code{gt_gating()}.
#'
#' @importFrom flowCore polygonGate rectangleGate quadGate filters
#' @importFrom flowWorkspace pData gh_pop_get_gate gh_get_pop_paths
#'   gh_pop_get_data gh_pop_is_negated
#' @importFrom grDevices chull
#' @importFrom methods as
#' @importFrom data.table fwrite
#'
#' @author Dillon Hammill, \email{Dillon.Hammill@ozette.com}
#'
#' @examples
#' \dontrun{
#' # gs contains variable BioSampleType
#' pData(gs)$BioSampleType <- rep(c("cells", "beads"), each = 4)
#' # create gatingTemplate object
#' gt <- generate_gate_template(
#'   gs,
#'   groupBy = "BioSampleType",
#'   gatingTemplate = TRUE,
#'   save_as = "gatingTemplate.csv"
#' )
#' # gt is a gatingTemplate object
#' gt
#' # apply gt to a new GatingSet
#' gt_gating(
#'   gt,
#'   gs_new
#' )
#' }
#'
#' @export
generate_gate_template <- function(x,
                                   groupBy = NA,
                                   nodes = NULL,
                                   gatingTemplate = FALSE,
                                   save_as = NULL,
                                   ...) {
  
  # NOTE: All samples in x must contain the same set of gates! We cannot
  # generate gatingTemplate entries for clusters (i.e. labelled events) as
  # these cannot be transferred to new GatingSets - only valid gate objects
  # (e.g. booleanFilter, rectangleGate, ellipsoidGate, quadrantGate etc.) are
  # supported.
  
  # experiment details
  pd <- pData(x)
  
  # extract full node paths
  nodes_full <- gh_get_pop_paths(
    x[[1]],
    path = "full"
  )
  
  # default nodes
  if(is.null(nodes)) {
    nodes <- nodes_full
  # manual nodes
  } else {
    if(!all(nodes %in% nodes_full)) {
      stop(
        "'nodes' must contain full paths to nodes existing in the ",
        class(x), "!"
      )
    }
  }
  
  # prepare groupBy - split samples into groups
  if(!all(is.na(groupBy))) {
    # split grouping variables
    if(length(groupBy) == 1 & grepl("\\:", groupBy)) {
      groupBy <- strsplit(
        groupBy,
        "\\:"
      )[[1]]
    }
    # check groupBy for valid variables
    if(!all(groupBy %in% colnames(pd))) {
      stop(
        "'groupBy' references variables that don't exist in pData(x)!"
      )
    }
  }
  
  # split details into groups by groupBy
  if(all(is.na(groupBy))) {
    pd_groups <- list(
      pd
    )
  # group order doesn't matter
  } else {
    pd_groups <- split(
      pd,
      pd[, groupBy],
      drop = TRUE,
      sep = " "
    )
  }
  
  # compute instrument limits - replace infinite co-ordinates
  rng <- apply(
    do.call(
      "rbind",
      lapply(
        seq_along(x),
        function(id) {
          cf <- gh_pop_get_data(
            x[[id]]
          )
          # determine range using instrument and data limits - data often lower
          rbind(
            range(cf, type = "instrument"),
            range(cf, type = "data")
          )
        }
      )
    ),
    2,
    "range"
  )
  
  # rename groupBy
  group_by <- groupBy
  
  # initialize gatingTemplate variables - CRAN R CMD CHECK NOTE
  parent <- NULL
  alias <- NULL
  pop <- NULL
  gating_method <- NULL
  gating_args <- NULL
  collapseDataForGating <- NULL
  groupBy <- NULL
  preprocessing_method <- NULL
  
  # generate a bare bones gatingTemplate 
  gt <- gh_generate_template(x[[1]])
  
  # split bare bones gatingTemplate by parent and dims
  gt_chunks <- split(
    gt,
    gt[, c("parent", "dims")],
    sep = "|",
    drop = TRUE
  )
  
  # order gt_chunks by nodes - ensures node order correct in gatingTemplate
  idx <- sapply(
    gt_chunks,
    function(gt_chunk) {
      parent <- gt_chunk$parent
      parent[parent == "root"] <- ""
      alias <- unlist(strsplit(gt_chunk$alias, ","))
      min(
        match(
          paste0(
            parent,
            "/",
            alias
          ),
          nodes
        )
      )
    }
  )
  
  # drop unwanted nodes
  gt_chunks <- gt_chunks[!is.na(idx)]
  idx <- idx[!is.na(idx)]
  
  # reorder gatingTemplate chunks
  gt_chunks <- gt_chunks[order(idx)]
  
  # construct gatingTemplate
  gt_new <- do.call(
    "rbind",
    structure(
      lapply(
        gt_chunks,
        function(gt_chunk) {
          # parent
          parent <- unique(gt_chunk$parent)
          # channels
          channels <- strsplit(unique(gt_chunk$dims), ",")[[1]]
          # alias
          alias <- gt_chunk$alias
          # REQUIRE: list of filters per population in each sample
          group_gate_list <- structure(
            lapply(
              pd_groups,
              function(pd_group) {
                # groupBy = NA -> all samples
                if(all(is.na(groupBy))) {
                  idx <- seq_along(x)
                # groupBy subset of samples
                } else {
                  idx <- match(pd_group$name, pd$name)
                }
                # extract gates per sample in group
                sample_gates <- lapply(
                  idx,
                  function(id) {
                    # extract gate objects
                    gate_list <- structure(
                      lapply(
                        alias,
                        function(pop) {
                          gate <- gh_pop_get_gate(
                            x[[id]],
                            paste0(
                              if(parent == "root") {
                                ""
                              } else {
                                parent
                              },
                              "/",
                              pop
                            )
                          )
                          # inherit negated flag
                          attributes(gate)$negated <- gh_pop_is_negated(
                            x[[id]],
                            paste0(
                              if(parent == "root") {
                                ""
                              } else {
                                parent
                              },
                              "/",
                              pop
                            )
                          )
                          return(gate)
                        }
                      ),
                      names = alias
                    )
                    # format quadrant gates
                    quad_idx <- which(
                      sapply(
                        gate_list,
                        function(gate){
                          if(inherits(gate, "rectangleGate")) {
                            any(grepl("quad", names(attributes(gate))))
                          } else {
                            FALSE
                          }
                        }
                      )
                    )
                    # quadrants located
                    if(length(quad_idx) > 0) {
                      # pull out quadrant rectangleGates
                      quads <- gate_list[quad_idx]
                      # drop quadrants from gate_list
                      gate_list <- gate_list[-quad_idx]
                      # quadrant co-ordinates
                      coords <- do.call(
                        "rbind",
                        lapply(
                          quads,
                          function(quad) {
                            structure(
                              c(quad@min, quad@max),
                              names = parameters(quad)
                            )
                          }
                        )
                      )
                      # quadrant center
                      coords <- unlist(
                        sapply(
                          channels,
                          function(z) {
                            unique(coords[, z][is.finite(coords[, z])])
                          }
                        )
                      )
                      names(coords) <- channels
                      # create quadrantGate
                      quads <- list(
                        quadGate(
                          filterId = paste0(
                            names(attributes(quads[[1]])[["quadrants"]]),
                            collapse = "|"
                          ),
                          .gate = coords
                        )
                      )
                      attributes(quads[[1]])$negated <- FALSE
                      gate_list <- c(quads, gate_list)
                    }
                    names(gate_list) <- sapply(
                      gate_list,
                      function(gate) {
                        gate@filterId
                      }
                    )
                    return(gate_list)
                  }
                )
                # construct gatingTemplate entries
                structure(
                  lapply(
                    names(sample_gates[[1]]),
                    function(gate) {
                      # extract unique gates across samples in group
                      gates <- unique(
                        lapply(
                          sample_gates,
                          `[[`,
                          gate
                        )
                      )
                      # samples in group have different gates
                      if(length(gates) > 1) {
                        # multiRangeGate or QuadGate - use first gate
                        if(inherits(gates[[1]], "multiRangeGate") |
                           inherits(gates[[1]], "quadGate")) {
                          return(gates[[1]])
                        # 1-D rectangleGate
                        } else if(inherits(gates[[1]], "rectangleGate") &
                                  length(parameters(gates[[1]])) == 1) {
                          # combine coords from each gate
                          coords <- do.call(
                            "rbind",
                            lapply(
                              gates,
                              function(pop) {
                                rbind(
                                  pop@min,
                                  pop@max
                                )
                              }
                            )
                          )
                        # coerce gates to polygon and chull
                        } else {
                          # combine coords from each gate
                          coords <- do.call(
                            "rbind",
                            lapply(
                              gates,
                              function(pop) {
                                as(pop, "polygonGate")@boundaries
                              }
                            )
                          )
                        }
                        # replace infinite co-ordinates with instrument limits
                        lapply(
                          colnames(coords),
                          function(chan) {
                            coords[, chan][
                              is.infinite(coords[, chan]) & coords[, chan] < 0
                            ] <<- min(rng[, chan])
                            coords[, chan][
                              is.infinite(coords[, chan]) & coords[, chan] > 0
                            ] <<- max(rng[, chan])
                          }
                        )
                        # 1-D gate - range
                        if(ncol(coords) == 1) {
                          # use range across all gates
                          chull <- rectangleGate(
                            .gate = apply(
                              coords,
                              2,
                              "range"
                            ),
                            filterId = gate
                          )
                        # 2-D gate - chull
                        } else {
                          # get convex hull
                          chull <- coords[chull(coords), ]
                          # chull is a rectangle
                          if(nrow(chull) < 3) {
                            # construct rectangle
                            chull <- rectangleGate(
                              chull,
                              filterId = gate
                            )
                            # convert to valid polygon
                            chull <- as(
                              chull, 
                              "polygonGate"
                            )
                          # chull is a polygon
                          } else {
                            # construct polygonGate
                            chull <- polygonGate(
                              .gate = chull,
                              filterId = gate
                            )
                          }
                        }
                        # inherit negated flag
                        attributes(chull)$negated <- 
                          attributes(gates[[1]])$negated
                        # return consensus gate
                        return(chull)
                      # boolean gates captured here - shared logic
                      } else {
                        return(gates[[1]])
                      }
                    }
                  ),
                  names = names(sample_gates[[1]])
                )
              }
            )
          )
          # create gatingTemplate entries
          do.call(
            "rbind",
            lapply(
              names(group_gate_list[[1]]),
              function(alias) {
                # gate flags
                bool <- FALSE
                quad <- FALSE
                negated <- FALSE
                # get gates from each group
                gates <- structure(
                  lapply(
                    group_gate_list,
                    function(gate_list) {
                      gate <- gate_list[[alias]]
                      # negated flag
                      negated <<- attributes(gate)$negated
                      # booleanFilter flag
                      if(inherits(gate, "booleanFilter")) {
                        bool <<- TRUE
                        return(
                          gate
                        )
                      # quadGates aren't wrapped in filters()
                      } else if(inherits(gate, "quadGate")) {
                        quad <<- TRUE
                        return(
                          gate_list[[alias]]
                        )
                      } else {
                        return(
                          filters(
                            gate_list[alias]
                          )
                        )
                      }
                    }
                  ),
                  names = names(group_gate_list)
                )
                # create gatingTemplate entry
                gt_entry <- as.data.table(gt_chunk[1, , drop = FALSE])
                # alias
                gate <- gsub("\\|", ",", alias)
                # pop
                if(negated) {
                  gt_entry[, pop := "-"]
                } else {
                  gt_entry[, pop := "+"]
                }
                # alias
                gt_entry[, alias := gate]
                # preprocessing method
                gt_entry[, preprocessing_method := "pp_gate_template"]
                # gating method
                gt_entry[, gating_method := "gate_template"]
                # collapse data - required or need gate for every sample
                gt_entry[, collapseDataForGating := TRUE]
                # groupBy
                if(all(is.na(group_by))) {
                  group_by <- NA
                } else {
                  group_by <- paste0(
                    group_by,
                    collapse = ":"
                  )
                }
                gt_entry[, groupBy := group_by]
                # booleanFilter
                if(bool) {
                  gt_entry[, groupBy := NA]
                  gt_entry[, preprocessing_method := NA]
                  gt_entry[, gating_method := "boolGate"]
                  gt_entry[, gating_args := gates[[1]]@deparse]
                  gt_entry[, collapseDataForGating := NA]
                # other gates handled above
                } else {
                  if(quad) {
                    gt_entry[, pop := "*"]
                  }
                  gt_entry[, gating_args := .argDeparser(
                    list(
                      gate = gates,
                      openCyto.minEvents  = -1
                    )
                  )]
                }
                return(gt_entry)
              }
            )
          )
        }
      ),
      names = names(gt_chunks)
    )
  )

  # write gatingTemplate to CSV file
  if(!is.null(save_as)) {
    fwrite(
      gt_new,
      file = save_as
    )
  }

  # convert to gatingTemplate object - only if its complete
  if(gatingTemplate & all(nodes %in% nodes_full)) {
    gt_new <- gatingTemplate(
      gt_new,
      ...
    )
  }

  # return constructed gatingTemplate
  return(gt_new)
  
}

setGeneric("preprocessing", function(x, y, ...) standardGeneric("preprocessing"))

#' apply a \link[openCyto:ppMethod-class]{ppMethod} to the \code{GatingSet}
#' 
#' @param x \code{ppMethod}
#' @param y \code{GatingSet} or \code{GatingSetList}
#' @param ... other arguments
#' 
#' @inheritParams .preprocessing
#' 
#' @aliases preprocessing,ppMethod,GatingSet-method
setMethod("preprocessing", signature = c("ppMethod", "GatingSet"),
    definition = function(x, y, ...) {
      .preprocessing(x,y,...)
    })
setMethod("preprocessing", signature = c("ppMethod", "GatingSetList"),
    definition = function(x, y, ...) {
      .preprocessing(x,y,...)
    })

update_list <- function (x, val) 
{
  if (is.null(x)) 
    x <- list()
  modifyList(x, val)
}
#' internal function (preprocessing)
#' 
#' @inheritParams .gating_gtMethod
#' @param gm: \code{gtMethod} object
#' @noRd 
.preprocessing <- function(x, y, gtPop, parent, gm
                            , mc.cores = 1, parallel_type = c("none", "multicore", "cluster"), cl = NULL
                            , ...) {
  requireNamespace("parallel")
  
  args <- parameters(x)
  # overwrite the golbal args with the local one
  args <- update_list(args,list(...))
  parallel_type <- match.arg(parallel_type)
  
  ppm <- paste0(".", names(x))
  if(!.isRegistered(ppm)){
    stop(sprintf("Can't gate using unregistered method %s",ppm))
  }
  groupBy <- groupBy(x)
  isCollapse <- isCollapse(x)
  dims <- dims(x)
  is_1d_gate <- any(is.na(dims))
  
  popAlias <- alias(gtPop)
  popName <- names(gtPop)
  popId <- gtPop@id
  
  gs_nodes <- basename(gs_pop_get_children(y[[1]], parent))
  if (length(gs_nodes) == 0 || !any(popAlias %in% gs_nodes)) {
    message("Preprocessing for '", popAlias, "'")
    
    parent_data <- gs_pop_get_data(y, parent)
    channels <- sapply(dims, function(channel)as.character(getChannelMarker(parent_data[[1, use.exprs = FALSE]], channel)$name))
    
    # when groupBy is set distribute the subset for each group to preprocessing function
    # otherwise, the entire data set is passed (which is different from the way we handle gating function)
    if (nchar(groupBy) > 0) {
      
      split_by <- as.character(groupBy)
      suppressWarnings(split_by_num <- as.numeric(groupBy))
      #split by every N samples
      if(!is.na(split_by_num)){
        nSamples <- length(parent_data)
        if(nSamples==1){
          split_by <- 1
        }else{
          split_by <-  rep(1:nSamples, each = split_by_num, length.out= nSamples)  
        }
        
      }else{
        
        #split by study variables
        pd <- pData(parent_data)
        split_by <- strsplit(split_by, ":")[[1]]
        # check variables exist 
        if(!all(split_by %in% colnames(pd))) {
          stop(
            "gatingTemplate references variables missing in the pData of the GatingSet: \n",
            paste0(
              split_by[!split_by %in% colnames(pd)],
              sep = "\n"
            )
          )
        }
        split_by <- apply(pd[, split_by, drop = FALSE], 1, function(i)paste(i, collapse = ":"))
        split_by <- as.character(split_by)
      }
      fslist <- split(parent_data, split_by)
    }else 
    {
      fslist <- list(parent_data)  
    } 
    
    # construct method call
    thisCall <- substitute(f1())
    thisCall[["X"]] <- quote(fslist)  #set data
    thisCall[["FUN"]] <- .resolve_plugin_function(ppm)  #set gating method
    thisCall[["channel"]] <- channels 
    thisCall[["gs"]] <- y
    thisCall[["gm"]] <- gm
    thisCall[["groupBy"]] <- groupBy
    thisCall[["isCollapse"]] <- isCollapse
    
    #we don't want to deal with subSampling in preprocessing
    #since it will increase the IO overhead in order to determine 
    #total number of events
    if(!(all(is.na(args)))){
      for (arg in names(args)) {
        thisCall[[arg]] <- args[[arg]]
      } 
    }
    
    if (parallel_type == "multicore") {
      message("Running in parallel mode with ", mc.cores, " cores.")
      thisCall[[1]] <- quote(mclapply)
      thisCall[["mc.cores"]] <- mc.cores
    }else if(parallel_type == "cluster"){
      if(is.null(cl))
        stop("cluster object 'cl' is empty!")
      thisCall[[1]] <- quote(parSapply)
      thisCall[["cl"]] <- cl
      thisCall[["SIMPLIFY"]] <- TRUE
    }else {
      thisCall[[1]] <- quote(lapply)  #select loop mode
      
    }
    
    res <- eval(thisCall)
    
    # when isCollapse == FALSE, the gating is done at each individual FCS sample
    # Thus we want to unlist the pps_res to FCS file level
    if(!isCollapse)
      res <- unlist2(res, recursive = FALSE)
      
    

  } else {
    message("Skip preprocessing! Population '", paste(popAlias, collapse = ","), "' already exists.")
    res <- NULL
  }
  
  
  
  res
  
}
# copied from AnnotationDbi that does not mangle the name
unlist2 <- function (x, recursive = TRUE, use.names = TRUE, what.names = "inherited") 
{
  ans <- unlist(x, recursive, FALSE)
  if (!use.names) 
    return(ans)
  if (!is.character(what.names) || length(what.names) != 1) 
    stop("'what.names' must be a single string")
  what.names <- match.arg(what.names, c("inherited", "full"))
  names(ans) <- unlist(make.name.tree(x, recursive, what.names), 
      recursive, FALSE)
  ans
}
# copied from AnnotationDbi that does not mangle the name
make.name.tree <- function (x, recursive, what.names) 
{
  if (!is.character(what.names) || length(what.names) != 1) 
    stop("'what.names' must be a single string")
  what.names <- match.arg(what.names, c("inherited", "full"))
  .make.name.tree.rec <- function(x, parent_name, depth) {
    if (length(x) == 0) 
      return(character(0))
    x_names <- names(x)
    if (is.null(x_names)) 
      x_names <- rep.int(parent_name, length(x))
    else if (what.names == "full") 
      x_names <- paste0(parent_name, x_names)
    else x_names[x_names == ""] <- parent_name
    if (!is.list(x) || (!recursive && depth >= 1L)) 
      return(x_names)
    if (what.names == "full") 
      x_names <- paste0(x_names, ".")
    lapply(seq_len(length(x)), function(i) .make.name.tree.rec(x[[i]], 
              x_names[i], depth + 1L))
  }
  .make.name.tree.rec(x, "", 0L)
}

#' @name <%= fun %>-deprecated
#' @usage <%= gsub("\n", "\n#' ", roxygen2:::wrapString(roxygen2:::function_usage(fun, formals(fun)))) %>
#' @seealso \code{\link{ourPkg-deprecated}}
#' @keywords internal

#' @templateVar old remove_pop
#' @templateVar new gs_remove_gating_method
#' @template template-depr_pkg
NULL
#' Reverse the action of gating methods applied via \code{gs_add_gating_method}
#' 
#' This function provides an easy way to remove the gates and nodes created by the most
#' recent call to \code{\link{gs_add_gating_method}} on the specified \code{GatingSet} or \code{GatingSetList},
#' with a separate history being maintained for each such object. \code{gs_remove_gating_method} allows 
#' for repeated use, effectively serving as a multi-level undo function for \code{gs_add_gating_method}.
#' 
#' @name gs_remove_gating_method
#' @aliases remove_pop
#' @param gs The \code{GatingSet} or \code{GatingSetList} for which the most recent
#' \code{gs_add_gating_method} call should be reversed.
#' 
#' @usage 
#' gs_remove_gating_method(gs)
#' 
#' @seealso \code{\link{gs_add_gating_method}} \code{\link{gs_add_gating_method_init}}
#' @examples
#' \dontrun{
#'  # add quad gates 
#'  gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
#'  # Remove the gates and nodes resulting from that gs_add_gating_method call
#'  gs_remove_gating_method(gs)
#' } 
#' @importFrom flowWorkspace identifier
#' @export
gs_remove_gating_method <- function(gs){
  if(!(identifier(gs) %in% names(add_pop_history$records))){
    stop(paste("No gs_add_gating_method calls have been made for this GatingSet or GatingSetList"))
  }
  this_record <- add_pop_history$records[[identifier(gs)]]
  old_ids <- this_record[[length(this_record)-1]]
  new_ids <- this_record[[length(this_record)]]
  new_ids <- new_ids[!(new_ids %in% old_ids)] # setdiff won't work here
  lapply(new_ids, function(x) gs_pop_remove(x, gs = gs))
  
  ## If that succeeded, remove the records
  # If it's a GatingSetList, remove the records from all of its GatingSets
  if(is(gs, "GatingSetList")){
    lapply(gs, function(x){
      add_pop_history$records[[identifier(x)]][[length(add_pop_history$records[[identifier(x)]])]] <- NULL
      # If back to initial state, clear the add_pop history
      if(length(add_pop_history$records[[identifier(x)]]) <= 1){
        add_pop_history$records[[identifier(x)]] <- NULL
      }
    })
  }
  # Pop off the record
  add_pop_history$records[[identifier(gs)]][[length(this_record)]] <- NULL
  # If back to initial state, clear the add_pop history
  if(length(add_pop_history$records[[identifier(gs)]]) <= 1){
    add_pop_history$records[[identifier(gs)]] <- NULL
  }
  
  invisible(this_record)
} 

#' @export
remove_pop <- function(gs){
  .Deprecated("gs_remove_gating_method")
  gs_remove_gating_method(gs)
}

library(data.table)
library(utils)
library(tools)
library(flowCore)
library(ncdfFlow)
gtFile <- system.file("extdata/gating_template/tcell.csv", package = "openCyto")
expectResults <- readRDS("expectResults.rds")
dataDir <- system.file("extdata",package="flowWorkspaceData")
gs_dir <- list.files(dataDir, pattern = "gs_manual",full = TRUE)
gs <- load_gs(gs_dir)
data("GvHD")

#' show method for fcFilter
#' 
#' @param object \code{fcFilter}
#' show method for fcFilter
#' @export 
setMethod("show", sig = c("fcFilter"),
          definition = function(object) {
  cat("fcFilter:")
  callNextMethod(object)
})

setGeneric("posteriors", function(x, y, ...) standardGeneric("posteriors"))
#' get posteriors from a \code{fcFilter} object
#' 
#' get posteriors from a \code{fcFilter} object
#' @param x \code{fcFilter}
#' @param y \code{character} or \code{missing} that specifiy which channel to look for
#' 
#' @aliases 
#' posteriors,fcFilter,ANY-method
#' posteriors,fcFilter,character-method 
setMethod("posteriors", sig = c("fcFilter", "ANY"),
          definition = function(x, y = "missing") {
  x@posteriors
})

setMethod("posteriors", sig = c("fcFilter", "character"),
          definition = function(x, y) {
  post <- posteriors(x)
  
  postNames <- names(post)
  ind <- match(y, postNames)
  if (is.na(ind)) {
    stop("FcFilter not found for:", y)
  } else {
    post[[ind]]
  }
})

setGeneric("priors", function(x, y, ...) standardGeneric("priors"))
#' get priors from a \code{fcFilter} object
#' 
#' @param x \code{fcFilter} object
#' @param y \code{character} specifying channel name. if \code{missing} then extract priors for all the channels 
#'
#' @inheritParams posteriors
#' @aliases 
#' priors,fcFilter,ANY-method
#' priors,fcFilter,character-method
setMethod("priors", sig = c("fcFilter", "ANY"),
          definition = function(x, y = "missing") {
  x@priors
})

setMethod("priors", sig = c("fcFilter", "character"),
          definition = function(x, y) {
  prior <- priors(x)
  
  priorNames <- names(prior)
  ind <- match(y, priorNames)
  if (is.na(ind)) {
    stop("FcFilter not found for:", y)
  } else {
    prior[[ind]]
  }
}) 

#' @templateVar old prior_flowClust
#' @templateVar new prior_flowclust
#' @template template-depr_pkg
NULL
#' Elicits data-driven priors from a flowSet object for specified channels
#'
#' We elicit data-driven prior parameters from a \code{flowSet} object for
#' specified channels. For each sample in the \code{flowSet} object, we apply the
#' given \code{prior_method} to elicit the priors parameters.
#' 
#' Currently, we have implemented only two methods. In the case that one channel
#' is given, we use the kernel-density estimator (KDE) approach for each sample
#' to obtain \code{K} peaks from which we elicit prior parameters. Otherwise,
#' if more than one channel is specified, we apply K-Means to each of the samples
#' in the \code{flowSet} and aggregate the clusters to elicit the prior
#' parameters.
#'
#' In the rare case that a prior covariance matrix is singular, we shrink the
#' eigenvalues of the matrix slightly to ensure that it is positive definite. For
#' instance, if the \code{flow_set} has two samples, this case can occur. The
#' amount of shrinkage is controlled in \code{shrink}.
#'
#' @name prior_flowclust
#' @aliases prior_flowClust
#' @param flow_set a \code{flowSet} object
#' @param channels a character vector containing the channels in the
#' \code{flowSet} from which we elicit the prior parameters for the Student's t
#' mixture
#' @param prior_method the method to elicit the prior parameters
#' @param K the number of mixture components to identify
#' @param nu0 prior degrees of freedom of the Student's t mixture components.
#' @param w0 the number of prior pseudocounts of the Student's t mixture components. (only the first element is used and the rest is ignored at the moment)
#' @param shrink the amount of eigenvalue shrinkage to add in the case the prior
#' covariance matrices are singular. See details.
#' @param ... Additional arguments passed to the prior elicitation method selected
#' @return list of the necessary prior parameters
#' @export 
#' @examples 
#' \dontrun{
#' library(flowCore)
#' data(GvHD)
#' prior_flowclust(GvHD[1:3], c("FSC-H", "SSC-H"))
#' }
prior_flowclust <- function(flow_set, channels, prior_method = c("kmeans"),
                            K = 2, nu0 = 4, w0 = c(10,10), shrink = 1e-6, ...) {
  #pass only the first element of w0 since it will be replicated ..
  #that said, the second element never gets used at this moment
  if (length(channels) == 1) {
    prior_list <- .prior_flowClust1d(flow_set = flow_set, channel = channels,
                                    K = K, nu0 = nu0, w0 = w0[1], ...)
  } else {
    prior_method <- match.arg(prior_method)
    if (prior_method == "kmeans") {
      prior_list <- .prior_kmeans(flow_set = flow_set, channels = channels, K = K,
                                 nu0 = nu0, w0 = w0[1], ...)  
    }
    # In the rare case a covariance matrix is singular, we shrink the eigenvalues
    # of the matrix. The amount of shrinkage is controlled in 'shrink'.

    prior_list$Lambda0 <- .aaply(prior_list$Lambda0, 1, function(cov_mat) {
      if (is(try(solve(cov_mat), silent = TRUE), "try-error")) {
        cov_mat <- cov_mat + shrink * diag(nrow(cov_mat))
      }
      cov_mat
    })
    prior_list$Lambda0 <- unname(prior_list$Lambda0)
    
    #Lambda0 should be Kxpxp.. but it's reduced to pxp when K=1
    #correct the dimensions here
    L0<-prior_list$Lambda0
    if(length(dim(L0))==2){
      dim(L0)<-c(1,dim(L0))
      prior_list$Lambda0<-L0
    }
    
    prior_list$Omega0 <- .aaply(prior_list$Omega0, 1, function(cov_mat) {
      if (class(try(solve(cov_mat), silent = TRUE)) == "try-error") {
        cov_mat <- cov_mat + shrink * diag(nrow(cov_mat))
      }
      cov_mat
    })
    prior_list$Omega0 <- unname(prior_list$Omega0)
  }
 
  prior_list
   
}

#' @export
prior_flowClust <- function(flow_set, channels, prior_method = c("kmeans"),
                            K = 2, nu0 = 4, w0 = c(10,10), shrink = 1e-6, ...){
  .Deprecated("prior_flowclust")
  prior_flowclust(flow_set, channels, prior_method, K, nu0, w0, shrink, ...)
}

#' Elicits data-driven priors from a flowSet object for a specified channel
#'
#' We elicit data-driven prior parameters from a \code{flowSet} object for a
#' specified channel. For each sample in the \code{flowSet} object, we apply a
#' kernel-density estimator (KDE) and identify its local maxima (peaks). 
#' We then aggregate these peaks to elicit a prior
#' parameters for each of \code{K} mixture components.
#'
#' Here, we outline the approach used for prior elicitation. First, we apply a
#' KDE to each sample and extract all of its peaks (local maxima). It is
#' important to note that different samples may have a different number of
#' peaks. Our goal then is to align the peaks before aggregating the information
#' across all samples.  To do this, we utilize a technique similar to the peak
#' probability contrasts (PPC) method from Tibshirani et al (2004). Effectively,
#' we apply hierarchical clustering to the peaks from all samples to find
#' clusters of peaks. We compute the sample mean and variance of the peaks
#' within each cluster to elicit the prior means and its hyperprior variance,
#' respectively, for a \code{\link{flowClust}} mixture component. We elicit the
#' prior variance for each mixture component by first assigning the observations
#' within each sample to the nearest prior mean. Then, we compute the variance
#' of the observations within each cluster. Finally, we average the variances
#' corresponding to each mixture component across all samples in the
#' \code{flowSet} object.
#'
#' Following Tibshirani et al. (2004), we cluster the peaks from each sample
#' using complete-linkage hierarchical clustering. The linkage type can be
#' changed via the \code{hclust_method} argument. This argument is passed
#' directly to \code{\link{hclust}}.
#'
#' To cluster the peaks, we must cut the hierarchical tree by selecting either a
#' value for \code{K} or by providing a height of the tree to cut. By default,
#' we cut the tree using as the height the median of the distances between
#' adjacent peaks within each sample. This value can be changed via the
#' \code{hclust_height} argument and, if provided, will be passed to
#' \code{\link{cutree}}. Also, by default, the number of mixture components
#' \code{K} is \code{NULL} and is ignored.  However, if \code{K} is provided,
#' then it has priority over \code{hclust_height} and is passed instead directly
#' to \code{\link{cutree}}.
#'
#' To ensure that the KDEs are smooth, we recommend that the bandwidth set in
#' the \code{adjust} argument be sufficiently large. We have defaulted this
#' value to 2. If the bandwidth is not large enough, the KDE may contain
#' numerous bumps, resulting in erroneous peaks.
#'
#' @references Tibshirani, R et al. (2004), "Sample classification from protein
#' mass spectrometry, by 'peak probability contrasts'," Bioinformatics, 20, 17,
#' 3034-3044. \url{http://bioinformatics.oxfordjournals.org/content/20/17/3034}.
#' @param flow_set a \code{flowSet} object
#' @param channel the channel in the \code{flowSet} from which we elicit the
#' prior parameters for the Student's t mixture
#' @param K the number of mixture components to identify. By default, this value
#' is \code{NULL} and determined automatically
#' @param hclust_height the height of the \code{\link{hclust}} tree of peaks,
#' where the should be cut By default, we use the median of the distances
#' between adjacent peaks. If a value is specified, we pass it directly to
#' \code{\link{cutree}}.
#' @param clust_method the method used to cluster peaks together when for prior
#' elicitation. By default, \code{kmeans} is used. However, if \code{K} is not
#' specified, \code{hclust} will be used instead.
#' @param hclust_method the agglomeration method used in the hierarchical
#' clustering. This value is passed directly to \code{\link{hclust}}. Default is
#' complete linkage.
#' @param artificial a numeric vector containing prior means for artificial
#' mixture components. The remaining prior parameters for the artificial
#' components are copied directly from the most informative prior component
#' elicited. If \code{NULL} (default), no artificial prior components are added.
#' @param nu0 prior degrees of freedom of the Student's t mixture components.
#' @param w0 the number of prior pseudocounts of the Student's t mixture components.
#' @param adjust the bandwidth to use in the kernel density estimation. See
#' \code{\link{density}} for more information.
#' @param min a numeric value that sets the lower bound for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param max a numeric value that sets the upper bound for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param vague \code{logical} Whether to elicit a vague prior. If \code{TRUE}, we first calculate the median of standard
#'                              deviations from all flowFrames. Then, we divide the overall standard
#'                              deviation by the number of groups to the scale the standard deviation.
#' @return list of prior parameters
#' @rdname prior_flowClust1d
#' @noRd 
.prior_flowClust1d <- function(flow_set, channel, K = NULL, hclust_height = NULL,
                              clust_method = c("kmeans", "hclust"),
                              hclust_method = "complete", artificial = NULL,
                              nu0 = 4, w0 = 10, adjust = 2, min = -200,
                              max = NULL, vague = TRUE) {

  channel <- as.character(channel)
  clust_method <- match.arg(clust_method)
  if (length(channel) != 1) {
    stop("There can be only 1...channel.")
  }

  if (!(is.null(min) && is.null(max))) {
    flow_set <- .truncate_flowset(flow_set, channels = channel, min = min, max = max)
  }
  # For each sample in 'flow_set', we identify the peaks after smoothing.
  peaks <- fsApply(flow_set, function(flow_frame, adjust) {
    x <- exprs(flow_frame)[, channel]
    peaks_found <- .find_peaks(x, adjust = adjust)[, "x"]
    
    # If K is specified and is smaller than the number of peaks found,
    # we keep only the K largest peaks from the sample.
    if (!is.null(K) && length(peaks_found) > K) {
      peaks_found <- peaks_found[seq_len(K)]
    }
    peaks_found
  }, adjust = adjust, simplify = FALSE)
  peaks_collapsed <- as.vector(do.call(c, peaks))

  # Remove any NAs that may have resulted in the peaks.
  peaks_collapsed <- peaks_collapsed[!is.na(peaks_collapsed)]

  # For each sample, we sort the peaks and find the distance between the
  # adjacent peaks.
  peaks_dist <- lapply(peaks, function(x) {
    diff(sort(x))
  })
  peaks_dist <- do.call(c, peaks_dist)

  if (is.null(K) || clust_method == "hclust") {
    # Applies hierarchical clustering to the peaks.
    hclust_out <- hclust(dist(peaks_collapsed), method = hclust_method)

    # To cluster the peaks, we must cut the hierarchical tree. By default, we
    # use the median of the distances between adjacent peaks.
    # By default, 'K' is 'NULL' and is ignored by 'cutree'. If 'K' is provided,
    # then it has priority over 'hclust_height'.
    if (is.null(hclust_height)) {
      hclust_height <- median(peaks_dist)
    }
    clust_labels <- factor(cutree(hclust_out, k = K, h = hclust_height))
    prior_means <- as.numeric(tapply(peaks_collapsed, clust_labels, mean))
  } else if (clust_method == "kmeans") {
    # Applies kmeans clustering to the peaks.
    kmeans_out <- kmeans(x = peaks_collapsed, centers = K, nstart = 10,algorithm="MacQueen")
    clust_labels <- factor(kmeans_out$cluster)
    prior_means <- as.vector(kmeans_out$centers)
  }
  K <- nlevels(clust_labels)

  if (vague) {
    prior_means <- sort(prior_means)

    # To elicit a vague prior, we first calculate the median of standard
    # deviations from all flowFrames. Then, we divide the overall standard
    # deviation by the number of groups to the scale the standard deviation. 
    sd_x <- fsApply(flow_set, function(flow_frame) {
      sd(exprs(flow_frame)[, channel])
    })
    sd_x <- median(sd_x, na.rm = TRUE) / K
    var_x <- sd_x^2

    hyperprior_vars <- rep.int(var_x, K)
    prior_vars <- rep.int(var_x, K)
                   
  } else {
    # For each cluster of peaks, we elicit the hyperprior variance (Omega0) for
    # the prior mean Mu0 by computing the variance of each cluster.
    hyperprior_vars <- as.numeric(tapply(peaks_collapsed, clust_labels, var))

    # Because the cluster labels assigned by the hierarchical clustering algorithm
    # are arbitrary, the 'prior_means' are not necessarily sorted. Here, we sort
    # them and then apply the new odering to the other prior parameters.
    prior_order <- order(prior_means)
    prior_means <- prior_means[prior_order]
    hyperprior_vars <- hyperprior_vars[prior_order]

    # Elicitation of prior variances for each mixture component:
    # For each flowFrame in the original flowSet object, we cluster its
    # observations by assigning them to the nearest mean in 'prior_means.' Then,
    # we compute the variance of the observations within each cluster. Finally, we
    # aggregate the variances across all of the flowFrame objects.
    prior_vars <- fsApply(flow_set, function(flow_frame, prior_means) {
      x <- exprs(flow_frame)[, channel]

      # To determine the nearest mean, we find the midpoints of the peaks and then
      # 'cut' the observations into the regions.
      peak_midpoints <- rowMeans(embed(prior_means, 2))
      labels <- cut(x, breaks = c(-Inf, peak_midpoints, Inf))
      tapply(x, labels, var)
    }, prior_means = prior_means)
    prior_vars <- as.numeric(colMeans(prior_vars, na.rm = TRUE))
  }

  # Here, we add any 'artificial' prior components, if provided, and increment
  # 'K' accordingly.
  if (!is.null(artificial)) {
    prior_df <- cbind.data.frame(Mu0 = prior_means, Omega0 = hyperprior_vars,
                                 Lambda0 = prior_vars)

    # For the remaining prior parameters for each artificial component, we select
    # minimum values of Omega0 and Lambda0.
    min_Omega0 <- min(prior_df$Omega0)
    min_Lambda0 <- min(prior_df$Lambda0)
    prior_df <- rbind(prior_df, cbind(Mu0 = artificial, Omega0 = min_Omega0,
                                      Lambda0 = min_Lambda0))

    prior_df <- prior_df[order(prior_df$Mu0), ]

    prior_means <- prior_df$Mu0
    hyperprior_vars <- prior_df$Omega0
    prior_vars <- prior_df$Lambda0

    K <- length(prior_means)
  }

  # Mu0 dimensions: K x p (p is the number of features. Here, p = 1 for 1D)
  Mu0 <- matrix(prior_means, K, 1)

  # Lambda0 dimensions: K x p x p
  Lambda0 <- array(prior_vars, c(K, 1, 1))

  # Omega0 dimensions: K x p x p
  Omega0 <- array(hyperprior_vars, c(K, 1, 1))

  # We assume that the degrees of freedom is the same for each mixture component.
  nu0 <- rep(nu0, K)

  # We assume that the prior probability of mixture component membership is equal
  # across all mixture components and use 10 pseudocounts.
  w0 <- rep(w0, K)

  list(Mu0 = Mu0, Lambda0 = Lambda0, Omega0 = Omega0, nu0 = nu0, w0 = w0)
}

#' Elicits data-driven priors from a flowSet object for specified channels using
#' the K-Means clustering algorithm
#'
#' We elicit data-driven prior parameters from a \code{flowSet} object for
#' specified channels. For each sample in the \code{flowSet} object, we apply
#' \code{kmeans} to obtain \code{K} clusters. From each cluster, we determine its
#' centroid and the sample covariance matrix. We then aggregate these two sample
#' moments across all samples for each cluster. 
#'
#' Because the cluster labels returned from \code{kmeans} are arbitrary, we align
#' the clusters based on the centroids that are closest to a randomly selected
#' reference sample. We apply the Hungarian algorithm implemented using the
#' \code{solve_LSAP} function from the \code{clue} package to assist with the
#' alignment.
#'
#' If each frame within \code{flow_set} has a large number of cells, the
#' computational costs of \code{kmeans} can be a burden. We provide the option
#' to randomly select \code{pct}, a percentage of the cells from each flow frame
#' to which \code{kmeans} is applied.
#'
#' @param flow_set a \code{flowSet} object
#' @param channels a character vector containing the channels in the
#' \code{flowSet} from which we elicit the prior parameters for the Student's t
#' mixture
#' @param K the number of mixture components to identify
#' @param nu0 prior degrees of freedom of the Student's t mixture components.
#' @param w0 the number of prior pseudocounts of the Student's t mixture
#' components.
#' @param nstart number of random starts used by \code{kmeans} algorithm
#' @param pct percentage of randomly selected cells in each \code{flowFrame}
#' that is used to elicit the prior parameters. The value should must be greater
#' than 0 and less than or equal to 1.
#' @param min a numeric vector that sets the lower bounds for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param max a numeric vector that sets the upper bounds for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param ... Additional arguments passed to \code{kmeans}
#' @return list of \code{flowClust} prior parameters
#' @rdname prior_kmeans
#' @noRd 
.prior_kmeans <- function(flow_set, channels, K, nu0 = 4, w0 = 10, nstart = 10,
                         pct = 0.1, min = NULL, max = NULL, ...) {

  channels <- as.character(channels)
  
  # Truncates flow_set before eliciting priors when necessary 
  if (!(is.null(min) && is.null(max))) {
    flow_set <- .truncate_flowset(flow_set, channels = channels, min = min,
                                 max = max)
  }
  # For each randomly selected sample in the flow_set, we apply K-means with to
  # find K clusters and retain additional summary statistics to elicit the prior
  # parameters for flowClust.
  num_samples <- length(flow_set)

  # The number of features, corresponding to the number of 'channels'
  p <- length(channels)
  
  kmeans_summary <- lapply(seq_len(num_samples), function(i) {
    # Grabs the channel data for the ith sample and randomly selects the
    # specified percentage of cells to which we apply 'kmeans'
    x <- as.matrix(exprs(flow_set[[i]])[, channels])
    x <- as.matrix(x[sample(seq_len(nrow(x)), pct * nrow(x)), ])

    # If there are too few rows in 'x', 'kmeans' will throw an error.
    # Instead, we do not elicit summary statistics from this sample and return NA.
    if (K >= nrow(x)) {
      warning("The number of clusters 'K' exceeds the number of subsampled rows.",
              "Try increasing the subsample 'pct'.")
      return(NA)
    }
    
    kmeans_out <- kmeans(x = x, centers = K, nstart = nstart, algorithm="MacQueen",...)

    cluster_centroids <- kmeans_out$centers
    cluster_sizes <- kmeans_out$size

    cluster_covs <- tapply(seq_len(nrow(x)), kmeans_out$cluster, function(i) {
      # In the case that a singleton cluster results, a vague covariance matrix
      # is reported: the covariance matrix of the entire data set.
      if (length(i) >= 2) {
        cov(as.matrix(x[i, ]))
      } else {
        warning("A singleton cluster resulted. Using vague covariance matrix.",
                "Try increasing the subsample 'pct'.")
        cov(x)
      }
    })

    list(centroids = cluster_centroids, sizes = cluster_sizes, covs = cluster_covs)
  })

  # Removes samples marked with NA that had too few observations.
  kmeans_summary <- kmeans_summary[!is.na(kmeans_summary)]

  # We select the flowFrame with the largest number of observations as the
  # reference sample.
  sample_sizes <- sapply(kmeans_summary, function(x) sum(x$sizes))
  ref_sample <- which.max(sample_sizes)
  kmeans_ref_sample <- kmeans_summary[[ref_sample]]
  kmeans_ref_sample$covs <- simplify2array(unname(kmeans_ref_sample$covs))

  # Converts covariance matrices to a 3D array (K x p x p)
  # We must treat the case of 1 channel as a special case to ensure the
  # dimensions of the 3-dimensional array is correct.
  if (p > 1) {
    kmeans_ref_sample$covs <- aperm(kmeans_ref_sample$covs, c(3, 1, 2))
  } else {
    kmeans_ref_sample$covs <- array(kmeans_ref_sample$covs, dim = c(p, 1, 1))
  }
  ref_centroids <- kmeans_ref_sample$centroids

  # Because the cluster labels returned from 'kmeans' are arbitrary, we align the
  # clusters based on the centroids that are closest to the reference sample.
  kmeans_summary <- kmeans_summary[-ref_sample]
  kmeans_centroids <- lapply(kmeans_summary, "[[", "centroids")
  
  # For each sample, we obtain the alignment indices.
    ref_indices <- lapply(kmeans_centroids, function(centroids) {
      dist_centroids <- as.matrix(dist(rbind(ref_centroids, centroids)))
      dist_centroids <- dist_centroids[seq_len(K), seq_len(K) + K,drop=FALSE]
      as.vector(solve_LSAP(dist_centroids))
    })
  # Now, we align the 'kmeans' summary information using the alignment indices.
  kmeans_summary <- mapply(function(kmeans_sample, align_idx) {
    kmeans_sample$centroids <- kmeans_sample$centroids[align_idx, ]
    kmeans_sample$sizes <- kmeans_sample$sizes[align_idx]
    kmeans_sample$covs <- unname(kmeans_sample$covs[align_idx])

    # Converts covariance matrices to a 3D array (K x p x p) As above,
    # we must treat the case of 1 channel as a special case to ensure
    # the dimensions of the 3-dimensional array is correct.
    kmeans_sample$covs <- simplify2array(kmeans_sample$covs)
    if (p > 1) {
      kmeans_sample$covs <- aperm(kmeans_sample$covs, c(3, 1, 2))
    } else {
      kmeans_sample$covs <- array(kmeans_sample$covs, dim = c(p, 1, 1))
    }

    kmeans_sample
  }, kmeans_summary, ref_indices, SIMPLIFY = FALSE)

  # Appends the reference sample to the 'kmeans' summary list.
  kmeans_summary <- c(kmeans_summary, list(kmeans_ref_sample))

  # Calculates the total number of observations assigned to each cluster across
  # all samples.
  kmeans_sizes <- do.call(rbind, lapply(kmeans_summary, "[[", "sizes"))
  kmeans_sizes_sum <- colSums(kmeans_sizes)

  # For each cluster after alignment, we extract a matrix of the centroids from
  # each sample.
  kmeans_centroids <- lapply(kmeans_summary, "[[", "centroids")
  centroids_split <- lapply(kmeans_centroids, split, f = seq_len(K))
  centroids_split <- lapply(seq_len(K), function(k) {
    do.call(rbind, lapply(centroids_split, function(x) x[[k]]))
  })

  # Next, for each cluster, we compute summary statistics of the centroids from
  # each sample. The statistics are weighted by the number of samples. This is
  # especially imporant when the number of samples is small and some of them
  # have few observations.
  weighted_stats <- lapply(centroids_split, cov.wt, wt = sample_sizes,
                           method = "ML")

  # Mu0 dimensions: K x p
  # Prior Mean
  # We aggregate the cluster centroids across all samples weighted by the
  # cluster sample sizes to elicit Mu0.
  centroids_aggregate <- lapply(weighted_stats, "[[", "center")
  Mu0 <- do.call(rbind, centroids_aggregate)
  Mu0 <- matrix(Mu0, K, p)

  # Omega0 dimensions: K x p x p
  # Hyperprior covariance matrix for the prior mean Mu0
  # For each cluster we calculate the weighted covariance matrix of the cluster
  # centroids across all samples to elicit Omega0.
  Omega0 <- lapply(weighted_stats, "[[", "cov")
  Omega0 <- array(do.call(rbind, lapply(Omega0, as.vector)), dim = c(K, p, p))

  # Lambda0 dimensions: K x p x p
  # Prior Covariance Matrix
  # For each cluster, we pool the covariance matrices from all samples to elicit
  # Lambda0.
  kmeans_scatter <- lapply(kmeans_summary, function(x) x$sizes * x$covs)
  Lambda0 <- Reduce("+", kmeans_scatter) / kmeans_sizes_sum
  Lambda0 <- array(Lambda0, c(K, p, p))

  # We assume that the degrees of freedom is the same for each mixture component.
  nu0 <- rep(nu0, K)

  # We assume that the prior probability of mixture component membership is equal
  # across all mixture components and use 10 pseudocounts.
  w0 <- rep(w0, K)

  list(Mu0 = Mu0, Lambda0 = Lambda0, Omega0 = Omega0, nu0 = nu0, w0 = w0)
}

#' Finds the local maxima (peaks) in the given vector after smoothing the data
#' with a kernel density estimator.
#'
#' First, we smooth the data using kernel density estimation (KDE) with the
#' \code{\link{density}} function. Then, we find every local maxima such that the
#' density is concave (downward).
#'
#' The \code{num_peaks} argument returns only the largest peaks. If
#' \code{num_peaks} is greater than the number of peaks found, then all the peaks
#' are returned, and a warning is issued.
#'
#' Effectively, we find the local maxima with a discrete analogue to a second
#' derivative applied to the KDE. For details, see this StackOverflow post:
#' \url{http://bit.ly/Zbl7LV}.
#'
#' @param x numeric vector
#' @param y numeric vector. If given, it is treated as the density values for
#' \code{x}. The length of \code{y} must equal that of \code{x}.
#' @param num_peaks the number of peaks to find. By default, all peaks are
#' returned. See details.
#' @param adjust the bandwidth to use in the kernel density estimation. See
#' \code{\link{density}} for more information.
#' @param ... additional arguments passed to the \code{\link{density}} function
#' @return a \code{data.frame} that contains the peaks(and their density heights) attained. The peaks are sorted in
#' descending order based on the density heights.
#' @examples
#' library(flowClust)
#' set.seed(42)
#' # 2 peaks with a minor hump
#' y <- flowStats::SimulateMixture(10000, c(.5, .3, .2), c(2, 5, 7), c(1, 1, 1), nu = 10)
#' plot(density(y))
#' peaks <- .find_peaks(y)
#' abline(v = peaks[, "x"], col = "red")
#' @noRd 
.find_peaks <- function(x, y = NULL, num_peaks = NULL, adjust = 2, plot = FALSE, ...) {
  x <- as.vector(x)

  if (length(x) < 2) {
    warning("At least 2 observations must be given in 'x' to find peaks.")
    return(NA)
  }

  if (is.null(y)) {
    dens <- density(x, adjust = adjust, ...)
  } else {
    y <- as.vector(y)
    if (length(x) != length(y)) {
      stop("The lengths of 'x' and 'y' must be equal.")
    }
    dens <- list(x = x, y = y)
  }

  # Discrete analogue to a second derivative applied to the KDE. See details.
  second_deriv <- diff(sign(diff(dens$y)))
  which_maxima <- which(second_deriv == -2) + 1

  # The 'density' function can consider observations outside the observed range.
  # In rare cases, this can actually yield peaks outside this range.  We remove
  # any such peaks.
  which_maxima <- which_maxima[findInterval(dens$x[which_maxima], range(x)) == 1]

  # Next, we sort the peaks in descending order based on the density heights.
  which_maxima <- which_maxima[order(dens$y[which_maxima], decreasing = TRUE)]
  
  # Returns the local maxima. If there are none, we return 'NA' instead.
  if (length(which_maxima) > 0) {
    peaks <- dens$x[which_maxima]
    if (is.null(num_peaks) || num_peaks > length(peaks)) {
      num_peaks <- length(peaks)
    }
    peaks <- peaks[seq_len(num_peaks)]
  } else {
    peaks <- NA
  }
  
  peaks <- data.frame(x = peaks, y = dens$y[which_maxima][seq_len(num_peaks)])
  if(plot){
    plot(dens, main = paste("adjust =" ,  adjust))
    points(peaks, ,col = "red")  
  }
  
  peaks  
}

#' Finds the local minima (valleys) in the given vector after smoothing the data
#' with a kernel density estimator.
#'
#' First, we smooth the data using kernel density estimation (KDE) with the
#' \code{\link{density}} function. Then, we find every local minima such that the
#' density is concave (downward).
#'
#' Effectively, we find the local minima with a discrete analogue to a second
#' derivative applied to the KDE. For details, see this StackOverflow post:
#' \url{http://bit.ly/Zbl7LV}.
#'
#' @param x numeric vector
#' @param y numeric vector. If given, it is treated as the density values for
#' \code{x}. The length of \code{y} must equal that of \code{x}.
#' @param num_valleys the number of valleys to find. By default, all valleys are
#' returned. See details.
#' @param adjust the bandwidth to use in the kernel density estimation. See
#' \code{\link{density}} for more information.
#' @param ... additional arguments passed to the \code{\link{density}} function
#' @return the values where the valleys are attained.
#' @examples
#' library(flowClust)
#' set.seed(42)
#' # 3 peaks and 2 valleys
#' y <- flowStats::SimulateMixture(10000, c(.25, .5, .25), c(1, 5, 9), c(1, 1, 1), nu = 10)
#' plot(density(y))
#' valleys <- .find_valleys(y)
#' abline(v = valleys, col = "red")
#' @noRd 
.find_valleys <- function(x, y = NULL, num_valleys = NULL, adjust = 2, ...) {

  x <- as.vector(x)

  if (length(x) < 2) {
    warning("At least 2 observations must be given in 'x' to find valleys.")
    return(NA)
  }
  
  if (is.null(y)) {
    dens <- density(x, adjust = adjust, ...)
  } else {
    y <- as.vector(y)
    if (length(x) != length(y)) {
      stop("The lengths of 'x' and 'y' must be equal.")
    }
    dens <- list(x = x, y = y)
  }

  # Discrete analogue to a second derivative applied to the KDE. See details.
  second_deriv <- diff(sign(diff(dens$y)))
  which_minima <- which(second_deriv == 2) + 1

  # The 'density' function can consider observations outside the observed range.
  # In rare cases, this can actually yield valleys outside this range. We remove
  # any such valleys.
  which_minima <- which_minima[findInterval(dens$x[which_minima], range(x)) == 1]

  # Next, we sort the valleys in descending order based on the density heights.
  which_minima <- which_minima[order(dens$y[which_minima], decreasing = FALSE)]

  # Returns the local minima. If there are none, we return 'NA' instead.
  if (length(which_minima) > 0) {
    valleys <- dens$x[which_minima]
    if (is.null(num_valleys) || num_valleys > length(valleys)) {
      num_valleys <- length(valleys)
    }
    valleys <- valleys[seq_len(num_valleys)]
  } else {
    valleys <- NA
  }
  valleys
}

# This file contains all wrapper methods for dispatching data and arguments to
# gating/preprocessing algorithms.


#' An adapter to connect the gating wrapper function with the \link{gating} method
#' 
#' It coerce the input (\code{flowSet}) to a single \code{flowFrame} and apply the gating wrapper function
#' then replicate the gates across samples from the \code{flowSet}.
#' 
#' @inheritParams .prior_flowclust
#' @param pp_res preprocessing result produced by the \code{preprocessing} method
#' @param gFunc \code{character} function name of the wrapper function to be invoked
#' @param popAlias \code{character} the population names that are used to determine how many gates to be expected from the gating function 
#' @param gFunc_args arguments to be passed to wrapper function('gFunc')
#' 
#' @return a \code{list} of \code{filter}s
#' @noRd 
.gating_adaptor <- function(fs, pp_res, gFunc, popAlias, channels, gFunc_args){
    require(openCyto)  #since it is going to be invoked by MPI, better load it
    #coercing
    sn <- sampleNames(fs)
    fr <- as(fs,"flowFrame")
    total <- nrow(fr)
    #parse the subSample argument from the gating function argument list
    subSample <- gFunc_args[["subSample"]]
    gFunc_args[["subSample"]] <- NULL #prevent it from passing down to the gFunc
    if(!is.null(subSample)){
      if(is.numeric(subSample)){
        if(subSample > 1){
          samp.ind <- sample.int(total, subSample)
          fr <- fr[samp.ind]
        }else if(subSample >0 ){
          samp.ind <- sample.int(total, subSample * total)
          fr <- fr[samp.ind]
        }else
          stop("invalid 'subSample' argument: ", subSample)
      }else
        stop("invalid 'subSample' argument: ", subSample)
    }
    
    openCyto.options <- getOption("openCyto")
    minEvents <- openCyto.options[["gating"]][["minEvents"]]
    # Allow rowwise specification of minEvents
    row_minEvents <- gFunc_args[["openCyto.minEvents"]]
    if(is.call(row_minEvents)) {
      row_minEvents <- eval(row_minEvents)
    }
    if(!is.null(row_minEvents)){
      minEvents <- row_minEvents
      gFunc_args[["openCyto.minEvents"]] <- NULL
    }
    
    if(is.null(minEvents))
      minEvents <- 0
    if(nrow(fr) <= minEvents){
      warning(paste(sn, collapse =","), ": Not enough events to proceed the data-driven gating!Returning a dummy gate instead.")
      
      #create dummy rectangleGate
      #TODO: move channels to ... to deprecate x,y channel
#      dots <- list(...)
#      channels <- dots$channels
      channels <- as.vector(na.omit(channels))
      nDim <- length(channels) 
      
      if(nDim ==  1)
        gate_coordinates <- list(c(-Inf, -Inf))
      else if(nDim ==  2)
        gate_coordinates <- list(c(-Inf, -Inf), c(-Inf, -Inf))
      else
        stop(nDim, " dimensional gating is not supported yet!")

      names(gate_coordinates) <- channels
      filterRes <- rectangleGate(gate_coordinates)
    
      #this is flowClust-specific operation, which
      # be abstracted out of this framework
      
      if(grepl("(flowClust|gate_flowclust)[\\._][12]d", gFunc))
        filterRes <- fcRectangleGate(filterRes, priors = list(), posts = list())
      
      nPop <- length(popAlias)
      filterResType <- ifelse(nPop == 1, "filter", "filters")
      if(filterResType == "filters"){
        filterRes <- filters(lapply(1:nPop, function(i)filterRes))
      }
      
    }else{
      if(!.isRegistered(gFunc)){
        stop(sprintf("Can't gate using unregistered method %s",gFunc))
      }
      gFunc_resolved <- .resolve_plugin_function(gFunc)
      filterRes <- try(do.call(gFunc_resolved, c(list(fr = fr
                            , pp_res = pp_res
                            , channels = channels
                            )
                        , gFunc_args
                        )
                      )              
          , silent = TRUE
          )
        
    }
      
    
    resType <- class(filterRes)    
    if(extends(resType, "filter")||extends(resType, "filters")||extends(resType, "logical")||extends(resType, "factor")){

      #replicate the filter across samples
      list(sapply(sampleNames(fs),function(i)filterRes, simplify = FALSE))      
    }else{
      stop("failed at ",paste0(sn), "\n", filterRes)
    }
#

}
#' wrapper for \link[singletGate]{singletGate}
#' 
#' @param pp_res not used
#' @param fr \code{flowFrame} object as a data input 
#' @param ... arguments to be passed to \link[singletGate]{singletGate}
#' @inheritParams .prior_flowclust
#' @return a \code{filter} object
#' @noRd 
.singletGate <- function(fr, pp_res = NULL, channels, ...) {
  

  # Creates a list of polygon gates based on the prediction bands at the minimum
  # and maximum x_channel observation using a robust linear model trained 

  gate_singlet(fr, area = channels[1], height = channels[2], ...)
}

#' boundary gating function
#' 
#' It essentially constructs a rectangle gate from input range (min, max)
#'
#' @description It is useful for filtering out very large signal (e.g. FSC,SSC, Time)
#'  
#' @inheritParams .singletGate
#' @param min,max the range input for constructing the \code{rectangleGate}
#' @param ... other arguments (not used.)
#' @return a \code{filter} object
#' @noRd 
.boundary <- function(fr, pp_res = NULL, channels, min = NULL, max = NULL, ...) {
  num_channels <- length(channels)
  
  if(!num_channels %in%1:2)
    stop("invalid number of channels for boundary!")
  
  
  if (is.null(min)) {
    min <- rep(-Inf, num_channels)
  }
  if (is.null(max)) {
    max <- rep(Inf, num_channels)
  }
  
  if (!(num_channels == length(min) && num_channels == length(max))) {
    stop("The lengths of 'min' and 'max' must match the number of 'channels' given.")
  }
  
  gate_coordinates <- lapply(seq_len(num_channels), function(i) {
        c(min[i], max[i])
      })
  names(gate_coordinates) <- channels
  
  rectangleGate(gate_coordinates)
}

#' wrapper for flowClust.1d
#' 
#' It does some parameter preprocessing before calling the flowClust.1d
#' 
#' @param fr \code{flowFrame} object as a data input 
#' @param ... arguments to be passed to \link{flowClust.1d}
#' @param xChannel must be \code{NA}.
#' @param yChannel the dimension used for gating
#'  
#' @inheritParams .gating_adaptor 
#' @inheritParams .prior_flowclust
#' 
#' @return a \code{filter} object
#' @noRd 
.gate_flowclust_1d <- function(fr, pp_res = NULL, channels,...) {
  
  if(length(channels) != 1)
    stop("invalid number of channels for gate_flowclust_1d!")
  prior <- pp_res
  
  priorList <- list()
  if(!is.null(prior))
    prior <- prior[[channels]]
  
  args <- list(...)
  # parse arguments for flowClust
  
  # If any of 'K, 'neg' or 'pos' are given, we extract them from the
  # 'args' and coerce them as integers. Otherwise, they are defaulted to
  # NULL.
  # NOTE: the named elements within 'args' are coerced to lowercase.
  K <- neg_cluster <- pos_cluster <- NULL
  
  if ("K" %in% names(args)) {
    K <- as.integer(args["K"])
  }
  if ("neg" %in% names(args)) {
    neg_cluster <- as.integer(args["neg"])
    args[["neg"]] <- NULL
  }
  if ("pos" %in% names(args)) {
    pos_cluster <- as.integer(args["pos"])
    args[["pos"]] <- NULL
  }
  
  # Examines the various cases of pos, neg and K.
  # In the case that K is not given, we automatically select it.
  # Exception: If both 'pos' and 'neg' are given, we set: K <- pos + neg
  if (is.null(K)) {
    if (!is.null(neg_cluster) && !is.null(pos_cluster)) {
      K <- neg_cluster + pos_cluster
    }
  } else {
    # If pos and neg are given, throw a warning and set: K <- pos + neg
    # In the case that one of pos and neg are given and either exceeds K,
    # we throw an error.         
    if (!is.null(neg_cluster) && !is.null(pos_cluster)) {
      warning("Values given for 'K', 'neg', and 'pos'. Setting K = neg + pos")
      K <- neg_cluster + pos_cluster
    } else if (!is.null(pos_cluster) && K < pos_cluster) {
      stop("The number of positive clusters exceeds 'K'.")
    } else if (!is.null(neg_cluster) && K < neg_cluster) {
      stop("The number of negative clusters exceeds 'K'.")
    }
  }
  
  args[["neg_cluster"]] <- neg_cluster
#        args[["pos_cluster"]] <- pos_cluster
  
#  args[["k"]] <- NULL
  args[["K"]] <- K
  
  # If 'min' and/or 'max' are given, we pass this value along to the
  # prior-elicitation method as well as flowClust. Otherwise, these values
  # are set to NULL.
  min_values <- -Inf
  max_values <- Inf
  if ("min" %in% names(args)) {
    min_values <- as.numeric(args["min"])
  }
  if ("max" %in% names(args)) {
    max_values <- as.numeric(args["max"])
  }                          
  
  # If the number of positive clusters is 0 and no cutpoint method has been
  # specified, we use the quantile gate by default.
  if (!is.null(pos_cluster) && pos_cluster == 0 && is.null(args[["cutpoint_method"]])) {
    args[["cutpoint_method"]] <- "quantile"
  }
  
  
    # 1d gate
  gate <- do.call("gate_flowclust_1d"
            ,args = c(list(fr = fr
                        ,params = channels
                        ,prior = prior
                        )
                      ,args
                    )
           )
  
}

.flowClust.1d <- .gate_flowclust_1d


#' wrapper for mindensity
#' 
#' It does some parameter preprocessing before calling the mindensity
#' 

#' @param ... arguments to be passed to \link{mindensity}
#' @inheritParams .gate_flowclust_1d 
#' 
#' @return a \code{filter} object
#' @noRd 
.gate_mindensity <- function(fr, pp_res = NULL, channels, ...) {
  
 if(length(channels) != 1)
   stop("invalid number of channels for mindensity!")
  gate <- gate_mindensity(fr, channel = channels, ...)
#  .gateToFilterResult(fr, yChannel, gate, positive)
  gate
}

.mindensity <- .gate_mindensity

#' wrapper for gate_flowclust_2d
#' 
#' It does some parameter preprocessing before calling the gate_flowclust_wd
#' 
#' @param ... arguments to be passed to \link{gate_flowclust_2d}
#' @inheritParams .gate_flowclust_1d 
#' 
#' @return a \code{filter} object
#' @noRd 
.gate_flowclust_2d <- function(fr, pp_res = NULL, channels, ...) {
  if(length(channels) != 2)
    stop("invalid number of channels for gate_flowclust_2d!")
  xChannel <- channels[1]
  yChannel <- channels[2]
  args <- list(...)
  # get the value of neg and pos
  if (is.element(c("K"), names(args))) {
    K <- as.integer(args["K"])
  } else {
    message("'K' argument is missing! Using default setting: K = 2")
    K <- 2
  
  }
  
  args[["K"]] <- K

  if (is.null(pp_res)) {
    usePrior <- "no"
    pp_res <- list(NA)
  } else {
    usePrior <- "yes"
  }
  
  do.call("gate_flowclust_2d"
      ,args = c(list(fr = fr
                    , xChannel = xChannel
                    , yChannel = yChannel
                    , usePrior = usePrior
                    ,prior = pp_res
                    )
                ,args
              )
        )
  
}

.flowClust.2d <- .gate_flowclust_2d

#' wrapper for rangeGate (deprecated)
#' 
#' It does some parameter preprocessing before calling the rangeGate
#' 
#' @param pp_res not used
#' @inheritParams .gate_flowclust_1d 
#' 
#' @return a \code{filter} object
#' @noRd 
.rangeGate <- function(fr, pp_res = NULL, channels,  ...) {
  
  .Defunct("mindensity")
}
#' wrapper for quantileGate
#' 
#' It does some parameter preprocessing before calling the quantileGate
#' 
#' @param ... arguments to be passed to \link{quantileGate}
#' @inheritParams .gate_flowclust_1d 
#' 
#' @return a \code{filter} object
#' @noRd 
.gate_quantile <- function(fr, pp_res = NULL, channels, ...) {
  
  gate_quantile(fr = fr, channel = channels, ...)
}

.quantileGate <- .gate_quantile

.gate_quad_tmix <- function(fr, pp_res = NULL, channels, ...) {
  if(length(channels) != 2)
    stop("invalid number of channels for quadGate.tmix!")
  gate_quad_tmix(fr, channels, ...)
}

.quadGate.tmix <- .gate_quad_tmix

.gate_quad_sequential <- function(fr, pp_res = NULL, channels, ...){
  quadrants <- gate_quad_sequential(fr, channels, ...)
  quadrants[[1]]@filterId <- paste0(channels[[1]], "-", channels[[2]], "+")
  quadrants[[2]]@filterId <- paste0(channels[[1]], "+", channels[[2]], "+")
  quadrants[[3]]@filterId <- paste0(channels[[1]], "+", channels[[2]], "-")
  quadrants[[4]]@filterId <- paste0(channels[[1]], "-", channels[[2]], "-")
  quadrants
}

.quadGate.seq <- .gate_quad_sequential

############################
# preprocessing wrappers
#########################
#' 
#' @inheritParams .prior_flowclust 
#' 
#' @return \code{NULL}
#' @noRd 
.warpSet <- function(fs, gs, gm, channels, groupBy, isCollapse, stains, ...){
  .Defunct()
  return (NULL)
 }
#'  wrapper for prior_flowclust
#' 
#'  This wrapper does some parameter preprocessing before calls \link{prior_flowclust}
#' 
#' @param fs \code{flowSet} or \code{ncdfFlowSet} object
#' @param gs \code{GatingSet}
#' @param gm \code{gtMethod}
#' @param xChannel,yChannel \code{character} specifying the dimensions of flow data used by \code{prior_flowclust}
#' @param prior_source \code{character} specifying the ancester node from where the prior is elicited.
#' @param neg,pos \code{numeric} specifying how many peaks are expected on positive and negative sides from 1d density profile
#' @inheritParams .prior_flowClust1d 
#' 
#' @return a \code{list} of priors, see \link{prior_flowClust} for more details
#' @noRd 
.prior_flowclust <- function(fs, gs, gm, channels, groupBy, isCollapse
     , prior_source = NULL
     , K = NULL
     , neg, pos
     , min, max
     , ...){
   prior_list <- list()
   nChannel <- length(channels)
   if(nChannel == 1){
     xChannel <- NULL
     yChannel <- channels
   }else if(nChannel == 2){
     xChannel <- channels[1]
     yChannel <- channels[2]  
   }else
     stop("invalid number of channels for prior_flowclust!")
   # prior estimation is done separately from flowClust routine because
   # .prior_flowClust1d requires the entire parent flowSet yet flowClust only
   # takes one flowFrame
   
   if (is.null(prior_source)) {
     prior_data <- fs
   } else {
     prior_data <- gs_pop_get_data(gs, prior_source)
   }
#    browser()
   
   if (names(gm) == "gate_flowclust_1d") {
     
     
     
     # If any of 'K, 'neg' or 'pos' are given, we extract them from the
     # 'args' and coerce them as integers. Otherwise, they are defaulted to
     # NULL.
     
     neg_cluster <- pos_cluster <- NULL
     
     if(!is.null(K)) {
       K <- as.integer(K)
     }
     if(!missing(neg)) {
       neg_cluster <- as.integer(neg)
     }
     if(!missing(pos)) {
       pos_cluster <- as.integer(pos)
     }
     
     # Examines the various cases of pos, neg and K.
     # In the case that K is not given, we automatically select it.
     # Exception: If both 'pos' and 'neg' are given, we set: K <- pos + neg
     if (is.null(K)) {
       if (!is.null(neg_cluster) && !is.null(pos_cluster)) {
         K <- neg_cluster + pos_cluster
       }
     } else {
       # If pos and neg are given, throw a warning and set: K <- pos + neg
       # In the case that one of pos and neg are given and either exceeds K,
       # we throw an error.         
       if (!is.null(neg_cluster) && !is.null(pos_cluster)) {
         warning("Values given for 'K', 'neg', and 'pos'. Setting K = neg + pos")
         K <- neg_cluster + pos_cluster
       } else if (!is.null(pos_cluster) && K < pos_cluster) {
         stop("The number of positive clusters exceeds 'K'.")
       } else if (!is.null(neg_cluster) && K < neg_cluster) {
         stop("The number of negative clusters exceeds 'K'.")
       }
     }
     
     # If 'min' and/or 'max' are given, we pass this value along to the
     # prior-elicitation method as well as flowClust. Otherwise, these values
     # are set to NULL.
     min_values <- -Inf
     max_values <- Inf
     if (!missing(min)) {
       min_values <- as.numeric(min)
     }
     if (!missing(max)) {
       max_values <- as.numeric(max)
     }                          
#     browser()
     if (!is.null(xChannel)) {
       prior_list[[xChannel]] <- prior_flowclust(flow_set = prior_data,
           channels = xChannel, K = K,
           min = min_values, max = max_values, ...)
     }
     
     prior_list[[yChannel]] <- prior_flowclust(flow_set = prior_data,
         channels = yChannel, K = K,
         min = min_values, max = max_values, ...)
     
   } else {
     # get the value of neg and pos
     if (!is.null(K)) {
       K <- as.integer(K)
     } else {
       message("'K' argument is missing in prior_flowclust! Using default setting: K = 2.\nYou should set this to the same value as 'K' in the call to flowClust.")
       K <- 2
     }
     
     prior_list <- prior_flowclust(flow_set = prior_data, channels = c(xChannel, 
             yChannel), K = K, ...)
   }
   if(isCollapse)
     prior_list
   else# when collapse is FALSE, gating is done at sample level, thus we need to replicate priors across samples so that it matches with the gating input
     sapply(sampleNames(prior_data), function(i)prior_list, simplify = FALSE)
   
}

.prior_flowClust <- .prior_flowclust

 #' preprocessing wrapper for .standardize_flowFrame
#' 
#' @param fs a \code{flowSet} object
#' @param xChannel not used
#' @param yChannel the channel to standardize
#' @param groupBy \code{character} tells whether 'groupBy' column in csv template has been set
#' @param isCollapse \code{logical} indicates whether the gating is done on collapsed data. If so, the prior does not need to be replicated across samples.
#' @noRd 
.standardize_flowset <- function(fs, gs, gm, channels, groupBy, isCollapse, ...) {
   if(length(channels) != 1)
     stop("invalid number of channels for standardize_flowset!")
   if(isCollapse)
     stop("'collapse = TRUE' is not applicable to 'standardize_flowset'!")
   
   isGroup <- nchar(groupBy) > 0
   
   if(isGroup){
#     if(pData(fs)$Center == "UCLA" && pData(fs)$Sample == "12828")
#       browser()
     #standardize multiple flowFrames within the sub-group
     transform_out <- fsApply(fs, .standardize_flowFrame, channel = channels, data = TRUE)
     
     # Creates a flowSet object from the transformed flowFrame objects
     fs <- flowSet(lapply(transform_out, function(x) x$flow_frame))
     collapsedFr <- as(fs, "flowFrame")
     
     # update each flow_frame with collapsed flowFrame
     transform_out <- lapply(transform_out, function(x) {
           x$flow_frame <- collapsedFr
           x
         })
     
   }else{
     #standarize within the single flowFrame
     # fs is now the entire data set
     # and we only pass the transformation back to gating function without the flow data (to save memory)
     transform_out <- fsApply(fs, .standardize_flowFrame, channel = channels, data = FALSE)
   }
   
   
   transform_out
 }
 
 
 

# Generated by cpp11: do not edit by hand

Cdqrls <- function(x, y, tol) {
  .Call(`_openCyto_Cdqrls`, x, y, tol)
}

collapseData <- function(mat_list, colnames) {
  .Call(`_openCyto_collapseData`, mat_list, colnames)
}

rlm_cpp <- function(x, y, maxit) {
  .Call(`_openCyto_rlm_cpp`, x, y, maxit)
}

solve_LSAP_cpp <- function(mat) {
  .Call(`_openCyto_solve_LSAP_cpp`, mat)
}

# coerce method for ncdfFlowList
# 
# overwrite the version defined in ncdfFlow package
# It is at least 2 times (for 2-channel-9-sample ncfs collapsing) faster than origina version.
# @param from \code{ncdfFlowList}
#' @importClassesFrom ncdfFlow ncdfFlowList
#' @importFrom methods coerce
setAs(from = "ncdfFlowList", to = "flowFrame", def = function(from){
      selectMethod("coerce", signature = c("ncdfFlowSet", "flowFrame"))(from)      
      
    })

# fast collapsing
# 
# It is revised version of flowSet coerce method,  
# 1. without the overhead of validity checking for parameters 
# 2. without the extra column 'Original'
# 3. concatenate matrices within c++ to avoid copying
fast_coerce <- function(from){
  if(length(from) == 1)
  	from[[1, returnType = "flowFrame"]]
  else {
  	thisFr <- from[[1, use.exprs = FALSE]]
  	
  	params <- parameters(thisFr)
  	colnames <- colnames(thisFr)
  	
  	mat_list <- fsApply(from, function(fr)exprs(fr), simplify = FALSE)
  	
  	
  	desc  <- list(description="Synthetic Frame",sampleNames=sampleNames(from))
  	new("flowFrame",exprs = collapseData(mat_list, colnames)
  			, parameters = params
  			, description = desc
  	)
  }
}  
#' @importFrom flowCore colnames
#' @importClassesFrom ncdfFlow ncdfFlowSet
setAs(from="ncdfFlowSet", to="flowFrame", def=function(from)
			fast_coerce(from)
)
setAs(from="cytoset", to="flowFrame", def=function(from){
			fast_coerce(from)
		})

#' @templateVar old gate_flowClust_1d
#' @templateVar new gate_flowclust_1d
#' @template template-depr_pkg
NULL
#' Applies flowClust to 1 feature to determine a cutpoint between the minimum
#' cluster and all other clusters.
#'
#' We cluster the observations in \code{fr} into \code{K} clusters.
#'
#' By default, the cutpoint is chosen to be the boundary of the first two
#' clusters. That is, between the first two cluster centroids, we find the
#' midpoint between the largest observation from the first cluster and the
#' smallest observations from the second cluster. Alternatively, if the
#' \code{cutpoint_method} is \code{min_density}, then the cutpoint is the point
#' at which the density between the first and second smallest cluster centroids
#' is minimum.
#'
#' @rdname gate_flowclust_1d
#' @aliases gate_flowclust_1d gate_flowClust_1d flowClust.1d
#' @param fr a \code{flowFrame} object
#' @param params \code{character} channel to be gated on
#' @param filterId A \code{character} string that identifies the filter created.
#' @param K the number of clusters to find
#' @param trans,min.count,max.count,nstart some flowClust parameters. see \code{\link{flowClust}}
#' @param prior list of prior parameters for the Bayesian
#' \code{\link{flowClust}}. If \code{NULL}, no prior is used.
#' @param criterion a character string stating the criterion used to choose the
#' best model. May take either "BIC" or "ICL". This argument is only relevant
#' when \code{K} is \code{NULL} or if \code{length(K) > 1}. The value selected
#' is passed to \code{\link{flowClust}}.
#' @param cutpoint_method How should the cutpoint be chosen from the fitted
#' \code{\link{flowClust}} model? See Details.
#' @param neg_cluster integer. The index of the negative cluster. The cutpoint
#' is computed between clusters \code{neg_cluster} and \code{neg_cluster + 1}.
#' @param cutpoint_min numeric value that sets a minimum thresold for the
#' cutpoint. If a value is provided, any cutpoint below this value will be set
#' to the given minimum value. If \code{NULL} (default), there is no minimum
#' cutpoint value.
#' @param cutpoint_max numeric value that sets a maximum thresold for the
#' cutpoint. If a value is provided, any cutpoint above this value will be set
#' to the given maximum value. If \code{NULL} (default), there is no maximum
#' cutpoint value.
#' @param min a numeric value that sets the lower bound for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param max a numeric value that sets the upper bound for data filtering. If
#' \code{NULL} (default), no truncation is applied.
#' @param quantile the quantile for which we will find the cutpoint using
#' the quantile \code{cutpoint_method}. If the \code{cutpoint_method} is not set
#' to \code{quantile}, this argument is ignored.
#' @param quantile_interval a vector of length 2 containing the end-points of
#' the interval of values to find the quantile cutpoint. If the
#' \code{cutpoint_method} is not set to \code{quantile}, this argument is
#' ignored.
#' @param plot logical value indicating that the fitted \code{\link{flowClust}}
#' model should be plotted along with the cutpoint
#' @param debug \code{logical} indicating whether to carry the prior and posterious with the gate
#'                               for debugging purpose. Default is FALSE.
#' @param ... additional arguments that are passed to \code{\link{flowClust}}
#' @return a \code{rectangleGate} object consisting of all values beyond the
#' cutpoint calculated
#' @export
#' @importFrom flowClust getEstimates tmixFilter dmvtmix dmvt flowClust
#' @examples
#' \dontrun{
#'  gate <- gate_flowclust_1d(fr, params = "APC-A", K =2) # fr is a flowFrame
#' }
gate_flowclust_1d <- function(fr, params, filterId = "", K = NULL 
                              , trans = 0 #no box transform
                              , min.count = -1, max.count = -1 #no flowClust data filering
                              , nstart = 1 #change kmeans nstart
                              , prior = NULL,
                              criterion = c("BIC", "ICL"),
                              cutpoint_method = c("boundary", "min_density",
                                                  "quantile", "posterior_mean", "prior_density"),
                              neg_cluster = 1, cutpoint_min = NULL,
                              cutpoint_max = NULL, min = NULL, max = NULL,
                              quantile = 0.99, quantile_interval = c(0, 10),
                              plot = FALSE, debug = FALSE, ...) {
  options("cores" = 1L) #suppress parallelism for 1d gating   
  cutpoint_method <- match.arg(cutpoint_method)
  
  # TODO: Determine if Bayesian flowClust works when 'K' is specified and has a
  # different length than the number of components given in 'prior'.
  # If not, add GitHub issue and then fix.
  
  # If 'K' is NULL, then 'K' is autoselected using either 'BIC' or 'ICL'.
  # The candidate values for 'K' range from 1 to the number of mixture components
  # given in the 'prior' list.
  if (is.null(K)) {
    criterion <- match.arg(criterion)
    
    if (!is.null(prior)) {
      K <- length(as.vector(prior$Mu0))
    } else {
      stop("Values for 'K' must be provided if no 'prior' is given.")
    }
  }
  
  usePrior <- ifelse(is.null(prior), "no", "yes")
  
  #  L<-list(...)
  #  if("useprior"%in%names(L)){
  #	  usePrior <- L["useprior"]
  #  	L$useprior<-NULL
  #  }
  
  # HACK: Circumvents a bug in flowClust.
  # TODO: Add an issue to the Github for flowClust to allow prior to be NULL.
  if (is.null(prior)) {
    prior <- list(NA)
  }
  
  # Filter out values less than the minimum and above the maximum, if they are
  # given. NOTE: These observations are removed from the 'flowFrame' locally and
  # are gated out only for the determining the gate.
  if (!(is.null(min) && is.null(max))) {
    fr <- .truncate_flowframe(fr, channels = params[1], min = min, max = max)
  }
  if (nrow(fr) < 2) {
    warning("Less than two observations are present in the given flowFrame.",
            "Constructing gate from prior...")
  }
  
  # Applies `flowClust` to the feature specified in the `params` argument using
  # the data given in `fr`. We use priors with hyperparameters given by the
  # elements in the `prior` list.
  # call via do.call. L contains the rest of the pairlist, after extracting the passed value of usePrior
  #  	tmix_filter<-do.call(tmixFilter,c(filterId=filterId,params[1],K=K,
  #                                      trans=trans,usePrior=usePrior,
  #                                      prior=list(prior),criterion=list(criterion),
  #                                      L))
  tmix_results <- flowClust(fr, varNames = params[1]
                            , K = K, trans = trans
                            , usePrior = usePrior
                            , prior = prior
                            , min.count = min.count, max.count = max.count
                            , nstart = nstart
                            , criterion = criterion
                            , ...)
  
  # In the case an error occurs when applying 'flowClust', the gate is
  # constructed from the density of the prior distributions. This error
  # typically occurs when there are less than 2 observations in the flow frame.
  if (class(tmix_results) != "try-error") {
    # To determine the cutpoint, we first sort the centroids so that we can
    # determine the second largest centroid.
    centroids_sorted <- sort(getEstimates(tmix_results)$locations)
    
    # Also, because the cluster labels are arbitrary, we determine the cluster
    # the cluster labels, sorted by the ordering of the cluster means.
    labels_sorted <- order(getEstimates(tmix_results)$locations)
    
    # Grabs the data matrix that is being gated.
    x <- exprs(fr)[, params[1]]
  } else {
    cutpoint_method <- "prior_density"
  }
  
  # Determines the cutpoint between clusters 1 and 2.
  if (cutpoint_method == "boundary") {
    # Choose the cutpoint as the boundary between the first two clusters.
    # First, we sort the data.
    order_x <- order(x)
    x_sorted <- x[order_x]
    labels <- flowClust::Map(tmix_results, rm.outliers = FALSE)[order_x]
    
    # Determine which observations are between the first two centroids and their
    # corresponding cluster labels.
    which_between <- which(centroids_sorted[neg_cluster] < x_sorted & x_sorted < centroids_sorted[neg_cluster + 1])
    x_between <- x_sorted[which_between]
    labels_between <- labels[which_between]
    
    # For the observations between the first two centroids, we find the last
    # observation that belongs to the first cluster and the first observation
    # that belongs to the second cluster. In the rare occurrence that no
    # observations from one of the clusters is between the labels, we set the
    # max/min index as NA. This results in the cutpoint being set to the max/min
    # observation on the boundary.
    which_cluster1 <- which(labels_between == labels_sorted[neg_cluster])
    which_cluster2 <- which(labels_between == labels_sorted[neg_cluster + 1])
    max_obs_cluster1 <- ifelse(length(which_cluster1) == 0, NA, max(which_cluster1))
    min_obs_cluster2 <- ifelse(length(which_cluster2) == 0, NA, min(which_cluster2))
    
    # We define the cutpoint to be the midpoint between the two clusters.
    cutpoint <- mean(x_between[c(max_obs_cluster1, min_obs_cluster2)], na.rm = TRUE)
    if (is.nan(cutpoint)) {
      cutpoint <- centroids_sorted[neg_cluster]
    }
  } else if (cutpoint_method == "min_density") {
    # Determine the minimum density value of the observations between clusters 1 and 2.
    # Sets the cutpoint at the observation that attained this minimum density.
    x_between <- x[centroids_sorted[neg_cluster] < x & x < centroids_sorted[neg_cluster + 1]]
    x_dens <- dmvtmix(x = as.matrix(x_between), object = tmix_results)
    cutpoint <- x_between[which.min(x_dens)]
  } else if (cutpoint_method == "quantile") {
    cutpoint <- .quantile_flowClust(p = quantile, object = tmix_results,
                                    interval = quantile_interval)
  } else if (cutpoint_method == "posterior_mean") {
    cutpoint <- centroids_sorted[neg_cluster]
  } else { # cutpoint_method == "prior_density"
    # The prior_density cutpoint is determined as the point at which the density
    # of the negative cluster becomes smaller than the density of its adjacent
    # cluster.
    prior_x <- as.matrix(seq(prior$Mu0[neg_cluster], prior$Mu0[neg_cluster + 1], length = 1000))
    
    prior_proportions <- with(prior, w0 / sum(w0))
    prior_y <- lapply(c(neg_cluster, neg_cluster + 1), function(k) {
      prior_density <- dmvt(x = prior_x, mu = prior$Mu0[k],
                            sigma = prior$Omega0[k], nu = 4)$value
      prior_proportions[k] * prior_density
    })
    prior_y <- do.call(cbind, prior_y)
    
    # For the two prior densities, we determine the cutpoint as the first value
    # where the second density is larger than the first.
    diff_densities <- apply(prior_y, 1, diff)
    cutpoint <- prior_x[which(diff_densities > 0)[1]]
  }
  
  # In some cases, we wish that a gating cutpoint not exceed some threshold, in
  # which case we allow the user to specify a minimum and/or maximum value for
  # the cutpoint. For instance, when constructing a debris gate for samples from
  # various batches, some samples may have had a debris gate applied already
  # while other samples have debris that should be removed. For the former, the
  # user may wish to avoid gating, in which case, we provide the option to
  # threshold the gate cutpoint.
  if (!is.null(cutpoint_min) && cutpoint < cutpoint_min) {
    cutpoint <- cutpoint_min
  }
  if (!is.null(cutpoint_max) && cutpoint > cutpoint_max) {
    cutpoint <- cutpoint_max
  }
  
  
  gate_coordinates <- list(c(cutpoint, Inf))
  
  names(gate_coordinates) <- params
  
  fres <- rectangleGate(gate_coordinates, filterId = filterId)
  if(debug){
    # Saves posterior point estimates
    # In the case that an error is thrown, the posterior is set to the prior
    # because no prior updating was performed.
    postList <- list()
    if (class(tmix_results) != "try-error") {
      posteriors <- list(mu = tmix_results@mu, lambda = tmix_results@lambda,
                         sigma = tmix_results@sigma, nu = tmix_results@nu, min = min(x)
                         ,w = tmix_results@w
                         , max = max(x))
    } else {
      posteriors <- prior
      posteriors$min <- NA
      posteriors$max <- NA
    }
    postList[[params[1]]] <- posteriors
    
    # Saves prior point estimates
    priorList <- list()
    priorList[[params[1]]] <- prior
    fres <- fcRectangleGate(fres, priorList, postList)
  }
  
  if (plot) {
    gate_pct <- round(100 * mean(x > cutpoint), 3)
    plot_title <- paste0(filterId, " (", gate_pct, "%)")
    plot(fr, tmix_results, main = plot_title, labels = FALSE)
    abline(v = centroids_sorted, col = rainbow(K))
    abline(v = cutpoint, col = "black", lwd = 3, lty = 2)
    
    if (!is.null(prior)) {
      x_dens <- seq(min(x), max(x), length = 1000)
      
      for(k in seq_len(K)) {
        prior_density <- with(prior, dnorm(x_dens, mean = Mu0[k], sd = sqrt(Omega0[k])))
        
        # Grab posterior estimates for the degrees of freedom (nu) and the
        # transformation parameters. Because these can either be of length 1 or
        # of length K, we grab the appropriate posterior estimates for the
        # current mixture component.
        nu <- ifelse(length(tmix_results@nu) == 1, tmix_results@nu, tmix_results@nu[k])
        lambda <- ifelse(length(tmix_results@lambda) == 1, tmix_results@lambda, tmix_results@lambda[k])        
        posterior_density <- dmvt(x_dens, mu = tmix_results@mu[k,],
                                  sigma = tmix_results@sigma[k,,], nu = nu,
                                  lambda = lambda)$value
        lines(x_dens, prior_density, col = rainbow(K)[k], lty = 2, lwd = 1)
        lines(x_dens, posterior_density, col = rainbow(K)[k], lwd = 1)
      }
    }
  }
  fres
  
}

#' @export
gate_flowClust_1d <- function(fr, params, filterId = "", K = NULL 
                              , trans = 0 #no box transform
                              , min.count = -1, max.count = -1 #no flowClust data filering
                              , nstart = 1 #change kmeans nstart
                              , prior = NULL,
                              criterion = c("BIC", "ICL"),
                              cutpoint_method = c("boundary", "min_density",
                                                  "quantile", "posterior_mean", "prior_density"),
                              neg_cluster = 1, cutpoint_min = NULL,
                              cutpoint_max = NULL, min = NULL, max = NULL,
                              quantile = 0.99, quantile_interval = c(0, 10),
                              plot = FALSE, debug = FALSE, ...) {
  .Deprecated("gate_flowclust_1d")
  gate_flowclust_1d(fr, params, filterId, K , trans, min.count, max.count
                    , nstart, prior, criterion, cutpoint_method, neg_cluster 
                    , cutpoint_min, cutpoint_max, min, max, quantile, quantile_interval
                    , plot, debug, ...)
}

#' @export 
flowClust.1d <- gate_flowClust_1d

#' @templateVar old gate_flowClust_2d
#' @templateVar new gate_flowclust_2d
#' @template template-depr_pkg
NULL
#' Automatic identification of a population of interest via flowClust based on
#' two markers
#'
#' We cluster the observations in \code{fr} into \code{K} clusters. We set the
#' cutpoint to be the point at which the density between the first and second
#' smallest cluster centroids is minimum.
#'
#' The cluster for the population of interest is selected as the one with
#' cluster centroid nearest the \code{target} in Euclidean distance. By default,
#' the largest cluster (i.e., the cluster with the largest proportion of
#' observations) is selected as the population of interest.
#'
#' We also provide the option of constructing a \code{transitional} gate from
#' the selected population of interest. The location of the gate can be
#' controlled with the \code{translation} argument, which translates the gate
#' along the major axis of the targest cluster as a function of the appropriate
#' chi-squared coefficient. The larger \code{translation} is, the more gate is
#' shifted in a positive direction. Furthermore, the width of the
#' \code{transitional} gate can be controlled with the \code{quantile} argument.
#'
#' The direction of the transitional gate can be controlled with the
#' \code{transitional_angle} argument. By default, it is \code{NULL}, and we use
#' the eigenvector of the \code{target} cluster that points towards the first
#' quadrant (has positive slope). If \code{transitional_angle} is specified, we
#' rotate the eigenvectors so that the angle between the x-axis (with the cluster
#' centroid as the origin) and the major eigenvector (i.e., the eigenvector with
#' the larger eigenvalue) is \code{transitional_angle}. 
#' So based on range that the angle falls in, the final rectangleGate will be constructed 
#' at the corresponding quadrant. i.e. Clockwise, [0,pi/2] UR, (pi/2, pi] LR,
#' (pi, 3/2 * pi] LL, (3/2 * pi, 2 * pi] UL
#'
#' @name gate_flowclust_2d
#' @aliases gate_flowclust_2d gate_flowClust_2d flowClust.2d
#' @param fr a \code{flowFrame} object
#' @param xChannel,yChannel \code{character} specifying channels to be gated on
#' @param filterId A \code{character} string that identifies the filter created.
#' @param K the number of clusters to find
#' @param usePrior Should we use the Bayesian version of \code{\link{flowClust}}?
#' Answers are "yes", "no", or "vague". The answer is passed along to
#' \code{\link{flowClust}}.
#' @param prior list of prior parameters for the Bayesian version of
#' \code{\link{flowClust}}. If \code{usePrior} is set to \code{no}, then the
#' list is unused.
#' @param trans,min.count,max.count,nstart some flowClust parameters. see \code{\link{flowClust}}
#' @param plot a logical value indicating if the fitted mixture model should be
#' plotted. By default, no.
#' @param target a numeric vector of length \code{2} (number of dimensions) containing the location of
#' the cluster of interest. See details.
#' @param transitional logical value indicating if a transitional gate should be
#' constructed from the target \code{\link{flowClust}} cluster. By default, no.
#' @param quantile the contour level of the target cluster from the
#' \code{\link{flowClust}} fit to construct the gate
#' @param translation a numeric value between 0 and 1 used to position a
#' transitional gate if \code{transitional = TRUE}. This argument is ignored if
#' \code{transitional = FALSE}. See details
#' @param transitional_angle the angle (in radians) of the transitional
#' gate. It is also used to determine which quadrant the final gate resides in.
#' See details. Ignored if \code{transitional = FALSE}. 
#' @param min A vector of length 2. Truncate observations less than this minimum
#' value. The first value truncates the \code{xChannel}, and the second value
#' truncates the \code{yChannel}. By default, this vector is \code{NULL} and is
#' ignored.
#' @param max A vector of length 2. Truncate observations greater than this
#' maximum value. The first value truncates the \code{xChannel}, and the second
#' value truncates the \code{yChannel}. By default, this vector is \code{NULL}
#' and is ignored.
#' @param ... additional arguments that are passed to \code{\link{flowClust}}
#' @return a \code{polygonGate} object containing the contour (ellipse) for 2D
#' gating.
#' @export
#' @examples
#' \dontrun{
#'  gate <- gate_flowclust_2d(fr, xChannel = "FSC-A", xChannel = "SSC-A", K = 3) # fr is a flowFrame
#' }
gate_flowclust_2d <- function(fr, xChannel, yChannel, filterId = "", K = 2,
                         usePrior = 'no', prior = list(NA)
                         , trans = 0 #no box transform
                         , min.count = -1, max.count = -1 #no flowClust data filering
                         , nstart = 1 #change kmeans nstart
                         , plot = FALSE, target = NULL, transitional = FALSE,
                         quantile = 0.9, translation = 0.25, transitional_angle = NULL,
                         min = NULL, max = NULL, ...) {
  options("cores" = 1L) ##suppress parallelism since it may lock the process due to the lack of resource when parallel gating was already using up the cores. 
  
  if (!is.null(target)) {
    target <- as.numeric(target)
    if (length(target) != 2) {
      warning("The 'target' location must be a numeric vector of length 2.
               Using largest cluster instead...")
      target <- NULL
    }    
  }

  # If specified, truncates all observations outside the 'min' and 'max' values.
  # NOTE: These observations are removed from the 'flowFrame' locally and are
  # gated out only for the determining the gate.
  if (!(is.null(min) && is.null(max))) {
    fr <- .truncate_flowframe(fr, channels = c(xChannel, yChannel), min = min,
                             max = max)
  }

  # If appropriate, we generate prior parameters for the Bayesian version of flowClust.
  if (usePrior == "yes" && identical(prior, list(NA))) {
    prior <- prior_flowclust(fr = fr, channels = c(xChannel, yChannel), K = K)
  }

  # Applies `flowClust` to the feature specified in the `params` argument using
  # the data given in `fr`. We use priors with hyperparameters given by the
  # elements in the list `prior`.
  
  tmix_results <- flowClust(fr, varNames = c(xChannel, yChannel)
                                , K = K, trans = trans
                                , usePrior = usePrior
                                , prior = prior
                                , min.count = min.count, max.count = max.count
                                , nstart = nstart
                                , ...)
                    
  # tmix_results <- filter(fr, tmix_filter)
  # In the case an error occurs when applying 'flowClust', the gate is
  # constructed from the prior distributions. Errors typically occur when there
  # are less than 2 observations in the flow frame.
  if (class(tmix_results) == "try-error") {
    tmix_results <- new("flowClust", varNames = c(xChannel, yChannel), K = K,
                        w = prior$w0, mu = prior$Mu0, sigma = prior$Lambda0,
                        nu = 4, prior = prior, ruleOutliers = c(0, quantile, quantile))
  }

  fitted_means <- getEstimates(tmix_results)$locations

  # By default, the cluster with the largest number of observations is
  # selected. Otherwise, the cluster centroid nearest to the 'target' is
  # selected.
  if (is.null(target)) {
    cluster_selected <- which.max(tmix_results@w)
  } else {
    target_dist <- as.matrix(dist(rbind(fitted_means, target)))
    target_dist <- tail(target_dist, n = 1)[seq_len(K)]
    cluster_selected <- which.min(target_dist)
  }

  if (!transitional) {
    flowClust_gate <- .getEllipseGate(filter = tmix_results, include = cluster_selected,
                                   quantile = quantile,trans=trans)
    
  } else {
    chisq_quantile <- qchisq(quantile, df = 2)
    tol <- sqrt(.Machine$double.eps)

    xbar <- tmix_results@mu[cluster_selected, ]
    Sigma <- tmix_results@sigma[cluster_selected, , ]

    Sigma_eigen <- eigen(Sigma, symmetric = TRUE)
    u1 <- Sigma_eigen$vectors[, 1]
    u2 <- Sigma_eigen$vectors[, 2]
    lambda1 <- Sigma_eigen$values[1]
    lambda2 <- Sigma_eigen$values[2]

    # Computes the angles of each eigenvector with the x-axis in terms of polar
    # coordinates. Note that each vector has magnitude 1, so the radius is 1.
    u1_angle <- atan2(u1[2], u1[1])
    if (u1_angle < 0) {
      u1_angle <- u1_angle + 2 * pi
    }
    u2_angle <- atan2(u2[2], u2[1])
    if (u2_angle < 0) {
      u2_angle <- u2_angle + 2 * pi
    }

    # We ensure that each eigenvector is pointing vertically. That is, we ensure
    # the angle between the x-axis and the eigenvector is between 0 and pi. If
    # If they are not, we rotate them by pi (i.e., 180 degrees).
    if (u1_angle > pi) {
      R <- .rotation_matrix(pi)
      u1 <- as.vector(R %*% u1)
    }
    if (u2_angle > pi) {
      R <- .rotation_matrix(pi)
      u2 <- as.vector(R %*% u2)
    }

    u1_angle <- atan2(u1[2], u1[1])
    if (u1_angle < 0) {
      u1_angle <- u1_angle + 2 * pi
    }
    u2_angle <- atan2(u2[2], u2[1])
    if (u2_angle < 0) {
      u2_angle <- u2_angle + 2 * pi
    }
    eigen_angles <- c(u1_angle, u2_angle)

    # If the transitional angle is not provided, we set it as the angle between
    # the x-axis as the eigenvector pointing towards the postive quadrant.
    # If the transitional angle is provided, we first calculate the angle between
    # the major eigenvector and the x-axis (with xbar as the origin). We then
    # construct a rotation matrix, where the angle applied is the difference
    # between the angle provided and the angle calculated. We then rotate the
    # eigenvectors u1 and u2 accordingly
    #
    # In both cases, we compute the axis from which the transitional gate is
    # constructed as well as the perpendicular axis. We ensure that 'axis_perp'
    # is pi/2 radians counterclockwise from 'axis', following the right-hand
    # rule.
    if (is.null(transitional_angle)) {

      # Determines which eigenvector points towards the positive quadrant
      which_pos_quadrant <- which(0 < eigen_angles & eigen_angles < pi/2)

      # Calculates the angle of the transitional gate as the angle of the
      # eigenvector pointing toward the positive quadrant
      transitional_angle <- eigen_angles[which_pos_quadrant]

      if (which_pos_quadrant == 1) {
        axis <- sqrt(lambda1 * chisq_quantile) * u1
        axis_perp <- sqrt(lambda2 * chisq_quantile) * u2
      } else {
        axis <- sqrt(lambda2 * chisq_quantile) * u2
        axis_perp <- sqrt(lambda1 * chisq_quantile) * u1
      }
    } else {
      # Rotation angle
      theta_u1 <- transitional_angle - eigen_angles[1]
      theta_u2 <- transitional_angle - (pi/2) - eigen_angles[2]

      # Rotation matrix
      R1 <- .rotation_matrix(theta_u1)
      R2 <- .rotation_matrix(theta_u2)

      # Rotates the eigenvectors
      u1 <- as.vector(R1 %*% u1)
      u2 <- as.vector(R2 %*% u2)

      axis <- sqrt(lambda1 * chisq_quantile) * u1
      axis_perp <- -sqrt(lambda2 * chisq_quantile) * u2
    }

    # The gate location is the frame of reference for the gate. If it is xbar,
    # then the frame of reference is the cross-section along the eigenvector
    # from top-left to bottom-right. We translate this reference as a function
    # of the appropriate chi-squared coefficient.
    gate_location <- xbar + translation * axis

    # To construct the gate, we have effectively shifted the eigenvector with the
    # negative slope. We then extend the gate both horizontally and vertically
    # to the maximum observed values in the horizontal and vertical directions.
    # NOTE: We extend the gate one standard deviation beyond the maximum values
    # observed to mimic a gate that extends without limit in the positive
    # directions. However, because flowCore cannot handle such a shape, we force
    # the gate to have the same shape for the observed data.
    x <- exprs(fr)[, xChannel]
    y <- exprs(fr)[, yChannel]

    x_min <- min(x) - sd(x)
    y_min <- min(y) - sd(y)
    x_max <- max(x) + sd(x)
    y_max <- max(y) + sd(y)

    # We construct the gate in clockwise fashion. The vertices of the gate
    # depend on the quadrant towards which the 'transitional_angle' is pointed.
    # No matter what the transitional gate's angle, the first and last vertices
    # will be the same
    first_vertex <- gate_location + axis_perp
    fifth_vertex <- gate_location - axis_perp
    
    if (0 <= transitional_angle && transitional_angle <= pi/2) {
      # First quadrant
      second_vertex <- c(first_vertex[1], y_max)
      third_vertex <- c(x_max, y_max)
      fourth_vertex <- c(x_max, fifth_vertex[2])
    } else if (pi/2 < transitional_angle && transitional_angle <= pi) {
      # Second quadrant
      second_vertex <- c(x_min, first_vertex[2])
      third_vertex <- c(x_min, y_max)
      fourth_vertex <- c(fifth_vertex[1], y_max)
    } else if (pi < transitional_angle && transitional_angle <= 3*pi/2) {
      # Third quadrant
      second_vertex <- c(first_vertex[1], y_min)
      third_vertex <- c(x_min, y_min)
      fourth_vertex <- c(x_min, fifth_vertex[2])
    } else {
      # Fourth quadrant
      second_vertex <- c(x_max, first_vertex[2])
      third_vertex <- c(x_max, y_min)
      fourth_vertex <- c(fifth_vertex[1], y_min)
    }
    polygon_gate <- rbind(first_vertex,
                          second_vertex,
                          third_vertex,
                          fourth_vertex,
                          fifth_vertex,
                          first_vertex)
    colnames(polygon_gate) <- c(xChannel, yChannel)
    flowClust_gate <- polygonGate(filterId = filterId, .gate = polygon_gate)
  }
  
  # List of posterior point estimates
  posteriors <- list(mu = tmix_results@mu, lambda = tmix_results@lambda,
                     sigma = tmix_results@sigma, nu = tmix_results@nu)

  if (plot) {
    plot(data = fr, tmix_results, main = filterId)

    if (transitional) {
      # The major and minor axes (eigenvectors) scaled by their respective
      # eigenvalues and the chi-squared quantile.    
      lines(rbind(xbar - axis, xbar + axis), col = "darkgreen")
      lines(rbind(xbar - axis_perp, xbar + axis_perp), col = "darkgreen")

      # Draws the polygon gate.
      lines(polygon_gate, col = "red")
      lines(rbind(gate_location - axis_perp, gate_location + axis_perp), col = "red")

      # Also, draws points at the vertices of the polygon gate.
      points(polygon_gate, col = "red", pch = 16)
    }
  }
  if(class(flowClust_gate) == "polygonGate")
    fcPolygonGate(flowClust_gate, prior, posteriors)
  else
    fcEllipsoidGate(flowClust_gate, prior, posteriors)
    
}

#' @export
gate_flowClust_2d <- function(fr, xChannel, yChannel, filterId = "", K = 2,
                              usePrior = 'no', prior = list(NA)
                              , trans = 0 #no box transform
                              , min.count = -1, max.count = -1 #no flowClust data filering
                              , nstart = 1 #change kmeans nstart
                              , plot = FALSE, target = NULL, transitional = FALSE,
                              quantile = 0.9, translation = 0.25, transitional_angle = NULL,
                              min = NULL, max = NULL, ...) {
  .Deprecated("gate_flowclust_2d")
  gate_flowclust_2d(fr, xChannel, yChannel, filterId, K, usePrior, prior, trans
                    , min.count, max.count, nstart, plot, target, transitional
                    , quantile, translation, transitional_angle, min, max, ...)
}


#' @export 
flowClust.2d <- gate_flowClust_2d

#' Determine the cutpoint by the events quantile.
#' 
#' It is possible that the cutpoint calculated by quantile function may not 
#' produce the exact the probability set by 'probs' argument if there are not enough
#' cell events to reach that precision. Sometime the difference could be significant.
#' 
#' @name gate_quantile
#' @aliases quantileGate
#' @param fr a \code{flowFrame} object
#' @param channel the channel from which the cytokine gate is constructed
#' @param filterId the name of the filter
#' @param plot whether to plot the gate result
#' @param probs probabilities passed to 'stats::quantile' function.
#' @param min a numeric value that sets the lower boundary for data filtering
#' @param max a numeric value that sets the upper boundary for data filtering
#' @param ... additional arguments passed to 'stats::quantile' function.
#' @return a \code{rectangleGate} 
#' @export
#' @examples
#' \dontrun{
#'  gate <- gate_quantile(fr, Channel = "APC-A", probs = 0.995) # fr is a flowFrame
#' }
gate_quantile <- function(fr, channel, probs = 0.999, plot = FALSE,
                         filterId = "", min = NULL, max = NULL, ...) {
   if (missing(channel) || length(channel) != 1) {
     stop("A single channel must be specified.")
   }
   
   # Filter out values less than the minimum and above the maximum, if they are
   # given.
   if (!(is.null(min) && is.null(max))) {
     fr <- .truncate_flowframe(fr, channels = channel, min = min,
         max = max)
   }
                       
  x <- exprs(fr)[, channel]
  cutpoint <- quantile(x, probs = probs, ...)

  gate_coordinates <- list(c(cutpoint, Inf))
  names(gate_coordinates) <- channel
  
  if (plot) {
    hist(x)
    plot(density(x))
    abline(v = cutpoint, col = "red")
    actual_probs <- sum(x>cutpoint)/length(x)
    text(y = 0.5, x = cutpoint, labels = paste("%:", actual_probs))
  }
  rectangleGate(gate_coordinates, filterId = filterId)
}
#' @templateVar old quantileGate
#' @templateVar new gate_quantile
#' @template template-depr_pkg
NULL
#' @export
quantileGate <- function(fr, channel, probs = 0.999, plot = FALSE,
                        filterId = "", min = NULL, max = NULL, ...){
  .Deprecated("gate_quantile")
  gate_quantile(fr, channel, probs, plot, filterId, min, max, ...)
}

#' Determines a cutpoint as the minimum point of a kernel density estimate
#' between two peaks
#'
#' We fit a kernel density estimator to the cells in the \code{flowFrame} and
#' identify the two largest peaks. We then
#' select as the cutpoint the value at which the minimum density is attained
#' between the two peaks of interest.
#'
#' In the default case, the two peaks of interest are the two largest peaks
#' obtained from the \code{link{density}} function. 
#'
#' In the special case that there is only one peak, we are conservative and set
#' the cutpoint as the \code{min(x)} if \code{positive} is \code{TRUE}, and the
#' \code{max(x)} otherwise.
#'
#' @name gate_mindensity
#' @aliases mindensity
#' @param fr a \code{flowFrame} object
#' @param channel TODO
#' @param filterId TODO
#' @param positive If \code{TRUE}, then the gate consists of the entire real
#' line to the right of the cutpoint. Otherwise, the gate is the entire real
#' line to the left of the cutpoint. (Default: \code{TRUE})
#' @param gate_range numeric vector of length 2. If given, this sets the bounds
#' on the gate applied. If no gate is found within this range, we set the gate to
#' the minimum value within this range if \code{positive} is \code{TRUE} and the
#' maximum value of the range otherwise.
#' @param min a numeric value that sets the lower boundary for data filtering
#' @param max a numeric value that sets the upper boundary for data filtering
#' @param peaks \code{numeric} vector. If not given , then perform peak detection first by .find_peaks
#' @param ... Additional arguments for peak detection.
#' @return a \code{rectangleGate} object based on the minimum density cutpoint
#' @export
#' @examples
#' \dontrun{
#'  gate <- gate_mindensity(fr, channel = "APC-A") # fr is a flowFrame
#' }
#' @importFrom flowCore %in% identifier filterDetails<-
gate_mindensity <- function(fr, channel, filterId = "", positive = TRUE,
                       gate_range = NULL, min = NULL, max = NULL,
                       peaks = NULL, ...) {
  
  if (missing(channel) || length(channel) != 1) {
    stop("A single channel must be specified.")
  }

  # Filter out values less than the minimum and above the maximum, if they are
  # given.
  if (!(is.null(min) && is.null(max))) {
    fr <- .truncate_flowframe(fr, channels = channel, min = min,
                                     max = max)
  }
  # Grabs the data matrix that is being gated.
  x <- exprs(fr)[, channel]
  
  if(is.null(peaks))
    peaks <- .find_peaks(x, ...)[, "x"]
  
  if (is.null(gate_range)) {
    gate_range <- c(min(x), max(x))
  } else {
    gate_range <- sort(gate_range)
  }

  
  
  # In the special case that there is only one peak, we are conservative and set
  # the cutpoint as min(x) if 'positive' is TRUE, and max(x) otherwise.
  if (length(peaks) == 1) {
    cutpoint <- ifelse(positive, gate_range[1], gate_range[2])
  } else {
    # The cutpoint is the deepest valley between the two peaks selected. In the
    # case that there are no valleys (i.e., if 'x_between' has an insufficient
    # number of observations), we are conservative and set the cutpoint as the
    # minimum value if 'positive' is TRUE, and the maximum value otherwise.
    valleys <- try(.find_valleys(x, ...), silent = TRUE)
    valleys <- .between_interval(x = valleys, interval = gate_range)

    if (any(is.na(valleys))) {
    #FIXME:currently it is still returning the first peak,
    #we want to pass density instead of x_between to 'min'
    #because x_between is the signal values
      cutpoint <- ifelse(positive, gate_range[1], gate_range[2])
    } else if (length(valleys) == 1) {
      cutpoint <- as.vector(valleys)
    } else if (length(valleys) > 1) {
      # If there are multiple valleys, we determine the deepest valley between
      # the two largest peaks.
      peaks <- sort(peaks[1:2])
      cutpoint <- .between_interval(valleys, peaks)[1]

      # If none of the valleys detected are between the two largest peaks, we
      # select the deepest valley.
      if (is.na(cutpoint)) {
        cutpoint <- valleys[1]
      }      
    }
  }
  gate_coordinates <- ifelse(positive, list(c(cutpoint, Inf)), list(c(-Inf, cutpoint)))
  
  names(gate_coordinates) <- channel
  
  rectangleGate(gate_coordinates, filterId = filterId)
  
}
#' @export
mindensity <- gate_mindensity


#' sequential quadrant gating function
#' 
#' The order of 1d-gating is determined so that the gates better capture the 
#' distributions of flow data.
#' 
#' @name gate_quad_sequential
#' @aliases quadGate.seq
#' @param fr \code{flowFrame}
#' @param channels \code{character} two channels used for gating
#' @param gFunc the name of the 1d-gating function to be used for either dimension
#' @param min a numeric vector that sets the lower bounds for data filtering
#' @param max a numeric vector that sets the upper bounds for data filtering
#' @param ... other arguments passed to \code{.find_peak} (e.g. 'num_peaks' and 'adjust').
#' @return a \code{filters} that contains four rectangleGates
#' @export 
gate_quad_sequential <- function(fr, channels, gFunc, min = NULL, max = NULL, ...){
  if (missing(channels) || length(channels) != 2) {
    stop("two channels must be specified.")
  }
  
  # Filter out values less than the minimum and above the maximum, if they are
  # given.
  if (!(is.null(min) && is.null(max))) {
    fr <- .truncate_flowframe(fr, channels = channels, min = min,max = max)
  }
  
  res <- sapply(channels, function(channel){
        
        x <- exprs(fr)[, channel]
        
        peaks <-.find_peaks(x,..., num_peaks = 2)
        #get peak and scores
        if(nrow(peaks)<2)
          score <- 0
        else
          score <- abs(diff(peaks[, "x"]))
        list(peaks = peaks[, "x"], score = score)
      }, simplify = FALSE)
  
  #order channel by scores
  channels.ordered <- names(sort(sapply(res, `[[`, "score"), decreasing = T))
  x.first <- all(channels == channels.ordered)
    
  #get first cut
  chnl <- channels.ordered[1]
  thisFunc <- function(fr, chnl, ...){
    thisCall <- substitute(f(data, channel
                          , peaks = p
                        , ...)
                    ,list(f=as.symbol(gFunc), data = fr
                        , channel = chnl
                      , p = res[[chnl]][["peaks"]]
                    )
    )  
  
    eval(thisCall)
  }
  
  g1 <- thisFunc(fr, chnl, ...)
  
  coord1 <- c(g1@min, g1@max)
  cut.1 <- coord1[!is.infinite(coord1)]
#  if(length(cut.1) == 0)
#    cut.1 <- Inf
  
  
  
  #get ind
  fres <- filter(fr, g1)
  ind <- as(fres, "logical")
  
  #gate on the second channel
  chnl <- channels.ordered[2]
  
  #+
  g2 <- thisFunc(fr[ind, ], chnl, ...)
  coord2 <- c(g2@min, g2@max)
  cut.2.pos <- coord2[!is.infinite(coord2)]
  
#  if(length(cut.2.pos) == 0)
#    cut.2.pos<- Inf
  #-
  g3 <- thisFunc(fr[!ind, ], chnl, ...)
  coord3 <- c(g3@min, g3@max)
  cut.2.neg <- coord3[!is.infinite(coord3)]
  
  if(x.first)
    coords <-list(q1 = list(c(-Inf, cut.1), c(cut.2.neg, Inf))
                  , q2 = list(c(cut.1, Inf), c(cut.2.pos, Inf))
                  , q3 = list(c(cut.1, Inf), c(-Inf, cut.2.pos))
                  , q4 = list(c(-Inf, cut.1), c(-Inf, cut.2.neg))
                  )
  else
    coords <-list(q1 = list(c(-Inf, cut.2.pos), c(cut.1, Inf))
                  , q2 = list(c(cut.2.pos, Inf), c(cut.1, Inf))
                  , q3 = list(c(cut.2.neg, Inf), c(-Inf, cut.1))
                  , q4 = list(c(-Inf, cut.2.neg), c(-Inf, cut.1))
                )
    
  
  gates <- lapply(coords, function(coord){
            
      names(coord) <- as.character(channels)
      rectangleGate(coord)
    })
  
  filters(gates)
}
#' @templateVar old quadGate.seq
#' @templateVar new gate_quad_sequential
#' @template template-depr_pkg
NULL
#' @export 
quadGate.seq <- function(fr, channels, gFunc, min = NULL, max = NULL, ...){
  .Deprecated("gate_quad_sequential")
  gate_quad_sequential(fr, channels, gFunc, min, max, ...)
}

##TODO: come up a more general design so that polygonGates can be derived from any two quadrants
#' quadGate based on flowClust::tmixFiler
#' 
#' This gating method identifies two quadrants (first, and third quadrants) by fitting the data with tmixture model.
#' It is particually useful when the two markers are not well resolved thus the regular quadGate method
#' based on 1d gating will not find the perfect cut points on both dimensions.
#' 
#' @name gate_quad_tmix
#' @aliases quadGate.tmix
#' @param fr \code{flowFrame}
#' @param channels \code{character} vector specifies two channels
#' @param usePrior see \link{gate_flowclust_2d}
#' @param K see \link{gate_flowclust_2d}
#' @param prior see \link{gate_flowclust_2d}
#' @param trans see \link{gate_flowclust_2d}
#' @param plot logical whether to plot flowClust clustering results
#' @param quantile1 \code{numeric} specifies the  quantile level(see 'level' in \link{flowClust}) for the first quadrant (x-y+)
#' @param quantile3 \code{numeric} specifies the  quantile level see 'level' in \link{flowClust} for third quadrant (x+y-)
#' @param ... other arguments passed to \link{flowClust}
#' @return a \code{filters} object that contains four \code{polygonGate}s following the order of (-+,++,+-,--)
#' @export 
gate_quad_tmix <- function(fr, channels, K, usePrior = "no", prior = list(NA)
    , quantile1 = 0.8, quantile3 = 0.8
    , trans = 0
    , plot = FALSE
    , ...){
  
  max.v <- .Machine$integer.max #.Machine$double.xmax double max no longer to be valid coordinates in the plot
  min.v <- - max.v
# browser()  
  # fit tmix models 
  tmix_filter <- tmixFilter(parameters = channels
      , K = K
      , usePrior = usePrior, prior = prior, trans = trans
      , ...)

  tmix_results <- try(filter(fr, tmix_filter), silent = F)
  if(plot)
    plot(tmix_results, data = fr, level = 0.85)
  #find the 1st and 3rd quads
  fitted_means <- getEstimates(tmix_results)$locations
  q1.ind <- which.max(fitted_means[,2])
  q3.ind <- which.max(fitted_means[,1])
  
  #construct polygon gates 
  q1.gate <- as(.getEllipseGate(filter = tmix_results
          , include = q1.ind,
          quantile = quantile1
          ,trans = 0), "polygonGate")
  q3.gate <- as(.getEllipseGate(filter = tmix_results
          , include = q3.ind,
          quantile = quantile3
          ,trans = 0), "polygonGate")
  #find intersection of 1,3 quad gates
  q1.bottom <- min(q1.gate@boundaries[, 2])
  q1.right <- max(q1.gate@boundaries[, 1])
  
  q3.top <- max(q3.gate@boundaries[, 2])
  q3.left <- min(q3.gate@boundaries[, 1])
  
  #two gates overlap on both dimensions
  p.BL <- c(q3.left, q1.bottom)
  p.TR <- c(q1.right, q3.top)
  if(q1.bottom >= q3.top){
    thisY <- mean(c(q1.bottom , q3.top))
    p.BL[2] <- p.TR[2] <- thisY
  }
  
  if(q3.left >= q1.right){
    thisX <- mean(c(q1.right , q3.left))
    p.BL[1] <- p.TR[1] <- thisX
  }
  
  
  #construct four quadGates based on these two points
  q1.coord <- matrix(c(min.v, max.v
          , min.v, p.BL[2]
          , p.BL
          , p.TR
          , p.TR[1], max.v
      )
      , 5, 2
      , byrow = TRUE
  )  
  colnames(q1.coord) <- channels    
  q1.g <- polygonGate(q1.coord, filterId = paste(paste0(channels, c("-", "+")), collapse = ""))
  
  
  q2.coord <- matrix(c(p.TR
          , max.v, p.TR[2]
          , max.v, max.v
          , p.TR[1], max.v
      )
      , 4, 2
      , byrow = TRUE
  )  
  colnames(q2.coord) <- channels    
  q2.g <- polygonGate(q2.coord, filterId = paste(paste0(channels, c("+", "+")), collapse = ""))
  
  q3.coord <- matrix(c(p.TR
          , max.v, p.TR[2]
          , max.v, min.v
          , p.BL[1], min.v
          ,p.BL
      )
      , 5, 2
      , byrow = TRUE
  )
  colnames(q3.coord) <- channels    
  q3.g <- polygonGate(q3.coord, filterId = paste(paste0(channels, c("+", "-")), collapse = ""))
  
  q4.coord <- matrix(c(min.v, p.BL[2]
          , p.BL
          , p.BL[1], min.v
          , min.v, min.v
      )
      , 4, 2
      , byrow = TRUE
  )
  colnames(q4.coord) <- channels    
  q4.g <- polygonGate(q4.coord, filterId = paste(paste0(channels, c("-", "-")), collapse = ""))
  
  filters(list(q1.g, q2.g, q3.g,q4.g))
}
#' @templateVar old quadGate.tmix
#' @templateVar new gate_quad_tmix
#' @template template-depr_pkg
NULL
#' @export 
quadGate.tmix <- function(fr, channels, K, usePrior = "no", prior = list(NA)
                          , quantile1 = 0.8, quantile3 = 0.8
                          , trans = 0
                          , plot = FALSE
                          , ...){
  .Deprecated("gate_quad_tmix")
  quad_gate_tmix(fr, channels, K, usePrior, prior, quantile1, quantile3, trans, plot, ...)
}

---
title: "How to write a csv gating template"
output: 
  html_document: 
    number_sections: true
    toc: true
    toc_float: true
    collapsed: false
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{How to write a csv gating template}
---


How to write a csv gating template
=======================================


```{r setup, include=FALSE}
library(knitr)
opts_chunk$set(out.extra='style="display:block; margin: auto"', fig.align="center", message = FALSE, warning = FALSE, knitr.table.format = "markdown")
assign("depthtrigger", 60, data.table:::.global)
```


The  **openCyto** package is uses a spreadsheet to compose the gating schemes. Basically each row corresponds to one population node
in the gating hierarchy tree. However sometime it is verbose to describe every single population.
So here we will explain how to make the template more succinct to easier to compose by using `pop` and `alias` pattern.      


## `pop` = "+/-"
For the 1d/2d gating function, we are normally interested in either positive(representing cell events within gate) or negative(or negated, representing cell events outside of the gate) by setting `pop` column 
in the form of `+` or `-`. But sometime we want to do the downstream gating for both.
By specifying `pop` as `+/-`, the `template parser` will expand it into two rows internally.
  
```{r init, echo=F}
library(openCyto)
library(data.table)

template_row <- data.table(alias = "TH", pop = "+-"
                          , parent = "cd3", dims = "cd4,cd8"
                          , gating_method = "mindensity", gating_args = ""
                          , collapseDataForGating = "TRUE", groupBy = "4"
                          , preprocessing_method = "", preprocessing_args = ""
                      )


```
For example, this row will be expanded automatically

```{r A+/-(before), echo = F, results = "asis"}
	this_row <- copy(template_row)	
	this_row[, pop := "+/-"]
	this_row[, alias := "*"]
	this_row[, dims := "cd4"]
	kable(this_row[, 1:6, with = F])
```

to two rows:
      
```{r A+/-(after), echo = F, results = "asis"}
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])
      
```

Note that the second row uses `refGate` which simply copies the gate coordinates computed by `mindensity` in the first row,
and assign the negative sign to the `pop` column indicating the population of interest is `cd4 negative`.  


## `pop` = "++"
Often time we need to apply 1d gating function on two dimensions separately and then 
use the two cutting points to construct `rectangleGate` to capture the cell events falling into one particular quadrant on the 2-d projections
For example, `T helper` cells are usually represented as `CD4+CD8-`.
Instead of writing three rows in the template, simply using `++` pattern in the `pop` column.  
e.g

```{r A+B+(before), echo = F, results = "asis"}
	this_row <- copy(template_row)	
	this_row[, pop := "+-"]
	this_row[, alias := "T helper"]
	kable(this_row[, 1:6, with = F])
```

And the template parser will take care of the expansion automatically.
      
```{r A+B+(after), echo = F, results = "asis"}
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])
```

As we see, first two rows do the actual gating by `mindensity` and the third row simply makes use the coordinates of that two 1d gates (`cd4+` and `cd8+`) and 
construct a `rectangleGate` (T helper) by using `refGate` as `gating_method`. And the `+` and `-` sign along with dimensions determines which quadrant to keep.  

## `pop` = "+/-+/-"
Apparently, we may want to get more than one quadrants by using the same mechanism.
For example, we can set `pop` to `+/-+/-` to keep all of four quadrants.

```{r A+/-B+/-(before), echo = F, results = "asis"}
	this_row <- copy(template_row)	
	this_row[, pop := "+/-+/-"]
	this_row[, alias := "*"]
	kable(this_row[, 1:6, with = F])
```

It will be expanded to six rows:
      
```{r A+/-B+/-(after), echo = F, results = "asis"}
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])
```

First two does the actual gating, and rest of four uses two 1d gates to construct four different `rectangleGate`s to represent
four different quadrants.

## Gating method that generates multiple populations 

### `pop` = "*" and `alias` = "A,B,C"

So far, we've been talking about the gating functions that only returns one gate object( `S4 class` that extends `flowCore::filter`).
If we want to apply the gating function(e.g. `curv2filter` or `flowClust::tmixFilter`) that returns more than one gates, 
we can set `pop` to `*` and specify multiple population names within `alias` with comma-separated characters. 

```{r multiPop(before), echo = F, results = "asis"}
	this_row <- copy(template_row)	
	this_row[, pop := "*"]
	this_row[, alias := "CD4,CD8"]
	this_row[, gating_method := "curv2gate"]
	kable(this_row[, 1:6, with = F])
```

Here we assume `curv1gate` always returns `two` gates in the order of `c("cd4", "cd8")`, then the population names in `alias` column
will be matched to these two gates and two `dummy_gate` rows are generated that simply serves as a reference to be used  
as `parent` node of the downstream gates.  
      
```{r multiPop(after), echo = F, results = "asis"}
      kable(openCyto:::.preprocess_row(this_row)[, 1:6, with = F])
```

###  `pop` = "*" and `alias` = "*"

If we don't know how many gates will be returned by `curv2gate` or the order of gates are undetermined, thus we will not able to 
name these populations. As long as they are not used as `parent` nodes for the further gating (i.e. `terminal gate`s), we can
simply set `alias` to `*` .

```{r multiPop_noExpand(before), echo = F, results = "asis"}
	this_row[, alias := "*"]
	kable(this_row[, 1:6, with = F])
```

This will not be expanded in the `openCyto` framework. However, multiple populations will be generated and added 
to the `GatingSet` object. They are named by the `filterId` slot of `filter` objects.

## Single row with multiple parents  (i.e. `parent` ="A,B,C")

If the same gating method (or simply `refGate`) needs to be applied to multiple parents, it is possible to write these into the single row. For example,

```{r multiple_parents, echo=F, results='asis'}
	this_row[, alias := "IL2+"]
  this_row[, pop := "+"]
  this_row[, parent := "cd4,cd8"]
  this_row[, dims := "IL2"]
  this_row[, gating_method := "refGate"]
  this_row[, gating_args := "cd3/IL2+"]
	kable(this_row[, 1:6, with = F])
```

This row is interpreted as copying the `IL2` gate that has been generated from `cd3` population and applying it to `cd4` and `cd8`, which is equivalent to the rows of below

```{r multiple_parents_expand, echo=F, results='asis'}
kable(openCyto:::.split_multi_parents(this_row)[, 1:6, with = F])
```

```{r finalizer, include=FALSE}
assign("depthtrigger", 3, data.table:::.global)
```

---
title: "OpenCyto: How to use different auto gating functions"
author: "Mike Jiang"
output: 
  html_document: 
    keep_md: true
    toc: true
    toc_float: true
vignette: >    
  %\VignetteIndexEntry{How to use different auto gating functions}    
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, echo=FALSE}
library(knitr)
opts_chunk$set(message = FALSE, warning = FALSE, fig.height= 2, fig.width= 3, fig.show = "hold")
```


Here we will illustrate how to choose and use the appropriate gating methods that are pre-registered in `openCyto` package. And users can always define their own `gating` algorithms and register them as the `plugin` functions in `openCyto` framework, see `?registerPlugins` for more details.

Note that all the function names illustrated below are prefixed with `.` indicating that they are simply the wrapper function registered in `openCyto`. The actual `gating engine` behind the wrapper can come from other packages (e.g. `flowCore`, `flowClust`). 
All these wrappers have these common interfaces:
* `fr`: a `flowFrame` object
* `pp_res`: an optional `pre-preocessing` result, which can be ignored in this document
* `channels`: channel names used for gating
* `...`: any other gating parameters pass on to the actual gating engine

```{r}
library(flowCore)
library(flowWorkspace)
library(openCyto)
library(ggcyto)

gs <- load_gs(system.file("extdata/gs_bcell_auto", package = "flowWorkspaceData"))
```
## 1D gating methods
### `mindensity`
The name of this gating function is self-explaining, that is to find the minimum as the cutpoint between negative and postive peaks in 1d density plot. It is fast,robust and extremely easy to use especially when there is a good separation between `+` and `-` populations/peaks.

For example, it is usually easy to gate on `CD3` channel and no need to supply any arguments to the method.

```{r}
fr <- gh_pop_get_data(gs[[2]], "Live", returnType = "flowFrame")
chnl <- "CD3"
g <- openCyto:::.mindensity(fr, channels = chnl)
autoplot(fr, chnl) + geom_gate(g)
autoplot(fr, chnl, "SSC-A") + geom_gate(g)
```

However, it may need some guidance when there are more than `2` major peaks/populations detected in densit profile.
```{r}
fr <- gh_pop_get_data(gs[[1]], "boundary", returnType = "flowFrame")
chnl <- "FSC-A"
g <- openCyto:::.mindensity(fr, channels = chnl)
mylimits <- ggcyto_par_set(limits = "instrument")
p <- autoplot(fr, chnl) + mylimits
p + geom_gate(g)
autoplot(fr, chnl, "SSC-A") + geom_gate(g)
```

Here we actually want to remove the `debris cells` that are represented by the first negative peak. But `mindensity` cuts between the second and third peaks since they are more predorminant. So we can simply specify a `range` that will limit the locations where the cut point should be placed. 
```{r}
g <- openCyto:::.mindensity(fr, channels = chnl, gate_range=c(7e4,1e5), adjust = 1.5)
p + geom_gate(g)
autoplot(fr, chnl, "SSC-A") + geom_gate(g)
```

And as shown, we also changed the `kernal density`  smoothing factor `adjust` from  `2`(default value set in `openCtyo`) to `1.5` to avoid over-smoothing.

Alternatively you can achieve the same effect by setting `min` or `max` to pre-filter the data before the `mindenstiy` works on it.
```{r}
g <- openCyto:::.mindensity(fr, channels = chnl, min = 7e4, max = 1e5)
p + geom_gate(g)
```

To choose one way or the other or combining both is highly dependent on how your data. The more contrains will give you more controls on how gating proceeds yet at cost of robustness of your gating pipeline sometime.


### `tailgate`
This gating method is used in the scenarios where there is only one major peak detected thus automatically disqualify the usage of `mindensity`. `tol` is to control how far the cut point should be placed away from the peak. This function is now housed in [cytoUtils](https://github.com/RGLab/cytoUtils).

```{r, eval = FALSE}
library(cytoUtils)
register_plugins(cytoUtils:::.tailgate, "gate_tail")
fr <- gh_pop_get_data(gs[[1]], "lymph", returnType = "flowFrame")
chnl <- "Live"
g <- openCyto:::.tailgate(fr, channels = chnl, tol = 0.05)
p <- autoplot(fr, chnl) + mylimits
p + geom_gate(g)
autoplot(fr, chnl, "SSC-A") + geom_gate(g)
```

### `quantileGate`
This method is an alternative to `tailgate` and it determines the cutpoint by the events quantile.
```{r}
g <- openCyto:::.quantileGate(fr, channels = chnl, probs = 0.99)
p <- autoplot(fr, chnl) + mylimits
p + geom_gate(g)
autoplot(fr, chnl, "SSC-A") + geom_gate(g)
```

This gating method is more commonly used in gating the `rare` populations when the target population is not prominent enough to stand out as the second peak. (e.g. `cytokine` gates in `ICS` assays.)

## 2D gating methods
### `boundary Gate`
It essentially constructs a rectangle gate from input range (min, max), which is useful for filtering out very extreme signals at the bounary.

```{r}
fr <- gh_pop_get_data(gs[[1]], "root", returnType = "flowFrame")
chnl <- c("FSC-A", "SSC-A")
g <- openCyto:::.boundary(fr, channels = chnl, min = c(0, 0), max=c(2.5e5,2.5e5))
p <- autoplot(fr, x = chnl[1], y = chnl[2])
p + geom_gate(g)
```

### `singletGate`
Use the `area` vs `height` to gate out the singlets. See details from `?singletGate`.
```{r}
fr <- read.FCS(system.file("extdata/CytoTrol_CytoTrol_1.fcs", package = "flowWorkspaceData"))
chnl <- c("FSC-A", "FSC-H")
g <- openCyto:::.singletGate(fr, channels = chnl)
p <- autoplot(fr, x = chnl[1], y = chnl[2])
p + geom_gate(g)
```

### `flowClust.2d`
`flowClust` package in itself is not limited to 2-dimensional gating. But here we are talking about a dedicated wrapper function `.flowClust.2d` from `openCyto` package that leverages `flowClust` clustering engine to work on `2D` cases specifically. You won't need to write the full name of the function in `csv` gating template, simply put `flowClust` in the `gating_method` column, and then the template parser will automatically dispatch to the right function.

```{r}
fr <- gh_pop_get_data(gs[[1]], "nonDebris", returnType = "flowFrame")
chnl <- c("FSC-A", "SSC-A")
g <- openCyto:::.flowClust.2d(fr, channels = chnl, K=2, target=c(1e5,5e4), quantile=0.95)
p <- autoplot(fr, x = chnl[1], y = chnl[2]) + mylimits
p + geom_gate(g)
```

`K` is to tell the algorithm how many major clusters/populations are expected in the 2d profile. `target` specify the mean/center of the target population to get, which doesn't have to be precise. If not supplied, flowClust will pick the most prominent cluster as the target, which would be the right choice in most cases.
`quantile` specify how large the `ellipse` should be. `pp_res` is used to provide the `prior` information for `flowClust`. (More details are in `?flowClust`)


### `Transitional gate`
`flowClust.2d` can optionally construct a `Transitional gate`, which is a speical kind of polygon gate with one edge placed diagonally that is often seen in `flowJo`. Here is an example:
```{r}
fr <- gh_pop_get_data(gs[[1]], "CD19andCD20", returnType = "flowFrame")
chnl <- c("CD38", "CD24")
g <- openCyto:::.flowClust.2d(fr, channels = chnl, K=6,transitional=TRUE,target=c(3.5e3,3.5e3), quantile=0.95,translation=0.15, pp_res = NULL)
p <- autoplot(fr, x = chnl[1], y = chnl[2]) + mylimits
p + geom_gate(g)
```

The rational behind the algorithm is beyond the scope of this document. Please see its detailed explainations in `?flowClust.2d`.

### `quadGate.tmix`
This gating method identifies two quadrants (first, and third quadrants) by fitting the data with tmixture model.
It is particually useful when the two markers are not well resolved thus the regular quadGate method
that is based on `1d` gating will not find the perfect cut points on both dimensions.
```{r}
gs <- load_gs(system.file("extdata/gs_DC_auto", package = "flowWorkspaceData"))
fr <- gh_pop_get_data(gs[[2]], "HLADR+", returnType = "flowFrame")
chnl <- c("CD11c", "CD123")
p <- autoplot(fr, chnl[1], chnl[2])
g <- openCyto:::.quadGate.tmix(fr, channels = chnl, K = 3, usePrior = "no")
p + geom_gate(g)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{openCyto.options}
\alias{openCyto.options}
\title{Some global options for openCyto
See examples for the meaning of these options and how to get/set them.}
\description{
Get/set some global options for openCyto
}
\examples{
opt <- getOption("openCyto")
#the threshold of minimum cell events required for the gating algorithm to proceed
opt[["gating"]][["minEvents"]]
#to change the threshold
opt[["gating"]][["minEvents"]] <- 100
options(openCyto = opt)

#switch off the validity check flags(Not recommended)
opt[["check.pop"]] <- FALSE
options(openCyto = opt)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/template_gating.R
\name{generate_gate_template}
\alias{generate_gate_template}
\title{Store gates from a GatingSet in a gatingTemplate for templated gating}
\usage{
generate_gate_template(
  x,
  groupBy = NA,
  nodes = NULL,
  gatingTemplate = FALSE,
  save_as = NULL,
  ...
)
}
\arguments{
\item{x}{a \code{GatingHierarchy} or \code{GatingSet} object.}

\item{groupBy}{a vector of variable names that exist in
\code{colnames(pData(x))} to split samples into groups before extracting
the gates specified by \code{nodes}, set to NA by default to produce
consensus gates across all samples.}

\item{nodes}{a vector of full node paths that exist in \code{x} for which
templated gating entries are required, set to all nodes in \code{x} by
default.}

\item{gatingTemplate}{logical to indicate whether a \code{gatingTemplate}
object should be returned, set to FALSE by default. \code{gatingTemplate}
objects can only be created when all nodes in \code{x} are specified.}

\item{save_as}{name of the CSV file to which the gatingTemplate entries
should be written, set to NULL to prevent the gatingTemplate from being
written to a CSV file.}

\item{...}{additional arguments passed to \code{\link{gatingTemplate-class}}.}
}
\value{
either a \code{data.table} or \code{gatingTemplate} object containing
  the requested gatingTemplate entries for downstream templated gating using
  \code{gt_gating()}.
}
\description{
\code{generate_gate_template()} iterates through samples groups in a
GatingSet to extract and store existing gates into an openCyto
\code{gatingTemplate} that can be easily applied to new GatingSets using
\code{gt_gating()}.
}
\examples{
\dontrun{
# gs contains variable BioSampleType
pData(gs)$BioSampleType <- rep(c("cells", "beads"), each = 4)
# create gatingTemplate object
gt <- generate_gate_template(
  gs,
  groupBy = "BioSampleType",
  gatingTemplate = TRUE,
  save_as = "gatingTemplate.csv"
)
# gt is a gatingTemplate object
gt
# apply gt to a new GatingSet
gt_gating(
  gt,
  gs_new
)
}

}
\author{
Dillon Hammill, \email{Dillon.Hammill@ozette.com}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtPopulation-methods.R
\name{names,gtPopulation-method}
\alias{names,gtPopulation-method}
\title{get population name}
\usage{
\S4method{names}{gtPopulation}(x)
}
\arguments{
\item{x}{\code{gtPopulation} object}
}
\description{
get population name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcFilter-methods.R
\name{priors,fcFilter,ANY-method}
\alias{priors,fcFilter,ANY-method}
\alias{priors,fcFilter,character-method}
\title{get priors from a \code{fcFilter} object}
\usage{
\S4method{priors}{fcFilter,ANY}(x, y = "missing")
}
\arguments{
\item{x}{\code{fcFilter} object}

\item{y}{\code{character} specifying channel name. if \code{missing} then extract priors for all the channels}
}
\description{
get priors from a \code{fcFilter} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{fcPolygonGate}
\alias{fcPolygonGate}
\title{constuctor for \code{fcPolygonGate}}
\usage{
fcPolygonGate(x, priors, posts)
}
\arguments{
\item{x}{a \code{polygonGate} object}

\item{priors}{a \code{list} storing priors}

\item{posts}{a \code{list} storing posteriors}
}
\description{
constuctor for \code{fcPolygonGate}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_quad_sequential}
\alias{gate_quad_sequential}
\alias{quadGate.seq}
\title{sequential quadrant gating function}
\usage{
gate_quad_sequential(fr, channels, gFunc, min = NULL, max = NULL, ...)
}
\arguments{
\item{fr}{\code{flowFrame}}

\item{channels}{\code{character} two channels used for gating}

\item{gFunc}{the name of the 1d-gating function to be used for either dimension}

\item{min}{a numeric vector that sets the lower bounds for data filtering}

\item{max}{a numeric vector that sets the upper bounds for data filtering}

\item{...}{other arguments passed to \code{.find_peak} (e.g. 'num_peaks' and 'adjust').}
}
\value{
a \code{filters} that contains four rectangleGates
}
\description{
The order of 1d-gating is determined so that the gates better capture the 
distributions of flow data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{gt_get_nodes}
\alias{gt_get_nodes}
\alias{getNodes}
\alias{getNodes,gatingTemplate-method}
\title{get nodes from \link[openCyto:gatingTemplate-class]{gatingTemplate} object}
\usage{
gt_get_nodes(
  x,
  y,
  order = c("default", "bfs", "dfs", "tsort"),
  only.names = FALSE
)
}
\arguments{
\item{x}{\code{gatingTemplate}}

\item{y}{\code{character} node index. When \code{missing}, return all the nodes}

\item{order}{\code{character} specifying the order of nodes. options are "default", "bfs", "dfs", "tsort"}

\item{only.names}{\code{logical} specifiying whether user wants to get the entire \code{gtPopulation} object or just the name of the population node}
}
\description{
get nodes from \link[openCyto:gatingTemplate-class]{gatingTemplate} object
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
gt_get_nodes(gt)[1:2]
gt_get_nodes(gt, only.names = TRUE)
gt_get_nodes(gt, "/nonDebris")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.table.R
\name{as.data.table}
\alias{as.data.table}
\alias{as.data.table.gatingTemplate}
\title{convert a gatingTemplate object to a data.table}
\usage{
\method{as.data.table}{gatingTemplate}(x, keep.rownames = FALSE)
}
\arguments{
\item{x}{gatingTemplate object}

\item{keep.rownames}{not used}
}
\value{
a data.table
}
\description{
It is the inverse function of gatingTemplate constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{parameters,gtMethod-method}
\alias{parameters,gtMethod-method}
\title{get parameters of the gating method/function}
\usage{
\S4method{parameters}{gtMethod}(object)
}
\arguments{
\item{object}{\code{gtMethod}}
}
\description{
get parameters of the gating method/function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcTree-methods.R
\name{getNodes,fcTree-method}
\alias{getNodes,fcTree-method}
\title{get nodes from \code{fcTree}}
\usage{
\S4method{getNodes}{fcTree}(x, y)
}
\arguments{
\item{x}{\code{fcTree}}

\item{y}{\code{character} node name}
}
\description{
get nodes from \code{fcTree}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{groupBy,gtMethod-method}
\alias{groupBy,gtMethod-method}
\title{get the grouping variable for the gating method}
\usage{
\S4method{groupBy}{gtMethod}(object)
}
\arguments{
\item{object}{\code{gtMethod}}
}
\description{
When specified, the flow data is grouped by the grouping variable (column names in pData).
Within each group, when \code{isCollapse} is set to TRUE, the gating method is applied to the collapsed data.
Otherwise, it is done indepentently for each indiviudal sample(flowFrame).
Grouping variable is also used by \code{preprocessing method}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{isCollapse,gtMethod-method}
\alias{isCollapse,gtMethod-method}
\title{get the flag that determines whether gating method is applied on collapsed data}
\usage{
\S4method{isCollapse}{gtMethod}(object)
}
\arguments{
\item{object}{\code{gtMethod}}
}
\value{
\code{logical}
}
\description{
When TRUE, the flow data(multiple flowFrames) is collapsed into one and the gating method is applied on the collapsed data.
Once the gate is generated, it is thenreplicated and applied to the each single flowFrame.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remove_pop.R
\name{gs_remove_gating_method}
\alias{gs_remove_gating_method}
\alias{remove_pop}
\title{Reverse the action of gating methods applied via \code{gs_add_gating_method}}
\usage{
gs_remove_gating_method(gs)
}
\arguments{
\item{gs}{The \code{GatingSet} or \code{GatingSetList} for which the most recent
\code{gs_add_gating_method} call should be reversed.}
}
\description{
This function provides an easy way to remove the gates and nodes created by the most
recent call to \code{\link{gs_add_gating_method}} on the specified \code{GatingSet} or \code{GatingSetList},
with a separate history being maintained for each such object. \code{gs_remove_gating_method} allows 
for repeated use, effectively serving as a multi-level undo function for \code{gs_add_gating_method}.
}
\examples{
\dontrun{
 # add quad gates 
 gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
 # Remove the gates and nodes resulting from that gs_add_gating_method call
 gs_remove_gating_method(gs)
} 
}
\seealso{
\code{\link{gs_add_gating_method}} \code{\link{gs_add_gating_method_init}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{dummyMethod-class}
\alias{dummyMethod-class}
\title{A class to represent a dummy gating method that does nothing but serves as reference to be refered by other population}
\description{
It is generated automatically by the csv template preprocessing to handle the gating function that returns multiple gates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{gatingTemplate-class}
\alias{gatingTemplate-class}
\alias{gatingTemplate}
\alias{gatingTemplate,character-method}
\alias{gatingTemplate,data.table-method}
\title{a class storing the gating method and population information in a graphNEL object}
\usage{
gatingTemplate(x, ...)

\S4method{gatingTemplate}{character}(
  x,
  name = "default",
  strict = TRUE,
  strip_extra_quotes = FALSE,
  ...
)

\S4method{gatingTemplate}{data.table}(
  x,
  name = "default",
  strict = TRUE,
  strip_extra_quotes = FALSE,
  ...
)
}
\arguments{
\item{x}{\code{character} csv file name or a \code{data.table}}

\item{...}{other arguments passed to \code{data.table::fread}}

\item{name}{\code{character} the label of the gating template}

\item{strict}{\code{logical} whether to perform validity check(special characters) on the alias column. By default it is(and should be) turned on for the regular template parsing. 
But sometime it is useful to turned it off to bypass the check for the dummy nodes(e.g. the csv template generated by 'gh_generate_template'  with some existing boolean gates that has '!' or ':' symbol).}

\item{strip_extra_quotes}{\code{logical} Extra quotes are added to strings by fread. This causes problems with parsing R strings to expressions in some cases. Default FALSE for usual behaviour. TRUE should be passed if parsing gating_args fails.}
}
\description{
Each cell population is stored in graph node and is connected with its parent population 
or its reference node for boolGate or refGate.

It parses the csv file that specifies the gating scheme for a particular staining pannel.
}
\details{
This csv must have the following columns:

'alias': a name used label the cell population, the path composed by the alias and its precedent nodes (e.g. /root/A/B/alias) has to be uniquely identifiable.
         So alias can not contain '/' character, which is reserved as path delimiter.
 
'pop': population patterns of '+/-` or '+/-+/-', which tells the algorithm which side (postive or negative) of 1d gate or which quadrant of 2d gate to be kept.
                        
'parent': the parent population alias, its path has to be uniquely identifiable.
 
'dims': characters seperated by comma specifying the dimensions(1d or 2d) used for gating. 
It can be either channel name or stained marker name (or the substrings of channel/marker names as long as they are uniquely identifiable.).
 
'gating_method': the name of the gating function (e.g. 'flowClust'). It is invoked by a wrapper function that has the identical function name prefixed with a dot.(e.g. '.flowClust')
    
'gating_args': the named arguments passed to gating function (Note that double quotes are often used as text delimiter by some csv editors. So try to use single quote instead if needed.)
 
'collapseDataForGating': When TRUE, data is collapsed (within groups if 'groupBy' specified) before gating and the gate is replicated across collapsed samples.
 When set FALSE (or blank),then 'groupBy' argument is only used by 'preprocessing' and ignored by gating.
   
'groupBy': If given, samples are split into groups by the unique combinations of study variable (i.e. column names of pData,e.g."PTID:VISITNO").
 when split is numeric, then samples are grouped by every N samples 

'preprocessing_method': the name of the preprocessing function(e.g. 'prior_flowclust'). It is invoked by a wrapper function that has the identical function name prefixed with a dot.(e.g. '.prior_flowclust')
 the preprocessing results are then passed to gating wrapper function through 'pps_res' argument.
      
'preprocessing_args': the named arguments passed to preprocessing function.
}
\examples{
\dontrun{ 
  gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
  plot(gt)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{fcFilterList}
\alias{fcFilterList}
\title{constuctor for \code{fcFilterList}}
\usage{
fcFilterList(x)
}
\arguments{
\item{x}{\code{list} of \code{fcFilter} (i.e. \code{fcPolygonGate} or \code{fcRectangleGate})}
}
\description{
constuctor for \code{fcFilterList}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/singletGate.R
\name{gate_singlet}
\alias{gate_singlet}
\title{Creates a singlet polygon gate using the prediction bands from a robust linear model}
\usage{
gate_singlet(
  x,
  area = "FSC-A",
  height = "FSC-H",
  sidescatter = NULL,
  prediction_level = 0.99,
  subsample_pct = NULL,
  wider_gate = FALSE,
  filterId = "singlet",
  maxit = 5,
  ...
)
}
\arguments{
\item{x}{a \code{\link[flowCore:flowFrame-class]{flowFrame}} object}

\item{area}{character giving the channel name that records the signal
intensity as peak area}

\item{height}{character giving the channel name that records the signal
intensity as peak heightchannel name of height}

\item{sidescatter}{character giving an optional channel name for the
sidescatter signal. By default, ignored.}

\item{prediction_level}{a numeric value between 0 and 1 specifying the level
to use for the prediction bands}

\item{subsample_pct}{a numeric value between 0 and 1 indicating the percentage
of observations that should be randomly selected from \code{x} to construct
the gate. By default, no subsampling is performed.}

\item{wider_gate}{logical value. If \code{TRUE}, the prediction bands used to
construct the singlet gate use the robust fitted weights, which increase
prediction uncertainty, especially for large FSC-A. This leads to wider gates,
which are sometimes desired.}

\item{filterId}{the name for the filter that is returned}

\item{maxit}{the limit on the number of IWLS iterations}

\item{...}{additional arguments (not used)}
}
\value{
a \code{\link[flowCore]{polygonGate}} object with the singlet gate
}
\description{
We construct a singlet gate by applying a robust linear model. By default, we model the forward-scatter height
(FSC-H)as a function of forward-scatter area (FSC-A). If \code{sidescatter}
is given, forward-scatter height is as a function of \code{area} +
\code{sidescatter} + \code{sidescatter / area}.
}
\details{
Because \code{rlm} relies on iteratively reweighted least
squares (IRLS), the runtime to construct a singlet gate is dependent in part
on the number of observations in \code{x}. To improve the runtime, we provide
an option to subsample randomly a subset of \code{x}. A percentage of
observations to subsample can be given in \code{subsample_pct}. By default, no
subsampling is applied.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mindensity2.R
\name{gate_mindensity2}
\alias{gate_mindensity2}
\alias{mindensity2}
\title{An improved version of mindensity used to determines a cutpoint as the minimum point 
of a kernel density estimate between two peaks.}
\usage{
gate_mindensity2(
  fr,
  channel,
  filterId = "",
  gate_range = NULL,
  min = NULL,
  max = NULL,
  peaks = NULL,
  ...
)
}
\arguments{
\item{fr}{a \code{flowFrame} object}

\item{channel}{the channel to operate on}

\item{filterId}{a name to refer to this filter}

\item{gate_range}{numeric vector of length 2. If given, this sets the bounds
on the gate applied.}

\item{min}{a numeric value that sets the lower boundary for data filtering}

\item{max}{a numeric value that sets the upper boundary for data filtering}

\item{peaks}{\code{numeric} vector. If not given , then perform peak detection first by .find_peaks}

\item{...}{Additional arguments for peak detection.}
}
\value{
a \code{rectangleGate} object based on the minimum density cutpoint
}
\description{
Analogous to the original openCyto::mindensity(), mindensity2 operates on a standard flowFrame. Its behavior is closely modeled on 
the original mindensity() whenever possible. However, the underlying peak-finding algorithm 
(improvedMindensity) behaves significantly differently.
}
\examples{
\dontrun{
 gate <- gate_mindensity2(fr, channel = "APC-A") # fr is a flowFrame
}
}
\author{
Greg Finak, Phu T. Van
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcFilter-methods.R
\name{posteriors,fcFilter,ANY-method}
\alias{posteriors,fcFilter,ANY-method}
\alias{posteriors,fcFilter,character-method}
\title{get posteriors from a \code{fcFilter} object}
\usage{
\S4method{posteriors}{fcFilter,ANY}(x, y = "missing")
}
\arguments{
\item{x}{\code{fcFilter}}

\item{y}{\code{character} or \code{missing} that specifiy which channel to look for}
}
\description{
get posteriors from a \code{fcFilter} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_flowclust_2d}
\alias{gate_flowclust_2d}
\alias{gate_flowClust_2d}
\alias{flowClust.2d}
\title{Automatic identification of a population of interest via flowClust based on
two markers}
\usage{
gate_flowclust_2d(
  fr,
  xChannel,
  yChannel,
  filterId = "",
  K = 2,
  usePrior = "no",
  prior = list(NA),
  trans = 0,
  min.count = -1,
  max.count = -1,
  nstart = 1,
  plot = FALSE,
  target = NULL,
  transitional = FALSE,
  quantile = 0.9,
  translation = 0.25,
  transitional_angle = NULL,
  min = NULL,
  max = NULL,
  ...
)
}
\arguments{
\item{fr}{a \code{flowFrame} object}

\item{xChannel, yChannel}{\code{character} specifying channels to be gated on}

\item{filterId}{A \code{character} string that identifies the filter created.}

\item{K}{the number of clusters to find}

\item{usePrior}{Should we use the Bayesian version of \code{\link{flowClust}}?
Answers are "yes", "no", or "vague". The answer is passed along to
\code{\link{flowClust}}.}

\item{prior}{list of prior parameters for the Bayesian version of
\code{\link{flowClust}}. If \code{usePrior} is set to \code{no}, then the
list is unused.}

\item{trans, min.count, max.count, nstart}{some flowClust parameters. see \code{\link{flowClust}}}

\item{plot}{a logical value indicating if the fitted mixture model should be
plotted. By default, no.}

\item{target}{a numeric vector of length \code{2} (number of dimensions) containing the location of
the cluster of interest. See details.}

\item{transitional}{logical value indicating if a transitional gate should be
constructed from the target \code{\link{flowClust}} cluster. By default, no.}

\item{quantile}{the contour level of the target cluster from the
\code{\link{flowClust}} fit to construct the gate}

\item{translation}{a numeric value between 0 and 1 used to position a
transitional gate if \code{transitional = TRUE}. This argument is ignored if
\code{transitional = FALSE}. See details}

\item{transitional_angle}{the angle (in radians) of the transitional
gate. It is also used to determine which quadrant the final gate resides in.
See details. Ignored if \code{transitional = FALSE}.}

\item{min}{A vector of length 2. Truncate observations less than this minimum
value. The first value truncates the \code{xChannel}, and the second value
truncates the \code{yChannel}. By default, this vector is \code{NULL} and is
ignored.}

\item{max}{A vector of length 2. Truncate observations greater than this
maximum value. The first value truncates the \code{xChannel}, and the second
value truncates the \code{yChannel}. By default, this vector is \code{NULL}
and is ignored.}

\item{...}{additional arguments that are passed to \code{\link{flowClust}}}
}
\value{
a \code{polygonGate} object containing the contour (ellipse) for 2D
gating.
}
\description{
We cluster the observations in \code{fr} into \code{K} clusters. We set the
cutpoint to be the point at which the density between the first and second
smallest cluster centroids is minimum.
}
\details{
The cluster for the population of interest is selected as the one with
cluster centroid nearest the \code{target} in Euclidean distance. By default,
the largest cluster (i.e., the cluster with the largest proportion of
observations) is selected as the population of interest.

We also provide the option of constructing a \code{transitional} gate from
the selected population of interest. The location of the gate can be
controlled with the \code{translation} argument, which translates the gate
along the major axis of the targest cluster as a function of the appropriate
chi-squared coefficient. The larger \code{translation} is, the more gate is
shifted in a positive direction. Furthermore, the width of the
\code{transitional} gate can be controlled with the \code{quantile} argument.

The direction of the transitional gate can be controlled with the
\code{transitional_angle} argument. By default, it is \code{NULL}, and we use
the eigenvector of the \code{target} cluster that points towards the first
quadrant (has positive slope). If \code{transitional_angle} is specified, we
rotate the eigenvectors so that the angle between the x-axis (with the cluster
centroid as the origin) and the major eigenvector (i.e., the eigenvector with
the larger eigenvalue) is \code{transitional_angle}. 
So based on range that the angle falls in, the final rectangleGate will be constructed 
at the corresponding quadrant. i.e. Clockwise, [0,pi/2] UR, (pi/2, pi] LR,
(pi, 3/2 * pi] LL, (3/2 * pi, 2 * pi] UL
}
\examples{
\dontrun{
 gate <- gate_flowclust_2d(fr, xChannel = "FSC-A", xChannel = "SSC-A", K = 3) # fr is a flowFrame
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{show,gtMethod-method}
\alias{show,gtMethod-method}
\title{show method for gtMethod}
\usage{
\S4method{show}{gtMethod}(object)
}
\arguments{
\item{object}{\code{gtMethod}
show method for gtMethod}
}
\description{
show method for gtMethod
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{fcTree-class}
\alias{fcTree-class}
\title{A class to represent a flowClust tree.}
\description{
It is a graphNEL used as a container to store priors and posteriors for each flowClust gate that can be
visualized for the purpose of fine-tunning parameters for flowClust algorithm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{dims,gtMethod-method}
\alias{dims,gtMethod-method}
\title{get gating method dimensions}
\usage{
\S4method{dims}{gtMethod}(x)
}
\arguments{
\item{x}{\code{gtMethod}}
}
\description{
get gating method dimensions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-methods.R
\name{gt_gating}
\alias{gt_gating}
\alias{gt_gating.gatingTemplate}
\alias{gating}
\alias{gating,gatingTemplate,GatingSet-method}
\alias{gt_gating,gatingTemplate,GatingSet-method}
\title{Applies a gatingTemplate to a GatingSet.}
\usage{
gt_gating(x, y, ...)
}
\arguments{
\item{x}{a \code{gatingTemplate} object}

\item{y}{a \code{GatingSet} object}

\item{...}{\itemize{
    \item{start}{ a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating process will start from. It is useful to quickly skip some gates and go directly to the target population in the testing run. Default is "root".}
    \item{stop.at}{ a \code{character} that specifies the population (correspoding to 'alias' column in csv template) where the gating prcoess will stop at. Default is NULL, indicating the end of gating tree.}
    \item{keep.helperGates}{a \code{logical} flag indicating whether to keep the intermediate helper gates that are automatically generated by openCyto. Default is TRUE.}
    \item{mc.cores}{ passed to \code{multicore} package for parallel computing}
    \item{parallel_type}{  \code{character} specifying the parallel type. The valid options are "none", "multicore", "cluster".}
    \item{cl}{ \code{cluster} object passed to \code{parallel} package (when \code{parallel_type} is "cluster")}
}}
}
\value{
Nothing. As the side effect, gates generated by gating methods are saved in \code{GatingSet}.
}
\description{
It loads the gating methods by topological order and applies them to \code{GatingSet}.
}
\examples{
\dontrun{
 gt <- gatingTemplate(file.path(path, "data/ICStemplate.csv"), "ICS")
 gs <- GatingSet(fs) #fs is a flowSet/ncdfFlowSet
 gt_gating(gt, gs)
 gt_gating(gt, gs, stop.at = "v") #proceed the gating until population 'v'
 gt_gating(gt, gs, start = "v") # start from 'v'
 gt_gating(gt, gs, parallel_type = "multicore", mc.cores = 8) #parallel gating using multicore
 #parallel gating by using cluster
 cl1 <- makeCluster (8, type = "MPI")
 gt_gating(gt, gs, parallel_type = "cluster", cl = cl1)
 stopCluster ( cl1 )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pluginFramework.R
\name{gt_list_methods}
\alias{gt_list_methods}
\alias{listgtMethods}
\title{Print a list of the registered gating methods}
\usage{
gt_list_methods()
}
\value{
Does not return anything. Prints a list of the available gating methods.
}
\description{
Print a list of the registered gating methods
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{ppMethod-class}
\alias{ppMethod-class}
\alias{ppMethod}
\title{A class to represent a preprocessing method.}
\description{
It extends \code{gtMethod} class.
}
\examples{
 \dontrun{
     gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
     ppMethod(gt, '3', '4')
}     
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{package}
\name{openCyto}
\alias{openCyto-package}
\alias{openCyto}
\title{Hierarchical Gating Pipeline for flow cytometry data}
\description{
Hierarchical Gating Pipeline for flow cytometry data.
}
\details{
openCyto is a package designed to facilitate the automated gating methods in sequential way 
to mimic the manual gating strategy.


\tabular{ll}{
Package: \tab openCyto\cr
Type: \tab Package\cr
Version: \tab 1.2.8\cr
Date: \tab 2014-04-10\cr
License: \tab GPL (>= 2)\cr
LazyLoad: \tab yes\cr
}
}
\examples{
\dontrun{gatingTemplate('test.csv')}
}
\seealso{
See \code{\link[openCyto]{gt_gating}}, 
\code{\link{gate_flowclust_1d}}, 
for an overview of gating functions.
}
\author{
Mike Jiang \email{wjiang2@fhcrc.org},
John Ramey \email{jramey@fhcrc.org},
Greg Finak \email{gfinak@fhcrc.org}

Maintainer: Mike Jiang \email{wjiang2@fhcrc.org}
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_mindensity}
\alias{gate_mindensity}
\alias{mindensity}
\title{Determines a cutpoint as the minimum point of a kernel density estimate
between two peaks}
\usage{
gate_mindensity(
  fr,
  channel,
  filterId = "",
  positive = TRUE,
  gate_range = NULL,
  min = NULL,
  max = NULL,
  peaks = NULL,
  ...
)
}
\arguments{
\item{fr}{a \code{flowFrame} object}

\item{channel}{TODO}

\item{filterId}{TODO}

\item{positive}{If \code{TRUE}, then the gate consists of the entire real
line to the right of the cutpoint. Otherwise, the gate is the entire real
line to the left of the cutpoint. (Default: \code{TRUE})}

\item{gate_range}{numeric vector of length 2. If given, this sets the bounds
on the gate applied. If no gate is found within this range, we set the gate to
the minimum value within this range if \code{positive} is \code{TRUE} and the
maximum value of the range otherwise.}

\item{min}{a numeric value that sets the lower boundary for data filtering}

\item{max}{a numeric value that sets the upper boundary for data filtering}

\item{peaks}{\code{numeric} vector. If not given , then perform peak detection first by .find_peaks}

\item{...}{Additional arguments for peak detection.}
}
\value{
a \code{rectangleGate} object based on the minimum density cutpoint
}
\description{
We fit a kernel density estimator to the cells in the \code{flowFrame} and
identify the two largest peaks. We then
select as the cutpoint the value at which the minimum density is attained
between the two peaks of interest.
}
\details{
In the default case, the two peaks of interest are the two largest peaks
obtained from the \code{link{density}} function. 

In the special case that there is only one peak, we are conservative and set
the cutpoint as the \code{min(x)} if \code{positive} is \code{TRUE}, and the
\code{max(x)} otherwise.
}
\examples{
\dontrun{
 gate <- gate_mindensity(fr, channel = "APC-A") # fr is a flowFrame
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fast_rlm.R
\name{fast_rlm}
\alias{fast_rlm}
\title{robust linear model using an M estimator}
\usage{
fast_rlm(x, y, maxit = 20)
}
\arguments{
\item{x}{matrix with first column as weight (default can be 1s), the rest columns are predict variable}

\item{y}{numeric vector as response}

\item{maxit}{maximum iterations}
}
\description{
rewritten in c++, till eval stats::lm.wfit r function in underlying cpp11 code 
It is internally used for singletGate, thus its output format may not be generic enough for common model fitting .
e.g. it doesn't take formula as input
}
\examples{
n <- 1e3
x <- seq_len(n)
y <- x * 2.5 - 1.3 + rnorm(n, sd = 30)
names(y) <- x
x <- cbind(1, x)
r2 <- fast_rlm(x, y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toggle.helperGates.R
\name{gt_toggle_helpergates}
\alias{gt_toggle_helpergates}
\alias{toggle.helperGates}
\alias{gt_get_helpergates}
\alias{get.helperGates}
\alias{gt_delete_helpergates}
\alias{delete.helperGates}
\title{toggle/delete the hidden flag of the helper gates}
\usage{
gt_toggle_helpergates(gt, gs)

gt_get_helpergates(gt, gs)

gt_delete_helpergates(gt, gs)
}
\arguments{
\item{gt}{gatingTemplate object}

\item{gs}{GatingSet}
}
\description{
The helper gates are defined as the referred gates in csv template. And all the chidlren of referred gates are also 
referred gates thus they are considered the helper gates and can usually be hidden to simply the final gating tree.
}
\details{
Note that delete action is NOT reversible.
}
\examples{
\dontrun{
gt <- gatingTemplate(gtFile)
#run the gating
gt_gating(gt, gs)
#hide the gates that are not of interest
gt_toggle_helpergates(gt, gs) 
#or simply remove them if you are sure they will not be useful in future
gt_delete_helpergates(gt, gs) 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{show,gatingTemplate-method}
\alias{show,gatingTemplate-method}
\title{show method for gatingTemplate}
\usage{
\S4method{show}{gatingTemplate}(object)
}
\arguments{
\item{object}{\code{gatingTemplate}}
}
\description{
show method for gatingTemplate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{fcPolygonGate-class}
\alias{fcPolygonGate-class}
\title{a concrete class that reprents the polygonGate generated by flowClust}
\description{
It stores priors and posteriors as well as the actual polygonGate.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_flowclust_1d}
\alias{gate_flowclust_1d}
\alias{gate_flowClust_1d}
\alias{flowClust.1d}
\title{Applies flowClust to 1 feature to determine a cutpoint between the minimum
cluster and all other clusters.}
\usage{
gate_flowclust_1d(
  fr,
  params,
  filterId = "",
  K = NULL,
  trans = 0,
  min.count = -1,
  max.count = -1,
  nstart = 1,
  prior = NULL,
  criterion = c("BIC", "ICL"),
  cutpoint_method = c("boundary", "min_density", "quantile", "posterior_mean",
    "prior_density"),
  neg_cluster = 1,
  cutpoint_min = NULL,
  cutpoint_max = NULL,
  min = NULL,
  max = NULL,
  quantile = 0.99,
  quantile_interval = c(0, 10),
  plot = FALSE,
  debug = FALSE,
  ...
)
}
\arguments{
\item{fr}{a \code{flowFrame} object}

\item{params}{\code{character} channel to be gated on}

\item{filterId}{A \code{character} string that identifies the filter created.}

\item{K}{the number of clusters to find}

\item{trans, min.count, max.count, nstart}{some flowClust parameters. see \code{\link{flowClust}}}

\item{prior}{list of prior parameters for the Bayesian
\code{\link{flowClust}}. If \code{NULL}, no prior is used.}

\item{criterion}{a character string stating the criterion used to choose the
best model. May take either "BIC" or "ICL". This argument is only relevant
when \code{K} is \code{NULL} or if \code{length(K) > 1}. The value selected
is passed to \code{\link{flowClust}}.}

\item{cutpoint_method}{How should the cutpoint be chosen from the fitted
\code{\link{flowClust}} model? See Details.}

\item{neg_cluster}{integer. The index of the negative cluster. The cutpoint
is computed between clusters \code{neg_cluster} and \code{neg_cluster + 1}.}

\item{cutpoint_min}{numeric value that sets a minimum thresold for the
cutpoint. If a value is provided, any cutpoint below this value will be set
to the given minimum value. If \code{NULL} (default), there is no minimum
cutpoint value.}

\item{cutpoint_max}{numeric value that sets a maximum thresold for the
cutpoint. If a value is provided, any cutpoint above this value will be set
to the given maximum value. If \code{NULL} (default), there is no maximum
cutpoint value.}

\item{min}{a numeric value that sets the lower bound for data filtering. If
\code{NULL} (default), no truncation is applied.}

\item{max}{a numeric value that sets the upper bound for data filtering. If
\code{NULL} (default), no truncation is applied.}

\item{quantile}{the quantile for which we will find the cutpoint using
the quantile \code{cutpoint_method}. If the \code{cutpoint_method} is not set
to \code{quantile}, this argument is ignored.}

\item{quantile_interval}{a vector of length 2 containing the end-points of
the interval of values to find the quantile cutpoint. If the
\code{cutpoint_method} is not set to \code{quantile}, this argument is
ignored.}

\item{plot}{logical value indicating that the fitted \code{\link{flowClust}}
model should be plotted along with the cutpoint}

\item{debug}{\code{logical} indicating whether to carry the prior and posterious with the gate
for debugging purpose. Default is FALSE.}

\item{...}{additional arguments that are passed to \code{\link{flowClust}}}
}
\value{
a \code{rectangleGate} object consisting of all values beyond the
cutpoint calculated
}
\description{
We cluster the observations in \code{fr} into \code{K} clusters.
}
\details{
By default, the cutpoint is chosen to be the boundary of the first two
clusters. That is, between the first two cluster centroids, we find the
midpoint between the largest observation from the first cluster and the
smallest observations from the second cluster. Alternatively, if the
\code{cutpoint_method} is \code{min_density}, then the cutpoint is the point
at which the density between the first and second smallest cluster centroids
is minimum.
}
\examples{
\dontrun{
 gate <- gate_flowclust_1d(fr, params = "APC-A", K =2) # fr is a flowFrame
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{gt_get_children}
\alias{gt_get_children}
\alias{getChildren}
\alias{getChildren,gatingTemplate,character-method}
\title{get children nodes}
\usage{
gt_get_children(obj, y)
}
\arguments{
\item{obj}{\code{gatingTemplate}}

\item{y}{\code{character} parent node path}
}
\description{
get children nodes
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))

gt_get_nodes(gt, "/nonDebris")
gt_get_children(gt, "/nonDebris") 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_pop.R, R/bayes-flowClust.R,
%   R/csvTemplate-parser.R, R/gating-functions.R, R/gating-methods.R,
%   R/gatingTemplate-methods.R, R/pluginFramework.R, R/remove_pop.R,
%   R/toggle.helperGates.R
\name{openCyto-deprecated}
\alias{openCyto-deprecated}
\title{Deprecated functions in package \pkg{openCyto}.}
\description{
\code{add_pop} --> \code{\link{gs_add_gating_method}}

\code{add_pop_init} --> \code{\link{gs_add_gating_method_init}}

\code{prior_flowClust} --> \code{\link{prior_flowclust}}

\code{templateGen} --> \code{\link{gh_generate_template}}

\code{gate_flowClust_1d} --> \code{\link{gate_flowclust_1d}}

\code{gate_flowClust_2d} --> \code{\link{gate_flowclust_2d}}

\code{quantileGate} --> \code{\link{gate_quantile}}

\code{quadGate.seq} --> \code{\link{gate_quad_sequential}}

\code{quadGate.tmix} --> \code{\link{gate_quad_tmix}}

\code{gating} --> \code{\link{gt_gating}}

\code{getNodes} --> \code{\link{gt_get_nodes}}

\code{getChildren} --> \code{\link{gt_get_children}}

\code{getParent} --> \code{\link{gt_get_parent}}

\code{getGate} --> \code{\link{gt_get_gate}}

\code{listgtMethods} --> \code{\link{gt_list_methods}}

\code{registerPlugins} --> \code{\link{register_plugins}}

\code{remove_pop} --> \code{\link{gs_remove_gating_method}}

\code{toggle.helperGates} --> \code{\link{gt_toggle_helpergates}}

\code{get.helperGates} --> \code{\link{gt_get_helpergates}}

\code{delete.helperGates} --> \code{\link{gt_delete_helpergates}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcTree-methods.R
\name{plot,fcTree,character-method}
\alias{plot,fcTree,character-method}
\title{plot the flowClust gating results}
\usage{
\S4method{plot}{fcTree,character}(x, y, channel = NULL, data = NULL, ...)
}
\arguments{
\item{x}{\code{fcTree}}

\item{y}{\code{character} node name in the \code{fcTree}}

\item{channel}{\code{character} specifying the channel.}

\item{data}{\code{GatingSet} that the \code{fcTree} is associated with}

\item{...}{other arguments}
}
\description{
This provides the priors and posteriors as well as the gates for the purpose of debugging flowClust gating algorithm
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{gtPopulation-class}
\alias{gtPopulation-class}
\title{A class to represent a cell population that will be generated by a gating method.}
\description{
A class to represent a cell population that will be generated by a gating method.
}
\section{Slots}{
 
\describe{

 \item{id}{ \code{numeric} unique ID that is consistent with node label of graphNEL in gating template} 
 \item{name}{ \code{character} the name of population}
 \item{alias}{ \code{character} the more user friendly name of population}
}
}

\examples{
 \dontrun{
     gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
      
     gt_get_nodes(gt, '2')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcFilterList-methods.R
\name{plot,fcFilterList,ANY-method}
\alias{plot,fcFilterList,ANY-method}
\title{plot a \code{fcFilterList}}
\usage{
\S4method{plot}{fcFilterList,ANY}(
  x,
  y,
  samples = NULL,
  posteriors = FALSE,
  xlim = NULL,
  ylim = NULL,
  node = NULL,
  data = NULL,
  breaks = 20,
  lwd = 1,
  ...
)
}
\arguments{
\item{x}{\code{fcFilterList}}

\item{y}{\code{character} channel name}

\item{samples}{\code{character} a vector of sample names to be plotted}

\item{posteriors}{\code{logical} indicating whether posteriors should be plotted}

\item{xlim, ylim}{scale settings for x,y axises}

\item{node}{\code{character} population name associated with the \code{fcFilterList}}

\item{data}{\code{GatingSet} object}

\item{breaks}{passed to \link{hist}}

\item{lwd}{line width}

\item{...}{other arguments passed to base \code{plot}}
}
\description{
It is usually called by \code{plot} method for \code{fcTree} instead of directly by users.
}
\examples{
\dontrun{ 
 env1<-new.env(parent=emptyenv())
 #gt is a gatingTemplate, gs is a GatingSet
 gt_gating(gt,gs,env1) #the flowClust gating results are stored in env1 
 plot(env1$fct,"nonDebris",post=T) #plot the priors as well as posteriors for the "nonDebris" gate
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add.R
\name{ocRectRefGate}
\alias{ocRectRefGate}
\title{constructor for ocRectRefGate}
\usage{
ocRectRefGate(rectGate, boolExprs)
}
\arguments{
\item{rectGate}{\code{rectangleGate}}

\item{boolExprs}{\code{character} boolean expression of reference nodes}
}
\description{
constructor for ocRectRefGate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayes-flowClust.R
\name{prior_flowclust}
\alias{prior_flowclust}
\alias{prior_flowClust}
\title{Elicits data-driven priors from a flowSet object for specified channels}
\usage{
prior_flowclust(
  flow_set,
  channels,
  prior_method = c("kmeans"),
  K = 2,
  nu0 = 4,
  w0 = c(10, 10),
  shrink = 1e-06,
  ...
)
}
\arguments{
\item{flow_set}{a \code{flowSet} object}

\item{channels}{a character vector containing the channels in the
\code{flowSet} from which we elicit the prior parameters for the Student's t
mixture}

\item{prior_method}{the method to elicit the prior parameters}

\item{K}{the number of mixture components to identify}

\item{nu0}{prior degrees of freedom of the Student's t mixture components.}

\item{w0}{the number of prior pseudocounts of the Student's t mixture components. (only the first element is used and the rest is ignored at the moment)}

\item{shrink}{the amount of eigenvalue shrinkage to add in the case the prior
covariance matrices are singular. See details.}

\item{...}{Additional arguments passed to the prior elicitation method selected}
}
\value{
list of the necessary prior parameters
}
\description{
We elicit data-driven prior parameters from a \code{flowSet} object for
specified channels. For each sample in the \code{flowSet} object, we apply the
given \code{prior_method} to elicit the priors parameters.
}
\details{
Currently, we have implemented only two methods. In the case that one channel
is given, we use the kernel-density estimator (KDE) approach for each sample
to obtain \code{K} peaks from which we elicit prior parameters. Otherwise,
if more than one channel is specified, we apply K-Means to each of the samples
in the \code{flowSet} and aggregate the clusters to elicit the prior
parameters.

In the rare case that a prior covariance matrix is singular, we shrink the
eigenvalues of the matrix slightly to ensure that it is positive definite. For
instance, if the \code{flow_set} has two samples, this case can occur. The
amount of shrinkage is controlled in \code{shrink}.
}
\examples{
\dontrun{
library(flowCore)
data(GvHD)
prior_flowclust(GvHD[1:3], c("FSC-H", "SSC-H"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{fcFilterList-class}
\alias{fcFilterList-class}
\title{a class that extends \code{filterList} class.}
\description{
Each filter in the filterList must extends the \code{fcFilter} class
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{gtSubsets-class}
\alias{gtSubsets-class}
\title{A class representing a group of cell populations.}
\description{
It extends \code{gtPopulation} class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{gt_get_parent}
\alias{gt_get_parent}
\alias{getParent}
\alias{getParent,gatingTemplate,character-method}
\title{get parent nodes}
\usage{
gt_get_parent(obj, y, isRef = FALSE)
}
\arguments{
\item{obj}{\code{gatingTemplate}}

\item{y}{\code{character} child node path}

\item{isRef}{\code{logical} whether show the reference node besides the parent node}
}
\description{
get parent nodes
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))

gt_get_nodes(gt, "/nonDebris")
gt_get_parent(gt, "/nonDebris/singlets") 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/csvTemplate-parser.R
\name{gh_generate_template}
\alias{gh_generate_template}
\alias{templateGen}
\title{generate a partially complete csv template from the existing gating hierarchy}
\usage{
gh_generate_template(gh)
}
\arguments{
\item{gh}{a \code{GatingHierarchy} likely parsed from a xml workspace}
}
\value{
a gating template in \code{data.frame} format that requires further edition after output to csv
}
\description{
To ease the process of replicating the existing (usually a manual one) gating schemes, 
this function populate an empty gating template with the 'alias', 'pop', 'parent' and 'dims' 
columns that exacted from an \code{GatingHierarchy}, and leave the other columns (e.g. `gating_method`) blank.
So users can make changes to that template instead of writing from scratch.
}
\examples{
library(flowWorkspace)
dataDir <- system.file("extdata",package="flowWorkspaceData")
gs <- load_gs(list.files(dataDir, pattern = "gs_manual",full = TRUE))
gh_generate_template(gs[[1]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing-method.R
\name{preprocessing,ppMethod,GatingSet-method}
\alias{preprocessing,ppMethod,GatingSet-method}
\title{apply a \link[openCyto:ppMethod-class]{ppMethod} to the \code{GatingSet}}
\usage{
\S4method{preprocessing}{ppMethod,GatingSet}(x, y, ...)
}
\arguments{
\item{x}{\code{ppMethod}}

\item{y}{\code{GatingSet} or \code{GatingSetList}}

\item{...}{other arguments}
}
\description{
apply a \link[openCyto:ppMethod-class]{ppMethod} to the \code{GatingSet}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_pop.R
\name{gs_add_gating_method}
\alias{gs_add_gating_method}
\alias{add_pop}
\title{apply a gating method to the \code{GatingSet}}
\usage{
gs_add_gating_method(
  gs,
  alias = "*",
  pop = "+",
  parent,
  dims = NA,
  gating_method,
  gating_args = NA,
  collapseDataForGating = NA,
  groupBy = NA,
  preprocessing_method = NA,
  preprocessing_args = NA,
  strip_extra_quotes = FALSE,
  ...
)
}
\arguments{
\item{gs}{GatingSet or GatingSetList}

\item{alias, pop, parent, dims, gating_method, gating_args, collapseDataForGating, groupBy, preprocessing_method, preprocessing_args}{see details in \link[openCyto:gatingTemplate-class]{gatingTemplate}}

\item{strip_extra_quotes}{\code{logical} Extra quotes are added to strings by fread. This causes problems with parsing R strings to expressions in some cases. Default FALSE for usual behaviour. TRUE should be passed if parsing gating_args fails.}

\item{...}{other arguments
\itemize{
     \item{mc.cores}{ passed to \code{multicore} package for parallel computing}
     \item{parallel_type}{  \code{character} specifying the parallel type. The valid options are "none", "multicore", "cluster".}
     \item{cl}{ \code{cluster} object passed to \code{parallel} package (when \code{parallel_type} is "cluster")}
     }}
}
\description{
When interacting with the existing gated data, this function provides an alternative way to interact with the GatingSet 
by supplying the gating description directly through arguments without the need to write the complete
csv gating template.
}
\details{
Calls to \code{gs_add_gating_method} can also be easily reversed with \code{\link{gs_remove_gating_method}}. Note, however, that it is not possible
to differentiate between different \code{GatingSet} objects loaded from the same directory with 
\code{\link[flowWorkspace]{load_gs}} within a session. Thus, to guarantee a clean history for \code{gs_remove_gating_method}, 
it is necessary to call \code{\link{gs_add_gating_method_init}} on the loaded \code{GatingSet} immediately after re-loading it. 
See the documentation for \code{\link{gs_add_gating_method_init}} for more details. 
This will not be an issue for \code{GatingSet} objects created directly using the constructor.
}
\examples{
\dontrun{
# add quad gates 
gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")

# polyfunctional gates (boolean combinations of exsiting marginal gates)
gs_add_gating_method(gs, gating_method = "polyFunctions", parent = "cd8", gating_args = "cd8/IFNg:cd8/IL2:cd8/TNFa")

#boolGate method
gs_add_gating_method(gs, alias = "IL2orIFNg", gating_method = "boolGate", parent = "cd4", gating_args = "cd4/IL2|cd4/IFNg") 
}
}
\seealso{
\code{\link{gs_remove_gating_method}} \code{\link{gs_add_gating_method_init}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add.R
\docType{class}
\name{ocRectangleGate-class}
\alias{ocRectangleGate-class}
\title{the class that carries event indices as well}
\description{
the class that carries event indices as well
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pluginFramework.R
\name{register_plugins}
\alias{register_plugins}
\alias{registerGatingFunction}
\alias{registerPlugins}
\title{Register a gating or preprocessing function with OpenCyto}
\usage{
register_plugins(fun = NA, methodName, dep = NA, ...)
}
\arguments{
\item{fun}{\code{function} to be registered}

\item{methodName}{\code{character} name of the gating or preprocessing method}

\item{dep}{\code{character} name of the library dependency required for the plugin method to work.}

\item{...}{other arguments
type \code{character} specifying the type of registering method. Should be either "gating" or "preprocessing".}
}
\value{
\code{logical} TRUE if successful and prints a message. FALSE otherwise.
}
\description{
Function registers a new gating or preprocessing method with openCyto so that it may be used in the 
csv template.
}
\details{
The \code{fun} argument should be a wrapper function definition for the gating or preprocessing method. 
                         Gating method must have formal arguments:

                          fr a \code{flowFrame}

                          pp_res a pre-processing result

                          xChannel \code{character} (optional)

                          yChannel \code{character} (required)

                          filterId \code{character}

                          ... ellipses for the additional parameters.

                         Preprocessing method must have formal arguments:

                         fs a \code{flowSet} that stores the flow data (could be subgrouped data if \code{groupBy} column is defined in the csv template

                         gs a \code{GatingSet}
 
                         gm a \code{gtMethod} object that stores the information from gating method

                         xChannel \code{character} (required)

                         yChannel \code{character} (required)

                          ... ellipses for the additional parameters.

The gating function must return a filter (i.e. polygonGate or other instance) from flowCore.
The preprocessing can return anything and it will be passed on to the gating function. So it is up to gating function to use and interpret the results of preprocessing.
Not all formal parameters need to be used. Additional arguments are passed via the ... and can be processed in the wrapper
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{plot,gatingTemplate,missing-method}
\alias{plot,gatingTemplate,missing-method}
\alias{plot,gatingTemplate,character-method}
\alias{plot,gatingTemplate-method}
\alias{plot,gatingTemplate,ANY-method}
\title{plot the gating scheme}
\usage{
\S4method{plot}{gatingTemplate,missing}(x, y, ...)
}
\arguments{
\item{x}{\code{gatingTemplate} object}

\item{y}{either \code{character} specifying the root node which can be used to visualize only the subgraph 
or \code{missing} which display the entire gating scheme}

\item{...}{other arguments 

    graphAttr, nodeAttr:  graph rendering attributes passed to \link[Rgraphviz:renderGraph]{renderGraph}
    showRef \code{logical}: whether to display the reference gates. Sometime it maybe helpful to 
                            hide all those reference gates which are not the cell population of interest 
                            and used primarily for generating other population nodes.}
}
\description{
plot the gating scheme using Rgraphviz
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
plot(gt) #plot entire tree
plot(gt, "lymph") #only plot the subtree rooted from "lymph"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcTree-methods.R
\name{getGate,fcTree,character-method}
\alias{getGate,fcTree,character-method}
\title{get gates saved in \code{fcTree}}
\usage{
\S4method{getGate}{fcTree,character}(obj, y, ...)
}
\arguments{
\item{obj}{\code{fcTree}}

\item{y}{\code{character} node name}

\item{...}{other arguments (not used)}
}
\description{
get gates saved in \code{fcTree}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gtMethod-methods.R
\name{names,gtMethod-method}
\alias{names,gtMethod-method}
\title{get gating method name}
\usage{
\S4method{names}{gtMethod}(x)
}
\arguments{
\item{x}{\code{gtMethod}}
}
\description{
get gating method name
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))

gtMthd <- gt_get_gate(gt, "/nonDebris/singlets",  "/nonDebris/singlets/lymph")
names(gtMthd) 
dims(gtMthd)
parameters(gtMthd)
isCollapse(gtMthd)
groupBy(gtMthd)

gtPop <- gt_get_nodes(gt, "/nonDebris/singlets/lymph/cd3/cd4+cd8-/CD38+")
names(gtPop)
alias(gtPop)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gatingTemplate-methods.R
\name{ppMethod,gatingTemplate,character-method}
\alias{ppMethod,gatingTemplate,character-method}
\title{get preprocessing method from the node}
\usage{
\S4method{ppMethod}{gatingTemplate,character}(obj, y, z)
}
\arguments{
\item{obj}{\code{gatingTemplate}}

\item{y}{\code{character} parent node path}

\item{z}{\code{character} child node path}
}
\description{
get preprocessing method from the node
}
\examples{
\dontrun{
gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
ppMethod(gt,  "/nonDebris/singlets",  "/nonDebris/singlets/lymph")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add.R
\docType{class}
\name{ocRectRefGate-class}
\alias{ocRectRefGate-class}
\title{special gate type that mix the rectangleGate with boolean gate}
\description{
special gate type that mix the rectangleGate with boolean gate
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_quantile}
\alias{gate_quantile}
\alias{quantileGate}
\title{Determine the cutpoint by the events quantile.}
\usage{
gate_quantile(
  fr,
  channel,
  probs = 0.999,
  plot = FALSE,
  filterId = "",
  min = NULL,
  max = NULL,
  ...
)
}
\arguments{
\item{fr}{a \code{flowFrame} object}

\item{channel}{the channel from which the cytokine gate is constructed}

\item{probs}{probabilities passed to 'stats::quantile' function.}

\item{plot}{whether to plot the gate result}

\item{filterId}{the name of the filter}

\item{min}{a numeric value that sets the lower boundary for data filtering}

\item{max}{a numeric value that sets the upper boundary for data filtering}

\item{...}{additional arguments passed to 'stats::quantile' function.}
}
\value{
a \code{rectangleGate}
}
\description{
It is possible that the cutpoint calculated by quantile function may not 
produce the exact the probability set by 'probs' argument if there are not enough
cell events to reach that precision. Sometime the difference could be significant.
}
\examples{
\dontrun{
 gate <- gate_quantile(fr, Channel = "APC-A", probs = 0.995) # fr is a flowFrame
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{boolMethod-class}
\alias{boolMethod-class}
\title{A class to represent a boolean gating method.}
\description{
It extends \code{refGate} class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CytoExploreR_wrappers.R
\name{CytoExploreR_exports}
\alias{CytoExploreR_exports}
\alias{CytoExploreR_.argDeparser}
\alias{CytoExploreR_.preprocess_csv}
\alias{CytoExploreR_.preprocess_row}
\title{CytoExploreR exports}
\usage{
CytoExploreR_.argDeparser(args, split = TRUE)

CytoExploreR_.preprocess_csv(dt, strict = TRUE)

CytoExploreR_.preprocess_row(this_row, strict = TRUE)
}
\description{
Exported wrappers of internal functions for use by CytoExploreR
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{gtMethod-class}
\alias{gtMethod-class}
\alias{gtMethod}
\title{A class to represent a gating method.}
\description{
A gating method object contains the specifics for generating the gates.
}
\section{Slots}{

 \describe{

     \item{name}{ a \code{character} specifying the name of the gating method}
     \item{dims}{ a \code{character} vector specifying the dimensions (channels or markers) of the gate}

     \item{args}{ a \code{list} specifying the arguments passed to gating function}

     \item{groupBy}{ a \code{character} or \code{integer} specifying how to group the data.
                  If \code{character}, group the data by the study variables (columns in \code{pData}).
                   If \code{integer},  group the data by every \code{N} samples.
                  }

     \item{collapse}{ a \code{logical} specifying wether to collapse the data within group before gating.
                 it is only valid when \code{groupBy} is specified}
 }
}

\examples{
 \dontrun{
     gt <- gatingTemplate(system.file("extdata/gating_template/tcell.csv",package = "openCyto"))
     gh_pop_get_gate(gt, '2', '3')
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add.R
\name{pop_add.ocRectangleGate}
\alias{pop_add.ocRectangleGate}
\alias{pop_add.ocRectRefGate}
\title{bypass the default flowWorkspace:::.addGate}
\usage{
\method{pop_add}{ocRectangleGate}(gate, gh, recompute, ...)

\method{pop_add}{ocRectRefGate}(gate, gh, recompute, ...)
}
\arguments{
\item{gate}{\code{ocRectangleGate} or \code{logicalFilterResult}}

\item{gh}{\code{GatingHierarchy} see \link[flowWorkspace]{add} in \code{flowWorkspace} package}

\item{recompute}{\code{logical} see \link[flowWorkspace]{add} in \code{flowWorkspace} package}

\item{...}{see \link[flowWorkspace]{add} in \code{flowWorkspace} package}
}
\description{
to support adding gate along with indices without loading flow data and computing

to support adding rectangleGate yet gating through boolean operations 
without loading flow data
}
\details{
however it is proven that logical indices are too big to be efficiently passed around
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{polyFunctions-class}
\alias{polyFunctions-class}
\title{A class to represent a polyFunctions gating method.}
\description{
It extends \code{boolMethod} class and will be expanded to multiple \code{boolMethod} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{fcFilter-class}
\alias{fcFilter-class}
\title{a virtual class that represents the gating result generated by flowClust gating function}
\description{
Bascially it extends flowCore 'filter classes to have extra slot to store priors and posteriors
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcFilter-methods.R
\name{show,fcFilter-method}
\alias{show,fcFilter-method}
\title{show method for fcFilter}
\usage{
\S4method{show}{fcFilter}(object)
}
\arguments{
\item{object}{\code{fcFilter}
show method for fcFilter}
}
\description{
show method for fcFilter
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\docType{class}
\name{refGate-class}
\alias{refGate-class}
\title{A class to represent a reference gating method.}
\description{
It extends \code{gtMethod} class.
}
\section{Slots}{

\describe{ 
 \item{refNodes}{ \code{character} specifying the reference nodes}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gating-functions.R
\name{gate_quad_tmix}
\alias{gate_quad_tmix}
\alias{quadGate.tmix}
\title{quadGate based on flowClust::tmixFiler}
\usage{
gate_quad_tmix(
  fr,
  channels,
  K,
  usePrior = "no",
  prior = list(NA),
  quantile1 = 0.8,
  quantile3 = 0.8,
  trans = 0,
  plot = FALSE,
  ...
)
}
\arguments{
\item{fr}{\code{flowFrame}}

\item{channels}{\code{character} vector specifies two channels}

\item{K}{see \link{gate_flowclust_2d}}

\item{usePrior}{see \link{gate_flowclust_2d}}

\item{prior}{see \link{gate_flowclust_2d}}

\item{quantile1}{\code{numeric} specifies the  quantile level(see 'level' in \link{flowClust}) for the first quadrant (x-y+)}

\item{quantile3}{\code{numeric} specifies the  quantile level see 'level' in \link{flowClust} for third quadrant (x+y-)}

\item{trans}{see \link{gate_flowclust_2d}}

\item{plot}{logical whether to plot flowClust clustering results}

\item{...}{other arguments passed to \link{flowClust}}
}
\value{
a \code{filters} object that contains four \code{polygonGate}s following the order of (-+,++,+-,--)
}
\description{
This gating method identifies two quadrants (first, and third quadrants) by fitting the data with tmixture model.
It is particually useful when the two markers are not well resolved thus the regular quadGate method
based on 1d gating will not find the perfect cut points on both dimensions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_pop.R
\name{gs_add_gating_method_init}
\alias{gs_add_gating_method_init}
\alias{add_pop_init}
\title{Clear history of \code{gs_add_gating_method} calls for a given \code{GatingSet} or \code{GatingSetList}}
\usage{
gs_add_gating_method_init(gs)
}
\arguments{
\item{gs}{a \code{GatingSet} or \code{GatingSetList}. Can be omitted to clean entire \code{gs_add_gating_method} history.}
}
\description{
Repeated calls to the \code{\link{load_gs}} method in the same session
will yield indistinguishable objects that can result in overlapping history
of \code{\link{gs_add_gating_method}} calls. This method allows for the history to be cleared
if the user would like to reload the \code{GatingSet} and start fresh. Calling
\code{gs_add_gating_method_init} without an argument will clear the entire \code{gs_add_gating_method} history.
}
\examples{
\dontrun{
# load in a GatingSet
gs <- load_gs(path)
# Add some nodes using gs_add_gating_method
gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
gs_add_gating_method(gs, gating_method = "polyFunctions", parent = "cd8", gating_args = "cd8/IFNg:cd8/IL2:cd8/TNFa")
# Remove the effect of the last gs_add_gating_method call using gs_remove_gating_method (note that the first call's effects remain)
gs_remove_gating_method(gs)
# Re-load the GatingSet to start over
gs <- load_gs(path)

# At this point, gs will still see the history of the first gs_add_gating_method call above
# which will cause problems for later calls to gs_remove_gating_method.
# To fix that, just call gs_add_gating_method_init() to start a clean history
gs_add_gating_method_init(gs)
# Now you can continue using gs_add_gating_method and gs_remove_gating_method from scratch
gs_add_gating_method(gs, gating_method = "mindensity", dims = "CCR7,CD45RA", parent = "cd4-cd8+", pop = "CCR7+/-CD45RA+/-")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{fcEllipsoidGate}
\alias{fcEllipsoidGate}
\title{constuctor for \code{fcEllipsoidGate}}
\usage{
fcEllipsoidGate(x, priors, posts)
}
\arguments{
\item{x}{a \code{ellipsoidGate} object}

\item{priors}{a \code{list} storing priors}

\item{posts}{a \code{list} storing posteriors}
}
\description{
constuctor for \code{fcEllipsoidGate}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllClasses.R
\name{fcTree}
\alias{fcTree}
\title{constructor of \code{fcTree}}
\usage{
fcTree(gt)
}
\arguments{
\item{gt}{a \code{gatingTemplate} object}
}
\description{
It adds an extra node data slot "fList"(which is a \code{filterList} object) to the \code{gatingTemplate}
}
